/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <assert.h>
#include <string.h>

#include "bip.h"

#include "orfa_network.h"
#include "orfa_debug.h"


/********************************
 * Non-exported BIP definitions *
 ********************************/

extern void bip_exit();
extern void dio_clean(void*, int);
extern char bip_nodename[][255];


/************
 * Register *
 ************/

void orfa_BIP_deregister_interval(void *start,
				  unsigned int length)
{
  dio_clean(start, (int)length);
}



/**************
 * Connecting *
 **************/

int orfa_BIP_network_init()
{
  bip_async_init();

  context.data_send_tag = REQUEST_DATA_TAG_FOR_NODE(bip_mynode);
  context.big_data_send_tag = REQUEST_BIG_DATA_TAG_FOR_NODE(bip_mynode);

  if ( context.data_send_tag >= NTAGS_LOCAL-1
       ||  context.big_data_send_tag >= NTAGS_LOCAL-1 ) {
    ERROR("This node %d needs tags %d-%d while %d are given\n",
	  bip_mynode,
	  context.data_send_tag, context.big_data_send_tag,
	  NTAGS_LOCAL);
    bip_exit();
    return -1;
  }

  return 0;
}


void orfa_BIP_network_finish()
{
  bip_exit();
}


int orfa_BIP_network_connect(char *server_name, struct connection *connection)
{
  int node = -1;
#ifndef FIXED_SERVER_NODE
  int i;
  LOG(3, "Connecting to server %s...\n", server_name);
  
  for(i=0; i<bip_numnodes; i++)
    {
      LOG(4, "Comparing name %s and %s\n", server_name, bip_nodename[i]);
      if ( !strcmp(server_name, bip_nodename[i]) )
	node = i;
    }
#else /* ~FIXED_SERVER_NODE */
  node = FIXED_SERVER_NODE;
#endif /* FIXED_SERVER_NODE */
  
  if ( node >= 0 )
    {
      bip_connect(node, PROTOCOL_TAG, PROTOCOL_TAG);
      bip_connect(node, context.data_send_tag, ANSWER_DATA_TAG);
      bip_connect(node, context.big_data_send_tag, ANSWER_DATA_TAG);
      connection->node = node;
      LOG(4, "Connected to server node %d\n", node);
      return 0;
    }
  else
    {
      ERROR("Didn't find server %s\n", server_name);
      return -1;
    }
}



/*******************
 * Sending/Recving *
 *******************/

/* this function provides recv buffers and sends request */
static int request_send(struct client_request *req)
{
  long ready_recv_id = -1;
  int in_buf_len = req->req.buffer_len;
  void *in_buf = req->in_buf;
  int out_buf_len = req->out_buf_len;
  void *out_buf = req->out_buf;

  if ( in_buf_len > 0 )
    {
      ALIGN_TO_WORD_SIZE(in_buf_len);
    }    

  /* prepare a readiness receive buffer if necessary */
  if ( in_buf_len > PREPROVIDED_BUFFER_LENGTH )
    {
      ready_recv_id = bip_tirecv(PROTOCOL_TAG, NULL, 0);
    }
  else
    {
      /* receiving answer buffer */
      req->answer_recv_id = bip_tirecv(PROTOCOL_TAG,
				       (int*)&req->ans,
				       sizeof(struct answer)/WORD_SIZE);
    }
  
  /* receiving data buffer if necessary */
  if ( out_buf_len > 0 )
    {
      ALIGN_TO_WORD_SIZE(out_buf_len);
      LOG(6, "Receiving %ld data at %p\n",
	  (long)out_buf_len, out_buf);

      req->data_recv_id = bip_tirecv(ANSWER_DATA_TAG,
				     (int*)out_buf,
				     out_buf_len/WORD_SIZE);
    }
  else
    {
      LOG(6, "No data to receive\n");
      req->data_recv_id = -1;
    }

  LOG(3, "Sending request type %d to node %d\n",
      (int)req->req.type, req->connection->node);

  /* Check REQ type */
  if ( req->req.type <= FIRST_REQ
       ||  req->req.type >= LAST_REQ )
    {
      ERROR("Bad request type %d !\n", (int)req->req.type);
      assert(0);
    }

  bip_tsend(req->connection->node,
	    PROTOCOL_TAG,
	    (int*)&req->req,
	    sizeof(struct request)/WORD_SIZE);
  LOG(4, "Request sent\n");

  /* process data if necessary */
  if ( in_buf_len > PREPROVIDED_BUFFER_LENGTH )
    {
      /* receiving readiness */
      LOG(3, "Waiting for server response\n");
      bip_rwait(ready_recv_id);
      
      /* receiving answer buffer */
      req->answer_recv_id = bip_tirecv(PROTOCOL_TAG,
				       (int*)&req->ans,
				       sizeof(struct answer)/WORD_SIZE);

      /* sending data */
      LOG(3, "Sending %d data at %p\n",
	  in_buf_len, in_buf);

      bip_tsend(req->connection->node,
		context.big_data_send_tag,
		(int*)req->in_buf,
		in_buf_len/WORD_SIZE);
    }
  else if ( in_buf_len >= 0 )
    {
      /* sending directly */
      LOG(3, "Sending %d data at %p\n",
	  in_buf_len, req->in_buf);

      bip_tsend(req->connection->node,
		context.data_send_tag,
		(int*)in_buf,
		in_buf_len/WORD_SIZE);
    }
  
  return 0;
}


/* this function does not allocate anything,
   it just fills things */
static int answer_recv(struct client_request *req)
{
  /* wait for the answer */
  bip_rwait(req->answer_recv_id);

#ifdef PROTOCOL_DEBUG
  /* check tag/pid */
  if ( req->ans.tag != req->req.tag
       ||  req->ans.pid != req->req.pid )
    {
      ERROR("Recv tag/pid %d/%d for %d/%d\n",
	    (int)req->ans.tag, (int)req->ans.pid,
	    (int)req->req.tag, (int)req->req.pid);
      assert(0);
    }
#endif /* PROTOCOL_DEBUG */

  /* check ANS type */
  LOG(5, "Received type %d\n",
      (int)req->ans.type);
  if ( req->ans.type <= FIRST_ANS
       ||  req->ans.type >= LAST_ANS )
    {
      ERROR("Bad answer type %d, impossible with BIP ?!\n",
	    (int)req->ans.type);
      assert(0);
    }

  /* wait for data if necessary */
  if ( req->data_recv_id >= 0 )
    {
      int rlen = bip_rwait(req->data_recv_id) * WORD_SIZE;
      LOG(5, "Data recv\n");

      if ( rlen != req->ans.buffer_len )
	{
	  LOG(4, "Received %d instead of %lld\n",
	      rlen, (long long)req->ans.buffer_len);

	  req->ans.buffer_len = rlen;
	}
    }

  return 0;
}


int orfa_BIP_network_call(struct client_request *req)
{
  int res = request_send(req);
  return res < 0 ? res : answer_recv(req);
}
