/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <asm/page.h>
#include <sched.h>
#include <string.h>
#include <assert.h>

#include "gm.h"

#include "orfa_network.h"
#include "orfa_client.h"
#include "orfa_regcache.h"
#include "orfa_debug.h"


/**********************
 * Static definitions *
 **********************/

/* id of the first port we'll try */
static int client_port_id;
/* the open port */
static struct gm_port *client_port = NULL;

/* some sizes we'll compute at the begining */
static unsigned int request_size, answer_size, preprovided_buffer_size;

/* static buffers for current request and answer */
static struct request *static_request;
static struct answer *static_answer;
/* static buffers for current data, if not to small */
static void *static_send_buffer, *static_recv_buffer;


#ifdef REGCACHE_ENABLED
# define STATIC_BUFFER_SIZE 0
#else /* REGCACHE_ENABLED */
# define STATIC_BUFFER_SIZE PAGE_SIZE
#endif /* ~REGCACHE_ENABLED */



/***************
 * Registering *
 ***************/

int orfa_GM_register_interval(void *start,
			 unsigned int length)
{
  LOG(7, "GM registering %d at %p\n", length, start);
  if ( gm_register_memory(client_port, start, length)
       == GM_SUCCESS )
    return 0;
  return -1;
}


void orfa_GM_deregister_interval(void *start,
				 unsigned int length)
{
  LOG(7, "GM deregistering %d at %p\n", length, start);
  gm_deregister_memory(client_port, start, length);
}

			  
/**************
 * Connecting *
 **************/

int orfa_GM_network_init()
{
  gm_status_t status;

  status = gm_init();

  if ( status != GM_SUCCESS )
    {
      ERROR("Cannot init GM (error %s)\n", gm_strerror(status));
      return -1;
    }
  LOG(1, "GM successfully launched\n");
  
  client_port_id = CLIENT_FIRST_PORT_ID;

  while ( client_port == NULL )
    {
      status = gm_open(&client_port,
		       CLIENT_BOARD_NUM, client_port_id,
		       "server_port", GM_API_VERSION);
      if ( status == GM_SUCCESS )
	{
	  LOG(1, "GM port %d successfully open\n", client_port_id);
	  break;
	}
      
      LOG(2, "GM cannot open port %d on board %d (error %s)\n",
	  client_port_id, CLIENT_BOARD_NUM, gm_strerror(status));
      
      if ( status != GM_BUSY )
        {
          ERROR("Unable to open a GM port\n");
          gm_finalize();
          return -1;
        }

      client_port = NULL;
      client_port_id++;
    }
  
  request_size = gm_min_size_for_length( sizeof(struct request) );  
  answer_size = gm_min_size_for_length( sizeof(struct answer) );  
  preprovided_buffer_size = gm_min_size_for_length(PREPROVIDED_BUFFER_LENGTH);

  static_request = gm_dma_malloc(client_port, sizeof(struct request));
  static_answer = gm_dma_malloc(client_port, sizeof(struct answer));

  static_send_buffer = gm_dma_malloc(client_port, STATIC_BUFFER_SIZE);
  static_recv_buffer = gm_dma_malloc(client_port, STATIC_BUFFER_SIZE);

  return 0;
}


void orfa_GM_network_finish()
{
  if ( client_port != NULL )
    {
      gm_dma_free(client_port, static_request);
      gm_dma_free(client_port, static_answer);
      
      gm_dma_free(client_port, static_send_buffer);
      gm_dma_free(client_port, static_recv_buffer);
      
      gm_close(client_port);
      client_port = NULL;
    }
  
  gm_finalize();
}


int orfa_GM_network_connect(char *server_name, struct connection *connection)
{
  unsigned int node_id;
  
  unsigned long server_port_id = SERVER_PORT_ID;
  char *end_name;
  char *end_port;
  char dummy;

  LOG(3, "Connecting to server %s...\n", server_name);
  
  if ( client_port == NULL )
    {
      LOG(3, "Network not working\n");
      goto error;
    }

  /* parse name and port */
  end_name = strrchr(server_name, ':');
  if ( end_name ) {
    server_port_id = strtol(end_name+1, &end_port, 10);
    if ( end_port && end_port != end_name+1 && *end_port ) {
      ERROR("Cannot extract port from %s\n", end_name+1);
      goto error;
    }
    /* remove :port temporarily */
    *end_name = '\0';
  }

#ifdef FIXED_SERVER_NODE
  node_id = FIXED_SERVER_NODE;
#else /* FIXED_SERVER_NODE */
  node_id = gm_host_name_to_node_id(client_port, server_name);
#endif /* FIXED_SERVER_NODE */

  /* restore name */
  *( end_name ? end_name : &dummy ) = ':';

  if ( node_id == GM_NO_SUCH_NODE_ID )
    {
      ERROR("Didn't find server %s\n", server_name);
      goto error;
    }

  /* set connection */
  connection->node_id = node_id;
  connection->port_id = server_port_id;
  LOG(4, "Got server id %d\n", node_id);
  return 0;

 error:
  return -1;
}



/***********
 * Sending *
 ***********/

static int to_send;


/* Request sending and resending */

static void send_request_callback(struct gm_port *port,
				  void *context,
				  gm_status_t status);

static void resume_send_request_callback(struct gm_port *port,
					 void *context,
					 gm_status_t status)
{
  struct client_request *req = (struct client_request*)context;

  ERROR("Resending request\n");
  gm_send_with_callback(port,
			static_request,
			request_size, sizeof(struct request),
			REQUEST_PRIORITY,
			req->connection->node_id,
			req->connection->port_id,
			send_request_callback, req);
}


static void send_request_callback(struct gm_port *port,
				  void *context,
				  gm_status_t status)
{
  struct client_request *req = (struct client_request*)context;

  switch ( status ) {
  case GM_SUCCESS: {
    LOG(5, "Request successfully sent\n");
    to_send--;
    break;
  }
  case GM_SEND_TIMED_OUT: {
    ERROR("Request send timed out\n");
    gm_resume_sending(port, REQUEST_PRIORITY,
		      req->connection->node_id,
		      req->connection->port_id,
		      resume_send_request_callback, req);
    break;
  }
  default: {
    ERROR("GM send buffer got error %s\n", gm_strerror(status));
    req->status = 1;
    break;
  }
  }
}


/* buffer sending and resending */

static void send_buffer_callback(struct gm_port *port,
				 void *context,
				 gm_status_t status);

static void resume_send_buffer_callback(struct gm_port *port,
					void *context,
					gm_status_t status)
{
  struct client_request *req = (struct client_request*)context;

  ERROR("Resending buffer\n");
  gm_send_with_callback(port,
			req->real_in_buf,
			req->in_buf_size,
			req->req.buffer_len,
			DATA_PRIORITY,
			req->connection->node_id,
			req->connection->port_id,
			send_buffer_callback, req);
}


static void send_buffer_callback(struct gm_port *port,
				 void *context,
				 gm_status_t status)
{
  struct client_request *req = (struct client_request*)context;

  switch ( status ) {
  case GM_SUCCESS: {
    LOG(5, "Buffer successfully sent\n");
    if ( req->real_in_buf != static_send_buffer )
      orfa_unuse_interval(req->real_in_buf, req->req.buffer_len);
    to_send--;
    break;
  }
  case GM_SEND_TIMED_OUT: {
    ERROR("Buffer send timed out\n");
    gm_resume_sending(port, DATA_PRIORITY,
		      req->connection->node_id,
		      req->connection->port_id,
		      resume_send_buffer_callback, req);
    break;
  }
  default: {
    ERROR("GM send buffer got error %s\n", gm_strerror(status));
    req->status = 1;
    break;
  }
  }
}


/* this function does not allocate anything,
   it just sends things */
static int request_send(struct client_request *req)
{
  gm_recv_event_t *gm_ev;

  LOG(3, "Sending request type %d to node %d port %d\n",
      (int)req->req.type,
      (int)req->connection->node_id,
      (int)req->connection->port_id);

  /* Check REQ type */
  if ( req->req.type <= FIRST_REQ
       ||  req->req.type >= LAST_REQ )
    {
      ERROR("Bad request type %d !\n",
	    (int)req->req.type);
      assert(0);
    }

  req->status = 0;

  memcpy(static_request, req, sizeof(struct request));
  gm_send_with_callback(client_port,
			static_request,
			request_size, sizeof(struct request),
			REQUEST_PRIORITY,
			req->connection->node_id, req->connection->port_id,
			send_request_callback, req);
  LOG(4, "Request sent\n");

  to_send = 1;

  /* Add a data buffer if necessary */
  if ( req->req.buffer_len > 0 )
    {
      int data_len = req->req.buffer_len;
        
      if ( data_len > STATIC_BUFFER_SIZE )
	{
	  orfa_use_interval(req->in_buf, data_len);
	  req->real_in_buf = req->in_buf;
	}
      else
	{
	  memcpy(static_send_buffer, req->in_buf, data_len);
	  req->real_in_buf = static_send_buffer;
	}

      req->in_buf_size = REQUEST_DATA_SIZE_OF_LENGTH(data_len);

      /* sending data */
      LOG(3, "Sending data\n");
      gm_send_with_callback(client_port,
			    req->real_in_buf,
			    req->in_buf_size,
			    data_len,
			    DATA_PRIORITY,
			    req->connection->node_id, req->connection->port_id,
			    send_buffer_callback, req);
      LOG(4, "Data sent\n");

      to_send++;
    }

  while ( to_send ) {
#if 1
    gm_ev = gm_receive(client_port);
#else
    gm_ev = gm_blocking_receive(client_port);
#endif
    switch ( gm_ntoh_u8(gm_ev->recv.type) ) {
#if 1
    case GM_NO_RECV_EVENT: {
      sched_yield();
      break;
    }
#endif
    default: {
      gm_unknown(client_port, gm_ev);

      /* handle send errors */
      if ( req->status )
	return -1;

      break;
    }
    }
  }
  
  return 0;
}



/***********
 * Recving *
 ***********/

/* this function provides buffers */
static void provide_receive_buffers(struct client_request *req)
{
  /* providing the buffer for the answer */
  gm_provide_receive_buffer(client_port, static_answer,
			    answer_size, REQUEST_PRIORITY);

  /* receiving data buffer if necessary */
  if ( req->out_buf_len > 0 )
    {
      if ( req->out_buf_len > STATIC_BUFFER_SIZE )
	{
	  req->real_out_buf = req->out_buf;
	  orfa_use_interval(req->out_buf, req->out_buf_len);
	}
      else
	{
	  req->real_out_buf = static_recv_buffer;
	}

      /* provide it for receiving */
      gm_provide_receive_buffer(client_port,
				req->real_out_buf,
				ANSWER_DATA_SIZE,
				DATA_PRIORITY);
      LOG(6, "Receiving %ld data at %p\n",
          (long)req->out_buf_len, req->real_out_buf);
    }
  else
    {
      req->real_out_buf = NULL;
    }
}



/* this function does not allocate anything,
   it just fills things */
static int answer_recv(struct client_request *req)
{
  gm_recv_event_t *gm_ev;
  int to_receive = 1;

  if ( req->real_out_buf )
    to_receive++;

  /* receive answer and data */
  while ( to_receive ) {
#if 1
    gm_ev = gm_receive(client_port);
#else
    gm_ev = gm_blocking_receive(client_port);
#endif
    switch ( gm_ntoh_u8(gm_ev->recv.type) ) {
    case REQUEST_RECV_EVENT:
    case DATA_RECV_EVENT: {
      to_receive--;
      break;
    }
#if 1
    case GM_NO_RECV_EVENT: {
      sched_yield();
      break;
    }
#endif
    default: {
      gm_unknown(client_port, gm_ev);
      break;
    }
    }
  }

  memcpy(&req->ans, static_answer, sizeof(struct answer));

#ifdef PROTOCOL_DEBUG
  /* check tag/pid */
  if ( req->ans.tag != req->req.tag
       ||  req->ans.pid != req->req.pid )
    {
      ERROR("Recv tag/pid %d/%d for %d/%d\n",
	    (int)req->ans.tag, (int)req->ans.pid,
	    (int)req->req.tag, (int)req->req.pid);
      assert(0);
    }
#endif /* PROTOCOL_DEBUG */

  /* check ANS type */
  LOG(5, "Received type %d\n", (int)req->ans.type);
  if ( req->ans.type <= FIRST_ANS
       ||  req->ans.type >= LAST_ANS )
    {
      ERROR("Bad answer type %d, impossible with GM ?!\n",
	    (int)req->ans.type);
      assert(0);
    }

  /* process receive buffer if there's one */
  if ( req->real_out_buf != NULL )
    {
      if ( req->real_out_buf == static_recv_buffer )
	memcpy(req->out_buf, static_recv_buffer, req->out_buf_len);
      else
	orfa_unuse_interval(req->out_buf, req->out_buf_len);
    }

  return 0;
}


int orfa_GM_network_call(struct client_request *req)
{
  int res;
  provide_receive_buffers(req);
  res = request_send(req);
  return res < 0 ? res : answer_recv(req);
}
