/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sched.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>

#include "orfa_network.h"
#include "orfa_client.h"
#include "orfa_debug.h"


#if (defined MAKE_WRAPPER) || (defined MAKE_LIB)
# include "orfa_real_calls.h"
#elif (defined MAKE_MOUNT)
# define REAL_CLOSE close
# define REAL_DUP2 dup2
# include <unistd.h>
#endif



/**************
 * Connecting *
 **************/

int orfa_TCP_network_connect(char *server_name, struct connection *connection)
{
  int client_socket;
  struct hostent *server_hostent;
  struct sockaddr_in server_address = { 0 };
  int magic;

  unsigned long server_port = SERVER_PORT;
  char *end_name;
  char *end_port;
  char dummy;

  LOG(3, "Connecting to server %s...\n", server_name);

  /* open an internet tcp socket */
  if ( (client_socket = socket(PF_INET, SOCK_STREAM, 0)) < 0 )
    {
      perror("openclient socket");
      goto error;
    }
  LOG(4, "Socket open\n");
  
  /* parse name and port */
  end_name = strrchr(server_name, ':');
  if ( end_name ) {
    server_port = strtol(end_name+1, &end_port, 10);
    if ( end_port && end_port != end_name+1 && *end_port ) {
      ERROR("Cannot extract port from %s\n", end_name+1);
      goto error_with_socket;
    }
    /* remove :port temporarily */
    *end_name = '\0';
  }

  /* get structure for remote host node on which server resides */  
  server_hostent = gethostbyname(server_name);

  /* restore name */
  *( end_name ? end_name : &dummy ) = ':';

  /* do not free server_hostent, it is allocated by malloc ! */
  if ( server_hostent == NULL )
    {
      herror(server_name);
      ERROR("Didn't find server %s\n", server_name);
      goto error_with_socket;
    }
  LOG(4, "Got hostent\n");

  /* fill the address of the server */
  server_address.sin_family = server_hostent->h_addrtype;
  server_address.sin_port = htons(server_port);
  memcpy(&server_address.sin_addr,
	 server_hostent->h_addr,
	 server_hostent->h_length);

  /* connect this socket to the server's Internet address */
  if( connect(client_socket,
	      (struct sockaddr*)&server_address,
	      sizeof(struct sockaddr_in)) < 0 )
    {
      perror("openclient connect");
      goto error_with_socket;
    }

  /* check accept has succeeded in server  */
  if ( !read(client_socket, &magic, sizeof(int)) ) {
    ERROR("Server may be overloaded. Connection failed.\n");
    goto error_with_socket;
  }

  /* check magic number */
  if ( magic != NETWORK_MAGIC ) {
    ERROR("Magic number incorrect\n");
    goto error_with_socket;
  }

  LOG(4, "Connected\n");

  /* make the socket close during exec */
  fcntl(client_socket, F_SETFD, FD_CLOEXEC);

  /* make the socket send packets without waiting */
  SOCKET_TCP_NODELAY(client_socket);
  
  /* set connection */
  connection->socket = client_socket;
  return 0;
  
 error_with_socket:
  close(client_socket);
 error:
  return -1;
}



/*******************
 * Sending/Recving *
 *******************/

/* this function does not allocate anything,
   it just sends things */
static int request_send(struct client_request *req)
{
  struct iovec vect[2];
  int vect_len;

  vect[0].iov_base = &req->req;
  vect[0].iov_len = sizeof(struct request);
  vect_len = 1;

#ifdef PROTOCOL_DEBUG
  {
#include <sys/poll.h>
    struct pollfd pfd = { fd: req->connection->socket, events: POLLIN, revents: 0};
    poll(&pfd, 1, 0);
    if ( pfd.revents )
      ERROR("events %d on socket %d\n",
	    pfd.revents, req->connection->socket);
  }
#endif /* PROTOCOL_DEBUG */

  LOG(3, "Sending type %d\n", (int)req->req.type);

  /* Check REQ type */
  if ( req->req.type <= FIRST_REQ
       ||  req->req.type >= LAST_REQ )
    {
      ERROR("Bad request type %d !\n", (int)req->req.type);
      assert(0);
    }

  /* Add a data buffer if necessary */
  if ( req->req.buffer_len > 0 )
    {
      vect[vect_len].iov_base = (void*)req->in_buf;
      vect[vect_len].iov_len = req->req.buffer_len;
      vect_len++;
      
      LOG(4, "Sending %ld data...\n", (long)req->req.buffer_len);
    }

#ifdef CHECK_SERVER
  {
#include <sys/poll.h>
    struct pollfd pfd = { fd: req->connection->socket, events: POLLIN, revents: 0 };
    poll(&pfd, 1, -1);
    if ( pfd.revents & (POLLERR|POLLHUP) )
      {
	ERROR("Server disconnected\n");
	return -1;
      }
  }
#endif /* CHECK_SERVER */

  /* Send the request */
  writev(req->connection->socket, vect, vect_len);

  return 0;
}


/* this function does not allocate anything,
   it just fills things */
static int answer_recv(struct client_request *req)
{
  int res;

  /* Recv the answer */
  res = recv(req->connection->socket, &req->ans, sizeof(struct answer), 0);

#ifdef CHECK_SERVER
  if ( !res )
    {
      ERROR("No data receive : Server disconnected\n");
      return -1;
    }
#endif /* CHECK_SERVER */
  
#ifdef PROTOCOL_DEBUG
  /* Check tag/pid */
  if ( req->ans.tag != req->req.tag
       ||  req->ans.pid != req->req.pid )
    {
      ERROR("Recv tag/pid %d/%d for %d/%d\n",
	    (int)req->ans.tag, (int)req->ans.pid,
	    (int)req->req.tag, (int)req->req.pid);
      assert(0);
    }
#endif /* PROTOCOL_DEBUG */

  /* check the type */
  LOG(5, "Received type %d\n", (int)req->ans.type);
  if ( req->ans.type <= FIRST_ANS
       ||  req->ans.type >= LAST_ANS )
    {
      ERROR("Bad answer type %d, impossible with TCP ?!\n",
	    (int)req->ans.type);
      assert(0);
    }

  /* And a data buffer if necessary */
  if ( req->ans.buffer_len > 0 )
    {
      int rlen = 0;
      int data_len = req->ans.buffer_len;
      
      LOG(4, "Receiving %d data...\n", data_len);
      
      while ( data_len > 0)
	{
	  rlen = recv(req->connection->socket, req->out_buf, data_len, 0);
	  LOG(5, "Recv %d/%d data\n", rlen, data_len);
	  (char*)req->out_buf += rlen;
	  data_len -= rlen;
	}
    }

  return 0;
}


int orfa_TCP_network_call(struct client_request *req)
{
  int res = request_send(req);
  return res < 0 ? res : answer_recv(req);
}
