/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sys/socket.h>
#include <sys/uio.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>

#include "orfa_network.h"
#include "orfa_client.h"
#include "orfa_debug.h"


#if (defined MAKE_WRAPPER) || (defined MAKE_LIB)
# include "orfa_real_calls.h"
#elif (defined MAKE_MOUNT)
# define REAL_CLOSE close
# include <unistd.h>
#endif



/**************
 * Connecting *
 **************/

static int client_socket = -1;


int orfa_UDP_network_init()
{
  if ( (client_socket = socket(PF_INET, SOCK_DGRAM , 0))
       < 0 )
    {
      perror("openclient socket");
      client_socket = -1;
      return -1;
    }
  
  LOG(2, "Now connected to network thru socket %d\n",
      client_socket);

  /* make the socket close during exec */
  fcntl(client_socket, F_SETFL, FD_CLOEXEC);

  return 0;
}


void orfa_UDP_network_finish()
{
 if ( client_socket != -1 )
   REAL_CLOSE(client_socket);
}


int orfa_UDP_network_connect(char *server_name, struct connection *connection)
{
  struct hostent *server_hostent;

  LOG(3, "Connecting to server %s...\n", server_name);
  
  if ( client_socket < 0 )
    {
      LOG(3, "Network not working\n");
      return -1;
    }

  /* get structure for remote host node on which server resides */  
  /* do not free server_hostent, it is allocated by malloc ! */
  if( (server_hostent = gethostbyname(server_name)) != NULL )
    {
      connection->ski.sin_family = server_hostent->h_addrtype;  /* should be AF_INET */
      connection->ski.sin_port = htons(SERVER_PORT);
      memcpy(&connection->ski.sin_addr,
	     server_hostent->h_addr,
	     server_hostent->h_length);
      LOG(4, "Got server\n");
      return 0;
    }
  else
    {
      herror(server_name);
      ERROR("Didn't find server %s\n", server_name);
      return -1;
    }
}



/*******************
 * Sending/Recving *
 *******************/

/* this function does not allocate anything,
   it just sends things */
static int request_send(struct client_request *req)
{
  struct iovec vect[2];
  struct msghdr message;

  vect[0].iov_base = &req->req;
  vect[0].iov_len = sizeof(struct request);

  message.msg_name = (struct sockaddr*)&req->connection->ski;
  message.msg_namelen = sizeof(struct sockaddr);
  message.msg_iov = vect;
  message.msg_iovlen = 1;
  message.msg_control = NULL;
  message.msg_controllen = 0;
  message.msg_flags = 0;

  LOG(3, "Sending type %d\n", (int)req->req.type);

  /* Check REQ type */
  if ( req->req.type <= FIRST_REQ
       ||  req->req.type >= LAST_REQ )
    {
      ERROR("Bad request type %d !\n",
	    (int)req->req.type);
      assert(0);
    }

  /* add a buffer if necessary */
  if ( req->req.buffer_len > 0 )
    {
      char *remaining_buf;
      size_t remaining_len;
      size_t to_send;
      
      remaining_buf = (char*)req->in_buf;
      remaining_len = req->req.buffer_len;

      LOG(3, "%ld data to send...\n",
	  (long)remaining_len);

      /* send the request and the begining of data */
      to_send = remaining_len > MAX_MESSAGE_LEN - sizeof(struct request)
	? MAX_MESSAGE_LEN - sizeof(struct request)
	: remaining_len;

      message.msg_iovlen = 2;
      vect[1].iov_base = remaining_buf;
      vect[1].iov_len = to_send;
      LOG(3, "Send request and begining of data\n");
      sendmsg(client_socket, &message, 0);

      while ( remaining_len > to_send ) {
	/* update buffer parameters */
	remaining_buf += to_send;
	remaining_len -= to_send;

	/* send next part of data */
	to_send = remaining_len > MAX_MESSAGE_LEN
	  ? MAX_MESSAGE_LEN
	  : remaining_len;
	sendto(client_socket,
	       remaining_buf, to_send,
	       0,
	       message.msg_name, message.msg_namelen);
	LOG(3, "Send %ld next data\n",
	    (long)to_send);
      }
    }
  else
    {
      /* just send a message */
      LOG(3, "Just send the request\n");
      sendmsg(client_socket, &message, 0);
    }
  
  return 0;
}


/* this function does not allocate anything,
   it just fills things */
static int answer_recv(struct client_request *req)
{
  struct iovec vect[2];
  struct msghdr message;

  vect[0].iov_base = &req->ans;
  vect[0].iov_len = sizeof(struct answer);

  message.msg_name = NULL;
  message.msg_namelen = 0;
  message.msg_iov = vect;
  message.msg_iovlen = 1;
  message.msg_control = NULL;
  message.msg_controllen = 0;
  message.msg_flags = 0;

  recvmsg(client_socket, &message, MSG_PEEK);

  /* check tag/pid, always activated for UDP ! */
  if ( req->ans.tag != req->req.tag
       ||  req->ans.pid != req->req.pid )
    {
      ERROR("Recv tag/pid %d/%d for %d/%d\n",
	    (int)req->ans.tag, (int)req->ans.pid,
	    (int)req->req.tag, (int)req->req.pid);
      assert(0);
    }

  /* Check the answer type */
  LOG(5, "Received type %d\n", (int)req->ans.type);
  if ( req->ans.type <= FIRST_ANS
       ||  req->ans.type >= LAST_ANS )
    {
      ERROR("Bad answer type %d received !\n",
	    (int)req->ans.type);
      req->ans.type = BAD_ANS;
      /* receive the message to empty the queue */
      recvmsg(client_socket, &message, 0);
      return -1;
    }

  /* Add a data buffer if necessary */
  if ( req->ans.buffer_len > 0 )
    {
      char *remaining_buf;
      size_t remaining_len;
      size_t to_recv;

      remaining_buf = req->out_buf;
      remaining_len = req->ans.buffer_len;

      LOG(3, "%ld data to recv...\n",
	  (long)remaining_len);

      /* recv the answer and the begining of data */
      to_recv = remaining_len > MAX_MESSAGE_LEN - sizeof(struct answer)
	? MAX_MESSAGE_LEN - sizeof(struct answer)
	: remaining_len;

      message.msg_iovlen = 2;
      vect[1].iov_base = remaining_buf;
      vect[1].iov_len = to_recv;
      LOG(3, "Recv the answer and begining of data\n");
      to_recv = recvmsg(client_socket, &message, 0) - sizeof(struct answer);
      
      while ( remaining_len > to_recv ) {
	/* update buffer parameters */      
	remaining_buf += to_recv;
	remaining_len -= to_recv;

	/* recv next part of data */
	to_recv = remaining_len > MAX_MESSAGE_LEN
	  ? MAX_MESSAGE_LEN
	  : remaining_len;
	to_recv = read(client_socket, remaining_buf, to_recv);
	LOG(3, "Recv %ld next data\n",
	    (long)to_recv);
      };
    }
  else
    {
      LOG(3, "Just receive the answer\n");
      /* just receive the whole message */    
      recvmsg(client_socket, &message, 0);
    }

  return 0;
}


int orfa_UDP_network_call(struct client_request *req)
{
  int res = request_send(req);
  return res < 0 ? res : answer_recv(req);
}
