/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <sys/mman.h>
#include <asm/atomic.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "orfa_real_calls.h"
#include "orfa_calls.h"
#include "orfa_portable.h"
#include "orfa_convert64.h"
#include "orfa_network.h"
#include "orfa_path.h"
#include "orfa_misc.h"
#include "orfa_debug.h"


/*********
 * Magic *
 *********/

#define DIR_MAGIC 0x14071789


/*********
 * Locks *
 *********/

#ifdef MULTITHREAD_SAFE_ENABLED

/* DIR_LIST_LOCK is used when modifying DIR list.
 * These operations may be long. We then use a mutex.
 * Usefull when multiple threads open and close dirs.
 */

/* DIR_LIST_SPIN_LOCK is a fast version of the previous one.
 * May only be used to ensure that a DIR won't be
 * closed by another thread while we are accessing it.
 */

static spinlock_t dir_list_spinlock = SPIN_LOCK_INIT;
#define DIR_LIST_SPIN_LOCK() spin_lock(&dir_list_spinlock)
#define DIR_LIST_SPIN_UNLOCK()  spin_unlock(&dir_list_spinlock)

#include <pthread.h>
static pthread_mutex_t dir_list_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
#define DIR_LIST_LOCK() pthread_mutex_lock(&dir_list_mutex)
#define DIR_LIST_UNLOCK() pthread_mutex_unlock(&dir_list_mutex)

/* DIR_LOCK used to protect the dir local data (cache, offset, ...).
 * This have to be a mutex because it has to be locked
 * for a long time during a network call. */

#define DIR_LOCK(dir) pthread_mutex_lock(&(dir)->dir_mutex)
#define DIR_UNLOCK(dir) pthread_mutex_unlock(&(dir)->dir_mutex)

#else

#define DIR_LIST_SPIN_LOCK() /* nothing */
#define DIR_LIST_SPIN_UNLOCK() /* nothing */

#define DIR_LIST_LOCK() /* nothing */
#define DIR_LIST_UNLOCK() /* nothing */

#define DIR_LOCK(dir) /* nothing */
#define DIR_UNLOCK(dir) /* nothing */

#endif


/***********
 * TYPEDEF *
 ***********/

/* now we add a dir_t and a real_dir_t
   dir_t will be used to replace DIR type
   used by classical opendir/readdir/... */

struct orfa_real_dir {
  /* This fields are fixed at open and will
   * also not require multithread protection. */
  server_t server;
  orfa_dir_id_t dir_id;
  /* This file descriptor is -1 when this structure
   * is not a mmapped file */
  int fd;
  /* This field may change when fork.
   * We use atomic modifications to ensure coherence. */
  int use_count;
};


struct orfa_dir {
  unsigned int magic;

  struct orfa_real_dir *real;

  off64_t current_seek;

#ifdef DIR_CACHE_ENABLED
# ifdef PORTABLE_ENABLED
  struct dirent64 *tmp_d64;
# endif /* PORTABLE_ENABLED */
  struct dirent *tmp_d32;
#else /* ~DIR_CACHE_ENABLED */
  struct dirent64 *tmp_d64;
  struct dirent *tmp_d32;
#endif /* ~DIR_CACHE_ENABLED */

#ifdef DIR_CACHE_ENABLED
  struct portable_dirent *cached_dirents;
  size_t cache_size;
  off64_t seek_begin_in_cache, seek_end_in_cache;
  int current_dirent_in_cache;
  int eof;
#endif

#ifdef MULTITHREAD_SAFE_ENABLED
  pthread_mutex_t dir_mutex;
#endif

  struct orfa_dir *next, *prev;
};



/************
 * Orfa_Dir *
 ************/

static orfa_dir_t *first_dir = NULL;


int is_orfa_dir(orfa_dir_t *dir)
{
  orfa_dir_t *tmp;
  int res = 0;

  DIR_LIST_SPIN_LOCK();

  tmp = first_dir;
  while ( tmp != NULL ) {
    if ( tmp == dir ) {
      res = 1;
      goto out;
    }
    tmp = tmp->next;
  }

 out:
  DIR_LIST_SPIN_UNLOCK();
  return res;
}



#ifdef DIR_CACHE_ENABLED
/*********
 * Cache *
 *********/

static INLINE void increase_cache_size(orfa_dir_t *dir)
{
#if ( DIRENTS_CACHE_INCREASING_FACTOR > 1 )
  dir->cache_size *= DIRENTS_CACHE_INCREASING_FACTOR;
  dir->cached_dirents
    = realloc(dir->cached_dirents, 
	      dir->cache_size * sizeof(struct portable_dirent) );
#endif
}


static INLINE void reset_cache(orfa_dir_t *dir)
{
  dir->cache_size = DEFAULT_DIRENTS_CACHE_SIZE;
  dir->cached_dirents
    = realloc(dir->cached_dirents,
	      dir->cache_size * sizeof(struct portable_dirent) );
  dir->seek_begin_in_cache = -1;
  dir->seek_end_in_cache = -1;
  dir->current_dirent_in_cache = -1;
  dir->eof = 0;
}
#endif /* DIR_CACHE_ENABLED */



/**********************
 * orfa_real_dir info *
 **********************/

int get_orfa_dir_server(orfa_dir_t *dir,
			server_t *server,
			orfa_dir_id_t *dir_id)
{
  DIR_LIST_SPIN_LOCK();
  
  DEBUG_ASSERT( dir->magic == DIR_MAGIC );

  *server = dir->real->server;
  *dir_id = dir->real->dir_id;

  DIR_LIST_SPIN_UNLOCK();

  return 0;
}



/************
 * Mmapping *
 ************/

static INLINE struct orfa_real_dir *mmap_real_dir(int fd)
{
  struct orfa_real_dir *f;
  f = (struct orfa_real_dir*) REAL_MMAP(0, sizeof(struct orfa_real_dir),
					PROT_READ | PROT_WRITE,
					MAP_SHARED, 
					fd, 0);
  /* no need to deregister interval after mmap */

  if ( !f )
    ERROR("Cannot mmap dir due to error %m\n");
  else
    LOG(5, "DIR fd %d mmapped at %p (%d bytes)\n",
	fd, f, sizeof(struct orfa_real_dir));

  return f;
}


/* This function is called before fork/exec to
 * mmap a dir. It is protected by DIR_LIST_MUTEX. */
static int mmap_dir(orfa_dir_t *dir)
{
  char local_file_name[PATH_MAX+1];
  int fd;
  struct orfa_real_dir *real_dir;
  int res;

  DEBUG_ASSERT(dir->real->fd < 0);

  /* this file name ensures two processes
   * or threads will not open the same file 
   */
  sprintf(local_file_name, ORFA_DIR_DESCR_FILE_PREFIX "%d", getpid());

  fd = REAL_OPEN(local_file_name, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
  if ( fd < 0 )
    return -1;

  /* unlink now */
  REAL_UNLINK(local_file_name);
  
  /* copy data to this file */
  res = REAL_WRITE(fd, dir->real, sizeof(struct orfa_real_dir));
  
  /* be sure of the size */
  if ( res != sizeof(struct orfa_real_dir) ) {
    REAL_CLOSE(fd);
    ERROR("No more space to allocate a dir (write returns %d, errno says %m)\n", res);
    errno = ENOMEM;
    return -1;
  }
  
  /* assures this FD will not be closed AFTER exec.
   * we will close it by our self */
  REAL_FCNTL(fd, F_SETFD, 0);
    
  /* mmap the orfa_real_dir and initialize it */
  real_dir = mmap_real_dir(fd);
  real_dir->fd = fd;
  
  /* update dir */
  free(dir->real);
  dir->real = real_dir;

  return 0;
}


static INLINE void munmap_real_dir(struct orfa_real_dir *dir)
{
  REAL_MUNMAP(dir, sizeof(struct orfa_real_dir));
  /* no need to deregister after THIS munmap because
   * we know that the previously mmapped area has not
   * been registered */
}


static INLINE struct orfa_real_dir *mmap_real_dir_to_close(int fd)
{
  return (struct orfa_real_dir*) REAL_MMAP(0, sizeof(struct orfa_real_dir),
				      PROT_READ | PROT_WRITE,
				      MAP_SHARED,
				      fd, 0);
  /* no need to deregister interval after mmap */
}


static INLINE void munmap_real_dir_to_close(struct orfa_real_dir *dir)
{
  REAL_MUNMAP(dir, sizeof(struct orfa_real_dir));
  /* no need to deregister after THIS munmap because
   * we know that the previously mmapped area has not
   * been registered */
}



/***********
 * Opendir *
 ***********/

orfa_dir_t *get_orfa_dir(orfa_dir_id_t dir_id, server_t server)
{
  struct orfa_real_dir *real_dir;
  struct orfa_dir *dir;

  /* get a real dir structure */
  real_dir = (struct orfa_real_dir*) malloc( sizeof(struct orfa_real_dir) );
  if ( !real_dir )
    return NULL;

  /* fill connection in the real dir */
  real_dir->dir_id = dir_id;
  real_dir->server = server;
  real_dir->fd = -1;
  real_dir->use_count = 1;

  /* get a dir structure and link it to real dir */
  dir = (orfa_dir_t*) malloc( sizeof(orfa_dir_t) );

  DIR_LIST_LOCK();

  if ( first_dir != NULL )
    first_dir->prev = dir;
  dir->next = first_dir;
  first_dir = dir;
  dir->prev = NULL;

  DIR_LIST_UNLOCK();
  
  dir->real = real_dir;

  /* fill dir with cache and seek */
#ifdef MULTITHREAD_SAFE_ENABLED
  {
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_FAST_NP);
    pthread_mutex_init(&dir->dir_mutex, &attr);
  }
#endif

  dir->current_seek = 0;

  /* allocate tmp dirents */
  dir->tmp_d32 = malloc( sizeof(struct dirent) );
#if (!defined DIR_CACHE_ENABLED  ||  defined PORTABLE_ENABLED)
  dir->tmp_d64 = malloc( sizeof(struct dirent64) );
#endif

#ifdef DIR_CACHE_ENABLED
  /* initialize the cache */
  dir->cached_dirents = NULL;
  reset_cache(dir);
#endif

  /* fix magic */
  dir->magic = DIR_MAGIC;

  return dir;
}



/************
 * Closedir *
 ************/

static void put_orfa_dir(orfa_dir_t *dir)
{
  int fd = dir->real->fd;

  /* orfa_closedir it. The dir cannot be
   * opened elsewhere because dup is impossible
   * and fork really duplicates the associated fd. */
  if ( atomic_dec_and_test((atomic_t*)&dir->real->use_count) )
    __orfa_closedir(dir->real->dir_id, dir->real->server);
      
#ifdef DIR_CACHE_ENABLED
  /* free current cache */
  free(dir->cached_dirents);
#endif

  free(dir->tmp_d32);
#if (!defined DIR_CACHE_ENABLED  ||  defined PORTABLE_ENABLED)
  free(dir->tmp_d64);
#endif

  /* munmap and cose if necessary */
  if ( fd >= 0 ) {
    munmap_real_dir(dir->real);
    REAL_CLOSE(fd);
  } else
    free(dir->real);

  /* unfix magic */
  dir->magic = 0;

  free(dir);
}


void put_all_orfa_dir()
{
  orfa_dir_t *dir;
  orfa_dir_t *next;

  DIR_LIST_LOCK();
  DIR_LIST_SPIN_LOCK();

  dir = first_dir;

  while ( dir != NULL ) {
    next = dir->next;
    put_orfa_dir(dir);
    dir = next;
  }
  first_dir = NULL;

  DIR_LIST_SPIN_UNLOCK();
  DIR_LIST_UNLOCK();
}


int orfa_closedir(orfa_dir_t *dir)
{
  
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return -1;
  }

  DIR_LIST_LOCK();
  DIR_LIST_SPIN_LOCK();

  /* release the dir structure */
  if ( dir->prev != NULL )
    dir->prev->next = dir->next;
  else
    first_dir = dir->next;

  if ( dir->next != NULL )
    dir->next->prev = dir->prev;

  put_orfa_dir(dir);

  DIR_LIST_SPIN_UNLOCK();
  DIR_LIST_UNLOCK();

  return 0;
}



/***********
 * Readdir *
 ***********/

#ifdef DIR_CACHE_ENABLED
static struct portable_dirent *orfa_portable_readdir64(orfa_dir_t *dir)
{
  struct portable_dirent *tmp_dirent;

  if ( dir->current_seek >= dir->seek_end_in_cache
       && dir->eof )
    {
      LOG(3, "EOF\n");
      return NULL;
    }

  if ( dir->current_seek < dir->seek_begin_in_cache
       || dir->current_seek >= dir->seek_end_in_cache )
    {
      int count;

      LOG(3, "Real ORFA readdir\n");

      count = __orfa_portable_readdir(dir->real->dir_id, dir->real->server,
				      dir->cached_dirents,
				      dir->cache_size,
				      dir->current_seek);
      
      if ( count == 0 )
	{
	  reset_cache(dir);
	  dir->eof = 1;
	  
	  LOG(3, "Nothing gotten\n");
	  return NULL;
	}
      else
	{
	  dir->seek_begin_in_cache = dir->current_seek;
	  dir->seek_end_in_cache = dir->cached_dirents[count-1].d_off;
	  dir->current_dirent_in_cache = 0;
      
	  if ( count != dir->cache_size )
	    dir->eof = 1;
	  else
	    {
	      dir->eof = 0;
	      increase_cache_size(dir);
	    }
	  
	  LOG(3, "Recv count %ld\n", (long)count);
	}
    }
  
  tmp_dirent = &dir->cached_dirents[dir->current_dirent_in_cache];

  dir->current_dirent_in_cache++;
  dir->current_seek = tmp_dirent->d_off;

  return tmp_dirent;
}


struct dirent *orfa_readdir(orfa_dir_t *dir)
{
  struct portable_dirent *pd;
  
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return NULL;
  }

  pd = orfa_portable_readdir64(dir);

  if ( pd )
    {
#if PORTABLE_ENABLED
      convert_dirent_portable_to_32(dir->tmp_d32, pd);
#else
      convert_dirent_64_to_32(dir->tmp_d32, pd);
#endif
      return dir->tmp_d32;
    }
  else
    return NULL;
}


struct dirent64 *orfa_readdir64(orfa_dir_t *dir)
{
  struct portable_dirent *pd;
  
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return NULL;
  }

  pd = orfa_portable_readdir64(dir);

  if ( pd )
    {
#if PORTABLE_ENABLED
      convert_dirent_portable_to_64(dir->tmp_d64, pd);
      return dir->tmp_d64;
#else
      return pd;
#endif
    }
  else
    return NULL;
}


int orfa_readdir_r(orfa_dir_t *dir,
		   struct dirent *entry,
		   struct dirent **result)
{
  int ret = 0;
  struct portable_dirent *pd;

  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return -1;
  }

  DIR_LOCK(dir);

  pd = orfa_portable_readdir64(dir);

  if ( pd )
    {
#if PORTABLE_ENABLED
      convert_dirent_portable_to_32(entry, pd);
#else
      convert_dirent_64_to_32(entry, pd);
#endif
      *result = entry;
      ret = sizeof(struct dirent);
    }
  else
    *result = NULL;

  DIR_UNLOCK(dir);

  return ret;
}


int orfa_readdir64_r(orfa_dir_t *dir,
		     struct dirent64 *entry,
		     struct dirent64 **result)
{
  int ret = 0;
  struct portable_dirent *pd;
  
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return -1;
  }

  DIR_LOCK(dir);

  pd = orfa_portable_readdir64(dir);

  if ( pd )
    {
#if PORTABLE_ENABLED
      convert_dirent_portable_to_64(entry, pd);
#else
      memcpy(entry, pd,
	     sizeof(struct dirent64));
#endif
      *result = entry;
      ret = sizeof(struct dirent64);
    }
  else
    *result = NULL;

  DIR_UNLOCK(dir);

  return ret;
}


#else /* ~DIR_CACHE_ENABLED */


struct dirent *orfa_readdir(orfa_dir_t *dir)
{
  struct portable_dirent pd;
  
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return NULL;
  }

  if ( __orfa_portable_readdir(dir->real->dir_id, dir->real->server,
			       &pd, 1, dir->current_seek) <= 0 )
    {
      LOG(3, "Nothing gotten\n");
      return NULL;
    }

#ifdef PORTABLE_ENABLED
  convert_dirent_portable_to_32(dir->tmp_d32, &pd);
#else
  convert_dirent_64_to_32(dir->tmp_d32, &pd);
#endif    

  dir->current_seek = dir->tmp_d32->d_off;

  return dir->tmp_d32;
}


struct dirent64 *orfa_readdir64(orfa_dir_t *dir)
{
  int count;
  
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return NULL;
  }

#ifdef PORTABLE_ENABLED
  {
    struct portable_dirent pd;
    count = __orfa_portable_readdir(dir->real->dir_id, dir->real->server,
				    &pd, 1, dir->current_seek);
    if ( count > 0 ) 
      convert_dirent_portable_to_64(dir->tmp_d64, &pd);
  }
#else
  count = __orfa_portable_readdir(dir->real->dir_id, dir->real->server,
				  dir->tmp_d64, 1, dir->current_seek);
#endif
  if ( count <= 0) {
    LOG(3, "Nothing gotten\n");
    return NULL;
  }

  dir->current_seek = dir->tmp_d64->d_off;

  return dir->tmp_d64;
}


int orfa_readdir_r(orfa_dir_t *dir,
		   struct dirent *entry,
		   struct dirent **result)
{
  struct portable_dirent pd;
  
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return -1;
  }

  if ( __orfa_portable_readdir(dir->real->dir_id, dir->real->server,
			       &pd, 1, dir->current_seek) <= 0 )
    {
      LOG(3, "Nothing gotten\n");
      return 0;
    }

#ifdef PORTABLE_ENABLED
  convert_dirent_portable_to_32(entry, &pd);
#else
  convert_dirent_64_to_32(entry, &pd);
#endif    
  *result = entry;

  dir->current_seek = entry->d_off;

  return sizeof(struct dirent);  
}


int orfa_readdir64_r(orfa_dir_t *dir,
		     struct dirent64 *entry,
		     struct dirent64 **result)
{
  int count;
  
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return -1;
  }

#ifdef PORTABLE_ENABLED
  {
    struct portable_dirent pd;
    count = __orfa_portable_readdir(dir->real->dir_id, dir->real->server,
				    &pd, 1, dir->current_seek);
    if ( count > 0 ) 
      convert_dirent_portable_to_64(entry, &pd);
  }
#else
  count = __orfa_portable_readdir(dir->real->dir_id, dir->real->server,
				  entry, 1, dir->current_seek);
#endif
  if ( count <= 0 ) {
    LOG(3, "Nothing gotten\n");
    return 0;
  }

  dir->current_seek = entry->d_off;

  return sizeof(struct dirent64);  
}


#endif /* ~DIR_CACHE_ENABLED */



/***********
 * Seeking *
 ***********/

void orfa_seekdir(orfa_dir_t *dir, off_t offset)
{
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return;
  }

  dir->current_seek = offset;

#ifdef DIR_CACHE_ENABLED
  reset_cache(dir);

  /* find the current dirent in the cache,
     if we are in the cache */
  if ( offset >= dir->seek_begin_in_cache
       &&  offset < dir->seek_end_in_cache )
    {
      int i=0;
      struct portable_dirent *pd;

      /* look for the dirent after the seek */
      while ( 1 ) {
	pd = &dir->cached_dirents[i];
	if ( offset <= pd->d_off )
	  break;
      }
      
      /* clean seek */
      dir->current_seek = pd->d_off;
      dir->current_dirent_in_cache = i+1;
    }
#endif /* DIR_CACHE_ENABLED */
}


off_t orfa_telldir(orfa_dir_t *dir)
{
  
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return -1;
  }

  return dir->current_seek;
}


void orfa_rewinddir(orfa_dir_t *dir)
{
  if ( dir->magic != DIR_MAGIC ) {
    errno = EBADF;
    return;
  }

  if ( __orfa_rewinddir(dir->real->dir_id, dir->real->server) >= 0 )
    {
      dir->current_seek = 0;
#ifdef DIR_CACHE_ENABLED
      reset_cache(dir);
#endif /* DIR_CACHE_ENABLED */
      LOG(3, "succeded\n");
    }
}


/********
 * Misc *
 ********/

void orfa_dir_before_fork()
{
  orfa_dir_t *dir = first_dir;

  DIR_LIST_LOCK();

  /* mmap dir */
  while ( dir != NULL ) {
    if ( dir->real->fd < 0 )
      mmap_dir(dir);
    dir = dir->next;
  }

  DIR_LIST_UNLOCK();
}


void orfa_dir_after_fork_child()
{
  orfa_dir_t *dir = first_dir;

  DIR_LIST_LOCK();

  while ( dir != NULL ) {
    atomic_inc((atomic_t*)&dir->real->use_count);
    dir = dir->next;
  }

  DIR_LIST_UNLOCK();
}


char *orfa_dir_before_exec()
{
  char *env, *current;
  orfa_dir_t *dir;
  int i=0;

  DIR_LIST_LOCK();

  /* find number of ORFA dir, and mmap it */
  dir = first_dir;
  while ( dir != NULL ) {
    if ( dir->real->fd < 0 )
      mmap_dir(dir);
    i++;
    dir = dir->next;
  }

  /* allocate the env and fill it */
  env = (char*) malloc( (EXEC_DIR_ENV_LEN+1+i*(FD_STRING_MAX_LEN+1)+1)*sizeof(char) );
  sprintf(env, EXEC_DIR_ENV "=");
  /* seek at the begining of the list */
  current = env + EXEC_DIR_ENV_LEN+1;

  /* add each dir */
  dir = first_dir;
  while ( dir != NULL ) {
    int res = sprintf(current, "%d,", dir->real->fd);
    current += res;
    dir = dir->next;
  }

  LOG(4, "Added env %s\n", env);

  return env;
}


void orfa_dir_after_exec_failed()
{
  unsetenv(EXEC_DIR_ENV);

  DIR_LIST_UNLOCK();
}


void orfa_dir_after_exec()
{
  char *p = getenv(EXEC_DIR_ENV);
  int fd;

  LOG(4, "Got dir env %s\n", p);

  DIR_LIST_LOCK();

  while ( *p != '\0' )
    {
      struct orfa_real_dir *real_dir;

      fd = strtol(p, &p, 10); /* 10 is the decimal base */
      p++;

      /* temporary mmap it during ORFA closing */
      real_dir = mmap_real_dir_to_close(fd);

      /* ORFA closedir */
      if ( atomic_dec_and_test((atomic_t*)&real_dir->use_count) ) {
	__orfa_closedir(real_dir->dir_id, real_dir->server);
      }
	  
      /* munmap and real close it */
      munmap_real_dir_to_close(real_dir);
      REAL_CLOSE(fd);
      
      LOG(5, "Closeonexec on dir %d\n", fd);
    }

  unsetenv(EXEC_DIR_ENV);

  DIR_LIST_UNLOCK();  
}
