/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "orfa_real_calls.h"
#include "orfa_network.h"
#include "orfa_calls.h"
#include "orfa_misc.h"
#include "orfa_debug.h"


/*********
 * Locks *
 *********/

/* FILE_ATTR_LOCK is used by seek/size functions.
 * May be used for forked processes which are accessing
 * the same mapped descriptor.
 */

#define FILE_ATTR_LOCK(f) spin_lock(&f->lock);
#define FILE_ATTR_UNLOCK(f) spin_unlock(&f->lock)


/* FILE_LIST_LOCK is used when modifying FD list.
 * These operations may be long. We then use a mutex.
 * Usefull when multiple threads open and close
 * descriptors.
 */

/* FILE_LIST_SPIN_LOCK is a fast version of the previous one.
 * May only be used to ensure that a descriptor won't be
 * closed by another thread while we are accessing it.
 */

#ifdef MULTITHREAD_SAFE_ENABLED

static spinlock_t file_list_spinlock = SPIN_LOCK_INIT;
#define FILE_LIST_SPIN_LOCK() spin_lock(&file_list_spinlock)
#define FILE_LIST_SPIN_UNLOCK()  spin_unlock(&file_list_spinlock)

#include <pthread.h>
static pthread_mutex_t file_list_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
#define FILE_LIST_LOCK() pthread_mutex_lock(&file_list_mutex)
#define FILE_LIST_UNLOCK() pthread_mutex_unlock(&file_list_mutex)

#else /* ~MULTITHREAD_SAFE_ENABLED */

#define FILE_LIST_SPIN_LOCK() /* nothing */
#define FILE_LIST_SPIN_UNLOCK() /* nothing */

#define FILE_LIST_LOCK() /* nothing */
#define FILE_LIST_UNLOCK() /* nothing */

#endif /* ~MULTITHREAD_SAFE_ENABLED */


/***********
 * TYPEDEF *
 ***********/

/* for each open ORFA file, a local file is created
   and mmapped in the caller process. in case of dup
   or fork, only the pointer to this file is copied
   so that the file is shared between process */

struct orfa_fd {

  /* These attributes may change during execution. */
  off64_t size;
  off64_t seek;
  int flags;

  /* this spinlock provides mutual exclusion for access
   * to previous fields. */
  spinlock_t lock;

  /* this field described whether this structure
   * is a mapped file or not */
  int is_mmapped;

  /* This field may change when fork or dup.
   * We use atomic modifications to ensure coherence. */
  atomic_t use_count;

  /* Remaining fields are fixed at open and will
   * also not require multithread protection. */
  server_t server;
  orfa_fd_id_t fd_id;

  /* the length of the following filename */
  size_t file_name_len;
  /* the length MUST be at the end */
  char file_name[0];

};


#define FLAG_MASK (O_APPEND | O_NONBLOCK | O_NDELAY | FASYNC | O_DIRECT)


/*******************
 * PROCESS GLOBALS *
 *******************/

static struct {

  /* Now we define an array to associate system fd
   * with their corresponding struct orfa_fd in the client.
   * It's associated with NULL for classical fd and
   * to the mmapping of the corresponding struct orfa_fd file
   * for a ORFA fd */
  struct orfa_fd *orfa_fd;

  /* The close_on_exec flag is specific to a fd number,
   * and not to a real fd */
  int close_on_exec;

  /* Doubly-linked list of dupped orfa_fd */
  int next_dupped, prev_dupped;

} local_fd[FD_MAX];



/***************************
 * FD Attributes Functions *
 ***************************/
/* These functions handles seek and size.
 * They are protected by a spinlock. */

/* this function returns seek and size */
int get_orfa_fd_params(int fd, off64_t *new_seek, off64_t *new_size)
{
  struct orfa_fd *orfa_fd;

  FILE_LIST_SPIN_LOCK();

  /* We check it is a valid struct orfa_fd */
  orfa_fd = local_fd[fd].orfa_fd;
  DEBUG_ASSERT( orfa_fd != NULL );

  FILE_ATTR_LOCK(orfa_fd);

  *new_seek = orfa_fd->seek;
  *new_size = orfa_fd->size;

  FILE_ATTR_UNLOCK(orfa_fd);
  FILE_LIST_SPIN_UNLOCK();

  return 0;
}


/* this function updates seek and size */
int set_orfa_fd_params(int fd, off64_t new_seek, off64_t new_size)
{
  struct orfa_fd *orfa_fd;

  FILE_LIST_SPIN_LOCK();

  /* We check it is a valid struct orfa_fd */
  orfa_fd = local_fd[fd].orfa_fd;
  DEBUG_ASSERT( orfa_fd != NULL );

  FILE_ATTR_LOCK(orfa_fd);

  orfa_fd->seek = new_seek;
  orfa_fd->size = new_size;

  FILE_ATTR_UNLOCK(orfa_fd);
  FILE_LIST_SPIN_UNLOCK();

  return 0;
}



/******************
 * Duplicate a FD *
 ******************/

/* Duplicate a FD assuming the target is free */
void orfa_dup(int oldfd, int newfd)
{
  /* take the mutex to keep list consistency */
  /* spinlock not required becausing not removing descriptor */
  FILE_LIST_LOCK();

  /* share the fd structure */
  local_fd[newfd].orfa_fd = local_fd[oldfd].orfa_fd;
  atomic_inc(&local_fd[newfd].orfa_fd->use_count);

  /* update doubly-linked list */
  local_fd[newfd].next_dupped = local_fd[oldfd].next_dupped;
  local_fd[newfd].prev_dupped = oldfd;
  local_fd[local_fd[oldfd].next_dupped].prev_dupped = newfd;
  local_fd[oldfd].next_dupped = newfd;

  /* unset the close_on_exec flag */
  local_fd[newfd].close_on_exec = 0;

  FILE_LIST_UNLOCK();
}



/***********************
 * FD Fields Functions *
 ***********************/
/* These functions may not suffer from multithreaded
 * accesses. But they assumes the FD will not be
 * closed during this call */

/* this function gives back informations concerning
   a fd assuming this is a struct orfa_fd */
int get_orfa_fd_server(int fd, server_t *server, orfa_fd_id_t *fd_id)
{
  struct orfa_fd *orfa_fd;

  FILE_LIST_SPIN_LOCK();

  /* We check it is a valid struct orfa_fd */
  orfa_fd = local_fd[fd].orfa_fd;
  DEBUG_ASSERT( orfa_fd != NULL );
  
  *server = orfa_fd->server;
  *fd_id = orfa_fd->fd_id;

  FILE_LIST_SPIN_UNLOCK();

  return 0;
}


/* this function gets the file name of a orfa_fd */
char *get_orfa_fd_file_name(int fd)
{
  struct orfa_fd *orfa_fd;
  char *name;

  FILE_LIST_SPIN_LOCK();

  /* We check it is a valid struct orfa_fd */
  orfa_fd = local_fd[fd].orfa_fd;
  DEBUG_ASSERT( orfa_fd != NULL );

  name = orfa_fd->file_name;

  FILE_LIST_SPIN_UNLOCK();

  return name;
}


int get_orfa_fd_flags(int fd)
{
  struct orfa_fd *orfa_fd;
  int flags;

  FILE_LIST_SPIN_LOCK();

  /* We check it is a valid struct orfa_fd */
  orfa_fd = local_fd[fd].orfa_fd;
  DEBUG_ASSERT( orfa_fd != NULL );

  flags = orfa_fd->flags;

  FILE_LIST_SPIN_UNLOCK();

  return flags;
}


/* this function sets the close_on_exec flag */
static void set_closeonexec_fd_flag(int fd, int flag)
{
  struct orfa_fd *orfa_fd;

  FILE_LIST_SPIN_LOCK();

  /* We check it is a valid struct orfa_fd */
  orfa_fd = local_fd[fd].orfa_fd;
  DEBUG_ASSERT( orfa_fd != NULL );

  if ( flag  &&  !local_fd[fd].close_on_exec )
    inc_orfa_server_use_count(orfa_fd->server);
  else if ( !flag  &&  local_fd[fd].close_on_exec )
    dec_orfa_server_use_count(orfa_fd->server);

  local_fd[fd].close_on_exec = flag;

  LOG(4, "Set closeonexec flag of fd %d at %d\n", fd, flag);

  FILE_LIST_SPIN_UNLOCK();
}


/* this function gets the close_on_exec flag */
static int get_closeonexec_fd_flag(int fd)
{
  int flag;

  FILE_LIST_SPIN_LOCK();

  /* We check it is a valid struct orfa_fd */
  DEBUG_ASSERT( local_fd[fd].orfa_fd != NULL );

  flag = local_fd[fd].close_on_exec;

  FILE_LIST_SPIN_UNLOCK();

  return flag;
}


int orfa_fd_fcntl(int fd, int cmd, long arg)
{
  int res = 0;

  switch ( cmd ) {
  case F_SETFD: {
    set_closeonexec_fd_flag(fd, arg);
    break;
  }
  case F_GETFD: {
    get_closeonexec_fd_flag(fd);
    break;
  }
  case F_DUPFD: {
    int fd2;
    fd2 = REAL_FCNTL(fd, cmd, arg);
    if ( fd2 >= 0 )
      orfa_dup(fd, fd2);
    res = fd2;
    break;
  }
  case F_SETFL: {
    struct orfa_fd *orfa_fd;
    FILE_LIST_SPIN_LOCK();
    orfa_fd = local_fd[fd].orfa_fd;
    orfa_fd->flags = (orfa_fd->flags & (~FLAG_MASK)) | ((int)arg & FLAG_MASK);
    FILE_LIST_SPIN_UNLOCK();
    break;
  }
  case F_GETFL: {
    struct orfa_fd *orfa_fd;
    FILE_LIST_SPIN_LOCK();
    orfa_fd = local_fd[fd].orfa_fd;
    res = orfa_fd->flags;
    FILE_LIST_SPIN_UNLOCK();
    break;
  }
  default: {
    res = REAL_FCNTL(fd, cmd, arg);
    break;
  }
  }

  return res;
}



/**********************
 * Mmapping functions *
 **********************/

/* This function is called when we need to map
 * a file descr (before fork/exec).
 * We assume the caller took FILE_LIST_LOCK. */
static INLINE int mmap_orfa_fd(int fd)
{
  struct orfa_fd *old_orfa_fd = local_fd[fd].orfa_fd;
  struct orfa_fd *new_orfa_fd;
  char local_file_name[PATH_MAX+1];
  int new_fd;
  int len;
  int dupped_fd=0;

  DEBUG_ASSERT( !old_orfa_fd->is_mmapped );

  /* this file name ensures two processes
   * or thread will not open the same file 
   */
  sprintf(local_file_name, ORFA_FD_FILE_PREFIX "%d", getpid());

  /* open the file */
  new_fd = REAL_OPEN(local_file_name,
		     O_CREAT | O_EXCL | O_TRUNC | O_RDWR,
		     S_IRWXU);
  if ( new_fd < 0 ) {
    ERROR("Local open of temp file failed on error %m\n");
    return -1;
  }

  /* unlink now */
  REAL_UNLINK(local_file_name);

  /* assures this FD will not be closed AFTER exec.
   * we will close it by our self if necessary */
  REAL_FCNTL(new_fd, F_SETFD, 0);
    
  /* compute bytes to copy */
  len = sizeof(struct orfa_fd) + local_fd[fd].orfa_fd->file_name_len + 1;

  /* put data in file */
  if ( REAL_WRITE(new_fd, old_orfa_fd, len) < len ) {
    ERROR("No more space to allocate a orfa_fd (errno %m)\n");
    errno = ENOMEM;
    return -1;
  }

  /* update local fd data */
  REAL_DUP2(new_fd, fd);
  REAL_CLOSE(new_fd);

  /* free old file descr */
  free(old_orfa_fd);

  /* mmap file */
  new_orfa_fd = (struct orfa_fd*) REAL_MMAP(0, len,
					    PROT_READ | PROT_WRITE,
					    MAP_SHARED, 
					    fd, 0);
  /* no need to deregister interval after mmap */

  if ( !new_orfa_fd ) {
    ERROR("Cannot mmap fd %d due to error %m\n", new_fd);
    return -1;
  }

  new_orfa_fd->is_mmapped = 1;
  local_fd[fd].orfa_fd = new_orfa_fd;

  LOG(5, "fd %d at %p now mmapped at %p (%d bytes)\n",
      fd, old_orfa_fd, new_orfa_fd, len);

  /* update dupped fd */
  dupped_fd = local_fd[fd].next_dupped;
  while ( dupped_fd != fd ) {

    /* update */
    local_fd[dupped_fd].orfa_fd = new_orfa_fd;
    REAL_DUP2(new_fd, dupped_fd);
    LOG(6, "Updated dupped fd %d\n", dupped_fd);

    dupped_fd = local_fd[dupped_fd].next_dupped;
  }

  return 0;
}


/* This function is used when reacquiring a FD after exec */
static INLINE struct orfa_fd *remmap_orfa_fd(int fd, int len)
{
  struct orfa_fd *f = (struct orfa_fd*) REAL_MMAP(0, len,
						  PROT_READ | PROT_WRITE,
						  MAP_SHARED, 
						  fd, 0);
  /* no need to deregister interval after mmap */

  if ( !f )
    ERROR("Cannot re-mmap fd due to error %m\n");
  else
    LOG(5, "old fd %d re-mmapped at %p (%d bytes)\n",
	fd, f, len);

  return f;
}


/* This function closes a FD. */
static INLINE void munmap_orfa_fd(struct orfa_fd *f)
{
  REAL_MUNMAP(f, sizeof(struct orfa_fd) + f->file_name_len);
  /* no need to deregister interval after THIS munmap
   * because we know the previously mmapped area has
   * not been registered. */
}


/* This function temporarily mmap a CLOEXEC FD
 * after an exec to close it in the server before
 * real closing in the client. */
static INLINE struct orfa_fd *remmap_orfa_fd_to_close(int fd)
{
  return (struct orfa_fd*) REAL_MMAP(0, sizeof(struct orfa_fd),
				     PROT_READ | PROT_WRITE,
				     MAP_SHARED,
				     fd, 0);
  /* no need to deregister interval after mmap */
}


/* This function releases the previous temporar mmapping
 * of an CLOEXEC FD. */
static INLINE void munmap_orfa_fd_to_close(struct orfa_fd *f)
{
  REAL_MUNMAP(f, sizeof(struct orfa_fd));
  /* no need to deregister interval after THIS munmap
   * because we know the previously mmapped area has
   * not been registered. */
}


/*********************
 * FD List Functions *
 *********************/
/* These functions are based on the FD list.
 * They require a lock protection.
 */

/* this function returns 1 for a struct orfa_fd
   and 0 for a non ORFA file descriptor */
int is_orfa_fd(int fd)
{
  int res;

  FILE_LIST_SPIN_LOCK();

  res = ( ( fd >= 0 )
	  && ( fd < FD_MAX )
	  && ( local_fd[fd].orfa_fd != NULL ) );

  FILE_LIST_SPIN_UNLOCK();

  return res;
}


/********************
 * Get a File Descr *
 ********************/

/* This function creates the struct orfa_fd file and stores
   datas into it.
   If necessary, the mmaped file will be created */
int get_orfa_fd(server_t server,
		orfa_fd_id_t fd_id,
		struct portable_stat *stats,
		const char *file_name,
		int flags)
{
  int fd;
  struct orfa_fd *orfa_fd;
  size_t file_name_len = strlen(file_name);

  /* take the mutex to keep list consistency */
  /* spinlock not required becausing not removing descriptor */
  FILE_LIST_LOCK();

  /* get a empty fd */
  fd = REAL_DUP(FAKE_FD_TO_DUP);

  /* initialize a new file descr */
  orfa_fd = malloc(sizeof(struct orfa_fd) + file_name_len + 1);
  orfa_fd->is_mmapped = 0;

  /* update fd table */
  local_fd[fd].orfa_fd = orfa_fd;
  local_fd[fd].close_on_exec = 0;
  local_fd[fd].next_dupped = local_fd[fd].prev_dupped = fd;

  /* file name */
  orfa_fd->file_name_len = file_name_len;
  strcpy(orfa_fd->file_name, file_name);
  
  /* server connection */
  orfa_fd->server = server;
  orfa_fd->fd_id = fd_id;
  inc_orfa_server_use_count(server);
  
  /* attributes */
  orfa_fd->size = stats->st_size;
  orfa_fd->seek = 0;
  spin_lock_init(&orfa_fd->lock);
  
  /* keep open flags */
  orfa_fd->flags = flags;
  
  atomic_set(&orfa_fd->use_count, 1);

  FILE_LIST_UNLOCK();

  return fd;
}


/* Recreate a orfa_fd after an exec
 * we assume its call is protected by file_mutex */
static /*INLINE*/ void reget_orfa_fd(int fd)
/* this INLINE is a gcc bug ! */
{
  struct orfa_fd tmp_orfa_fd;
  struct orfa_fd *orfa_fd;
  int len;

  REAL_READ(fd, &tmp_orfa_fd, sizeof(struct orfa_fd));
  len = sizeof(struct orfa_fd) + tmp_orfa_fd.file_name_len + 1;

  orfa_fd = remmap_orfa_fd(fd, len);

  local_fd[fd].orfa_fd = orfa_fd;
  local_fd[fd].close_on_exec = 0;
  local_fd[fd].next_dupped = local_fd[fd].prev_dupped = fd;
}


/* Recreate a orfa_fd fd2 after an exec
 * dupping it from fd */
static INLINE void redup_orfa_fd(int fd, int fd2)
{
  local_fd[fd2].orfa_fd = local_fd[fd].orfa_fd;
  /* update doubly-linked list */
  local_fd[fd2].next_dupped = local_fd[fd].next_dupped;
  local_fd[fd2].prev_dupped = fd;
  local_fd[local_fd[fd].next_dupped].prev_dupped = fd2;
  local_fd[fd].next_dupped = fd2;
}


static void INLINE closexec_dupped_orfa_fd(int fd, int mmapped_fd) {
  struct orfa_fd *orfa_fd;
  int res;

  orfa_fd = local_fd[mmapped_fd].orfa_fd;

  res = atomic_dec_and_test(&orfa_fd->use_count);
  DEBUG_ASSERT( res != 0 );

  REAL_CLOSE(fd);

  LOG(5, "Have to close %d, still %d links open\n", fd, res);
}


static void INLINE closexec_orfa_fd(int fd) {
  struct orfa_fd tmp_orfa_fd;

  REAL_READ(fd, &tmp_orfa_fd, sizeof(struct orfa_fd));

  if ( atomic_read(&tmp_orfa_fd.use_count) > 1 ) {
    struct orfa_fd *orfa_fd;
    int res;

    orfa_fd = remmap_orfa_fd_to_close(fd);

    res = atomic_dec_and_test(&orfa_fd->use_count);
    DEBUG_ASSERT( res != 0 );

    munmap_orfa_fd_to_close(orfa_fd);
  }
  REAL_CLOSE(fd);
}



/***********
 * Closing *
 ***********/

/* This function releases a struct orfa_fd by
   munmapping, closing and removing it.
   Real removing will only occur when all
   open fd will be closed for this file. */
int put_orfa_fd(int fd)
{
  struct orfa_fd *orfa_fd;

  /* take the mutex to keep list consistency */
  FILE_LIST_LOCK();

  /* spinlock required during removing */
  FILE_LIST_SPIN_LOCK();
  orfa_fd = local_fd[fd].orfa_fd;
  local_fd[fd].orfa_fd = NULL;

  if ( !local_fd[fd].close_on_exec )
    dec_orfa_server_use_count(orfa_fd->server);

  if ( local_fd[fd].next_dupped == fd )
    {
      /* no more direct reference to this orfa_fd now.
       * it's safe to release the spinlock.
       * this is required to ensure that __orfa_close()
       * will not block in get_orfa_fd_server().
       */
      FILE_LIST_SPIN_UNLOCK();

      /* if we are the last process using this orfa_fd, close it */
      if ( atomic_dec_and_test(&orfa_fd->use_count) )
	__orfa_close(orfa_fd->fd_id, orfa_fd->server);

      /* unmap or free */
      if ( orfa_fd->is_mmapped )
	munmap_orfa_fd(orfa_fd);
      else
	free(orfa_fd);
      /* no need to update doubly-linked list
       * because it will reinitialized when reused
       */
    }
  else
    {
      /* update doubly-linked list */
      int prev = local_fd[fd].prev_dupped;
      int next = local_fd[fd].next_dupped;
      local_fd[prev].next_dupped = next;
      local_fd[next].prev_dupped = prev;

      /* no more direct reference to this orfa_fd now.
       * it's safe to release the spinlock so that
       */
      FILE_LIST_SPIN_UNLOCK();

      /* decrease use_count.
       * it can't be zero since this fd has been duplicated.
       */
      atomic_dec(&orfa_fd->use_count);
    }

  /* close the fd */
  REAL_CLOSE(fd);

  FILE_LIST_UNLOCK();

  return 0;
}


void put_all_orfa_fd()
{
  int fd;
  for(fd=0; fd<FD_MAX; fd++)
    if ( is_orfa_fd(fd) )
      put_orfa_fd(fd);
}


int orfa_close(int fd)
{
  return put_orfa_fd(fd);
}



/******************
 * Fork Functions *
 ******************/

void orfa_fd_before_fork()
{
  int fd;

  /* lock the list during fork */
  /* spinlock not required because not removing descriptor */
  FILE_LIST_LOCK();

  for(fd=0; fd<FD_MAX; fd++)
    if ( is_orfa_fd(fd) )
      if ( !local_fd[fd].orfa_fd->is_mmapped )
	mmap_orfa_fd(fd);

  FILE_LIST_UNLOCK();
}


void orfa_fd_after_fork_child()
{
  int fd;

  /* lock the list during fork */
  /* spinlock not required because not removing descriptor */
  FILE_LIST_LOCK();

  for(fd=0; fd<FD_MAX; fd++)
    if ( is_orfa_fd(fd) )
      atomic_inc(&(local_fd[fd].orfa_fd->use_count));

  FILE_LIST_UNLOCK();
}



/**********************
 * Exec Env Functions *
 **********************/

char *orfa_fd_before_exec()
{
  char *env, *current;
  int fd;
  int i=0;
  int fds[FD_MAX] = { 0 };
  int dupped_fd;

  /* lock the list during exec */
  /* spinlock not required because not removing descriptor */
  FILE_LIST_LOCK();

  /* find number of ORFA fd */
  for(fd=0; fd<FD_MAX; fd++)
    if ( is_orfa_fd(fd) ) {
      if ( !local_fd[fd].orfa_fd->is_mmapped )
	mmap_orfa_fd(fd);
      i++;
    }

  /* allocate the env and fill it */
  env = (char*) malloc( (EXEC_FD_ENV_LEN+1+i*(FD_STRING_MAX_LEN+2)+1)*sizeof(char) );
  sprintf(env, EXEC_FD_ENV "=");
  /* seek at the begining of the list */
  current = env + EXEC_FD_ENV_LEN+1;

  /* let's try each fd */
  for(fd=0; fd<FD_MAX; fd++) {
    int res;

    /* if already put, or not ORFA, skip it */
    if ( fds[fd]  || !is_orfa_fd(fd) )
      continue;
    
    /* add dupped first */
    dupped_fd = local_fd[fd].next_dupped;
    while ( 1 ) {
      fds[dupped_fd] = 1;
      if ( local_fd[dupped_fd].close_on_exec )
	res = sprintf(current, "%dc", dupped_fd);
      else
	res = sprintf(current, "%dk", dupped_fd);
      current += res;
      if ( dupped_fd == fd )
	break;
      *(current++) = ',';
      dupped_fd = local_fd[dupped_fd].next_dupped;
    }
    *(current++) = ';';
  }
  *(current++) = '\0';

  LOG(4, "Added env %s\n", env);

  return env;
}


/* This function is called after an exec to cancel it */
void orfa_fd_after_exec_failed()
{
  unsetenv(EXEC_FD_ENV);

  FILE_LIST_UNLOCK();
}


/* This function re-creates FD from env after exec */
void orfa_fd_after_exec()
{
  char *p = getenv(EXEC_FD_ENV);
  int fd;
  int to_dup = 0;
  int prev_fd = 0;
  int to_keep = 0;

  LOG(4, "Got fd env %s\n", p);

  /* lock the list during init after exec */
  /* spinlock not required because not removing descriptor */
  FILE_LIST_LOCK();

  while ( *p != '\0' ) {
    fd = strtol(p, &p, 10); /* 10 is the decimal base */
    
    to_keep = ( *(p++) == 'k' );

    if ( to_keep ) { /* keep this fd */
      if ( to_dup ) {
	redup_orfa_fd(prev_fd, fd);
	LOG(5, "Redup fd %d from fd %d\n", fd, prev_fd);
      } else {
	reget_orfa_fd(fd);
	LOG(5, "Reget fd %d\n", fd);
      }
    } else { /* close this fd */
      if ( to_dup ) {
	closexec_dupped_orfa_fd(fd, prev_fd);
      } else {
	closexec_orfa_fd(fd);
      }
      LOG(5, "Closeonexec on fd %d\n", fd);
    }

    /* dealing with dup to get next one or not */
    if ( *(p++) == ',' ) {
      /* if not keeping this one, do not changing to_dup/prev_fd */
      if ( to_keep ) {
	to_dup = 1;
	prev_fd = fd;
      }
    } else {
      to_dup = 0;
    }
  }

  unsetenv(EXEC_FD_ENV);

  FILE_LIST_UNLOCK();
}
