/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * This file is designed to be link with FUSE
 * Copyright (C) 2001  Miklos Szeredi (mszeredi@inf.bme.hu)
 */

#include <stdlib.h>
#include <string.h>

#include "fuse.h"

#include "orfa_network.h"
#include "orfa_portable.h"
#include "orfa_convert64.h"
#include "orfa_debug.h"


/* General informations about this process */
struct context context;

/* Server */
static int server = -1;


static int orfa_open(pid_t pid, uid_t uid, gid_t gid,
		     const char *path, int flags, uint64_t *fd)
{
  struct client_request req;
  orfa_fd_id_t res;
  struct stat64 unused;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = OPEN_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.req.data.open.oflags = convert_oflags_to_portable(flags);
  req.req.data.open.mode = 0;
  req.in_buf = (void*)path;
  req.out_buf = &unused;
  req.out_buf_len = sizeof(struct stat64);

  LOG(2, "Send open of %s\n", path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.ptr_value;

  if ( res == 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Received fd %llx\n", (long long)res);
      *fd = res;
      return 0;
    }
}


static int orfa_close(pid_t pid, uid_t uid, gid_t gid, uint64_t fd)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = CLOSE_REQ;
  req.req.buffer_len = -1;
  req.req.data.close.fd_id = fd;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send close for fd %llx\n", (long long)fd);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_read(pid_t pid, uid_t uid, gid_t gid,
		     uint64_t fd, char *buf, size_t size, off_t offset)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = PREAD_REQ;
  req.req.buffer_len = -1;
  req.req.data.pread.fd_id = fd;
  req.req.data.pread.nbyte = size;
  req.req.data.pread.offset = offset;
  req.in_buf = NULL;
  req.out_buf = buf;
  req.out_buf_len = size;

  LOG(2, "Send read for fd %llx\n", (long long)fd);

  orfa_network_call(server,  &req);

  res = req.ans.return_value.ssize_t_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Read %d\n", res);
      return res;
    }
}


static int orfa_write(pid_t pid, uid_t uid, gid_t gid,
		      uint64_t fd, const char *buf, size_t size, off_t offset)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = PWRITE_REQ;
  req.req.buffer_len = size;
  req.req.data.pwrite.fd_id = fd;
  req.req.data.pwrite.offset = offset;
  req.in_buf = (void*)buf;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send write for fd %llx\n", (long long)fd);

  orfa_network_call(server, &req);

  res = req.ans.return_value.ssize_t_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Written %d\n", res);
      return res;
    }
}


static int orfa_getattr(pid_t pid, uid_t uid, gid_t gid,
			const char *path, struct stat *stbuf)
{
  struct client_request req;
  int res;
  struct portable_stat pst;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = STAT_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.in_buf = (void*)path;
  req.out_buf = &pst;
  req.out_buf_len = sizeof(struct portable_stat);

  LOG(2, "Send stat for file %s\n", path);

  orfa_network_call(server, &req);

#if PORTABLE_ENABLED
  convert_stat_portable_to_32(stbuf, &pst);
#else
  convert_stat_64_to_32(stbuf, &pst);
#endif

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Recv stat\n");
      return 0;
    }
}


static int orfa_statfs(pid_t pid, uid_t uid, gid_t gid, struct statfs *st)
{
  struct client_request req;
  int res;
  char *path = "/";
  struct portable_statfs pst;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = STATFS_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.in_buf = (void*)path;
  req.out_buf = &pst;
  req.out_buf_len = sizeof(struct portable_statfs);

  LOG(2, "Send statfs for file %s\n", path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

#if PORTABLE_ENABLED
  convert_statfs_portable_to_32(st, &pst);
#else
  convert_statfs_64_to_32(st, &pst);
#endif

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Recv statfs\n");
      return 0;
    }
}


static int orfa_chmod(pid_t pid, uid_t uid, gid_t gid,
			const char *path, mode_t mode)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = CHMOD_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.req.data.chmod.mode = mode;
  req.in_buf = (void*)path;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send chmod for file %s\n", path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_chown(pid_t client_pid, uid_t client_uid, gid_t client_gid,
		      const char *path, uid_t uid, gid_t gid)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = client_pid;
#endif
  req.req.euid = client_uid;
  req.req.egid = client_gid;
  req.req.type = CHOWN_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.req.data.chown.owner = uid;
  req.req.data.chown.group = gid;
  req.in_buf = (void*)path;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send chown for file %s\n", path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_truncate(pid_t pid, uid_t uid, gid_t gid,
			 const char *path, off_t off)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = TRUNC_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.req.data.trunc.length = off;
  req.in_buf = (void*)path;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send trunc for file %s at %ld\n",
      path, off);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_mknod(pid_t pid, uid_t uid, gid_t gid,
		      const char *path, mode_t mode, dev_t dev)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = MKNOD_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.req.data.mknod.mode = mode;
  req.req.data.mknod.dev = dev;
  req.in_buf = (void*)path;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send mknod for file %s mode %d dev %d\n",
      path, mode, (int)dev);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;
 
  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_mkdir(pid_t pid, uid_t uid, gid_t gid,
		      const char *path, mode_t mode)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = MKDIR_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.req.data.mkdir.mode = mode;
  req.in_buf = (void*)path;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send mkdir for file %s mode %d\n",
      path, mode);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_link(pid_t pid, uid_t uid, gid_t gid,
		     const char *old_path, const char *new_path)
{
  struct client_request req;
  int res;
  char *in_buf;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = LINK_REQ;
  req.req.buffer_len = strlen(old_path)+strlen(new_path)+2;
  req.req.data.link.part_change = strlen(old_path)+1;
  
  in_buf = (char*) malloc(req.req.buffer_len);
  strcpy(in_buf,
	 old_path);
  strcpy(in_buf + req.req.data.link.part_change,
	 new_path);

  req.in_buf = in_buf;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send link from %s to %s\n",
      old_path, new_path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(in_buf);

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_rename(pid_t pid, uid_t uid, gid_t gid,
		       const char *old_path, const char *new_path)
{
  struct client_request req;
  int res;
  char *in_buf;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = RENAME_REQ;
  req.req.buffer_len = strlen(old_path)+strlen(new_path)+2;
  req.req.data.rename.part_change = strlen(old_path)+1;
  
  in_buf = (char*) malloc(req.req.buffer_len);
  strcpy(in_buf,
	 old_path);
  strcpy(in_buf + req.req.data.rename.part_change,
	 new_path);

  req.in_buf = in_buf;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send rename from %s to %s\n",
      old_path, new_path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(in_buf);

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_symlink(pid_t pid, uid_t uid, gid_t gid,
			const char *old_path, const char *new_path)
{
  struct client_request req;
  int res;
  char *in_buf;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = SYMLINK_REQ;
  req.req.buffer_len = strlen(old_path)+strlen(new_path)+2;
  req.req.data.symlink.part_change = strlen(old_path)+1;
  
  in_buf = (char*) malloc(req.req.buffer_len);
  strcpy(in_buf,
	 old_path);
  strcpy(in_buf + req.req.data.symlink.part_change,
	 new_path);

  req.in_buf = in_buf;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send symlink from %s to %s\n",
      old_path, new_path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(in_buf);

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_readlink(pid_t pid, uid_t uid, gid_t gid,
			 const char *old_path, char *buf, size_t bufsize)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = READLINK_REQ;
  req.req.buffer_len = strlen(old_path)+1;
  req.req.data.readlink.bufsiz = bufsize;
  req.in_buf = (void*)old_path;
  req.out_buf = buf;
  req.out_buf_len = bufsize;

  LOG(2, "Send readlink for file %s\n", old_path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_unlink(pid_t pid, uid_t uid, gid_t gid, const char *path)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = UNLINK_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.in_buf = (void*)path;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send unlink for file %s\n", path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_rmdir(pid_t pid, uid_t uid, gid_t gid, const char *path)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = RMDIR_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.in_buf = (void*)path;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send rmdir for file %s\n", path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_utime(pid_t pid, uid_t uid, gid_t gid,
		      const char *path, struct utimbuf *buf)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = UTIME_REQ;
  req.req.buffer_len = strlen(path)+1;
  if ( buf )
    {
      req.req.data.utime.time_defined = 1;
      req.req.data.utime.actime = buf->actime;
      req.req.data.utime.modtime = buf->modtime;
    }
  else
    {
      req.req.data.utime.time_defined = 0;
    }
  req.in_buf = (void*)path;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send utime for file %s\n", path);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_fsync(pid_t pid, uid_t uid, gid_t gid, uint64_t fd)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = FSYNC_REQ;
  req.req.buffer_len = -1;
  req.req.data.fsync.fd_id = fd;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send fsync for fd %llx\n", (long long)fd);
  
  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}



static int orfa_fdatasync(pid_t pid, uid_t uid, gid_t gid, uint64_t fd)
{
  struct client_request req;
  int res;

#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = FDATASYNC_REQ;
  req.req.buffer_len = -1;
  req.req.data.fsync.fd_id = fd;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send fdatasync for fd %llx\n", (long long)fd);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Succeded\n");
      return 0;
    }
}


static int orfa_getdir(pid_t pid, uid_t uid, gid_t gid,
			 const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
{
  orfa_dir_id_t dir;
  struct client_request req;
  int res = 0;
  size_t count = 0;
  size_t readdir_max = READDIR_MAX_DEFAULT;
  off64_t seek = 0;

  /* OPENDIR */
#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = OPENDIR_REQ;
  req.req.buffer_len = strlen(path)+1;
  req.in_buf = (void*)path;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send opendir of %s\n", path);

  orfa_network_call(server, &req);

  dir = req.ans.return_value.ptr_value;

  if ( dir == (orfa_dir_id_t)0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Received dir %llx\n", dir);
    }

  /* READDIR */
  count = 0;
  while ( 1 )
    {
      int gotten;
      int indix = 0;
      struct portable_dirent *dirents =
	orfa_aligned_malloc(readdir_max 
			    * sizeof(struct portable_dirent));

#ifdef PROTOCOL_DEBUG
      req.req.pid = pid;
#endif
      req.req.euid = uid;
      req.req.egid = gid;
      req.req.type = READDIR_REQ;
      req.req.buffer_len = -1;
      req.req.data.readdir.dir_id = dir;
      req.req.data.readdir.count = readdir_max;
      req.req.data.readdir.offset = seek;
      req.in_buf = NULL;
      req.out_buf = dirents;
      req.out_buf_len = readdir_max * sizeof(struct portable_dirent);

      LOG(2, "Send readdir for %d entries in dir %llx\n", readdir_max, dir);

      orfa_network_call(server, &req);

      gotten = req.ans.return_value.int_value;

      LOG(3, "Recv %d entries\n", gotten);

      /* nothing gotten */
      if ( gotten <= 0 )
	goto readdir_out;

      /* something gotten */
      while ( indix < gotten )
	{
	  struct portable_dirent *d = &dirents[indix];
	  filler(h, d->d_name, d->d_type);
	  indix++;
	}

      /* not everything gotten */
      if (gotten < readdir_max)
	goto readdir_out;

      seek = dirents[gotten-1].d_off;
      count += gotten;
      readdir_max *= READDIR_MAX_INCREASING_FACTOR;
      continue;

      /* go out of the loop */
    readdir_out:
      free(dirents);
      break;
    }

  /* CLOSEDIR */
#ifdef PROTOCOL_DEBUG
  req.req.pid = pid;
#endif
  req.req.euid = uid;
  req.req.egid = gid;
  req.req.type = CLOSEDIR_REQ;
  req.req.buffer_len = -1;
  req.req.data.closedir.dir_id = dir;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send closedir for dir %llx\n", dir);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      int error = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %d\n", error);
      return -error;
    }
  else
    {
      LOG(3, "Recv %d\n", res);
      return 0;
    }
}


static int orfa_init(char *arg)
{
  LOG(3, "Connecting to server %s...\n", arg);
  
  server = orfa_network_get_server(arg);
  if ( server < 0 )
    {
      ERROR("Unable to connect to server %s\n", arg);
      exit(-1);
    }

  return 0;
}


static void orfa_finish()
{
  orfa_network_finish();
}


struct fuse_operations orfa_fuse_operations = {
    open:	orfa_open,
    close:      orfa_close,
    read:	orfa_read,
    write:	orfa_write,
    getattr:	orfa_getattr,
    statfs:	orfa_statfs,
    chmod:	orfa_chmod,
    chown:	orfa_chown,
    truncate:	orfa_truncate,
    mknod:	orfa_mknod,
    mkdir:	orfa_mkdir,
    link:	orfa_link,
    rename:     orfa_rename,
    symlink:	orfa_symlink,
    readlink:	orfa_readlink,
    unlink:	orfa_unlink,
    rmdir:	orfa_rmdir,
    utime:	orfa_utime,
    fsync:      orfa_fsync,
    fdatasync:  orfa_fdatasync,
    getdir:     orfa_getdir,
    init:       NULL,
    init_with_arg: orfa_init,
    finish:     orfa_finish,
};


int main(int argc, char *argv[])
{
  fuse_main(argc, argv, &orfa_fuse_operations);
  return 0;
}
