/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __ORFA_CALLS_H__
#define __ORFA_CALLS_H__


#include <utime.h>

#include "orfa_portable.h"
#include "orfa_symbols.h"
#include "orfa_dir.h"


/* This files exports all orfa_func even some
 * of these are not defined in remote_calls.c
 * but in file_descr.c or dir.c.
 * This file also contains some real_orfa_func
 * which are called by file_descr.c or dir.c
 */


extern int orfa_open(const char *pathname,
		     int flags,
		     mode_t mode);
extern int orfa_close(int fd);

extern ssize_t orfa_read(int fd,
			 void *buf,
			 size_t count);
extern ssize_t orfa_pread64(int fd,
			    void *buf,
			    size_t count,
			    off64_t offset);
extern ssize_t orfa_write(int fd,
			  const void *buf,
			  size_t count); 
extern ssize_t orfa_pwrite64(int fd,
			     const void *buf,
			     size_t count,
			     off64_t offset);

extern int orfa_ftruncate64(int fd,
			    off64_t length);
extern int orfa_truncate64(const char *path,
			   off64_t length);


extern off64_t orfa_lseek64(int fd,
			    off64_t offset,
			    int whence);

extern int orfa_fstat64(int fd,
			struct portable_stat *buf);
extern int orfa_stat64(const char *file_name,
		       struct portable_stat *buf);
extern int orfa_lstat64(const char *file_name,
			struct portable_stat *buf);

extern int orfa_fstatfs64(int fd,
			  struct portable_statfs *buf);
extern int orfa_statfs64(const char *file_name,
			 struct portable_statfs *buf);

extern int orfa_mknod(const char *pathname,
		      mode_t mode,
		      dev_t dev);
extern int orfa_mkdir(const char *pathname,
		      mode_t mode);
extern int orfa_rmdir(const char *pathname);
extern int orfa_unlink(const char *pathname);
extern int orfa_rename(const char *oldpath,
		       const char *newpath);

extern int orfa_access(const char *pathname,
		       int mode);
extern int orfa_fchmod(int fildes,
		       mode_t mode);
extern int orfa_chmod(const char *path,
		      mode_t mode);

extern int orfa_fchown(int fd,
		       uid_t owner,
		       gid_t group);
extern int orfa_chown(const char *path,
		      uid_t owner,
		      gid_t group);
extern int orfa_lchown(const char *path,
		       uid_t owner,
		       gid_t group);

extern int orfa_flock(int fd,
		      int operation);

extern int orfa_link(const char *oldpath, 
		     const char *newpath);
extern int orfa_symlink(const char *oldpath, 
			const char *newpath);
extern int orfa_readlink(const char *path,
			 char *buf,
			 size_t bufsiz);

extern int orfa_fsync(int fd);
extern int orfa_fdatasync(int fd);

extern int orfa_utime(const char *filename,
		      const struct utimbuf *buf);

extern int orfa_fcntl(int fd, int cmd, long arg);

extern orfa_dir_t *orfa_opendir(const char *name);
extern int orfa_closedir(orfa_dir_t *dir);

extern struct dirent64 *orfa_readdir64(orfa_dir_t *dir);
extern struct dirent *orfa_readdir(orfa_dir_t *dir);
extern int orfa_readdir64_r(orfa_dir_t *dir,
			    struct dirent64 *entry,
			    struct dirent64 **result);
extern int orfa_readdir_r(orfa_dir_t *dir,
			  struct dirent *entry,
			  struct dirent **result);

extern void orfa_seekdir(orfa_dir_t *dir, off_t offset);
extern off_t orfa_telldir(orfa_dir_t *dir);
extern void orfa_rewinddir(orfa_dir_t *dir);

extern ssize_t orfa_sendfile64(int out_fd, int in_fd,
			       off64_t *offset, size_t count);

extern caddr_t orfa_mmap64(void *start,
			   size_t length,
			   int prot,
			   int flags,
			   int fd,
			   off64_t offset);
extern void *orfa_mremap(void *old_address,
			 size_t old_size,
			 size_t new_size,
			 unsigned long flags);
extern int orfa_munmap(void *start,
		       size_t length);

extern int orfa_execve(const char *path,
		       char *const argv[],
		       char *const envp[]);
#ifdef HAVE_FEXECVE
extern int orfa_fexecve(int fd,
			char *const argv[],
			char *const envp[]);
#endif

#ifdef HAVE_LIBACL
#include <sys/acl.h>
extern acl_t orfa_acl_get_fd(int fd);
extern int orfa_acl_set_fd(int fd, acl_t acl);
extern int orfa_acl_extended_fd(int fd);
extern acl_t orfa_acl_get_file(const char *path, acl_type_t type);
extern int orfa_acl_set_file(const char *path, acl_type_t type, acl_t acl);
extern int orfa_acl_extended_file(const char *path);
extern int orfa_acl_delete_def_file(const char *path);
#endif

/* This call is used by file_descr.c */
extern int __orfa_close(orfa_fd_id_t fd_id, server_t server);

/* This call is used by dir.c */
extern int __orfa_closedir(orfa_dir_id_t dir_id, server_t server);
extern int __orfa_portable_readdir(orfa_dir_id_t dir_id, server_t server,
				   void *buf, orfa_off_t count, orfa_off_t offset);
extern int __orfa_rewinddir(orfa_dir_id_t dir_id, server_t server);


#endif /* __REMOTE_CALLS_H__ */
