/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __ORFA_DEBUG_H__
#define __ORFA_DEBUG_H__

#include <stdio.h>


#define LOG_FD 2
#define LOG_FILE stderr



/********************************
 * MT-safe specific definitions *
 *********************************/

#if MULTITHREAD_SAFE_ENABLED

/* defining the LOG mutex */
#include <pthread.h>
extern pthread_mutex_t LOG_mutex;
#define INIT_LOGS() pthread_mutex_t LOG_mutex = PTHREAD_MUTEX_INITIALIZER

#else

#define INIT_LOGS() /* nothing */

#endif



/***********************************
 * Syscall include for FAST macros *
 * *********************************/

#include <sys/syscall.h>
#include <unistd.h>



/**********
 * Errors *
 **********/

/* ERROR macro : always written */

#ifdef MULTITHREAD_SAFE_ENABLED

#define ERROR(args...)                        \
do {                                          \
pthread_mutex_lock(&LOG_mutex);               \
fprintf(stderr, "ORFA: " args);               \
pthread_mutex_unlock(&LOG_mutex);             \
} while ( 0 )

#else

#define ERROR(args...) fprintf(LOG_FILE, "ORFA: " args)

#endif


/* FAST_ERROR macro : always written,
   using syscall to prevent stack using */

#ifdef MULTITHREAD_SAFE_ENABLED

#define FAST_ERROR(arg)                    \
do {                                       \
int debug_int;                             \
pthread_mutex_lock(&LOG_mutex);            \
syscall(SYS_write, 2, "ORFA: ", 11);       \
for(debug_int=0;                           \
     *(((char*)arg)+debug_int) != '\0';    \
      debug_int++);                        \
syscall(SYS_write, 2, arg, debug_int);     \
pthread_mutex_unlock(&LOG_mutex);          \
} while ( 0 )
#else

#define FAST_ERROR(arg)                        \
do {                                           \
int debug_int;                                 \
syscall(SYS_write, LOG_FD, "ORFA: ", 11);      \
for(debug_int=0;                               \
     *(((char*)arg)+debug_int) != '\0';        \
      debug_int++);                            \
syscall(SYS_write, LOG_FD, arg, debug_int);    \
} while ( 0 )

#endif




#if ( CLIENT_DEBUG_LEVEL <= 0 )
/************
 * No debug *
 ************/

#define DEBUG_ASSERT(x) do { /* nothing */ } while ( 0 )
#define LOG(level, args...) do { /* nothing */ } while ( 0 )
#define FAST_LOG(level, args...) do { /* nothing */ } while ( 0 )


#else /* CLIENT_DEBUG_LEVEL > 0 */
/*********
 * Debug *
 *********/

#include <assert.h>
#define DEBUG_ASSERT(x) assert(x)


/* Classical LOG macro */

#ifdef MULTITHREAD_SAFE_ENABLED

#define LOG(level, args...)                  \
do {                                         \
if ( level < CLIENT_DEBUG_LEVEL ) {          \
int debug_int;                               \
pthread_mutex_lock(&LOG_mutex);              \
for(debug_int=1;debug_int<level;debug_int++) \
fprintf(LOG_FILE, "..");                     \
fprintf(LOG_FILE, args);                     \
pthread_mutex_unlock(&LOG_mutex);            \
} } while ( 0 )

#else

#define LOG(level, args...)                  \
do {                                         \
if ( level < CLIENT_DEBUG_LEVEL ) {          \
int debug_int;                               \
for(debug_int=1;debug_int<level;debug_int++) \
fprintf(LOG_FILE, "..");                     \
fprintf(LOG_FILE, args);                     \
} } while ( 0 )

#endif


/* FAST_LOG macro : written using syscall to
   prevent stack using or looping */

#ifdef MULTITHREAD_SAFE_ENABLED

#define FAST_LOG(level, arg)                 \
do {                                         \
if ( level < CLIENT_DEBUG_LEVEL ) {          \
int debug_int;                               \
pthread_mutex_lock(&LOG_mutex);              \
for(debug_int=1;debug_int<level;debug_int++) \
  syscall(SYS_write, LOG_FD, "..", 2);       \
for(debug_int=0;                             \
     *(((char*)arg)+debug_int) != '\0';      \
      debug_int++);                          \
syscall(SYS_write, LOG_FD, arg, debug_int);  \
pthread_mutex_unlock(&LOG_mutex);            \
} } while ( 0 )

#else

#define FAST_LOG(level, arg)                 \
do {                                         \
if ( level < CLIENT_DEBUG_LEVEL ) {          \
int debug_int;                               \
for(debug_int=1;debug_int<level;debug_int++) \
  syscall(SYS_write, LOG_FD, "..", 2);       \
for(debug_int=0;                             \
     *(((char*)arg)+debug_int) != '\0';      \
      debug_int++);                          \
syscall(SYS_write, LOG_FD, arg, debug_int);  \
} } while ( 0 )

#endif



#endif /* CLIENT_DEBUG_LEVEL */



#endif /* __ORFA_DEBUG_H__ */
