/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <errno.h>

#include "orfa_protocol.h"
#include "orfa_network.h"
#include "orfa_client.h"
#include "orfa_server.h"
#include "orfa_debug.h"


/* this variable equals :
 * -1 if network has not been initialized yet ;
 * 0 if network has been initialized but locked by _exit, exec or fork
 * 1 if network has been initialized and is free.
 */
int orfa_network_ready = -1;


/********
 * Lock *
 ********/
 
#ifdef MULTITHREAD_SAFE_ENABLED

/* This lock is used to protect the server list.
 * This have to be a mutex because it has to be locked
 * for a long time during exec/fork. */

#include <pthread.h>
static pthread_mutex_t network_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
static pthread_cond_t network_cond = PTHREAD_COND_INITIALIZER;

#define NETWORK_LOCK() pthread_mutex_lock(&network_mutex);

#define NETWORK_UNLOCK()              \
do {                                  \
pthread_cond_signal(&network_cond);   \
pthread_mutex_unlock(&network_mutex); \
} while ( 0 )

#define NETWORK_TAKE()                            \
do {                                              \
pthread_mutex_lock(&network_mutex);               \
while ( orfa_network_ready != 1 )                 \
pthread_cond_wait(&network_cond, &network_mutex); \
} while ( 0 )

#define NETWORK_RELEASE() pthread_mutex_unlock(&network_mutex);

#else /* MULTITHREAD_SAFE_ENABLED */

#define NETWORK_LOCK() /* nothing */
#define NETWORK_UNLOCK() /* nothing */
#define NETWORK_TAKE() /* nothing */
#define NETWORK_RELEASE() /* nothing */

#endif /* MULTITHREAD_SAFE_ENABLED */



/*******************
 * Message tagging *
 *******************/

int current_tag = 0;

#ifdef MULTITHREAD_SAFE_ENABLED
/* assume the mutex is used around call to this macro */
#define GET_CURRENT_TAG() current_tag++
#else
#define GET_CURRENT_TAG() current_tag++
#endif



/*****************
 * General calls *
 *****************/

int orfa_network_init()
{
  int res = 0;

  NETWORK_LOCK();

  /* check if the network has already
   * been initialized */
  if ( orfa_network_ready != -1 )
    {
      NETWORK_UNLOCK();
      return 0;
    }

  /* real initialization */
  res = __orfa_network_init();

  if ( res < 0 )
    errno = E_NETWORK_INIT_FAILED;
  else
    orfa_network_ready = 1;

  NETWORK_UNLOCK();

  return res;
}


void orfa_network_finish()
{
  NETWORK_LOCK();

  /* check whether the network
   * has really been initialized yet */
  if ( orfa_network_ready < 0 )
    return;

  /* stop the network */
  orfa_network_ready = 0;
  __orfa_network_finish();
}


int orfa_network_connect(char *server_name, struct connection *connection)
{
  if ( orfa_network_ready < 0 )
    if ( orfa_network_init() < 0 )
      {
	ERROR("Client connect failed due to network init failed\n");
	return -1;
      }

  return __orfa_network_connect(server_name, connection);
}


int orfa_network_get_server(char *server_name)
{
  int server;
   
  server = find_orfa_server(server_name);
  
  if ( server >= 0 ) {
    /* there is a structure for this server */
    LOG(3, "Client already connected to server %s\n",
	server_name);
    free(server_name);
  } else {
    struct connection connection;

    /* try to connect */
    if ( orfa_network_connect(server_name, &connection) < 0 ) {
      free(server_name);
      errno = E_CONNECT_SERVER_FAILED;
      return -1;
    }

    /* the server_node exists, we can allocate corresponding structure */
    server = add_orfa_server(server_name);

    /* and fill it */
    set_orfa_server_connection(server, &connection);
#ifdef CONNECTION_ID_ENABLED
    set_orfa_server_connection_id(server,
				  orfa_network_connection_call(server, 0,
							       CONNECTION_OP_NEW));
#endif  
  }

  return server;
}


/* This function launches the real call to server
 * without checking anything
 */
static INLINE int orfa_network_real_call(int server, struct client_request *req)
{
  int res;

#if (defined MAKE_WRAPPER) || (defined MAKE_LIB)
  req->req.euid = context.euid;
  req->req.egid = context.egid;
#ifdef PROTOCOL_DEBUG
  req->req.pid = context.pid;
#endif /* PROTOCOL_DEBUG */
#endif /* (defined MAKE_WRAPPER) || (defined MAKE_LIB) */
#ifdef PROTOCOL_DEBUG
  req->req.tag = GET_CURRENT_TAG();
#endif /* PROTOCOL_DEBUG */

  req->connection = get_orfa_server_connection(server);

  /* wait for the network to be ready */
  NETWORK_TAKE();
  res = __orfa_network_call(req);
  NETWORK_RELEASE();

  return res;
}


/* This function is used for usual server calls.
 * It fills the request and checks the answer carefully.
 */
int orfa_network_call(int server, struct client_request *req)
{
  int res;

  if ( orfa_network_ready < 0 )
    if ( orfa_network_init() < 0 )
      {
	ERROR("Client network call failed due to network init failed\n");
	return -1;
      }

#ifdef CONNECTION_ID_ENABLED
  req->req.connection_id = get_orfa_server_connection_id(server);
#endif

  res = orfa_network_real_call(server, req);

  if ( res < 0 )
    {
      req->ans.error = E_REMOTE_CALL_FAILED;

      if ( req->req.type == OPENDIR_REQ )
	req->ans.return_value.ptr_value = 0;
      else if ( req->req.type == PREAD_REQ  ||  req->req.type == PWRITE_REQ )
	req->ans.return_value.ssize_t_value = -1;
      else
	req->ans.return_value.int_value = -1;
    }

  return res;
}


#ifdef CONNECTION_ID_ENABLED
/* This functions makes a special call
 * to ask the server for a connection_id*
 */
connection_id_t orfa_network_connection_call(int server,
					connection_id_t id,
					int op)
{
  struct client_request req;

  req.req.type = CONNECTION_REQ;
  req.req.buffer_len = -1;
  req.req.connection_id = id;
  req.req.data.connection.op = op;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = -1;

  LOG(4, "Calling server to get connection_id %llx\n", id);

  orfa_network_real_call(server, &req);

  if ( op == CONNECTION_OP_NEW )
    LOG(5, "Got connection_id %llx\n",
	(long long)req.ans.return_value.ptr_value);

  return (connection_id_t)req.ans.return_value.ptr_value;
}
#endif /* CONNECTION_ID_ENABLED */



#if (defined MAKE_WRAPPER) || (defined MAKE_LIB)
/*******************************
 * What to do during exec/fork *
 *******************************/

void orfa_network_before_fork()
{
  NETWORK_LOCK();

  if ( orfa_network_ready < 0 )
    return;

  /* stop the network */
  orfa_network_ready = 0;
  __orfa_network_before_fork();
}


void orfa_network_after_fork_parent()
{
  if ( orfa_network_ready < 0 )
    return;

  /* restart the network */
  __orfa_network_after_fork_parent();
  orfa_network_ready = 1;

  NETWORK_UNLOCK();
}


void orfa_network_after_fork_child()
{
  if ( orfa_network_ready < 0 )
    return;

  /* restart the network */
  __orfa_network_after_fork_child();
  orfa_server_after_fork_child();
  orfa_network_ready = 1;

  NETWORK_UNLOCK();
}


void orfa_network_before_exec()
{
  NETWORK_LOCK();

  if ( orfa_network_ready < 0 )
    return;

  /* stop the network */
  orfa_network_ready = 0;
  __orfa_network_before_exec();
}


void orfa_network_after_exec_failed()
{
  if ( orfa_network_ready < 0 )
    return;


  /* restart the network */
  __orfa_network_after_exec_failed();
  orfa_network_ready = 1;

  NETWORK_UNLOCK();
}

#endif /* (defined MAKE_WRAPPER) || (defined MAKE_LIB) */
