/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "orfa_calls.h"
#include "orfa_real_calls.h"
#include "orfa_fd.h"
#include "orfa_dir.h"
#include "orfa_path.h"
#include "orfa_network.h"
#include "orfa_portable.h"
#include "orfa_misc.h"
#include "orfa_debug.h"


/* This file contains all orfa calls, that is, their
 * network call core. When something special has to be
 * done before going on the network, the orfa_func is
 * defined elsewhere (as orfa_close is file_descr.c,
 * or some orfa_*dir in dir.c) and this files only
 * defines __orfa_func which only contains the
 * network core.
 */


int orfa_open(const char *pathname, int flags, mode_t mode)
{
  char *orfa_server_name;
  server_t server;
  char *path_on_server;

  struct portable_stat stats;
  int fd;
  orfa_fd_id_t fd_id;

  struct client_request req;
  
  orfa_parse_path(pathname, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = OPEN_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.req.data.open.oflags = convert_oflags_to_portable(flags);
  req.req.data.open.mode = mode & ~context.umask;
  req.in_buf = path_on_server;
  req.out_buf = &stats;
  req.out_buf_len = sizeof(struct portable_stat);

  LOG(2, "Send open of %s\n", path_on_server);

  orfa_network_call(server, &req);

  fd_id = req.ans.return_value.ptr_value;

  if ( fd_id == 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
      fd = -1;
    }
  else
    {
      LOG(3, "Received fd %llx size %lld\n",
 	  (long long)fd_id, (long long)stats.st_size);
      
      fd = get_orfa_fd(server, fd_id,
		       &stats, pathname,
		       flags);
      /* we assume fd is defined because file_descr number
	 is first defined by real open of shared file */

      LOG(4, "Storing directory name\n");
    }
  free(path_on_server);

  return fd;
}


int __orfa_close(orfa_fd_id_t fd_id, server_t server)
{
  struct client_request req;
  int res;

  req.req.type = CLOSE_REQ;
  req.req.buffer_len = -1;
  req.req.data.close.fd_id = fd_id;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send close for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv %d\n", res);
    }

  return res;
}


ssize_t orfa_read(int fd, void *buf, size_t count)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  off64_t seek, size;
  server_t server;
  ssize_t res;

  get_orfa_fd_server(fd, &server, &fd_id);
  get_orfa_fd_params(fd, &seek, &size);

  req.req.type = PREAD_REQ;
  req.req.buffer_len = -1;
  req.req.data.pread.fd_id = fd_id;
  req.req.data.pread.nbyte = count;
  req.req.data.pread.offset = seek;
  req.in_buf = NULL;
  req.out_buf = buf;
  req.out_buf_len = count;

  LOG(2, "Send read for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.ssize_t_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv %ld\n", (long)res);

      seek += res;
      set_orfa_fd_params(fd, seek, size);
    }

  return res;
}


ssize_t orfa_pread64(int fd, void *buf, size_t count, off64_t offset)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  server_t server;
  ssize_t res;

  get_orfa_fd_server(fd, &server, &fd_id);

  req.req.type = PREAD_REQ;
  req.req.buffer_len = -1;
  req.req.data.pread.fd_id = fd_id;
  req.req.data.pread.nbyte = count;
  req.req.data.pread.offset = offset;
  req.in_buf = NULL;
  req.out_buf = buf;
  req.out_buf_len = count;

  LOG(2, "Send pread64 for fd %llx\n", (long long)fd_id);

  orfa_network_call(server,&req);

  res = req.ans.return_value.ssize_t_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv res %ld\n", (long)res);
    }

  return res;
}


ssize_t orfa_write(int fd, const void *buf, size_t count)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  off64_t seek, size;
  server_t server;
  ssize_t res;

  get_orfa_fd_server(fd, &server, &fd_id);
  get_orfa_fd_params(fd, &seek, &size);

  req.req.type = PWRITE_REQ;
  req.req.buffer_len = count;
  req.req.data.pwrite.fd_id = fd_id;
  req.req.data.pwrite.offset = seek;
  req.in_buf = (void*)buf;
  req.out_buf = NULL;
  req.out_buf_len = 0;
  
  LOG(2, "Send write for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.ssize_t_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv %ld\n", (long)res);

      if ( is_append_orfa_fd(fd) )
	{
	  size += res;
	  seek = size;
	}
      else
	{
	  seek += res;
	  if ( seek > size )
	    size = seek;
	}
      set_orfa_fd_params(fd, seek, size);
    }

  return res;
}


ssize_t orfa_pwrite64(int fd, const void *buf, size_t count, off64_t offset)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  off64_t seek, size;
  server_t server;
  ssize_t res;

  get_orfa_fd_server(fd, &server, &fd_id);
  get_orfa_fd_params(fd, &seek, &size);

  req.req.type = PWRITE_REQ;
  req.req.buffer_len = count;
  req.req.data.pwrite.fd_id = fd_id;
  req.req.data.pwrite.offset = offset;
  req.in_buf = (void*)buf;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send pwrite64 for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.ssize_t_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv %ld\n", (long)res);

      if ( is_append_orfa_fd(fd) )
	{
	  size += res;	  
	}
      else
	{
	  if ( offset + res > size )
	    size = res + offset;
	}
      set_orfa_fd_params(fd, seek, size);
    }

  return res;
}


off64_t orfa_lseek64(int fd, off64_t offset, int whence)
{
  off64_t seek, size;
  get_orfa_fd_params(fd, &seek, &size);

  switch ( whence )
    {
    case SEEK_SET:
      {
	LOG(3, "Seek in %d at %lld from the begining\n",
	    fd, (long long)offset);
	seek = offset;
	break;
      }
    case SEEK_CUR:
      {
	LOG(3, "Seek in %d at %lld from current => to %lld\n",
	    fd, (long long)offset, (long long)(offset+seek));
	seek += offset;
	break;
      }
    case SEEK_END:
      {
	if ( !context.concurrent_mode )
	  {
	    /* If the non concurrent mode environment variable is set,
	       we assume that the file cannot be modified by another 
	       process, so we know its size. */
	    LOG(3, "Seekend starting at the end in NON-CONCURRENT mode\n");
	  }
	else
	  {
	    int res;
	    struct portable_stat stats;

	    /* If the non concurrent mode environment variable is NOT set,
	       we have to ask the server to update our status of this file
	       and then use the given size to seek */
	    LOG(3, "Seekend starting at the end in NON-CONCURRENT mode\n");

	    res = orfa_fstat64(fd, &stats);

	    if ( res < 0 )
	      {
		errno = EIO;
		return (off64_t)-1;
	      }

	    size = stats.st_size;
	  }

	seek = size + offset;
	break;
      }

    default:
      {
	errno = EINVAL;
	return (off64_t)-1;
      }
    }

  set_orfa_fd_params(fd, seek, size);

  return seek;
}


int orfa_ftruncate64(int fd, off64_t length)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  server_t server;
  int res;

  get_orfa_fd_server(fd, &server, &fd_id);

  req.req.type = FTRUNC_REQ;
  req.req.buffer_len = -1;
  req.req.data.ftrunc.fd_id = fd_id;
  req.req.data.ftrunc.length = length;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send ftrunc for fd %llx at %lld\n",
      (long long)fd_id, (long long)length);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv ftruncate\n");
    }

  return res;
}


int orfa_truncate64(const char *path, off64_t length)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(path, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = TRUNC_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.req.data.trunc.length = length;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send trunc for file %s at %lld\n",
      path_on_server, (long long)length);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv truncate\n");
    }

  return res;
}


int orfa_fstat64(int fd, struct portable_stat *pst)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  server_t server;
  int res;

  get_orfa_fd_server(fd, &server, &fd_id);

  req.req.type = FSTAT_REQ;
  req.req.buffer_len = -1;
  req.req.data.fstat.fd_id = fd_id;
  req.in_buf = NULL;
  req.out_buf = pst;
  req.out_buf_len = sizeof(struct portable_stat);

  LOG(2, "Send fstat for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv fstat with mode %d\n", (int)pst->st_mode);
    }

  return res;
}


int orfa_stat64(const char *file_name, struct portable_stat *pst)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(file_name, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = STAT_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.in_buf = path_on_server;
  req.out_buf = pst;
  req.out_buf_len = sizeof(struct portable_stat);

  LOG(2, "Send stat for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv stat with mode %d\n", (int)pst->st_mode);
    }

  return res;
}


int orfa_lstat64(const char *file_name, struct portable_stat *pst)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(file_name, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = LSTAT_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.in_buf = path_on_server;
  req.out_buf = pst;
  req.out_buf_len = sizeof(struct portable_stat);

  LOG(2, "Send lstat for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv stat with mode %d\n", (int)pst->st_mode);
    }

  return res;
}


int orfa_fstatfs64(int fd, struct portable_statfs *pst)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  server_t server;
  int res;

  get_orfa_fd_server(fd, &server, &fd_id);

  req.req.type = FSTATFS_REQ;
  req.req.buffer_len = -1;
  req.req.data.fstatfs.fd_id = fd_id;
  req.in_buf = NULL;
  req.out_buf = pst;
  req.out_buf_len = sizeof(struct portable_statfs);

  LOG(2, "Send fstatfs for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv fstatfs\n");
    }

  return res;
}


int orfa_statfs64(const char *file_name, struct portable_statfs *pst)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(file_name, &orfa_server_name, &path_on_server);
  
  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = STATFS_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.in_buf = path_on_server;
  req.out_buf = pst;
  req.out_buf_len = sizeof(struct portable_statfs);

  LOG(2, "Send statfs for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv statfs\n");
    }

  return res;
}


int orfa_mknod(const char *pathname, mode_t mode, dev_t dev)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(pathname, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = MKNOD_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.req.data.mknod.mode = mode;
  req.req.data.mknod.dev = dev;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send mknod for file %s mode %d dev %d\n",
      path_on_server, mode, (int)dev);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }
  
  return res;
}


int orfa_mkdir(const char *pathname, mode_t mode)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(pathname, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = MKDIR_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.req.data.mkdir.mode = mode;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send mkdir for file %s mode %d\n", path_on_server, mode);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_rmdir(const char *pathname)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(pathname, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = RMDIR_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send rmdir for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_unlink(const char *pathname)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(pathname, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = UNLINK_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send unlink for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_access(const char *pathname, int mode)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(pathname, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = ACCESS_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.req.data.access.mode = mode;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send access for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_fchmod(int fildes, mode_t mode)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  server_t server;
  int res;

  get_orfa_fd_server(fildes, &server, &fd_id);

  req.req.type = FCHMOD_REQ;
  req.req.buffer_len = -1;
  req.req.data.fchmod.fd_id = fd_id;
  req.req.data.fchmod.mode = mode;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send fchmod for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_chmod(const char *path, mode_t mode)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(path, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = CHMOD_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.req.data.chmod.mode = mode;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send chmod for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_fchown(int fd, uid_t owner, gid_t group)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  server_t server;
  int res;

  get_orfa_fd_server(fd, &server, &fd_id);

  req.req.type = FCHOWN_REQ;
  req.req.buffer_len = -1;
  req.req.data.fchown.fd_id = fd_id;
  req.req.data.fchown.owner = owner;
  req.req.data.fchown.group = group;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send fchown for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_chown(const char *path, uid_t owner, gid_t group)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(path, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = CHOWN_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.req.data.chown.owner = owner;
  req.req.data.chown.group = group;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send chown for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_lchown(const char *path, uid_t owner, gid_t group)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(path, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = LCHOWN_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.req.data.chown.owner = owner;
  req.req.data.chown.group = group;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send lchown for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_flock(int fd, int operation)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  server_t server;
  int res;

  get_orfa_fd_server(fd, &server, &fd_id);

  req.req.type = FLOCK_REQ;
  req.req.buffer_len = -1;
  req.req.data.flock.fd_id = fd_id;
  req.req.data.flock.operation = operation;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send flock for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}

int orfa_rename(const char *oldpath, const char *newpath)
{
  char *old_path_on_server, *new_path_on_server;
  char *old_orfa_server_name, *new_orfa_server_name;
  int oldserver, newserver;

  struct client_request req;
  int res;

  orfa_parse_path(oldpath,
		  &old_orfa_server_name,
		  &old_path_on_server);
  
  if ( (oldserver=orfa_network_get_server(old_orfa_server_name)) < 0 )
    {
      free(old_path_on_server);
      return -1;
    }
  
  orfa_parse_path(newpath,
		  &new_orfa_server_name,
		  &new_path_on_server);

  if ( (newserver=orfa_network_get_server(new_orfa_server_name)) < 0 )
    {
      free(new_path_on_server);
      return -1;
    }
  
  if ( newserver == oldserver )
    {
      char *in_buf;

      req.req.type = RENAME_REQ;
      req.req.buffer_len = strlen(old_path_on_server)+strlen(new_path_on_server)+2;
      req.req.data.rename.part_change = strlen(old_path_on_server)+1;
      
      /* in_buf need to be registered */
      in_buf = (char*) orfa_aligned_malloc(req.req.buffer_len);
      strcpy(in_buf,
	     old_path_on_server);
      strcpy(in_buf + req.req.data.rename.part_change,
	     new_path_on_server);
      req.in_buf = in_buf;
      req.out_buf = NULL;
      req.out_buf_len = 0;

      LOG(2, "Send rename from %s to %s\n",
	  old_path_on_server, new_path_on_server);

      orfa_network_call(newserver, &req);

      res = req.ans.return_value.int_value;

      free(old_path_on_server);
      free(new_path_on_server);
      free(in_buf);
    }
  else
    {
      errno = EXDEV;
      LOG(2, "Cross server rename impossible\n");
      return -1;
    }

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Rename succeded\n");
    }

  return res;
}


int orfa_link(const char *oldpath,
		const char *newpath)
{
  char *old_path_on_server, *new_path_on_server;
  char *old_orfa_server_name, *new_orfa_server_name;
  int oldserver, newserver;

  struct client_request req;
  int res;

  orfa_parse_path(oldpath,
		  &old_orfa_server_name,
		  &old_path_on_server);

  if ( (oldserver=orfa_network_get_server(old_orfa_server_name)) < 0 )
    {
      free(old_path_on_server);
      return -1;
    }
  
  orfa_parse_path(newpath,
		  &new_orfa_server_name,
		  &new_path_on_server);

  if ( (newserver=orfa_network_get_server(new_orfa_server_name)) < 0 )
    {
      free(new_path_on_server);
      return -1;
    }
  
  if ( newserver == oldserver )
    {
      char *in_buf;

      req.req.type = LINK_REQ;
      req.req.buffer_len = strlen(old_path_on_server)+strlen(new_path_on_server)+2;
      req.req.data.link.part_change = strlen(old_path_on_server)+1;
      
      in_buf = (char*) orfa_aligned_malloc(req.req.buffer_len);
      strcpy(in_buf,
	     old_path_on_server);
      strcpy(in_buf + req.req.data.link.part_change,
	     new_path_on_server);
      req.in_buf = in_buf;
      req.out_buf = NULL;
      req.out_buf_len = 0;

      LOG(2, "Send link from %s to %s\n",
	  old_path_on_server, new_path_on_server);

      orfa_network_call(newserver, &req);

      res = req.ans.return_value.int_value;

      free(old_path_on_server);
      free(new_path_on_server);
      free(in_buf);
    }
  else
    {
      LOG(2, "Cross server link\n");
      req.ans.error = EXDEV;
      res = -1;
    }

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Link succeded\n");
    }

  return res;
}


int orfa_symlink(const char *oldpath,
		   const char *newpath)
{
  char *new_path_on_server;
  char *new_orfa_server_name;
  int newserver;

  struct client_request req;
  int res;
  char *in_buf;

  orfa_parse_path(newpath,
		  &new_orfa_server_name,
		  &new_path_on_server);
  
  if ( (newserver=orfa_network_get_server(new_orfa_server_name)) < 0 )
    {
      free(new_path_on_server);
      return -1;
    }

  req.req.type = SYMLINK_REQ;
  req.req.buffer_len = strlen(oldpath)+strlen(new_path_on_server)+2;
  req.req.data.symlink.part_change = strlen(oldpath)+1;

  in_buf = (char*) orfa_aligned_malloc(req.req.buffer_len);
  strcpy(in_buf,
	 oldpath);
  strcpy(in_buf + req.req.data.symlink.part_change,
	 new_path_on_server);
  req.in_buf = in_buf;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send symlink from %s to %s\n",
      oldpath, new_path_on_server);

  orfa_network_call(newserver, &req);

  res = req.ans.return_value.int_value;

  free(new_path_on_server);
  free(in_buf);
  
  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Symlink succeded\n");
    }

  return res;
}


int orfa_readlink(const char *path,
		  char *buf, size_t bufsiz)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(path, &orfa_server_name, &path_on_server);
  
  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = READLINK_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.req.data.readlink.bufsiz = bufsiz;
  req.in_buf = path_on_server;
  req.out_buf = buf;
  req.out_buf_len = bufsiz;

  LOG(2, "Send readlink for file %s\n", path_on_server);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_fsync(int fd)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  server_t server;
  int res;

  get_orfa_fd_server(fd, &server, &fd_id);

  req.req.type = FSYNC_REQ;
  req.req.buffer_len = -1;
  req.req.data.fsync.fd_id = fd_id;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send fsync for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_fdatasync(int fd)
{
  struct client_request req;
  orfa_fd_id_t fd_id;
  server_t server;
  int res;

  get_orfa_fd_server(fd, &server, &fd_id);

  req.req.type = FDATASYNC_REQ;
  req.req.buffer_len = -1;
  req.req.data.fdatasync.fd_id = fd_id;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;
  
  LOG(2, "Send fdatasync for fd %llx\n", (long long)fd_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_utime(const char *filename,
		 const struct utimbuf *buf)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  struct client_request req;
  int res;

  orfa_parse_path(filename, &orfa_server_name, &path_on_server);

  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return -1;
    }

  req.req.type = UTIME_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  if ( buf )
    {
      req.req.data.utime.time_defined = 1;
      req.req.data.utime.actime = buf->actime;
      req.req.data.utime.modtime = buf->modtime;
    }
  else
    {
      req.req.data.utime.time_defined = 0;
    }
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send utime for file %s\n", path_on_server);

  orfa_network_call(server, &req);
  
  res = req.ans.return_value.int_value;

  free(path_on_server);

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Succeded\n");
    }

  return res;
}


int orfa_fcntl(int fd, int cmd, long arg)
{
  if ( cmd == F_SETFL ) {
    struct client_request req;
    orfa_fd_id_t fd_id;
    server_t server;
    int res;

    get_orfa_fd_server(fd, &server, &fd_id);

    req.req.type = FCNTL_SETFL_REQ;
    req.req.buffer_len = -1;
    req.req.data.fcntl_setfl.fd_id = fd_id;
    req.req.data.fcntl_setfl.flags = (mode_t)arg;
    req.in_buf = NULL;
    req.out_buf = NULL;
    req.out_buf_len = 0;

    LOG(2, "Send fcntl_setfl for fd %llx with flags %d\n",
	(long long)fd_id, (int)arg);

    orfa_network_call(server, &req);
    
    res = req.ans.return_value.int_value;
    
    if ( res < 0 )
      {
	errno = convert_errno_from_portable(req.ans.error);
	LOG(3, "Failed received, error %m\n");
      }
    else
      {
	LOG(3, "Succeded\n");
	orfa_fd_fcntl(fd, cmd, arg);
      }

    return res;

  } else
    return orfa_fd_fcntl(fd, cmd, arg);
}


orfa_dir_t *orfa_opendir(const char *name)
{
  char *path_on_server;
  char *orfa_server_name;
  server_t server;

  orfa_dir_id_t dir_id;
  orfa_dir_t *dir;

  struct client_request req;

  orfa_parse_path(name, &orfa_server_name, &path_on_server);
  
  if ( (server=orfa_network_get_server(orfa_server_name)) < 0 )
    {
      free(path_on_server);
      return NULL;
    }

  req.req.type = OPENDIR_REQ;
  req.req.buffer_len = strlen(path_on_server)+1;
  req.in_buf = path_on_server;
  req.out_buf = NULL;
  req.out_buf_len = 0;
  
  LOG(2, "Send opendir of %s\n", path_on_server);

  orfa_network_call(server, &req);

  dir_id = req.ans.return_value.ptr_value;

  if ( dir_id == 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
      dir = NULL;
    }
  else
    {
      LOG(3, "Received dir %llx\n",
 	  (long long)dir_id);
      
      dir = get_orfa_dir(dir_id, server);
    }
  free(path_on_server);

  return dir;
}


int __orfa_closedir(orfa_dir_id_t dir_id, server_t server)
{
  struct client_request req;
  int res;

  req.req.type = CLOSEDIR_REQ;
  req.req.buffer_len = -1;
  req.req.data.closedir.dir_id = dir_id;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send closedir for dir %llx\n",
      (long long)dir_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;
 
  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv %d\n", res);
    }

  return res;
}


int __orfa_portable_readdir(orfa_dir_id_t dir_id, server_t server,
			    void *buf, orfa_off_t count, orfa_off_t offset)
{
  struct client_request req;
  int res;

  req.req.type = READDIR_REQ;
  req.req.buffer_len = -1;
  req.req.data.readdir.dir_id = dir_id;
  req.req.data.readdir.count = count;
  req.req.data.readdir.offset = offset;
  req.in_buf = NULL;
  req.out_buf = buf;
  req.out_buf_len = count * sizeof(struct portable_dirent);

  LOG(2, "Send readdir for %ld entries in dir %llx\n",
      (long)count,
      (long long)dir_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.ssize_t_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv %d\n", res);
    }

  return res;
}


int __orfa_rewinddir(orfa_dir_id_t dir_id, server_t server)
{
  struct client_request req;
  int res;

  req.req.type = REWINDDIR_REQ;
  req.req.data.readdir.dir_id = dir_id;
  req.in_buf = NULL;
  req.out_buf = NULL;
  req.out_buf_len = 0;

  LOG(2, "Send rewinddir for dir %llx\n",
      (long long)dir_id);

  orfa_network_call(server, &req);

  res = req.ans.return_value.int_value;

  if ( res < 0 )
    {
      errno = convert_errno_from_portable(req.ans.error);
      LOG(3, "Failed received, error %m\n");
    }
  else
    {
      LOG(3, "Recv %d\n", res);
    }

  return res;
}


ssize_t orfa_sendfile64(int out_fd, int in_fd,
			  off64_t *offset, size_t count)
{
  errno = ENOSYS;
  return -1;
}


caddr_t orfa_mmap64(void *start,
		      size_t length,
		      int prot,
		      int flags,
		      int fd,
		      off64_t offset)
{
  errno = ENOSYS;
  return NULL;
}


void *orfa_mremap(void *old_address,
		    size_t old_size,
		    size_t new_size,
		    unsigned long flags)
{
  errno = ENOSYS;
  return NULL;
}


int orfa_munmap(void *start,
		  size_t length)
{
  errno = ENOSYS;
  return -1;
}


int orfa_execve(const  char *filename,
		  char *const argv [],
		  char *const envp[])
{
  struct portable_stat stat;
  int orfa_fd;
  int local_fd_to_fill, local_fd_to_exec;
  char tmp_local_filename[PATH_MAX+1];
  char proc_self_fd_path[PATH_MAX+1];
  void *buffer;
  int res;

  /* stat orfa file */
  if ( orfa_stat64(filename, &stat) < 0 )
    {
      LOG(5, "Stat for file %s failed\n", filename);
      errno = ENOENT;
      return -1;
    }
  /* check read/exec permission */
  if ( access_test(context.euid, context.egid,
		   &stat, R_OK | X_OK) < 0 )
    {
      LOG(5, "Read|Exec access for file %s failed\n", filename);
      errno = EACCES;
      return -1;
    }

  /* read orfa data */
  buffer = malloc(stat.st_size);
  orfa_fd = orfa_open(filename, O_RDONLY, 0);
  res = orfa_pread64(orfa_fd, buffer, stat.st_size, 0);
  orfa_close(orfa_fd);

  if ( res != stat.st_size )
    {
      ERROR("Failed to read orfa file\n");
      errno = EIO;
      goto out;
    }

  /* We are going to create a local copy of the remote file
   * It will be unlinked after creation.
   * We first need a writable descriptor to fill the copy
   * with remote file content.
   * We then need a readonly descriptor to avoid ETXTBSY error
   * during exec.
   *
   * This is protected by all mutex provided by network/dir/...
   * called before exec.
   */

  sprintf(tmp_local_filename, ORFA_EXEC_FILE_PREFIX "%d", getpid());

  /* writable descriptor */
  local_fd_to_fill = REAL_OPEN(tmp_local_filename,
			       O_CREAT | O_EXCL | O_TRUNC | O_WRONLY,
			       S_IRWXU);

  /* descriptor that will be executed */
  local_fd_to_exec = REAL_OPEN(tmp_local_filename,
			       O_RDONLY);

  /* unlink */
  REAL_UNLINK(tmp_local_filename);

  /* fill with data */
  res = REAL_WRITE(local_fd_to_fill, buffer, stat.st_size);
  REAL_CLOSE(local_fd_to_fill);
  if ( res != stat.st_size )
    {
      REAL_CLOSE(local_fd_to_exec);
      ERROR("No enough space on local device to get orfa file\n");
      errno = EIO;
      goto out;
    }

  /* set close_on_exec flag */
  REAL_FCNTL(local_fd_to_exec, F_SETFL, 1);

  /* this could be done through fexecve, when available */
  sprintf(proc_self_fd_path, "/proc/self/fd/%d", local_fd_to_exec);
  res = REAL_EXECVE(proc_self_fd_path, argv, envp);

  /* never come here except if failed */
  ERROR("execve returns %d (%m)\n", res);

  REAL_CLOSE(local_fd_to_exec);

 out:
  /* free all */
  free(buffer);

  return res;
}


#ifdef HAVE_FEXECVE

int orfa_fexecve(int fd,
		 char *const argv[],
		 char *const envp[])
{
  struct portable_stat stat;
  int local_fd_to_fill, local_fd_to_exec;
  char tmp_local_filename[PATH_MAX+1];
  void *buffer;
  int res;

  /* stat orfa file */
  if ( orfa_fstat64(fd, &stat) < 0 )
    {
      LOG(5, "Stat for fd %d failed\n", fd);
      errno = ENOENT;
      return -1;
    }
  /* check read/exec permission */
  if ( access_test(context.euid, context.egid,
		   &stat, R_OK | X_OK) < 0 )
    {
      LOG(5, "Read|Exec access for fd %d failed\n", fd);
      errno = EACCES;
      return -1;
    }

  /* read orfa data */
  buffer = malloc(stat.st_size);
  res = orfa_pread64(fd, buffer, stat.st_size, 0);

  if ( res != stat.st_size )
    {
      ERROR("Failed to read orfa file\n");
      errno = EIO;
      goto out;
    }

  /* We are going to create a local copy of the remote file
   * It will be unlinked after creation.
   * We first need a writable descriptor to fill the copy
   * with remote file content.
   * We then need a descriptor with orfa_fd open flags
   * (generally O_RDONLY to avoir ETXTBSY error during exec).
   *
   * This is protected by all mutex provided by network/dir/...
   * called before exec.
   */

  /* temporary filename */
  sprintf(tmp_local_filename, ORFA_EXEC_FILE_PREFIX "%d", getpid());

  /* writable descriptor */
  local_fd_to_fill = REAL_OPEN(tmp_local_filename,
			       O_CREAT | O_EXCL | O_TRUNC | O_WRONLY,
			       S_IRWXU);

  /* descriptor that will be executed.
   * with same flags that orfa_fd.
   */
  local_fd_to_exec = REAL_OPEN(tmp_local_filename,
			       get_orfa_fd_flags(fd));

  /* unlink */
  REAL_UNLINK(tmp_local_filename);

  /* fill with data */
  res = REAL_WRITE(local_fd_to_fill, buffer, stat.st_size);
  REAL_CLOSE(local_fd_to_fill);
  if ( res != stat.st_size )
    {
      REAL_CLOSE(local_fd_to_exec);
      ERROR("No enough space on local device to get orfa file\n");
      errno = EIO;
      goto out;
    }

  /* set close_on_exec flag */
  REAL_FCNTL(local_fd_to_exec, F_SETFL, 1);

  /* point of no return */
  res = REAL_FEXECVE(local_fd_to_exec, argv, envp);

  /* never come here except if failed */
  ERROR("execve returns %d (%m)\n", res);

  REAL_CLOSE(local_fd_to_exec);

 out:
  /* free all */
  free(buffer);

  return res;
}

#endif /* HAVE_FEXECVE */


#ifdef HAVE_LIBACL

#include <sys/acl.h>

acl_t orfa_acl_get_fd(int fd)
{
  errno = ENOTSUP;
  return (acl_t)NULL;
}


int orfa_acl_set_fd(int fd, acl_t acl)
{
  errno = ENOTSUP;
  return -1;
}


int orfa_acl_extended_fd(int fd)
{
  errno = ENOTSUP;
  return -1;
}


acl_t orfa_acl_get_file(const char *path, acl_type_t type)
{
  errno = ENOTSUP;
  return (acl_t)NULL;
}


int orfa_acl_set_file(const char *path, acl_type_t type, acl_t acl)
{
  errno = ENOTSUP;
  return -1;
}


int orfa_acl_extended_file(const char *path)
{
  errno = ENOTSUP;
  return -1;
}


int orfa_acl_delete_def_file(const char *path)
{
  errno = ENOTSUP;
  return -1;
}

#endif /* HAVE_LIBACL */
