/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "orfa_fd.h"
#include "orfa_calls.h"
#include "orfa_real_calls.h"
#include "orfa_client.h"
#include "orfa_misc.h"
#include "orfa_debug.h"



/********
 * Lock *
 ********/
 
#ifdef MULTITHREAD_SAFE_ENABLED

/* This lock is used to protect current_orfa_pwd access.
 * This have to be a mutex because it has to be locked
 * for a long time during exec/fork. */

#include <pthread.h>
static pthread_mutex_t pwd_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
#define PWD_LOCK() pthread_mutex_lock(&pwd_mutex)
#define PWD_UNLOCK() pthread_mutex_unlock(&pwd_mutex)

#else

#define PWD_LOCK() /* nothing */
#define PWD_UNLOCK() /* nothing */

#endif



/************
 * ORFA PWD *
 ************/

static char *current_orfa_pwd = (char*)NULL;


/* this function returns 1 if the current pwd is ORFA */
int is_orfa_pwd()
{
  return ( current_orfa_pwd != NULL );
}



/*****************
 * Simplify path *
 *****************/

#ifndef PATH_SIMPLIFY_DISABLED
/* this function finds the previous '/' before
   the given one in the given string */
static char *find_previous_slash(char *current_slash,
				 char *full_string)
{
  if ( current_slash == full_string )
    {
      return full_string;
    }
  else
    {
      char *tmp = current_slash - 1;
      while ( *tmp != '/' )
	tmp--;
      return tmp;
    }
}
#endif /* ~PATH_SIMPLIFY_DISABLED */


/* this function copy a source path
   in a preallocated target path by removing
   . .. or // */
static void simplify_path(const char *source, char *target, char *current_ptr_in_target)
{
#ifdef PATH_SIMPLIFY_DISABLED

  strcpy(current_ptr_in_target, source);

#else /* ~PATH_SIMPLIFY_DISABLED */

  char *current_ptr_in_source = (char*)source;
  /* points to the first character after the current '/' */
  char *next_slash_in_source;
  /* points to the next '/' */
  char *end_of_source = strchr(source, '\0');
  /* points to the '\0' at the end */

  while ( 1 )
    {
      next_slash_in_source = strchr(current_ptr_in_source, '/');

      if ( next_slash_in_source == NULL )
	{
	  if ( current_ptr_in_source == end_of_source )
	    /* it's "/" at the end */
	    {
	      LOG(5, "Simplified / at the end\n");
	      *(current_ptr_in_target-1) = '\0';
	      break;
	    }
	  else if ( current_ptr_in_source+1 == end_of_source
		    && *current_ptr_in_source == '.' )
	    /* it's "/." at the end */
	    {
	      LOG(5, "Simplified /. at the end\n");
	      *(current_ptr_in_target-1) = '\0';
	      break;
	    }
	  else if ( current_ptr_in_source+2 == end_of_source
		    && *current_ptr_in_source == '.'
		    && *(current_ptr_in_source+1) == '.' )
	    /* it's "/.." at the end */
	    {
	      LOG(5, "Simplified /.. at the end\n");
	      current_ptr_in_target
		= 1 + find_previous_slash(current_ptr_in_target-1,
					  target);
	      *(current_ptr_in_target-1) = '\0';
	      break;
	    }
	  else
	    /* it's "/xyz" at the end */
	    {
	      LOG(5, "Simplified usual case at the end\n");
	      strcpy(current_ptr_in_target, current_ptr_in_source);
	      break;
	    }
	}
      else
	{
	  if ( current_ptr_in_source == next_slash_in_source )
	    /* it's "//" */
	    {
	      LOG(5, "Simplified //\n");
	      current_ptr_in_source++;
	      continue;
	    }
	  else if ( current_ptr_in_source+1 == next_slash_in_source
		    && *current_ptr_in_source == '.' )
	    /* it's "/./" */
	    {
	      LOG(5, "Simplified /./\n");
	      current_ptr_in_source += 2;
	      continue;
	    }
	  else if ( current_ptr_in_source+2 == next_slash_in_source
		    && *current_ptr_in_source == '.'
		    && *(current_ptr_in_source+1) == '.' )
	    /* it's "/../" */
	    {
	      LOG(5, "Simplified /../\n");
	      current_ptr_in_source += 3;
	      current_ptr_in_target
		= 1 + find_previous_slash(current_ptr_in_target-1,
					  target);
	    }
	  else
	    /* it's "/xyz/' */
	    {
	      LOG(5, "Simplified usual case\n");
	      strncpy(current_ptr_in_target, current_ptr_in_source,
		      next_slash_in_source-current_ptr_in_source+1);
	      current_ptr_in_target += next_slash_in_source-current_ptr_in_source+1;
	      current_ptr_in_source = next_slash_in_source+1;
	    }
	}
    }

  LOG(4, "Simplified in \"%s\"\n", target);

#endif /* ~PATH_SIMPLIFY_DISABLED */
}



/******************************
 * Is this path a ORFA path ? *
 ******************************/

/* this function returns 1 if the given path begins
   with ORFA_PATH_PREFIX else 0 */
int is_orfa_pathname(char *path, char **simplified_path)
{
  char *full_path;
  int path_len = strlen(path);

  LOG(4, "Is ORFA \"%s\"\n", path);

  PWD_LOCK();

  if ( *path == '/' )
    {
      full_path = (char*) malloc( (path_len+1) * sizeof(char) );
      *full_path = '/';
      simplify_path(path+1, full_path, full_path+1);
    }
  else
    {
      int pwd_len;
      char *pwd;
      char real_pwd[PATH_MAX];
      /* get the pwd */
      if ( current_orfa_pwd )
	{
	  pwd = current_orfa_pwd;
	}
      else
	{
	  REAL_GETCWD(real_pwd, PATH_MAX);
	  pwd = real_pwd;
	}
      /* allocate the new path buffer */
      pwd_len = strlen(pwd);
      full_path = (char*) malloc( (path_len+1+pwd_len+1) * sizeof(char) );
      /* fill the pwd */
      strcpy(full_path, pwd);
      full_path[pwd_len] = '/';
      /* simplify the end */
      simplify_path(path, full_path, full_path+pwd_len+1);
    }

  /* replace "" by "/" */
  if ( *full_path == '\0' )
    strcpy(full_path, "/");


  /* Update the path */
  *simplified_path = full_path;
  LOG(4, "The real simplified path is \"%s\"\n", *simplified_path);

  PWD_UNLOCK();

  if ( !strncmp(ORFA_PATH_PREFIX, full_path,
		ORFA_PATH_PREFIX_LENGTH) )
    /* the path begins with ORFA_PATH_PREFIX */
    return 1;
  else
    /* the path is an absolute local path */
    return 0;
}



/********************************
 * Extract ORFA server and path *
 ********************************/

/* This function allocates and filles server_name 
   and path_on_server with given path.
   We assume the path begins with ORFA_PATH_PREFIX
   and has been simplified. */
void orfa_parse_path(const char *path, char **server_name, char **path_on_server)
{
  char *first_slash;
  int len;
  
  first_slash = strchr(path+ORFA_PATH_PREFIX_LENGTH, '/');
  if ( first_slash == NULL )
    len = strlen(path) - ORFA_PATH_PREFIX_LENGTH;
  else
    len = first_slash - path - ORFA_PATH_PREFIX_LENGTH;
  
  if ( first_slash == NULL )
    {
      *path_on_server = orfa_aligned_malloc(2);
      strcpy(*path_on_server, "/");
    }
  else
    {
      *path_on_server = orfa_aligned_malloc(strlen(first_slash)+1);
      strcpy(*path_on_server, first_slash);
    }
  
  *server_name = (char *) malloc( (len+1) * sizeof(char) );
  strncpy(*server_name,
	  path + ORFA_PATH_PREFIX_LENGTH, 
	  len);
  (*server_name)[len] = '\0';
  LOG(4, "Pathname recognize : file %s on server %s\n",
      *path_on_server, *server_name);
}



/*********
 * Chdir *
 *********/

void unorfa_chdir()
{
  PWD_LOCK();

  if ( current_orfa_pwd )
    {
      free(current_orfa_pwd);
      current_orfa_pwd = (char*)NULL;
    }

  PWD_UNLOCK();
}


int orfa_fchdir(int fd)
{
  char *file_name;
  int err;
  struct portable_stat stats;
  int len;

  /* get file name */
  file_name = get_orfa_fd_file_name(fd);

  /* get stats */
  orfa_fstat64(fd, &stats);
  LOG(4, "Stat mode %d (%d) for fd %d\n",
      (int)stats.st_mode, S_IFDIR, fd);
  
  /* does this path exist ? */
  err = ENOTDIR;
  if ( !S_ISDIR(stats.st_mode) )
    goto error;

  /* chdir uses euid/egid to test access rights */
  err = EACCES;
  if ( access_test(context.euid, context.egid,
		   &stats, X_OK) )
    goto error;

  /* access granted */
  unorfa_chdir();

  len = strlen (file_name) + 1;
  current_orfa_pwd = malloc (len);
  memcpy (current_orfa_pwd, file_name, len);

  /* avoid current directory from being busy */
  REAL_CHDIR("/");

  LOG(3, "succeeded\n");
  LOG(4, "Now in \"%s\"\n", current_orfa_pwd);
  return 0;

 error:
  LOG(3, "failed\n");
  return -1;
}


int orfa_chdir(const char *path)
{
  int err;
  struct portable_stat stats;
  int len;

  /* get stats */
  orfa_stat64(path, &stats);
  LOG(4, "Stat mode %d (%d) for file %s\n",
      (int)stats.st_mode, S_IFDIR, path);
  
  /* does this path exist ? */
  err = ENOTDIR;
  if ( !S_ISDIR(stats.st_mode) )
    goto error;

  /* chdir uses euid/egid to test access rights */
  err = EACCES;
  if ( access_test(context.euid, context.egid,
		   &stats, X_OK) )
    goto error;

  /* access granted */
  unorfa_chdir();

  len = strlen (path) + 1;
  current_orfa_pwd = malloc (len);
  memcpy (current_orfa_pwd, path, len);
  
  /* avoid current directory from being busy */
  REAL_CHDIR("/");
  
  LOG(3, "succeeded\n");
  LOG(4, "Now in \"%s\"\n", current_orfa_pwd);
      
  return 0;
  
 error:
  LOG(3, "failed\n");
  return -1;
}


char *orfa_getcwd(char *buf, size_t size)
{
  PWD_LOCK();

  if ( strlen(current_orfa_pwd) > size )
    {
      errno = ERANGE;
      buf = NULL;
    }
  else
    {
      strcpy(buf, current_orfa_pwd);
    }
  
  PWD_UNLOCK();

  return buf;
}



/********
 * Exec *
 ********/

char *orfa_pwd_before_exec()
{
  char *env;

  env = (char*) malloc((EXEC_PWD_ENV_LEN+1+PATH_MAX+1)
		       * sizeof(char));
  strcpy(env, EXEC_PWD_ENV "=");

  PWD_LOCK();

  if ( is_orfa_pwd() )
    orfa_getcwd(env+EXEC_PWD_ENV_LEN+1, PATH_MAX+1);

  LOG(4, "Added new env %s\n", env);

  return env;
}


void orfa_pwd_after_exec_failed()
{
  unsetenv(EXEC_PWD_ENV);

  PWD_UNLOCK();
}


void orfa_pwd_after_exec()
{
  char *env = getenv(EXEC_PWD_ENV);
  LOG(4, "Got pwd env %s\n", env);
  if ( *env )
    {
      orfa_chdir(env);
    }
  unsetenv(EXEC_PWD_ENV);
}
