/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <string.h>
#include <fcntl.h>

#include "orfa_portable.h"
#include "orfa_debug.h"


void convert_stat_portable_to_64(struct stat64 *st,
				 struct portable_stat *pst)
{
  st->st_dev = pst->st_dev;
  st->st_ino = pst->st_ino;
  st->st_mode = pst->st_mode;
  st->st_nlink = pst->st_nlink;
  st->st_uid = pst->st_uid;
  st->st_gid = pst->st_gid;
  st->st_rdev = pst->st_rdev;
  st->st_size = pst->st_size;
  st->st_blksize = pst->st_blksize;
  st->st_blocks = pst->st_blocks;
  st->st_atime = pst->orfa_st_atime;
  st->st_mtime = pst->orfa_st_mtime;
  st->st_ctime = pst->orfa_st_ctime;
}


void convert_stat_portable_to_32(struct stat *st,
				 struct portable_stat *pst)
{
  st->st_dev = pst->st_dev;
  st->st_ino = pst->st_ino;
  st->st_mode = pst->st_mode;
  st->st_nlink = pst->st_nlink;
  st->st_uid = pst->st_uid;
  st->st_gid = pst->st_gid;
  st->st_rdev = pst->st_rdev;
  st->st_size = pst->st_size;
  st->st_blksize = pst->st_blksize;
  st->st_blocks = pst->st_blocks;
  st->st_atime = pst->orfa_st_atime;
  st->st_mtime = pst->orfa_st_mtime;
  st->st_ctime = pst->orfa_st_ctime;
}


void convert_statfs_portable_to_64(struct statfs64 *st,
				   struct portable_statfs *pst)
{
  st->f_type = pst->f_type;
  st->f_bsize = pst->f_bsize;
  st->f_blocks = pst->f_blocks;
  st->f_bfree = pst->f_bfree;
  st->f_bavail = pst->f_bavail;
  st->f_files = pst->f_files;
  st->f_ffree = pst->f_ffree;
  st->f_fsid.__val[0] = pst->f_fsid.__val[0];
  st->f_fsid.__val[1] = pst->f_fsid.__val[1];
  st->f_namelen = pst->f_namelen;
  st->f_spare[0] = pst->f_spare[0];
  st->f_spare[1] = pst->f_spare[1];
  st->f_spare[2] = pst->f_spare[2];
  st->f_spare[3] = pst->f_spare[3];
  st->f_spare[4] = pst->f_spare[4];
  st->f_spare[5] = pst->f_spare[5];
}


void convert_statfs_portable_to_32(struct statfs *st,
				   struct portable_statfs *pst)
{
  st->f_type = pst->f_type;
  st->f_bsize = pst->f_bsize;
  st->f_blocks = pst->f_blocks;
  st->f_bfree = pst->f_bfree;
  st->f_bavail = pst->f_bavail;
  st->f_files = pst->f_files;
  st->f_ffree = pst->f_ffree;
  st->f_fsid.__val[0] = pst->f_fsid.__val[0];
  st->f_fsid.__val[1] = pst->f_fsid.__val[1];
  st->f_namelen = pst->f_namelen;
  st->f_spare[0] = pst->f_spare[0];
  st->f_spare[1] = pst->f_spare[1];
  st->f_spare[2] = pst->f_spare[2];
  st->f_spare[3] = pst->f_spare[3];
  st->f_spare[4] = pst->f_spare[4];
  st->f_spare[5] = pst->f_spare[5];
}


void convert_dirent_portable_to_64(struct dirent64 *d,
				   struct portable_dirent *pd)
{
  d->d_ino = pd->d_ino;
  d->d_off = pd->d_off;
  d->d_reclen = pd->d_reclen;
  d->d_type = pd->d_type;
  strcpy(d->d_name, pd->d_name);
}


void convert_dirent_portable_to_32(struct dirent *d,
				   struct portable_dirent *pd)
{
  d->d_ino = pd->d_ino;
  d->d_off = pd->d_off;
  d->d_reclen = pd->d_reclen;
  d->d_type = pd->d_type;
  strcpy(d->d_name, pd->d_name);
}


orfa_oflags_t convert_oflags_to_portable(int oflags)
{
  orfa_oflags_t poflags = 0;

  switch ( oflags & O_ACCMODE )
    {
    case O_RDONLY: poflags = P_O_RDONLY; break;
    case O_WRONLY: poflags = P_O_WRONLY; break;
    case O_RDWR: poflags = P_O_RDWR; break;
    }
    
  if ( oflags & O_CREAT ) poflags |= P_O_CREAT;
  if ( oflags & O_EXCL ) poflags |= P_O_EXCL;
  if ( oflags & O_NOCTTY ) poflags |= P_O_NOCTTY;
  if ( oflags & O_TRUNC ) poflags |= P_O_TRUNC;

  if ( oflags & O_APPEND ) poflags |= P_O_APPEND;
  if ( oflags & O_NONBLOCK ) poflags |= P_O_NONBLOCK;
  if ( oflags & O_NDELAY ) poflags |= P_O_NDELAY;
  if ( oflags & O_SYNC ) poflags |= P_O_SYNC;

  if ( oflags & O_FSYNC ) poflags |= P_O_FSYNC;
  if ( oflags & O_ASYNC ) poflags |= P_O_ASYNC;
  if ( oflags & O_DSYNC ) poflags |= P_O_DSYNC;
  if ( oflags & O_RSYNC ) poflags |= P_O_RSYNC;

  if ( oflags & O_NOFOLLOW ) poflags |= P_O_NOFOLLOW;
  if ( oflags & O_DIRECTORY ) poflags |= P_O_DIRECTORY;
  if ( oflags & O_LARGEFILE ) poflags |= P_O_LARGEFILE;
  if ( oflags & O_DIRECT ) poflags |= P_O_DIRECT;

  LOG(6, "Converted flags %d to %lld\n",
      (int)oflags, (long long)poflags);

  return poflags;
}


int convert_errno_from_portable(orfa_error_t perror)
{
  return (int)perror;
}
