/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include <asm/page.h>

#include "orfa_debug.h"
#include "orfa_network.h"


# define PAGE_LEN PAGE_SIZE


/*************
 * Debugging *
 *************/

#if REGCACHE_DEBUG_ENABLED

# define REG_PAGES_MAX 10

static unsigned int reg_pages = 0;

static INLINE int __register_pages(void *addr, unsigned int pages)
{
  LOG(5, "Trying to register %d pages at %p\n", pages, addr);
  LOG(5, "Already %d/%d pages registered\n", reg_pages, REG_PAGES_MAX);
  if ( reg_pages + pages > REG_PAGES_MAX )
    {
      LOG(5, "Registering failed\n");
      return -1;
    }
  reg_pages += pages;
  return __orfa_register_interval(addr, pages*PAGE_LEN);
}

static INLINE void __deregister_pages(void *addr, unsigned int pages)
{
  LOG(5, "Deregistering %d pages at %p\n", pages, addr);
  LOG(5, "Still %d/%d pages registered\n", reg_pages-pages, REG_PAGES_MAX);
  reg_pages -= pages;
  __orfa_deregister_interval(addr, pages*PAGE_LEN);
}

#else /* ~REGCACHE_DEBUG_ENABLED */

static INLINE int __register_pages(void *addr, unsigned int pages)
{
  LOG(5, "Trying to register %d pages at %p\n", pages, addr);
  return __orfa_register_interval(addr, pages*PAGE_LEN);
}

static INLINE void __deregister_pages(void *addr, unsigned int pages)
{
  LOG(5, "Deregistering %d pages at %p\n", pages, addr);
  __orfa_deregister_interval(addr, pages*PAGE_LEN);
}

#endif /* REGCACHE_DEBUG_ENABLED */



/***************************
 * Page entry in the cache *
 ***************************/

struct regcache_entry
{
  void *addr;
  unsigned int refcount;
  int registered;
  struct regcache_entry *lru_prev, *lru_next;
  struct regcache_entry *hash_prev, *hash_next; 
};



/************
 * LRU list *
 ************/

static struct regcache_entry *lru_head = NULL;
static struct regcache_entry *lru_tail = NULL;


static INLINE void lru_insert(struct regcache_entry *entry)
{
  if ( !lru_head )
    {
      lru_head = entry;
      entry->lru_prev = NULL;
    }
  else
    {
      lru_tail->lru_next = entry;
      entry->lru_prev = lru_tail;
    }

  entry->lru_next = NULL;
  lru_tail = entry;
}

static INLINE void lru_remove(struct regcache_entry *entry)
{
  if ( entry == lru_head )
    lru_head = entry->lru_next;
  else
    entry->lru_prev->lru_next = entry->lru_next;

  if ( entry == lru_tail )
    lru_tail = entry->lru_prev;
  else
    entry->lru_next->lru_prev = entry->lru_prev;
}


static void lru_reuse(struct regcache_entry *entry)
{
  if ( entry != lru_tail )
    {
      lru_remove(entry);
      lru_insert(entry);
    }
}



/**************
 * Hash table *
 **************/

#define HASH_TABLE_SIZE 4096

static struct regcache_entry *entry_hash_table[HASH_TABLE_SIZE]
= {NULL};


static INLINE unsigned int hash_func(void *addr)
{
  return ((unsigned long)addr/PAGE_LEN)%HASH_TABLE_SIZE;
}


static INLINE void hash_insert(struct regcache_entry *entry)
{
  unsigned int key = hash_func(entry->addr);
  struct regcache_entry **entry_list = &entry_hash_table[key];

  entry->hash_next = *entry_list;
  if ( *entry_list )
    (*entry_list)->hash_prev = entry;

  *entry_list = entry;
  entry->hash_prev = NULL;
}


static INLINE void hash_remove(struct regcache_entry *entry)
{
  unsigned int key = hash_func(entry->addr);
  struct regcache_entry **entry_list = &entry_hash_table[key];

  if ( entry->hash_prev )
    entry->hash_prev->hash_next = entry->hash_next;
  else
    *entry_list = entry->hash_next;

  if ( entry->hash_next )
    entry->hash_next->hash_prev = entry->hash_prev;
}


static struct regcache_entry *hash_find(void *addr)
{
  unsigned int key = hash_func(addr);
  struct regcache_entry *entry = entry_hash_table[key];

  while ( entry != NULL )
    {
      if ( entry->addr == addr )
	return entry;
      entry = entry->hash_next;
    }

  return NULL;
}



/********************
 * Entry processing *
 ********************/

static INLINE struct regcache_entry *add_entry(void *addr)
{
  struct regcache_entry *entry;

  entry = (struct regcache_entry*)
    malloc(sizeof(struct regcache_entry));
  entry->addr = addr;
  entry->refcount = 0;
  entry->registered = 0;
  hash_insert(entry);
  lru_insert(entry);

  return entry;
}    


static INLINE void remove_entry(struct regcache_entry *entry)
{
  hash_remove(entry);
  lru_remove(entry);
  free(entry);
}



/**********************
 * Garbage collecting *
 **********************/

static void garbage_collecting(int required_pages)
{
  struct regcache_entry *entry = lru_head;
  struct regcache_entry *next;
  void *contig_unreg_pages_start = NULL;
  unsigned int contig_unreg_pages = 0;

  while ( entry != NULL  &&  required_pages > 0 )
    {
      LOG(6, "Garbaging page at %p with count %d ?\n",
	  entry->addr, entry->refcount);

      next = entry->lru_next;

      if ( entry->refcount )
	/* cannot release this entry */
	{
	  if ( contig_unreg_pages_start )
	    /* release the previous sequence */
	    {
	      LOG(6, "Entry not free, deregister previous sequence\n");
	      __deregister_pages(contig_unreg_pages_start,
				 contig_unreg_pages);
	      contig_unreg_pages_start = NULL;
	    }
	      
	  entry = next;
	  continue;
	}

      required_pages--;

      if ( !contig_unreg_pages_start )
	/* start a new sequence */
	{
	  LOG(6, "Start a new sequence\n");
	  contig_unreg_pages_start = entry->addr;
	  contig_unreg_pages = 1;
	}
      else if ( entry->addr == contig_unreg_pages_start + contig_unreg_pages*PAGE_LEN )
	/* go on in the previous sequence */
	{
	  LOG(6, "Go on the sequence\n");
	  contig_unreg_pages++;
	}
      else
	/* finish the previous sequence and start a new one */
	{
	  LOG(6, "Finish this sequence and start a new one\n");
	  __deregister_pages(contig_unreg_pages_start,
			     contig_unreg_pages);
	  required_pages -= contig_unreg_pages;
	  contig_unreg_pages_start = entry->addr;
	  contig_unreg_pages = 1;	  
	}

      /* remove the entry */
      hash_remove(entry);
      lru_remove(entry);
      free(entry);
      entry = next;
    }

  if ( contig_unreg_pages_start )
    /* release the last sequence if it's not already done */
    {
      LOG(6, "Deregistering the last sequence\n");
      __deregister_pages(contig_unreg_pages_start,
			 contig_unreg_pages);
    }
}


/******************
 * User functions *
 ******************/

void orfa_regcache_deregister_interval(void *start,
				       unsigned int len)
{
  void *aligned_start = (void*)
    ((unsigned long)start & ~(PAGE_LEN-1));
  void *end = start + len - 1;

  void *current_addr = aligned_start;
  struct regcache_entry *current_entry;

  void *sequence_start = NULL;
  int sequence_len = 0;

  LOG(5, "Deregister interval %d at %p\n", len, start);

  while ( current_addr <= end )
    {
      /* find the corresponding entry */
      current_entry = hash_find(current_addr);

      if ( current_entry != NULL
	   &&  !current_entry->refcount
	   &&  current_entry->registered )
	/* we have to really deregister this page */
	{
	  if ( sequence_start != NULL )
	    /* add it at the end of the sequence */
	    {
	      sequence_len++;
	    }
	  else
	    /* start a new sequence */
	    {
	      sequence_start = current_addr;
	      sequence_len = 1;
	    }
	}
      else
	/* end the previous sequence by deregistering it */
	{
	  if ( sequence_start != NULL )
	    {
	      __deregister_pages(sequence_start,
				 sequence_len);
	      sequence_start = NULL;
	    }
	}

      /* now, remove the entry */
      if ( current_entry  &&  !current_entry->refcount )
	remove_entry(current_entry);

      /* go to the next page */
      current_addr += PAGE_LEN;
    }
 
  /* end the last sequence if necessary */
  if ( sequence_start != NULL )
    {
      __deregister_pages(sequence_start,
			 sequence_len);
    }
}


int orfa_regcache_use_interval(void *start,
			       unsigned int len)
{
  void *aligned_start = (void*)
    ((unsigned long)start & ~(PAGE_LEN-1));
  void *end = start + len - 1;

  void *current_addr = aligned_start;
  struct regcache_entry *current_entry;

  void *sequence_start = NULL;
  int sequence_len = 0;

  LOG(5, "Use interval %d at %p\n", len, start);

  while ( current_addr <= end )
    {
      /* find the corresponding entry */
      current_entry = hash_find(current_addr);
      if ( !current_entry )
	current_entry = add_entry(current_addr);
      else
	lru_reuse(current_entry);
      

      if ( current_entry->registered )
	/* this entry is already registered */
	{
	  if ( sequence_start )
	    /* end the previous sequence */  
	    {
	      if ( __register_pages(sequence_start,
				    sequence_len) < 0 )
		{
		  garbage_collecting((end-current_addr+PAGE_LEN-1)/PAGE_LEN);
		  if ( __register_pages(sequence_start,
					sequence_len) < 0 )
		    {
		      /* here we should cancel all :
		       - decrease refcounts from the begining
		       - put registered to 0 from the begining of the sequence */

		      return -1;
		    }
		}
	    }
	}
      else
	/* we have to register it */
	{
	  if ( sequence_start )
	    /* go on the sequence */
	    {
	      sequence_len++;
	    }
	  else
	    /* start a new sequence */
	    {
	      sequence_start = current_addr;
	      sequence_len = 1;
	    }

	  /* register it */
	  current_entry->registered = 1;
	}

      
      /* increase the refcount */
      current_entry->refcount++;
      
      /* go to the next page */
      current_addr += PAGE_LEN;
    }


  if ( sequence_start )
    /* end the last sequence */  
    {
      if ( __register_pages(sequence_start,
			    sequence_len) < 0 )
	{
	  garbage_collecting((end-current_addr+PAGE_LEN-1)/PAGE_LEN);
	  if ( __register_pages(sequence_start,
				sequence_len) < 0 )
	    {
	      /* here we should cancel all :
		 - decrease refcounts from the begining
		 - put registered to 0 from the begining of the sequence */
	      
	      return -1;
	    }
	}
    }

  return 0;
}


void orfa_regcache_unuse_interval(void *start,
			     unsigned int len)

{
  void *aligned_start = (void*)
    ((unsigned long)start & ~(PAGE_LEN-1));
  void *end = start + len - 1;

  void *current_addr = aligned_start;
  struct regcache_entry *current_entry;

  LOG(5, "Unuse interval %d at %p\n", len, start);

  while ( current_addr <= end )
    {
      current_entry = hash_find(current_addr);
      DEBUG_ASSERT( current_entry != NULL );

      current_entry->refcount--;

      current_addr += PAGE_LEN;
    }
}
