/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <asm/param.h>
#include <string.h>
#include <stdlib.h>

#include "orfa_common_network.h"
#include "orfa_network.h"
#include "orfa_server.h"
#include "orfa_real_calls.h"
#include "orfa_debug.h"



/********
 * Lock *
 ********/

#ifdef MULTITHREAD_SAFE_ENABLED

/* This lock is used to protect the server list.
 * This have to be a mutex because it has to be locked
 * for a long time during exec/fork. */

#include <pthread.h>
static pthread_mutex_t server_list_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
#define SERVER_LIST_LOCK() pthread_mutex_lock(&server_list_mutex)
#define SERVER_LIST_UNLOCK() pthread_mutex_unlock(&server_list_mutex)

#else

#define SERVER_LIST_LOCK() /* nothing */
#define SERVER_LIST_UNLOCK() /* nothing */

#endif



/***************
 * Server List *
 ***************/

struct server {
  char *name;
  struct connection connection;
  int use_count;
#ifdef CONNECTION_ID_ENABLED
  connection_id_t connection_id;
#endif
};
 

static struct server **servers = NULL;
static int server_max = DEFAULT_SERVER_MAX;
static int server_nr = 0;


int add_orfa_server(char *name)
{
  struct server *new = (struct server*) malloc( sizeof(struct server) );
  int i;

  new->name = name;
  new->use_count = 0;
  
  SERVER_LIST_LOCK();

  /* make the array big enough */
  if ( !servers )
    servers = (struct server**) malloc(server_max * sizeof(struct server*));
  else if (server_nr == server_max) {
    server_max *= 2;
    servers = (struct server**) realloc(servers,
					server_max * sizeof(struct server*));
  }
  DEBUG_ASSERT( servers != NULL );

  /* get a new struct server */
  i = server_nr++;
  servers[i] = new;

  SERVER_LIST_UNLOCK();

  return i;
}


int find_orfa_server(char *name)
{
  int i;
  SERVER_LIST_LOCK();
  for(i=0; i<server_nr; i++)
    if ( !strcmp(servers[i]->name, name) )
      {
	SERVER_LIST_UNLOCK();
	return i;
      }
  SERVER_LIST_UNLOCK();
  return -1;
}


void set_orfa_server_connection(int i, struct connection *connection)
{
  DEBUG_ASSERT( i>=0 && i<server_nr );
  memcpy(&servers[i]->connection, connection, sizeof(struct connection));
}


struct connection *get_orfa_server_connection(int i)
{
  DEBUG_ASSERT( i>=0 && i<server_nr );
  return &servers[i]->connection;
}  


void inc_orfa_server_use_count(int i)
{
  DEBUG_ASSERT( i>=0 && i<server_nr );
  servers[i]->use_count++;
}  


void dec_orfa_server_use_count(int i)
{
  DEBUG_ASSERT( i>=0 && i<server_nr );
  servers[i]->use_count--;
}



#ifdef CONNECTION_ID_ENABLED

/* fork has an id but needs to inc it
 * exec has an id but doesn't need to inc it
 * no id doesn't need to enc it */
void set_orfa_server_connection_id(int i, connection_id_t id)
{
  servers[i]->connection_id = id;
  LOG(5, "Set connection_id %llx\n", id);
}


connection_id_t get_orfa_server_connection_id(int i)
{
  DEBUG_ASSERT( i>=0 && i<server_nr );
  return servers[i]->connection_id;
}

#endif /* CONNECTION_ID_ENABLED */




#if (defined MAKE_WRAPPER) || (defined MAKE_LIB)
/*******************************
 * What to do during exec/fork *
 *******************************/


/* Before exec, we put server datas in the environment */
char *orfa_server_before_exec()
{
  char *env = NULL;
  char *current;
  int i=0;

  env = (char*) malloc( (EXEC_SERVER_ENV_LEN+1
			 +server_nr*(MAXHOSTNAMELEN+1
#ifdef CONNECTION_ID_ENABLED
				     +UINTMAX_STRING_MAX_LEN+1
#endif
				     )
			 +1)
			* sizeof(char));

  strcpy(env, EXEC_SERVER_ENV "=");

  /* seek at the begining of the list */
  current = env + EXEC_SERVER_ENV_LEN+1;

  SERVER_LIST_LOCK();

  /* add each server */
  for(i=0; i<server_nr; i++)
    if ( servers[i]->use_count ) {
      int res = 0;
#ifdef CONNECTION_ID_ENABLED
      orfa_network_connection_call(i, servers[i]->connection_id,
				   CONNECTION_OP_BEFORE_EXEC);
      res = sprintf(current, "%llx:",
		    servers[i]->connection_id);
      current += res;
#endif

      res = sprintf(current, "%s;",
		    servers[i]->name);
      current += res;
    }

  LOG(4, "Added env %s\n", env);
  
  return env;
}


/* if exec failed, nothing to do */
void orfa_server_after_exec_failed()
{
#ifdef CONNECTION_ID_ENABLED
  int i;
  for(i=0; i<server_nr; i++)
    if ( servers[i]->use_count )
      orfa_network_connection_call(i, servers[i]->connection_id,
				   CONNECTION_OP_EXEC_FAILED);
#endif

  unsetenv(EXEC_SERVER_ENV);

  SERVER_LIST_UNLOCK();
}


/* if exec succeeded, we recreate servers from environment */
void orfa_server_after_exec()
{
  char *p = getenv(EXEC_SERVER_ENV);

  LOG(4, "Got server env %s\n", p);
  
  SERVER_LIST_LOCK();

  while ( *p )
    {
      int i;
      char *sep;
      char *name = malloc( MAXHOSTNAMELEN * sizeof(char) );
#ifdef CONNECTION_ID_ENABLED
      connection_id_t connection_id;
#endif

      /* scan next part of the environment variable */
 
#ifdef CONNECTION_ID_ENABLED
      DEBUG_ASSERT( sscanf(p, "%llx:", &connection_id) == 1 );
      p = strchr(p, ':') + 1;
#endif

      sep = strchr(p, ';');
      strncpy(name, p, sep-p);
      LOG(5, "Readd server %s\n", name);

      p = sep+1;

      /* add the server */
      i = add_orfa_server(name);
      orfa_network_connect(name, &servers[i]->connection);
#ifdef CONNECTION_ID_ENABLED
      set_server_connection_id(i, connection_id);
      orfa_network_connection_call(i, connection_id,
				   CONNECTION_OP_EXEC_DONE);
#endif      
    }

  unsetenv(EXEC_SERVER_ENV);

  SERVER_LIST_UNLOCK();
}


void orfa_server_after_fork_child() {
  int i;
  for(i=0; i<server_nr; i++) {
#ifdef NETWORK_TCP
    REAL_CLOSE(servers[i]->connection.socket);
    orfa_network_connect(servers[i]->name, &servers[i]->connection);
#endif
#ifdef CONNECTION_ID_ENABLED
    set_server_connection_id(i,
			     orfa_network_connection_call(i, servers[i]->connection_id,
							  CONNECTION_OP_FORK_DONE));
#endif
  }
}


#endif /* (defined MAKE_WRAPPER) || (defined MAKE_LIB) */
