/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/resource.h>

#include "orfa_shared.h"
#include "orfa_path.h"
#include "orfa_fd.h"
#include "orfa_dir.h"
#include "orfa_real_calls.h"
#include "orfa_debug.h"


/***********
 * Context *
 ***********/

struct context context;


/********
 * Fork *
 ********/

void orfa_before_fork()
{
  orfa_network_before_fork();
  orfa_fd_before_fork();
  orfa_dir_before_fork();
}

void orfa_after_fork_failed()
{
  orfa_network_after_fork_parent();      
}

void orfa_after_fork_parent()
{
  orfa_network_after_fork_parent();    
}

void orfa_after_fork_child()
{
  /* fixes its general informations */
  context.pid = getpid();
  
  /* fork succeded and this is the child
   * the child duplicates fd and dir */
  orfa_fd_after_fork_child();
  orfa_dir_after_fork_child();
  
  /* start network */
  orfa_network_after_fork_child();
}


/********
 * Exec *
 ********/

void orfa_before_exec(char **envp, int *lenp, char ***new_envpp)
{
  char **ptr = envp;
  char *fd_env=NULL, *dir_env=NULL, *pwd_env=NULL, *server_env=NULL;
  char **new_envp;
  int len = 0;

  /* lock the network */
  orfa_network_before_exec();
  
  if ( orfa_network_ready != 1 )
    {
      /* get environment variables */
      pwd_env = orfa_pwd_before_exec();
      fd_env = orfa_fd_before_exec();
      dir_env = orfa_dir_before_exec();
      server_env = orfa_server_before_exec();
      
      /* create new environment */
      for(; *ptr; ptr++, len++);
      new_envp = (char**) malloc( (len+5) * sizeof(char*) );
      memcpy(new_envp, envp, len * sizeof(char*));
      new_envp[len] = fd_env;
      new_envp[len+1] = dir_env;
      new_envp[len+2] = pwd_env;
      new_envp[len+3] = server_env;
      new_envp[len+4] = (char*)NULL;
      *new_envpp = new_envp;
      *lenp = len;
    }
}

void orfa_after_exec_failed(int old_len, char **new_envp)
{
  if ( orfa_network_ready )
    {
      /* delete environment */
      free(new_envp[old_len]);
      free(new_envp[old_len+1]);
      free(new_envp[old_len+2]);
      free(new_envp[old_len+3]);
      free(new_envp);
      
      /* clean all */
      orfa_dir_after_exec_failed();
      orfa_fd_after_exec_failed();
      orfa_pwd_after_exec_failed();
      orfa_server_after_exec_failed();
    }
  
  /* unlock the network */
  orfa_network_after_exec_failed();
}

/* This function is never called by the application itself.
 * The library calls it during _init if necessary.
 */
static void orfa_after_exec()
{
  /* the library was initialized before exec,
   * process structure rebuilding */
  orfa_server_after_exec();
  orfa_pwd_after_exec();
  orfa_fd_after_exec();
  orfa_dir_after_exec();
}


/*****************
 * Init and Fini *
 *****************/

void orfa_init()
{
  if ( getenv(EXEC_FD_ENV) != NULL )
    orfa_after_exec();
  else
    { /* it's the first time the library is used in this process,
       * initialize it */

      /* get/set the maximal FD number */
      struct rlimit rlim;
      rlim.rlim_cur = FD_MAX;
      rlim.rlim_max = FD_MAX;
      if ( setrlimit(RLIMIT_NOFILE, &rlim) < 0 ) {
	getrlimit(RLIMIT_NOFILE, &rlim);
	rlim.rlim_cur = rlim.rlim_max;
	setrlimit(RLIMIT_NOFILE, &rlim);
	ERROR("Cannot setrlimit ENOFILE to %d, reverted to %ld\n",
	      FD_MAX, (long)rlim.rlim_max);
      }
    }

  /* get some informations */
  orfa_update_pid(getpid());
  orfa_update_euid(geteuid());
  orfa_update_egid(getegid());
  orfa_update_concurrent_mode(getenv(NON_CONCURRENT_MODE_ENV)==NULL);
  
  /* modify umask to get previous value */
  orfa_update_umask(umask(0000));
  /* reinstall previous umask */
  REAL_UMASK(context.umask);
}


void orfa_fini()
{
  put_all_orfa_fd();
  put_all_orfa_dir();
  orfa_network_finish();  
}
