/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>

#include "orfa_shared.h"
#include "orfa_path.h"
#include "orfa_fd.h"
#include "orfa_dir.h"
#include "orfa_symbols.h"
#include "orfa_real_calls.h"
#include "orfa_calls.h"
#include "orfa_convert64.h"
#include "orfa_portable.h"
#include "orfa_misc.h"
#include "orfa_debug.h"


/***************************
 * Interception principles *
 ***************************/

/* Each function is simply overriden by creating
 * a same name function. Calling the original libc
 * function is done via the symbol "real_func" which
 * is initialized using the dlfcn library and the
 * "CHECK_REAL_CALLS" macro.
 *
 * Functions can have differents symbols:
 * - "func" is the standard symbol ;
 * - "_func" xor "__func" may exist to provide
 *    an easy access to the original function
 *    of an intercepted function ;
 * - "__libc_func" provides an access to the
 *   original function if "func" AND "_func"/"__func"
 *   are intercepted (used in libpthread).
 *
 * This library intercepts ALL these symbols
 * testing if they exists for the non-common ones.
 * "HAVE_FUNC" are defined in symbols/symbols.h
 * during the begining of compilation.
 *
 * fstat/lstat/... and mknod are protected by
 * "HAVE_FUNC" tests because they often are inlined
 * link to underlying "__fxstat" or "__xmknod" functions.
 * 
 * "func" redefinition uses "real_func" call whereas
 * "__libc_func" redefinition uses "real___libc_func" call.
 * It prevents loops when underlying library redefines
 * "func" using "__libc_func".
 *
 * When calling orfa_func in "__libc_func", we have to
 * ensure that only real___libc_calls will be used.
 * So we used the lowest real_*_call in internal function
 * used by orfa_func.
 */



/**************************
 * Initialization for LOG *
 **************************/

INIT_LOGS();



/****************************
 * Process Handling Mutexes *
 ****************************/

/* this counts nested *fork calls */
static int fork_calls = 0;


#ifdef MULTITHREAD_SAFE_ENABLED

#include <pthread.h>
/* this mutex MUST be recursive to accept
 * calls to __libc_fork inside fork and so on */
static pthread_mutex_t process_handle_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
#define PROCESS_LOCK() pthread_mutex_lock(&process_handle_mutex)
#define PROCESS_UNLOCK() pthread_mutex_unlock(&process_handle_mutex)

#else

#define PROCESS_LOCK() /* nothing */
#define PROCESS_UNLOCK() /* nothing */

#endif


/*****************
 * INIT and FINI *
 *****************/


#if TEST_ENABLED
static struct timeval tv1, tv2;
#endif /* TEST_ENABLED */


int _init( void )
{
  FAST_LOG(1, "Initializing library...\n");

  CHECK_REAL_CALLS();

  orfa_init();

#if TEST_ENABLED
  if ( orfa_network_init() )
    {
      FAST_ERROR("Client network init failed\n");
      exit(-1);
    }

  gettimeofday(&tv1, NULL);
#endif

  FAST_LOG(1, "Library initialized\n");

  return 0;
}


int _fini( void )
{
  PROCESS_LOCK();

#if TEST_ENABLED
  gettimeofday(&tv2, NULL);
  ERROR("Execution duration %f us\n",
	(tv2.tv_sec-tv1.tv_sec)*1000000.+tv2.tv_usec-tv1.tv_usec);
#endif

  FAST_LOG(1, "Exiting library\n");

  orfa_fini();

  return 0;
}



/********
 * OPEN *
 ********/

int open(const char *pathname, int flags, ...)
{
  char *new_path;
  int res;

  va_list ap;
  va_start(ap, flags);
  
  if ( is_orfa_pathname(pathname, &new_path) )
    {
      LOG(1, "ORFA open of %s flags %d\n",
	  pathname, flags);
      res = orfa_open(new_path, flags, va_arg(ap, mode_t));
    }
  else
    {
      FAST_LOG(1, "Classical open\n");
      CHECK_REAL_CALLS();
      res = real_open(new_path, flags, va_arg(ap, mode_t));
    }

  va_end(ap);

  free(new_path);
  return res;
}


#if (defined HAVE___OPEN || defined HAVE__OPEN)
#ifdef HAVE___OPEN
int __open(const char *pathname, int flags, ...)
#elif HAVE__OPEN
int _open(const char *pathname, int flags, ...)
#endif
{
  char *new_path;
  int res;

  va_list ap;
  va_start(ap, flags);
  
  if ( is_orfa_pathname(pathname, &new_path) )
    {
      LOG(1, "ORFA _open of %s flags %d\n",
	  pathname, flags);
      res = orfa_open(new_path, flags, va_arg(ap, mode_t));
    }
  else
    {
      FAST_LOG(1, "Classical _open\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___OPEN
      res = real___open(new_path, flags, va_arg(ap, mode_t));
#elif HAVE__OPEN
      res = real__open(new_path, flags, va_arg(ap, mode_t));
#endif
    }

  va_end(ap);

  free(new_path);
  return res;  
}
#endif


#ifdef HAVE___LIBC_OPEN
int __libc_open(const char *pathname, int flags, ...)
{
  char *new_path;
  int res;

  va_list ap;
  va_start(ap, flags);
  
  if ( is_orfa_pathname(pathname, &new_path) )
    {
      LOG(1, "ORFA __libc_open of %s flags %d\n",
	  pathname, flags);
      res = orfa_open(new_path, flags, va_arg(ap, mode_t));
    }
  else
    {
      FAST_LOG(1, "Classical __libc_open\n");
      CHECK_REAL_CALLS();
      res = real___libc_open(new_path, flags, va_arg(ap, mode_t));
    }

  va_end(ap);

  free(new_path);
  return res;
}
#endif



/**********
 * OPEN64 *
 **********/

#ifdef HAVE_OPEN64
int open64(const char *name, int flags, ...)
{
  char *new_path;
  int res;

  va_list ap;
  va_start(ap, flags);

  if ( is_orfa_pathname(name, &new_path) )
    {
      LOG(1, "ORFA open64 of %s flags %d\n",
	  name, flags);
      res = orfa_open(new_path, flags | O_LARGEFILE,
		      va_arg(ap, mode_t));
    }
  else
    {
      FAST_LOG(1, "Classical open64\n");
      CHECK_REAL_CALLS();
      res = real_open64(new_path, flags,
			va_arg(ap, mode_t));
    }

  va_end(ap);

  free(new_path);
  return res;
}
#endif


#if (defined HAVE___OPEN64 || defined HAVE__OPEN64)
#ifdef HAVE___OPEN64
int __open64(const char *name, int flags, ...)
#elif HAVE__OPEN64
int _open64(const char *name, int flags, ...)
#endif
{
  char *new_path;
  int res;

  va_list ap;
  va_start(ap, flags);

  if ( is_orfa_pathname(name, &new_path) )
    {
      LOG(1, "ORFA _open64 of %s flags %d\n",
	  name, flags);
      res = orfa_open(new_path, flags | O_LARGEFILE,
		      va_arg(ap, mode_t));
    }
  else
    {
      FAST_LOG(1, "Classical _open64\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___OPEN64
      res = real___open64(new_path, flags,
			va_arg(ap, mode_t));
#elif HAVE__OPEN64
      res = real__open64(new_path, flags,
			va_arg(ap, mode_t));
#endif
    }

  va_end(ap);

  free(new_path);
  return res;
}
#endif


#ifdef HAVE___LIBC_OPEN64
int __libc_open64(const char *name, int flags, ...)
{
  char *new_path;
  int res;

  va_list ap;
  va_start(ap, flags);

  if ( is_orfa_pathname(name, &new_path) )
    {
      LOG(1, "ORFA __libc_open64 of %s flags %d\n",
	  name, flags);
      res = orfa_open(new_path, flags | O_LARGEFILE,
		      va_arg(ap, mode_t));
    }
  else
    {
      FAST_LOG(1, "Classical __libc_open64\n");
      CHECK_REAL_CALLS();
      res = real___libc_open64(new_path, flags,
			       va_arg(ap, mode_t));
    }

  va_end(ap);

  free(new_path);
  return res;
}
#endif



/*********
 * CREAT *
 *********/

#ifdef HAVE_CREAT
int creat(const char *pathname, mode_t mode)
{
  return open(pathname, O_WRONLY | O_CREAT | O_TRUNC, mode);
}
#endif


#ifdef HAVE_CREAT64
int creat64(const char *path, mode_t mode)
{
  return open64(path, O_WRONLY | O_CREAT | O_TRUNC, mode);
}
#endif



/*********
 * CLOSE *
 *********/

int close(int fd)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA close of fd %d\n", fd);
      return orfa_close(fd);
    }
  else
    {
      FAST_LOG(1, "Classical close\n");
      CHECK_REAL_CALLS();
      return real_close(fd);
    }
}


#if (defined HAVE___CLOSE || defined HAVE__CLOSE)
#ifdef HAVE___CLOSE
int __close(int fd)
#elif HAVE__CLOSE
int _close(int fd)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA close of fd %d\n", fd);
      return orfa_close(fd);
    }
  else
    {
      FAST_LOG(1, "Classical close\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___CLOSE
      return real___close(fd);
#elif HAVE__CLOSE
      return real__close(fd);
#endif
    }  
}
#endif


#ifdef HAVE___LIBC_CLOSE
int __libc_close(int fd)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_close of fd %d\n", fd);
      return orfa_close(fd);
    }
  else
    {
      FAST_LOG(1, "Classical __libc_close\n");
      CHECK_REAL_CALLS();
      return real___libc_close(fd);
    }
}
#endif


/********
 * READ *
 ********/

ssize_t read(int fd, void *buf, size_t count)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA read\n");
      return orfa_read(fd, buf, count);
    }
  else
    {
      FAST_LOG(1, "Classical read\n");
      CHECK_REAL_CALLS();
      return real_read(fd, buf, count);
    }
}


#if (defined HAVE___READ || defined HAVE__READ)
#ifdef HAVE___READ
ssize_t __read(int fd, void *buf, size_t count)
#elif HAVE__READ
ssize_t _read(int fd, void *buf, size_t count)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _read\n");
      return orfa_read(fd, buf, count);
    }
  else
    {
      FAST_LOG(1, "Classical _read\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___READ
      return real___read(fd, buf, count);
#elif HAVE__READ
      return real__read(fd, buf, count);
#endif
    }
}
#endif


#ifdef HAVE___LIBC_READ
ssize_t __libc_read(int fd, void *buf, size_t count)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_read\n");
      return orfa_read(fd, buf, count);
    }
  else
    {
      FAST_LOG(1, "Classical __libc_read\n");
      CHECK_REAL_CALLS();
      return real___libc_read(fd, buf, count);
    }
}
#endif



/*********
 * PREAD *
 *********/

#ifdef HAVE_PREAD
ssize_t pread(int fd, void *buf, size_t count, off_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA pread\n");
      return orfa_pread64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical pread\n");
      CHECK_REAL_CALLS();
      return real_pread(fd, buf, count, offset);
    }
}
#endif


#if (defined HAVE___PREAD || defined HAVE__PREAD)
#ifdef HAVE___PREAD
ssize_t __pread(int fd, void *buf, size_t count, off_t offset)
#elif HAVE__PREAD
ssize_t _pread(int fd, void *buf, size_t count, off_t offset)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _pread\n");
      return orfa_pread64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical _pread\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___PREAD
      return real___pread(fd, buf, count, offset);
#elif HAVE__PREAD
      return real__pread(fd, buf, count, offset);
#endif
    }
}
#endif


#ifdef HAVE___LIBC_PREAD
ssize_t __libc_pread(int fd, void *buf, size_t count, off_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_pread\n");
      return orfa_pread64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical __libc_pread\n");
      CHECK_REAL_CALLS();
      return real___libc_pread(fd, buf, count, offset);
    }
}
#endif



/***********
 * PREAD64 *
 ***********/

#ifdef HAVE_PREAD64
ssize_t pread64(int fd, void *buf, size_t count, off64_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA pread64\n");
      return orfa_pread64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical pread64\n");
      CHECK_REAL_CALLS();
      return real_pread64(fd, buf, count, offset);
    }
}
#endif


#if (defined HAVE___PREAD64 || defined HAVE__PREAD64)
#ifdef HAVE___PREAD64
ssize_t __pread64(int fd, void *buf, size_t count, off64_t offset)
#elif HAVE__PREAD64
ssize_t _pread64(int fd, void *buf, size_t count, off64_t offset)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _pread64\n");
      return orfa_pread64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical _pread64\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___PREAD64
      return real___pread64(fd, buf, count, offset);
#elif HAVE__PREAD64
      return real__pread64(fd, buf, count, offset);
#endif
    }
}
#endif


#ifdef HAVE___LIBC_PREAD64
ssize_t __libc_pread64(int fd, void *buf, size_t count, off64_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_pread64\n");
      return orfa_pread64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical __libc_pread64\n");
      CHECK_REAL_CALLS();
      return real___libc_pread64(fd, buf, count, offset);
    }
}
#endif



/*********
 * WRITE *
 *********/

ssize_t write(int fd, const void *buf, size_t count)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA write\n");
      return orfa_write(fd, buf, count);
    }
  else
    {
      if ( fd != LOG_FD )
	FAST_LOG(1, "Classical write\n");
      CHECK_REAL_CALLS();
      return real_write(fd, buf, count);
    }
}


#if (defined HAVE___WRITE || defined HAVE__WRITE)
#ifdef HAVE___WRITE
ssize_t __write(int fd, const void *buf, size_t count)
#elif HAVE__WRITE
ssize_t _write(int fd, const void *buf, size_t count)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _write\n");
      return orfa_write(fd, buf, count);
    }
  else
    {
      if ( fd != LOG_FD )
	FAST_LOG(1, "Classical _write\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___WRITE
      return real___write(fd, buf, count);
#elif HAVE__WRITE
      return real__write(fd, buf, count);
#endif
    }
}
#endif


#ifdef HAVE___LIBC_WRITE
ssize_t __libc_write(int fd, const void *buf, size_t count)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_write\n");
      return orfa_write(fd, buf, count);
    }
  else
    {
      if ( fd != LOG_FD )
	FAST_LOG(1, "Classical __libc_write\n");
      CHECK_REAL_CALLS();
      return real___libc_write(fd, buf, count);
    }
}
#endif



/**********
 * PWRITE *
 **********/

#ifdef HAVE_PWRITE
ssize_t pwrite(int fd, const void *buf, size_t count, off_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA pwrite\n");
      return orfa_pwrite64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical pwrite\n");
      CHECK_REAL_CALLS();
      return real_pwrite(fd, buf, count, offset);
    }
}
#endif


#if (defined HAVE___PWRITE || defined HAVE__PWRITE)
#ifdef HAVE___PWRITE
ssize_t __pwrite(int fd, const void *buf, size_t count, off_t offset)
#elif HAVE__PWRITE
ssize_t _pwrite(int fd, const void *buf, size_t count, off_t offset)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _pwrite\n");
      return orfa_pwrite64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical _pwrite\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___PWRITE
      return real___pwrite(fd, buf, count, offset);
#elif HAVE__PWRITE
      return real__pwrite(fd, buf, count, offset);
#endif
    }
}
#endif


#ifdef HAVE___LIBC_PWRITE
ssize_t __libc_pwrite(int fd, const void *buf, size_t count, off_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_pwrite\n");
      return orfa_pwrite64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical __libc_pwrite\n");
      CHECK_REAL_CALLS();
      return real___libc_pwrite(fd, buf, count, offset);
    }
}
#endif



/************
 * PWRITE64 *
 ************/

#ifdef HAVE_PWRITE64
ssize_t pwrite64(int fd, const void *buf, size_t count, off64_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA pwrite\n");
      return orfa_pwrite64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical pwrite\n");
      CHECK_REAL_CALLS();
      return real_pwrite64(fd, buf, count, offset);
    }
}
#endif


#if (defined HAVE___PWRITE64 || defined HAVE__PWRITE64)
#ifdef HAVE___PWRITE64
ssize_t __pwrite64(int fd, const void *buf, size_t count, off64_t offset)
#elif HAVE__PWRITE64
ssize_t _pwrite64(int fd, const void *buf, size_t count, off64_t offset)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _pwrite\n");
      return orfa_pwrite64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical _pwrite\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___PWRITE64
      return real___pwrite64(fd, buf, count, offset);
#elif HAVE__PWRITE64
      return real__pwrite64(fd, buf, count, offset);
#endif
    }
}
#endif


#ifdef HAVE___LIBC_PWRITE64
ssize_t __libc_pwrite64(int fd, const void *buf, size_t count, off64_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_pwrite\n");
      return orfa_pwrite64(fd, buf, count, offset);
    }
  else
    {
      LOG(1, "Classical __libc_pwrite\n");
      CHECK_REAL_CALLS();
      return real___libc_pwrite64(fd, buf, count, offset);
    }
}
#endif



/*********
 * LSEEK *
 *********/

off_t lseek(int fd, off_t offset, int whence)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA lseek\n");
      return (off_t)orfa_lseek64(fd, offset, whence);
    }
  else
    {
      LOG(1, "Classical lseek\n");
      CHECK_REAL_CALLS();
      return real_lseek(fd, offset, whence);
    }
}


#if (defined HAVE___LSEEK || defined HAVE__LSEEK)
#ifdef HAVE___LSEEK
off_t __lseek(int fd, off_t offset, int whence)
#elif HAVE__LSEEK
off_t _lseek(int fd, off_t offset, int whence)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _lseek\n");
      return (off_t)orfa_lseek64(fd, offset, whence);
    }
  else
    {
      LOG(1, "Classical _lseek\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___LSEEK
      return real___lseek(fd, offset, whence);
#elif HAVE__LSEEK
      return real__lseek(fd, offset, whence);
#endif
    }
}
#endif


#ifdef HAVE___LIBC_LSEEK
off_t __libc_lseek(int fd, off_t offset, int whence)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_lseek\n");
      return (off_t)orfa_lseek64(fd, offset, whence);
    }
  else
    {
      LOG(1, "Classical __libc_lseek\n");
      CHECK_REAL_CALLS();
      return real___libc_lseek(fd, offset, whence);
    }
}
#endif



/***********
 * LSEEK64 *
 ***********/

#ifdef HAVE_LSEEK64
off64_t lseek64(int fd, off64_t offset, int whence)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA lseek64\n");
      return orfa_lseek64(fd, offset, whence);
    }
  else
    {
      LOG(1, "Classical lseek64\n");
      CHECK_REAL_CALLS();
      return real_lseek64(fd, offset, whence);
    }
}
#endif


#if (defined HAVE___LSEEK64 || defined HAVE__LSEEK64)
#ifdef HAVE___LSEEK64
off_t __lseek64(int fd, off64_t offset, int whence)
#elif HAVE__LSEEK64
off_t _lseek64(int fd, off64_t offset, int whence)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _lseek64\n");
      return orfa_lseek64(fd, offset, whence);
    }
  else
    {
      LOG(1, "Classical _lseek64\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___LSEEK64
      return real___lseek64(fd, offset, whence);
#elif HAVE__LSEEK64
      return real__lseek64(fd, offset, whence);
#endif
    }
}
#endif


#ifdef HAVE___LIBC_LSEEK64
off_t __libc_lseek64(int fd, off64_t offset, int whence)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_lseek64\n");
      return orfa_lseek64(fd, offset, whence);
    }
  else
    {
      LOG(1, "Classical __libc_lseek64\n");
      CHECK_REAL_CALLS();
      return real___libc_lseek64(fd, offset, whence);
    }
}
#endif



/****************
 * Various Seek *
 ****************/

#ifdef HAVE_LLSEEK
loff_t llseek(int fd, loff_t ofs, int whence)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA llseek\n");
      return (loff_t)orfa_lseek64(fd, ofs, whence);
    }
  else
    {
      LOG(1, "Classical llseek\n");
      CHECK_REAL_CALLS();
      return real_llseek(fd, ofs, whence);
    }
}
#endif


#ifdef HAVE__IO_FILE_SEEK
/* Overriding fseek implies to override the following function.
   It looks like if it is the only special case like this */
_IO_off64_t
_IO_file_seek(struct _IO_FILE *fp, _IO_off64_t offset, int dir)
{
  if ( is_orfa_fd(fp->_fileno) )
    {
      LOG(1, "ORFA _IO_file_seek\n");
      return (_IO_off64_t) orfa_lseek64(fp->_fileno, offset, dir);
    }
  else
    {
      LOG(1, "Classical _IO_file_seek\n");
      CHECK_REAL_CALLS();
      return real__IO_file_seek(fp, offset, dir);
    }
}
#endif



/*********
 * TRUNC *
 *********/

#ifdef HAVE_FTRUNCATE64
int ftruncate64(int fd, off64_t length)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA ftruncate64 of %d at %lld\n",
	  fd, (long long)length);
      return orfa_ftruncate64(fd, length);
    }
  else
    {
      LOG(1, "Classical ftruncate64 of %d at %lld\n",
	  fd, (long long)length);
      CHECK_REAL_CALLS();
      return real_ftruncate64(fd, length);
    }
}
#endif


int ftruncate(int fd, off_t length)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA ftruncate of %d at %ld\n",
	  fd, (long)length);
      return orfa_ftruncate64(fd, length);
    }
  else
    {
      LOG(1, "Classical ftruncate of %d at %ld\n",
	  fd, (long)length);
      CHECK_REAL_CALLS();
      return real_ftruncate(fd, length);
    }
}


#ifdef HAVE_TRUNCATE64
int truncate64(const char *path, off64_t length)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA truncate64 of %s at %lld\n",
	  path, (long long)length);
      res = orfa_truncate64(new_path, length);
    }
  else
    {
      LOG(1, "Classical truncate64 of %s at %lld\n",
	  path, (long long)length);
      CHECK_REAL_CALLS();
      res = real_truncate64(new_path, length);
    }
  free(new_path);

  return res;
}
#endif


int truncate(const char *path, off_t length)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA truncate of %s at %ld\n",
	  path, (long)length);
      res = orfa_truncate64(new_path, length);
    }
  else
    {
      LOG(1, "Classical truncate64 of %s at %ld\n",
	  path, (long)length);
      CHECK_REAL_CALLS();
      res = real_truncate(new_path, length);
    }
  free(new_path);

  return res;
}



/*********
 * FSTAT *
 *********/

#ifdef HAVE___FXSTAT
int __fxstat(int ver, int fd, struct stat *buf)
#else
int fstat(int fd, struct stat *buf)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      int res;
      LOG(1, "ORFA fstat\n");
      PORTABLE_CALL32(fstat, stat, fd, buf, res);
      return res;
    }
  else
    {
      LOG(1, "Classical fstat\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___FXSTAT
      return real___fxstat(ver, fd, buf);
#else
      return real_fstat(fd, buf);
#endif
    }
}


#if (!defined HAVE___FXSTAT && (defined HAVE___FSTAT || defined HAVE__FSTAT))
#ifdef HAVE___FSTAT
int __fstat(int fd, struct stat *buf)
#elif HAVE__FSTAT
int _fstat(int fd, struct stat *buf)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      int res;
      LOG(1, "ORFA _fstat\n");
      PORTABLE_CALL32(fstat, stat, fd, buf, res);
      return res;
    }
  else
    {
      LOG(1, "Classical _fstat\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___FSTAT
      return real___fstat(fd, buf);
#elif HAVE__FSTAT
      return real__fstat(fd, buf);
#endif
    }
}
#endif



/***********
 * FSTAT64 *
 ***********/

#ifdef HAVE___FXSTAT64
int __fxstat64(int ver, int fd, struct stat64 *buf)
#else
int fstat64(int fd, struct stat64 *buf)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      int res;
      LOG(1, "ORFA fstat64\n");
      PORTABLE_CALL64(fstat, stat, fd, buf, res);
      return res;
    }
  else
    {
      LOG(1, "Classical fstat64\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___FXSTAT64
      return real___fxstat64(ver, fd, buf);
#else
      return real_fstat64(fd, buf);
#endif
    }
}


#if (!defined HAVE___FXSTAT64 && (defined HAVE___FSTAT64 || defined HAVE__FSTAT64))
#ifdef HAVE___FSTAT64
int __fstat64(int fd, struct stat64 *buf)
#elif HAVE__FSTAT64
int _fstat64(int fd, struct stat64 *buf)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      int res;
      LOG(1, "ORFA _fstat64\n");
      PORTABLE_CALL64(fstat, stat, fd, buf, res);
      return res;
    }
  else
    {
      LOG(1, "Classical _fstat64\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___FSTAT64
      return real___fstat64(fd, buf);
#elif HAVE__FSTAT64
      return real__fstat64(fd, buf);
#endif
    }
}
#endif



/********
 * STAT *
 ********/

#ifdef HAVE___XSTAT
int __xstat(int ver, const char *file_name, struct stat *buf)
#else
int stat(const char *file_name, struct stat *buf)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA stat for file %s\n", file_name);
      PORTABLE_CALL32(stat, stat, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical stat for file %s\n", file_name);
      CHECK_REAL_CALLS();
#ifdef HAVE___XSTAT
      res = real___xstat(ver, new_path, buf);
#else
      res = real_stat(new_path, buf);
#endif
    }
  free(new_path);

  return res;
}


#if (!defined HAVE___XSTAT && (defined HAVE___STAT || defined HAVE__STAT))
#ifdef HAVE___STAT
int __stat(const char *file_name, struct stat *buf)
#elif HAVE__STAT
int _stat(const char *file_name, struct stat *buf)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA _stat for file %s\n", file_name);
      PORTABLE_CALL32(stat, stat, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical _stat for file %s\n", file_name);
      CHECK_REAL_CALLS();
#ifdef HAVE___STAT
      res = real___stat(new_path, buf);
#elif HAVE__STAT
      res = real__stat(new_path, buf);
#endif
    }
  free(new_path);

  return res;
}
#endif



/**********
 * STAT64 *
 **********/

#ifdef HAVE___XSTAT64
int __xstat64(int ver, const char *file_name, struct stat64 *buf)
#else
int stat64(const char *file_name, struct stat64 *buf)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA stat64 for file %s\n", file_name);
      PORTABLE_CALL64(stat, stat, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical stat64 for file %s\n", file_name);
      CHECK_REAL_CALLS();
#ifdef HAVE___XSTAT64
      res = real___xstat64(ver, new_path, buf);
#else
      res = real_stat64(new_path, buf);
#endif
    }
  free(new_path);

  return res;
}


#if (!defined HAVE___XSTAT64 && (defined HAVE___STAT64 || defined HAVE__STAT64))
#ifdef HAVE___STAT64
int __stat64(const char *file_name, struct stat64 *buf)
#elif HAVE__STAT64
int _stat64(const char *file_name, struct stat64 *buf)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA _stat64 for file %s\n", file_name);
      PORTABLE_CALL64(stat, stat, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical _stat64 for file %s\n", file_name);
      CHECK_REAL_CALLS();
#ifdef HAVE___STAT64
      res = real___stat64(new_path, buf);
#elif HAVE__STAT64
      res = real__stat64(new_path, buf);
#endif
    }
  free(new_path);

  return res;
}
#endif



/*********
 * LSTAT *
 *********/

#ifdef HAVE___LXSTAT
int __lxstat(int ver, const char *file_name, struct stat *buf)
#else
int lstat(const char *file_name, struct stat *buf)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA lstat for file %s\n", file_name);
      PORTABLE_CALL32(lstat, stat, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical lstat for file %s\n", file_name);
      CHECK_REAL_CALLS();
#ifdef HAVE___LXSTAT
      res = real___lxstat(ver, new_path, buf);
#else
      res = real_lstat(new_path, buf);
#endif
    }
  free(new_path);

  return res;
}


#if (!defined HAVE___LXSTAT && (defined HAVE___LSTAT || defined HAVE__LSTAT))
#ifdef HAVE___LSTAT
int __lstat(const char *file_name, struct stat *buf)
#elif HAVE__LSTAT
int _lstat(const char *file_name, struct stat *buf)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA _lstat for file %s\n", file_name);
      PORTABLE_CALL32(lstat, stat, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical _lstat for file %s\n", file_name);
      CHECK_REAL_CALLS();
#ifdef HAVE___LSTAT
      res = real___lstat(new_path, buf);
#elif HAVE__LSTAT
      res = real__lstat(new_path, buf);
#endif
    }
  free(new_path);

  return res;
}
#endif



/***********
 * LSTAT64 *
 ***********/

#ifdef HAVE___LXSTAT64
int __lxstat64(int ver, const char *file_name, struct stat64 *buf)
#else
int lstat64(const char *file_name, struct stat64 *buf)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA lstat64 for file %s\n", file_name);
      PORTABLE_CALL64(lstat, stat, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical lstat64 for file %s\n", file_name);
      CHECK_REAL_CALLS();
#ifdef HAVE___LXSTAT64
      res = real___lxstat64(ver, new_path, buf);
#else
      res = real_lstat64(new_path, buf);
#endif
    }
  free(new_path);

  return res;
}


#if (!defined HAVE___LXSTAT64 && (defined HAVE___LSTAT64 || defined HAVE__LSTAT64))
#ifdef HAVE___LSTAT64
int __lstat64(const char *file_name, struct stat64 *buf)
#elif HAVE__LSTAT64
int _lstat64(const char *file_name, struct stat64 *buf)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA lstat64 for file %s\n", file_name);
      PORTABLE_CALL64(lstat, stat, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical lstat64 for file %s\n", file_name);
      CHECK_REAL_CALLS();
#ifdef HAVE___LSTAT64
      res = real___lstat64(new_path, buf);
#elif HAVE__LSTAT64
      res = real__lstat64(new_path, buf);
#endif
    }
  free(new_path);

  return res;
}
#endif



/***********
 * FSTATFS *
 ***********/

int fstatfs(int fd, struct statfs *buf)
{
  if ( is_orfa_fd(fd) )
    {
      int res;
      LOG(1, "ORFA fstatfs for fd %d\n", fd);
      PORTABLE_CALL32(fstatfs, statfs, fd, buf, res);
      return res;
    }
  else
    {
      LOG(1, "Classical fstatfs for fd %d\n", fd);
      CHECK_REAL_CALLS();
      return real_fstatfs(fd, buf);
    }
}


#ifdef HAVE_FSTATFS64
int fstatfs64(int fd, struct statfs64 *buf)
{
  if ( is_orfa_fd(fd) )
    {
      int res;
      LOG(1, "ORFA fstatfs64 for fd %d\n", fd);
      PORTABLE_CALL64(fstatfs, statfs, fd, buf, res);
      return res;
    }
  else
    {
      LOG(1, "Classical fstatfs64 for fd %d\n", fd);
      CHECK_REAL_CALLS();
      return real_fstatfs64(fd, buf);
    }
}
#endif



/**********
 * STATFS *
 **********/

int statfs(const char *file_name, struct statfs *buf)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA statfs for file %s\n", file_name);
      PORTABLE_CALL32(statfs, statfs, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical statfs for file %s\n", file_name);
      CHECK_REAL_CALLS();
      res = real_statfs(new_path, buf);
    }
  free(new_path);

  return res;
}


#ifdef HAVE_STATFS64
int statfs64(const char *file_name, struct statfs64 *buf)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(file_name, &new_path) )
    {
      LOG(1, "ORFA statfs64 for file %s\n", file_name);
      PORTABLE_CALL64(statfs, statfs, new_path, buf, res);
    }
  else
    {
      LOG(1, "Classical statfs64 for file %s\n", file_name);
      CHECK_REAL_CALLS();
      res = real_statfs64(new_path, buf);
    }
  free(new_path);

  return res;
}
#endif

      

/*********
 * MKNOD *
 *********/

#ifdef HAVE___XMKNOD
int __xmknod(int ver, const char *pathname, mode_t mode, dev_t *dev)
#else
int mknod(const char *pathname, mode_t mode, dev_t dev)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(pathname, &new_path) )
    {
      LOG(1, "ORFA mknod %s\n", pathname);
#ifdef HAVE___XMKNOD
      res = orfa_mknod(new_path, mode, *dev);
#else
      res = orfa_mknod(new_path, mode, dev);
#endif
    }
  else
    {
      LOG(1, "Classical mknod %s\n", pathname);
      CHECK_REAL_CALLS();
#ifdef HAVE___XMKNOD
      res = real___xmknod(ver, new_path, mode, dev);
#else
      res = real_mknod(new_path, mode, dev);
#endif
    }
  free(new_path);

  return res;
}


#if (!defined HAVE___XMKNOD && (defined HAVE___MKNOD || defined HAVE__MKNOD))
#ifdef HAVE___MKNOD
int __mknod(const char *pathname, mode_t mode, dev_t dev)
#elif HAVE__MKNOD
int _mknod(const char *pathname, mode_t mode, dev_t dev)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(pathname, &new_path) )
    {
      LOG(1, "ORFA _mknod %s\n", pathname);
      res = orfa_mknod(new_path, mode, dev);
    }
  else
    {
      LOG(1, "Classical _mknod %s\n", pathname);
      CHECK_REAL_CALLS();
#ifdef HAVE___MKNOD
      res = real___mknod(new_path, mode, dev);
#elif HAVE__MKNOD
      res = real__mknod(new_path, mode, dev);
#endif
    }
  free(new_path);

  return res;
}
#endif



/*********
 * MKDIR *
 *********/

int mkdir(const char *pathname, mode_t mode)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(pathname, &new_path) )
    {
      LOG(1, "ORFA mkdir %s\n", pathname);
      res = orfa_mkdir(new_path, mode);
    }
  else
    {
      LOG(1, "Classical mkdir %s\n", pathname);
      CHECK_REAL_CALLS();
      res = real_mkdir(new_path, mode);
    }  
  free(new_path);

  return res;
}



/*********
 * RMDIR *
 *********/

int rmdir(const char *pathname)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(pathname, &new_path) )
    {
      LOG(1, "ORFA rmdir %s\n", pathname);
      res = orfa_rmdir(new_path);
    }
  else
    {
      LOG(1, "Classical rmdir %s\n", pathname);
      CHECK_REAL_CALLS();
      res = real_rmdir(new_path);
    }
  free(new_path);

  return res;
}
  


/**********
 * UNLINK *
 **********/

int unlink(const char *pathname)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(pathname, &new_path) )
    {
      LOG(1, "ORFA unlink %s\n", pathname);
      res = orfa_unlink(new_path);
    }
  else
    {
      LOG(1, "Classical unlink %s\n", pathname);
      CHECK_REAL_CALLS();
      res = real_unlink(new_path);
    }
  free(new_path);

  return res;
}



/**********
 * RENAME *
 **********/

int rename(const char *oldpath, const char *newpath)
{
  char *new_newpath, *new_oldpath;
  int res;

  if ( is_orfa_pathname(oldpath, &new_oldpath) )
    {
      if ( is_orfa_pathname(newpath, &new_newpath) )
	{
	  LOG(1, "ORFA rename %s in %s\n", oldpath, newpath);
	  res = orfa_rename(new_oldpath, new_newpath);
	}
      else
	{
	  LOG(1, "Cross device rename impossible\n");
	  errno = EXDEV;
	  res = -1;
	}
    }
  else
    {
      if ( is_orfa_pathname(newpath, &new_newpath) )
	{
	  LOG(1, "Cross device rename impossible\n");
	  errno = EXDEV;
	  res = -1;
	}
      else
	{
	  LOG(1, "Classical rename %s in %s\n", oldpath, newpath);
	  CHECK_REAL_CALLS();
  	  res = real_rename(new_oldpath, new_newpath);
	}
    }
  free(new_oldpath);
  free(new_newpath);

  return res;
}



/********
 * LINK *
 ********/

int link(const char *oldpath, const char *newpath)
{
  char *new_newpath, *new_oldpath;
  int res;

  if ( is_orfa_pathname(oldpath, &new_oldpath) )
    {
      if ( is_orfa_pathname(newpath, &new_newpath) )
	{
	  LOG(1, "ORFA link\n");
	  res = orfa_link(new_oldpath, new_newpath);
	}
      else
	{
	  LOG(1, "Cross fs link\n");
	  errno = EXDEV;
	  res = -1;
	}
    }
  else
    {
      if ( is_orfa_pathname(newpath, &new_newpath) )
	{
	  LOG(1, "Cross fs link\n");
	  errno = EXDEV;
	  res = -1;
	}
      else
	{
	  LOG(1, "Classical link\n");
	  CHECK_REAL_CALLS();
  	  res = real_link(new_oldpath, new_newpath);
	}
    }
  free(new_oldpath);
  free(new_newpath);

  return res;
}


int symlink(const char *oldpath, const char *newpath)
{
  char *new_newpath;
  int res;

  if ( is_orfa_pathname(newpath, &new_newpath) )
    {
      LOG(1, "ORFA symlink\n");
      res = orfa_symlink(oldpath, new_newpath);
    }
  else
    {
      LOG(1, "Classical symlink\n");
      CHECK_REAL_CALLS();
      res = real_symlink(oldpath, new_newpath);
    }
  free(new_newpath);

  return res;
}


int readlink(const char *path, char *buf, size_t bufsiz)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA readlink\n");
      res = orfa_readlink(new_path, buf, bufsiz);
    }
  else
    {
      LOG(1, "Classical readlink\n");
      CHECK_REAL_CALLS();
      res = real_readlink(new_path, buf, bufsiz);
    }
  free(new_path);

  return res;
}



/**********
 * ACCESS *
 **********/

int access(const char *pathname, int mode)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(pathname, &new_path) )
    {
      LOG(1, "ORFA access %s mode %d\n", pathname, mode);
      res = orfa_access(new_path, mode);
    }
  else
    {
      LOG(1, "Classical access %s mode %d\n", pathname, mode);
      CHECK_REAL_CALLS();
      res = real_access(new_path, mode);
    }
  free(new_path);

  return res;
}


int euidaccess(const char *pathname, int mode)
{
  char *new_path;
  int res = 0;

  if ( is_orfa_pathname(pathname, &new_path) )
    {
      struct portable_stat st;

      /* check file existency and get stat */
      res = orfa_stat64(new_path, &st);
      if ( res < 0 )
	return -1;
      
      /* check access rights */
      res = access_test(context.euid, context.egid,
			&st, mode);
      if ( res < 0 )
	errno = EACCES;
    }
  else
    {
      LOG(1, "Classical euidaccess %s mode %d\n", pathname, mode);
      CHECK_REAL_CALLS();
      res = real_euidaccess(new_path, mode);
    }
  free(new_path);

  return res;
}



/*********
 * CHMOD *
 *********/

int fchmod(int fildes, mode_t mode)
{
  if ( is_orfa_fd(fildes) )
    {
      LOG(1, "ORFA fchmod\n");
      return orfa_fchmod(fildes, mode);
    }
  else
    {
      LOG(1, "Classical fchmod\n");
      CHECK_REAL_CALLS();
      return real_fchmod(fildes, mode);
    }
}


int chmod(const char *path, mode_t mode)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA chmod %s mode %d\n", path, mode);
      res = orfa_chmod(new_path, mode);
    }
  else
    {
      LOG(1, "Classical chmod %s mode %d\n", path, mode);
      CHECK_REAL_CALLS();
      res = real_chmod(new_path, mode);
    }
  free(new_path);

  return res;
}



/*********
 * CHOWN *
 *********/

int fchown(int fd, uid_t owner, gid_t group)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA fchown\n");
      return orfa_fchown(fd, owner, group);
    }
  else
    {
      LOG(1, "Classical fchown\n");
      CHECK_REAL_CALLS();
      return real_fchown(fd, owner, group);
    }  
}


int chown(const char *path, uid_t owner, gid_t group)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA chown %s owner %d group %d\n", path, owner, group);
      res = orfa_chown(new_path, owner, group);
    }
  else
    {
      LOG(1, "Classical chown %s owner %d group %d\n", path, owner, group);
      CHECK_REAL_CALLS();
      res = real_chown(new_path, owner, group);
    }
  free(new_path);

  return res;
}


int lchown(const char *path, uid_t owner, gid_t group)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA lchown %s owner %d group %d\n", path, owner, group);
      res = orfa_lchown(new_path, owner, group);
    }
  else
    {
      LOG(1, "Classical lchown %s owner %d group %d\n", path, owner, group);
      CHECK_REAL_CALLS();
      res = real_lchown(new_path, owner, group);
    }
  free(new_path);

  return res;
}



/*********
 * FCNTL *
 *********/

int fcntl(int fd, int cmd, ...)
{
  int res;

  va_list ap;
  va_start(ap, cmd);
  
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA fcntl\n");
      res = orfa_fcntl(fd, cmd, va_arg(ap, long int));
    }
  else
    {
      LOG(1, "Classical fcntl\n");
      CHECK_REAL_CALLS();
      res = real_fcntl(fd, cmd, va_arg(ap, long int));
    }

  va_end(ap);

  return res;
}


#if (defined HAVE___FCNTL || defined HAVE__FCNTL)
#ifdef HAVE___FCNTL
int __fcntl(int fd, int cmd, ...)
#elif HAVE__FCNTL
int _fcntl(int fd, int cmd, ...)
#endif
{
  int res;

  va_list ap;
  va_start(ap, cmd);
  
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _fcntl\n");
      res = orfa_fcntl(fd, cmd, va_arg(ap, long int));
    }
  else
    {
      LOG(1, "Classical _fcntl\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___FCNTL
      res = real___fcntl(fd, cmd, va_arg(ap, long int));
#elif HAVE__FCNTL
      res = real__fcntl(fd, cmd, va_arg(ap, long int));
#endif
    }

  va_end(ap);

  return res;
}
#endif


#ifdef HAVE___LIBC_FCNTL
int __libc_fcntl(int fd, int cmd, ...)
{
  int res;

  va_list ap;
  va_start(ap, cmd);
  
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_fcntl\n");
      res = orfa_fcntl(fd, cmd, va_arg(ap, long int));
    }
  else
    {
      LOG(1, "Classical __libc_fcntl\n");
      CHECK_REAL_CALLS();
      res = real___libc_fcntl(fd, cmd, va_arg(ap, long int));
    }

  va_end(ap);

  return res;
}
#endif



/*********
 * FLOCK *
 *********/

int flock(int fd, int operation)
{
  if ( is_orfa_fd(fd) )
    {
//      LOG(1, "ORFA flock\n");
//      return orfa_flock(fd, operation);
      ERROR("Flock support broken\n");
      errno = ENOSYS;
      return -1;
    }
  else
    {
      LOG(1, "Classical flock\n");
      CHECK_REAL_CALLS();
      return real_flock(fd, operation);
    }
}



/**********
 * FCHDIR *
 **********/

int fchdir(int fd)
{
   CHECK_REAL_CALLS();
  
 if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA fchdir\n");
      return orfa_fchdir(fd);
    }
  else
    {
      int res;
      LOG(1, "Classical chdir\n");
      res = real_fchdir(fd);
      if ( !res )
	unorfa_chdir();

      return res;
    }
}



/*********
 * CHDIR *
 *********/

int chdir(const char *path)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA chdir\n");
      res = orfa_chdir(new_path);
    }
  else
    {
      LOG(1, "Classical chdir\n");
      CHECK_REAL_CALLS();
      res = real_chdir(new_path);
      if ( !res )
	unorfa_chdir();
    }
  free(new_path);

  return res;  
}



/**********
 * GETCWD *
 **********/

char *getcwd(char *buf, size_t size)
{
  if ( is_orfa_pwd() )
    {
      LOG(1, "Getting ORFA pwd\n");
      return orfa_getcwd(buf, size);
    }
  else
    {
      LOG(1, "Getting classical pwd\n");
      CHECK_REAL_CALLS();
      return real_getcwd(buf, size);
    }
}



/***********
 * OPENDIR *
 ***********/

DIR *opendir(const char *name)
{
  char *new_path;
  void *res;

  if ( is_orfa_pathname(name, &new_path) )
    {
      LOG(1, "ORFA opendir\n");
      res = (DIR*)orfa_opendir(new_path);
    }
  else
    {
      LOG(1, "Classical opendir\n");
      CHECK_REAL_CALLS();
      res = real_opendir(new_path);
    }
  free(new_path);

  return res;
}


int closedir(DIR *dir)
{
  if ( is_orfa_dir((orfa_dir_t*)dir) )
    {
      LOG(1, "ORFA closedir\n");
      return orfa_closedir((orfa_dir_t*)dir);
    }
  else
    {
      LOG(1, "Classical closedir\n");
      CHECK_REAL_CALLS();
      return real_closedir(dir);
    }
}



/***********
 * READDIR *
 ***********/

struct dirent *readdir(DIR *dir)
{
  if ( is_orfa_dir((orfa_dir_t*)dir) )
    {
      LOG(1, "ORFA readdir\n");
      return orfa_readdir((orfa_dir_t*)dir);
    }
  else
    {
      LOG(1, "Classical readdir\n");
      CHECK_REAL_CALLS();
      return real_readdir(dir);
    }
}


#ifdef HAVE_READDIR_R
int readdir_r(DIR *dir, struct dirent *entry, struct dirent **result)
{
  if ( is_orfa_dir((orfa_dir_t*)dir) )
    {
      LOG(1, "ORFA readdir_r\n");
      return orfa_readdir_r((orfa_dir_t*)dir, entry, result);
    }
  else
    {
      LOG(1, "Classical readdir\n");
      CHECK_REAL_CALLS();
      return real_readdir_r(dir, entry, result);
    }
}
#endif


#ifdef HAVE_READDIR64
struct dirent64 *readdir64(DIR *dir)
{
  if ( is_orfa_dir((orfa_dir_t*)dir) )
    {
      LOG(1, "ORFA readdir64\n");
      return orfa_readdir64((orfa_dir_t*)dir);
    }
  else
    {
      LOG(1, "Classical readdir64\n");
      CHECK_REAL_CALLS();
      return real_readdir64(dir);
    }
}
#endif


#ifdef HAVE_READDIR64_R
int readdir64_r(DIR *dir, struct dirent64 *entry, struct dirent64 **result)
{
  if ( is_orfa_dir((orfa_dir_t*)dir) )
    {
      LOG(1, "ORFA readdir64_r\n");
      return orfa_readdir64_r((orfa_dir_t*)dir, entry, result);
    }
  else
    {
      LOG(1, "Classical readdir64\n");
      CHECK_REAL_CALLS();
      return real_readdir64_r(dir, entry, result);
    }
}
#endif



/***********
 * SEEKDIR *
 ***********/

void seekdir(DIR *dir, off_t offset)
{
  if ( is_orfa_dir((orfa_dir_t*)dir) )
    {
      LOG(1, "ORFA Seekdir\n");
      orfa_seekdir((orfa_dir_t*)dir, offset);
    }
  else
    {
      LOG(1, "Classical Seekdir\n");
      CHECK_REAL_CALLS();
      real_seekdir(dir, offset);
    }
}


off_t telldir(DIR *dir)
{
  if ( is_orfa_dir((orfa_dir_t*)dir) )
    {
      LOG(1, "ORFA Telldir\n");
      return orfa_telldir((orfa_dir_t*)dir);
    }
  else
    {
      LOG(1, "Classical Telldir\n");
      CHECK_REAL_CALLS();
      return real_telldir(dir);
    }
}


void rewinddir(DIR *dir)
{
  if ( is_orfa_dir((orfa_dir_t*)dir) )
    {
      LOG(1, "ORFA Rewinddir\n");
      orfa_rewinddir((orfa_dir_t*)dir);
    }
  else
    {
      LOG(1, "Classical Rewinddir\n");
      CHECK_REAL_CALLS();
      real_rewinddir(dir);
    }
}



/*********
 * UTIME *
 *********/

int utime(const char *filename, const struct utimbuf *buf)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(filename, &new_path) )
    {
      LOG(1, "ORFA utime\n");
      res = orfa_utime(new_path, buf);
    }
  else
    {
      LOG(1, "Classical utime\n");
      CHECK_REAL_CALLS();
      res = real_utime(new_path, buf);
    }
  free(new_path);

  return res;
}



/**********
 * UTIMES *
 **********/

#if (defined __GLIBC_MINOR__) && (__GLIBC_MINOR__ < 2 )
int utimes(const char *filename, struct timeval *tvp)
#else
int utimes(const char *filename, const struct timeval *tvp)
#endif
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(filename, &new_path) )
    {
      LOG(1, "ORFA utime\n");
      if ( tvp != NULL )
	{
	  struct utimbuf buf;
	  buf.actime = tvp[0].tv_sec;
	  buf.modtime = tvp[1].tv_sec; 
	  res = orfa_utime(new_path, &buf);
	}
      else
	res = orfa_utime(new_path, NULL);
    }
  else
    {
      LOG(1, "Classical utimes\n");
      CHECK_REAL_CALLS();
      res = real_utimes(new_path, tvp);
    }
  free(new_path);

  return res;
}



/*********
 * FSYNC *
 *********/

int fsync(int fd)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA fsync\n");
      return orfa_fsync(fd);
    }
  else
    {
      LOG(1, "Classical fsync\n");
      CHECK_REAL_CALLS();
      return real_fsync(fd);
    }
}


#if (defined HAVE___FSYNC || defined HAVE__FSYNC)
#ifdef HAVE___FSYNC
int __fsync(int fd)
#elif HAVE__FSYNC
int _fsync(int fd)
#endif
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA _fsync\n");
      return orfa_fsync(fd);
    }
  else
    {
      LOG(1, "Classical _fsync\n");
      CHECK_REAL_CALLS();
#ifdef HAVE___FSYNC
      return real___fsync(fd);
#elif HAVE__FSYNC
      return real__fsync(fd);
#endif
    }
}
#endif


#ifdef HAVE___LIBC_FSYNC
int __libc_fsync(int fd)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA __libc_fsync\n");
      return orfa_fsync(fd);
    }
  else
    {
      LOG(1, "Classical __libc_fsync\n");
      CHECK_REAL_CALLS();
      return real___libc_fsync(fd);
    }
}
#endif


#ifdef HAVE_FDATASYNC
int fdatasync(int fd)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA fdatasync\n");
      return orfa_fdatasync(fd);
    }
  else
    {
      LOG(1, "Classical fdatasync\n");
      CHECK_REAL_CALLS();
      return real_fdatasync(fd);
    }
}
#endif



/*******
 * DUP *
 *******/

int dup(int oldfd)
{
  if ( is_orfa_fd(oldfd) )
    {
      int newfd;
      LOG(1, "ORFA dup of %d\n", oldfd);
      CHECK_REAL_CALLS();  
      newfd = real_dup(oldfd);

      /* update ORFA fd structures */
      if ( newfd >= 0 )
	orfa_dup(oldfd, newfd);

      return newfd;
    }
  else
    {
      LOG(1, "Classical dup of %d\n", oldfd);
      CHECK_REAL_CALLS();  
      return real_dup(oldfd);
    }
}


#if (defined HAVE___DUP || defined HAVE__DUP)
#ifdef HAVE___DUP
int __dup(int oldfd)
#elif HAVE__DUP
int _dup(int oldfd)
#endif
{
  if ( is_orfa_fd(oldfd) )
    {
      int newfd;
      LOG(1, "ORFA _dup of %d\n", oldfd);
      CHECK_REAL_CALLS();  
#ifdef HAVE___DUP
      newfd = real___dup(oldfd);
#elif HAVE__DUP
      newfd = real__dup(oldfd);
#endif

      /* update ORFA fd structures */
      if ( newfd >= 0 )
	orfa_dup(oldfd, newfd);

      return newfd;
    }
  else
    {
      LOG(1, "Classical _dup of %d\n", oldfd);
      CHECK_REAL_CALLS();  
#ifdef HAVE___DUP
      newfd = real___dup(oldfd);
#elif HAVE__DUP
      newfd = real__dup(oldfd);
#endif
    }
}
#endif



/********
 * DUP2 *
 ********/

int dup2(int oldfd, int newfd)
{
  int res;

  if ( is_orfa_fd(oldfd) )
    {
      LOG(1, "ORFA dup2 of %d in %d\n", oldfd, newfd);
      CHECK_REAL_CALLS();
      res = real_dup2(oldfd, newfd);

      if ( res < 0 )
	return res;

      /* check if we replace an old ORFA fd */
      if ( is_orfa_fd(newfd) )
	orfa_close(newfd);

      /* update ORFA fd structures */
      orfa_dup(oldfd, newfd);
    }
  else
    {
      LOG(1, "Classical dup2 of %d in %d\n", oldfd, newfd);
      CHECK_REAL_CALLS();
      res = real_dup2(oldfd, newfd);

      if ( res < 0 )
	return res;

      /* check if we replace an old ORFA fd */
      if ( is_orfa_fd(newfd) )
	orfa_close(newfd);
    }  
  return res;
}


#if (defined HAVE___DUP2 || defined HAVE__DUP2)
#ifdef HAVE___DUP2
int __dup2(int oldfd, int newfd)
#elif HAVE__DUP2
int _dup2(int oldfd, int newfd)
#endif
{
  int res;

  if ( is_orfa_fd(oldfd) )
    {
      LOG(1, "ORFA _dup2 of %d in %d\n", oldfd, newfd);
      CHECK_REAL_CALLS();
#ifdef HAVE___DUP2
      res = real___dup2(oldfd, newfd);
#elif HAVE__DUP2
      res = real__dup2(oldfd, newfd);
#endif

      if ( res < 0 )
	return res;

      /* check if we replace an old ORFA fd */
      if ( is_orfa_fd(newfd) )
	orfa_close(newfd);

      /* update ORFA fd structures */
      orfa_dup(oldfd, newfd);
    }
  else
    {
      LOG(1, "Classical _dup2 of %d in %d\n", oldfd, newfd);
      CHECK_REAL_CALLS();
#ifdef HAVE___DUP2
      res = real___dup2(oldfd, newfd);
#elif HAVE__DUP2
      res = real__dup2(oldfd, newfd);
#endif

      if ( res < 0 )
	return res;

      /* check if we replace an old ORFA fd */
      if ( is_orfa_fd(newfd) )
	orfa_close(newfd);
    }  
  return res;
}
#endif



/************
 * SENDFILE *
 ************/

#ifdef HAVE_SENDFILE
ssize_t sendfile(int out_fd, int in_fd, off_t *offset, size_t count)
{
  if ( is_orfa_fd(in_fd)  ||  is_orfa_fd(out_fd) )
    {
      int res;
      off64_t offset64;
      LOG(1, "ORFA sendfile\n");
      res = orfa_sendfile64(out_fd, in_fd, &offset64, count);
      *offset = (off64_t)offset64;
      return res;
    }
  else
    {
      LOG(1, "Classical sendfile\n");
      return real_sendfile(out_fd, in_fd, offset, count);
    }
}
#endif


#ifdef HAVE_SENDFILE64
ssize_t sendfile64(int out_fd, int in_fd, off64_t *offset, size_t count)
{
  if ( is_orfa_fd(in_fd)  ||  is_orfa_fd(out_fd) )
    {
      LOG(1, "ORFA sendfile64\n");
      return orfa_sendfile64(out_fd, in_fd, offset, count);
    }
  else
    {
      LOG(1, "Classical sendfile64\n");
      return real_sendfile64(out_fd, in_fd, offset, count);
    }
}
#endif



/********
 * MMAP *
 ********/

void *mmap(void *start, size_t length, int prot, int flags, int fd, off_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA mmap\n");
      return orfa_mmap64(start, length, prot, flags, fd, offset);
    }
  else
    {
      FAST_LOG(1, "Classical mmap\n");
      CHECK_REAL_CALLS();
      return real_mmap(start, length, prot, flags, fd, offset);
    }
}


#ifdef HAVE_MMAP64
caddr_t mmap64(void *start, size_t length, int prot, int flags, int fd, off64_t offset)
{
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA mmap64\n");
      return orfa_mmap64(start, length, prot, flags, fd, offset);
    }
  else
    {
      LOG(1, "Classical mmap64\n");
      CHECK_REAL_CALLS();
      return real_mmap64(start, length, prot, flags, fd, offset);
    }
}
#endif


void *mremap(void *old_address, size_t old_size, size_t new_size, unsigned long flags)
{
  void *res; 
  
  /* if ( is_orfa_mmap(old_address) ) */
  LOG(1, "Classical mremap\n");

  CHECK_REAL_CALLS();
  res = real_mremap(old_address, old_size, new_size, flags);

#if REGCACHE_ENABLED
  if ( res != (void*)-1 )
    orfa_clear_interval(old_address, old_size);
#endif

  return res;
}


int munmap(void *start, size_t length)
{
  int res;

  /* if ( is_orfa_mmap(old_address) ) */
  FAST_LOG(1, "Classical munmap\n");

  CHECK_REAL_CALLS();
  res = real_munmap(start, length);

#if REGCACHE_ENABLED
  if ( res != -1 )
    orfa_clear_interval(start, length);  
#endif

  return res;
}



/*********
 * MSYNC *
 *********/

int msync(const void *start, size_t length, int flags)
{
  /* if ( is_orfa_mmap(old_address) ) */
  LOG(1, "Classical msync\n");

  CHECK_REAL_CALLS();
  return real_msync(start, length, flags);
}


#if (defined HAVE___MSYNC || defined HAVE__MSYNC)
#ifdef HAVE___MSYNC
int __msync(const void *start, size_t length, int flags)
#elif HAVE__MSYNC
int _msync(const void *start, size_t length, int flags)
#endif
{
  /* if ( is_orfa_mmap(old_address) ) */
  LOG(1, "Classical msync\n");

  CHECK_REAL_CALLS();
#ifdef HAVE___MSYNC
  return real___msync(start, length, flags);
#elif HAVE__MSYNC
  return real__msync(start, length, flags);
#endif
}
#endif


#ifdef HAVE___LIBC_MSYNC
int __libc_msync(const void *start, size_t length, int flags)
{
  /* if ( is_orfa_mmap(old_address) ) */
  LOG(1, "Classical msync\n");

  CHECK_REAL_CALLS();
  return real___libc_msync(start, length, flags);
}
#endif



/********
 * EXEC *
 ********/

/* ETXTBSY may not occur since we actually exec a local copy
 * of the remote file.
 */

int execve(const  char  *filename,
	   char *const argv[],
	   char *const envp[])
{
  char *new_path;
  int res;
  char **new_envp=NULL;
  int len = 0;

  PROCESS_LOCK();

  orfa_before_exec((char**)envp, &len, &new_envp);
  
  if ( is_orfa_pathname(filename, &new_path) )
    {
      LOG(1, "ORFA execve of %s\n", filename);
      res = orfa_execve(new_path, argv, new_envp);
    }
  else
    {
      LOG(1, "Classical execve of %s\n", filename);
      CHECK_REAL_CALLS();
      res = real_execve(new_path, argv, new_envp);
    }
  LOG(3, "Failed on error %d\n", errno);

  orfa_after_exec_failed(len, new_envp);

  PROCESS_UNLOCK();
  
  free(new_path);
  return res;
}


#ifdef HAVE_FEXECVE

/* fexecve is implemented in the Glibc as
 * __execve("/proc/self/fd/xxx", ...) which has the expected
 * behavior.
 * We get a local copy of the remote file and use fexecve on it.
 */

/* ETXTBSY may not occur since we actually exec a local copy
 * of the remote file.
 */

int fexecve(int fd, char *const argv[], char *const envp[])
{
  int res;
  char **new_envp=NULL;
  int len = 0;

  PROCESS_LOCK();
  
  orfa_before_exec((char**)envp, &len, &new_envp);
  
  if ( is_orfa_fd(fd) )
    {
      LOG(1, "ORFA fexecve of %d\n", fd);
      res = orfa_fexecve(fd, argv, new_envp);
    }
  else
    {
      LOG(1, "Classical fexecve of %d\n", fd);
      CHECK_REAL_CALLS();
      res = real_fexecve(fd, argv, new_envp);
    }
  LOG(3, "Failed on error %d\n", errno);

  orfa_after_exec_failed(len, new_envp);

  PROCESS_UNLOCK();
  
  return res;
}
#endif /* HAVE_FEXECVE */


int execv(const char *path, char *const argv[])
{
  return execve(path, argv, environ);
}


/* execvp is a call to exec => we don't need to redefine it */



/*********
 * EXECL *
 *********/

#include <stdarg.h>

int execle(const char *path, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  const char *const *envp;
  unsigned int i;
  va_list args;
  argv[0] = arg;
  
  va_start (args, arg);

  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
        {
          const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));
 
//#ifndef _STACK_GROWS_UP
          if ((char *) nptr + argv_max == (char *) argv)
            {
              /* Stack grows down.  */
              argv = (const char **) memcpy (nptr, argv, i);
              argv_max += i;
            }
          else
//#endif
//#ifndef _STACK_GROWS_DOWN
            if ((char *) argv + i == (char *) nptr)
            /* Stack grows up.  */
            argv_max += i;
          else
//#endif
            /* We have a hole in the stack.  */
            argv = (const char **) memcpy (nptr, argv, i);
        }
 
      argv[i] = va_arg (args, const char *);
    }
 
  envp = va_arg (args, const char *const *);
  va_end (args);
 
  return execve (path, (char *const *) argv, (char *const *) envp);  
}


int execl(const char *path, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  unsigned int i;
  va_list args;
 
  argv[0] = arg;
 
  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
        {
          const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));
 
//#ifndef _STACK_GROWS_UP
          if ((char *) nptr + argv_max == (char *) argv)
            {
              /* Stack grows down.  */
              argv = (const char **) memcpy (nptr, argv, i);
              argv_max += i;
            }
          else
//#endif
//#ifndef _STACK_GROWS_DOWN
            if ((char *) argv + i == (char *) nptr)
            /* Stack grows up.  */
            argv_max += i;
          else
//#endif
            /* We have a hole in the stack.  */
            argv = (const char **) memcpy (nptr, argv, i);
        }
 
      argv[i] = va_arg (args, const char *);
    }
  va_end (args);
 
  return execv (path, (char *const *) argv);
}


/* execlp is a call to execvp => we don't need to redefine it */



/********
 * FORK *
 ********/

pid_t fork()
{
  pid_t res;

  PROCESS_LOCK();
  fork_calls++;

  /* if inside another fork, do nothing but fork */
  if ( fork_calls > 1 ) {
    /* check real_calls already done but other fork */
    res = real_fork();    
    goto done;
  }

  orfa_before_fork();

  /* now fork */
  CHECK_REAL_CALLS();
  res = real_fork();

  /* error, parent or child ? */
  if ( res < 0 )
    orfa_after_fork_failed();
  else if ( res )
    orfa_after_fork_parent();
  else
    orfa_after_fork_child();

 done:
  fork_calls--;
  PROCESS_UNLOCK();

  return res;
}


#if (defined HAVE___FORK || defined HAVE__FORK)
#ifdef HAVE___FORK
pid_t __fork()
#elif HAVE__FORK
pid_t _fork()
#endif
{
  pid_t res;

  PROCESS_LOCK();
  fork_calls++;

  /* if inside another fork, do nothing but fork */
  if ( fork_calls > 1 ) {
    /* check real_calls already done but other fork */
#ifdef HAVE___FORK
    res = real___fork();
#elif HAVE__FORK
    res = real__fork();
#endif
    goto done;
  }

  orfa_before_fork();

  /* now fork */
  CHECK_REAL_CALLS();
#ifdef HAVE___FORK
  res = real___fork();
#elif HAVE__FORK
  res = real__fork();
#endif

  /* error, parent or child ? */
  if ( res < 0 )
    orfa_after_fork_failed();
  else if ( res )
    orfa_after_fork_parent();
  else
    orfa_after_fork_child();

 done:
  fork_calls--;
  PROCESS_UNLOCK();

  return res;
}
#endif


#ifdef HAVE___LIBC_FORK
pid_t __libc_fork()
{
  pid_t res;

  PROCESS_LOCK();
  fork_calls++;

  /* if inside another fork, do nothing but fork */
  if ( fork_calls > 1 ) {
    /* check real_calls already done but other fork */
    res = real___libc_fork();    
    goto done;
  }

  orfa_before_fork();

  /* now fork */
  CHECK_REAL_CALLS();
  res = real___libc_fork();

  /* error, parent or child ? */
  if ( res < 0 )
    orfa_after_fork_failed();
  else if ( res )
    orfa_after_fork_parent();
  else
    orfa_after_fork_child();

 done:
  fork_calls--;
  PROCESS_UNLOCK();

  return res;
}
#endif



/*********
 * VFORK *
 *********/

/* Because of inherent full VM sharing in vfork,
 * we cannot use a real_vfork() symbol.
 *
 * Actually, vfork() shares the stack between processes.
 * Thus it is impossible to let the child quit the
 * original vfork() and come back in our vfork()
 * because it could corrupt the parent stack.
 *
 * We just use our fork() which ensures compatibility
 * despite the lost of vfork's performance gain. */

pid_t vfork()
{
  return fork();
}


#if (defined HAVE___VFORK || defined HAVE__VFORK)
#ifdef HAVE___VFORK
pid_t __vfork()
#elif HAVE__VFORK
pid_t _vfork()
#endif
{
#ifdef HAVE___FORK
  return __fork();
#elif HAVE__FORK
  return _fork();
#endif
}
#endif



/***********
 * __CLONE *
 ***********/

#ifdef MULTITHREAD_SAFE_ENABLED
#include <sched.h>
int __clone(int (*fn) (void *arg),
	    void *child_stack,
	    int flags,
	    void *arg)
{
  int res;

  if ( (flags & (CLONE_VM|CLONE_FILES|CLONE_FS))
       != (CLONE_VM|CLONE_FILES|CLONE_FS) ) {
    /* CLONE_FILES shares file table
     * CLONE_FS shares FS root, umask and pwd
     * These two things are emulated in VM
     * and thus must be used with CLONE_VM,
     * and reciprocally.
     *
     * Actually CLONE_FILES may be used without CLONE_VM
     * since process are not supposed to read random
     * file descriptor.
     * We still disable this ability for now.
     */
    
    errno = ENOSYS;
    FAST_ERROR("CLONE_FILES/FS/VM must be used together.\n");
    return -1;
  }
    
  if ( flags&CLONE_VFORK ) {
    /* CLONE_VFORK is not supported for the same reason
     * than vfork is emulated with fork */
    FAST_LOG(3, "Cannot clone with CLONE_VFORK. Disabling it.\n");
    flags &= ~CLONE_VFORK;
  }

  if ( flags& CLONE_VM ) {
    /* This seems to be a thread creation. */
    CHECK_REAL_CALLS();
    res = real___clone(fn, child_stack, flags, arg);
  } else {
    /* This seems to be a fork.*/
    /* Disabled for now because we have to check some stack overflow */
    errno = ENOSYS;
    return -1;
  }

  return res;
}
#endif /* MULTITHREAD_SAFE_ENABLED */



/********
 * SBRK *
 ********/

#if ( defined REGCACHE_ENABLED  &&  defined MALLOC_ENABLED )
extern void *__sbrk(ptrdiff_t increment);

void *sbrk(ptrdiff_t increment)
{
  void *res;

  FAST_LOG(1, "Sbrk\n");
   
  /* We cannot use a "real_sbrk" defined using dlfcn
     because dlfcn need sbrk for internal malloc.
     Hopefully, in a few centuries, mmap will have made
     sbrk disappeared */
  res = __sbrk(increment);

  if ( increment < 0  &&  res != (void*)-1 )
    orfa_clear_interval(res, -increment);

  return res;
}
#endif /* REGCACHE_ENABLED && MALLOC_ENABLED */



/********
 * EUID *
 ********/

int setuid(uid_t new_uid)
{
  int res = real_setuid(new_uid);
  if ( !res )
    orfa_update_euid(new_uid);
  return res;
}


int seteuid(uid_t new_euid)
{
  int res = real_seteuid(new_euid);
  if ( !res )
    orfa_update_euid(new_euid);
  return res;
}


int setreuid(uid_t new_ruid, uid_t new_euid)
{
  int res = real_setreuid(new_ruid, new_euid);
  if ( !res )
    orfa_update_euid(new_euid);
  return res;
}


int setresuid(uid_t new_ruid, uid_t new_euid, uid_t new_suid)
{
  int res = real_setresuid(new_ruid, new_euid, new_suid);
  if ( !res )
    orfa_update_euid(new_euid);
  return res;
}



/********
 * EGID *
 ********/

int setgid(gid_t new_gid)
{
  int res = real_setgid(new_gid);
  if ( !res )
    orfa_update_egid(new_gid);
  return res;
}


int setegid(gid_t new_egid)
{
  int res = real_setegid(new_egid);
  if ( !res )
    orfa_update_egid(new_egid);
  return res;
}


int setregid(gid_t new_rgid, gid_t new_egid)
{
  int res = real_setregid(new_rgid, new_egid);
  if ( !res )
    orfa_update_egid(new_egid);
  return res;
}


int setresgid(gid_t new_rgid, gid_t new_egid, gid_t new_sgid)
{
  int res = real_setresgid(new_rgid, new_egid, new_sgid);
  if ( !res )
    orfa_update_egid(new_egid);
  return res;
}



/*********
 * UMASK *
 *********/

mode_t umask(mode_t umask) {
  orfa_update_umask(umask);
  return real_umask(umask);
  /* always succeed */
}



/************
 * GETDENTS *
 ************/

/* getdents is a problematic function because the struct dirent
 * it uses looks different from the one used by readdir.
 * getdents remains on linux/dirent.h which is the internal
 * direntry structure, while readdir remains on dirent.h which
 * is user level glibc-provided direntry structure.
 * getdents is currently not exported by glibc, we do not
 * implement a ORFA version.
 */

#ifdef HAVE_GETDENTS
int getdents(unsigned int fd, void *dirp, unsigned int count)
{
  if ( is_orfa_fd(fd) )
    {
      ERROR("Getdents deprecated. No ORFA implementation.\n");
      errno = ENOSYS;
      return -1;
    }
  else
    {
      LOG(1, "Classical getdents\n");
      CHECK_REAL_CALLS();
      return real_getdents(fd, dirp, count);
    }
}
#endif


/* getdents64 is currently exported by the libc and it seems
 * that the struct dirent64 it uses is the same than the one
 * returned by readdir64.
 * We choose to not implement getdents64 before knowing exactly
 * whether readdir(64)/getdents(64) are supposed to use same
 * dirent(64) structure or not.
 */

#ifdef HAVE_GETDENTS64
int getdents64(unsigned int fd, void *dirp, unsigned int count)
{
  CHECK_REAL_CALLS();
  
  if ( is_orfa_fd(fd) )
    {
      ERROR("Getdents64 deprecated. No ORFA implementation.\n");
      errno = ENOSYS;
      return -1;
    }
  else
    {
      LOG(1, "Classical getdents64\n");
      return real_getdents64(fd, dirp, count);
    }
}
#endif


/*******
 * ACL *
 *******/

#ifdef HAVE_LIBACL

#include <sys/acl.h>


acl_t acl_get_fd(int fd)
{
  if ( is_orfa_fd(fd) ) 
    {
      LOG(1, "ORFA acl_get_fd for %d\n", fd);
      return orfa_acl_get_fd(fd);
    }
  else
    {
      LOG(1, "Classical acl_get_fd for %d\n", fd);
      CHECK_REAL_CALLS();
      return real_acl_get_fd(fd);
    }
}


int acl_set_fd(int fd, acl_t acl)
{
  if ( is_orfa_fd(fd) ) 
    {
      LOG(1, "ORFA acl_set_fd for %d\n", fd);
      return orfa_acl_set_fd(fd, acl);
    }
  else
    {
      LOG(1, "Classical acl_set_fd for %d\n", fd);
      CHECK_REAL_CALLS();
      return real_acl_set_fd(fd, acl);
    }
}


int acl_extended_fd(int fd)
{
  if ( is_orfa_fd(fd) ) 
    {
      LOG(1, "ORFA acl_extended_fd for %d\n", fd);
      return orfa_acl_extended_fd(fd);
    }
  else
    {
      LOG(1, "Classical acl_extended_fd for %d\n", fd);
      CHECK_REAL_CALLS();
      return real_acl_extended_fd(fd);
    }
}


acl_t acl_get_file(const char *path, acl_type_t type)
{
  char *new_path;
  acl_t res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA acl_get_file for %s\n", path);
      res = orfa_acl_get_file(new_path, type);
    }
  else
    {
      LOG(1, "Classical acl_get_file for %s\n", path);
      CHECK_REAL_CALLS();
      res = real_acl_get_file(new_path, type);
    }

  free(new_path);
  return res;
}


int acl_set_file(const char *path, acl_type_t type, acl_t acl)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA acl_set_file for %s\n", path);
      res = orfa_acl_set_file(new_path, type, acl);
    }
  else
    {
      LOG(1, "Classical acl_set_file for %s\n", path);
      CHECK_REAL_CALLS();
      res = real_acl_set_file(new_path, type, acl);
    }

  free(new_path);
  return res;
}


int acl_extended_file(const char *path)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA acl_extended_file for %s\n", path);
      res = orfa_acl_extended_file(new_path);
    }
  else
    {
      LOG(1, "Classical acl_extended_file for %s\n", path);
      CHECK_REAL_CALLS();
      res = real_acl_extended_file(new_path);
    }

  free(new_path);
  return res;
}


int acl_delete_def_file(const char *path)
{
  char *new_path;
  int res;

  if ( is_orfa_pathname(path, &new_path) )
    {
      LOG(1, "ORFA acl_delete_def_file for %s\n", path);
      res = orfa_acl_delete_def_file(new_path);
    }
  else
    {
      LOG(1, "Classical acl_delete_def_file for %s\n", path);
      CHECK_REAL_CALLS();
      res = real_acl_delete_def_file(new_path);
    }

  free(new_path);
  return res;
}
#endif /* HAVE_LIBACL */
