/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __COMMON_NETWORK_H__
#define __COMMON_NETWORK_H__


/* check if NETWORK is well defined */
#if ( NETWORK_TCP + NETWORK_UDP + NETWORK_GM + NETWORK_BIP + NETWORK_FAKE != 1 )
# error "You must define one communication type !"
#endif /* TCP or UDP or GM or BIP or FAKE */



/*******
 * TCP *
 *******/
#if ( defined NETWORK_TCP )

#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/tcp.h>

/* a connection is a socket */
struct connection {
  int socket;
};

/* the server's port */
#define SERVER_PORT 8256

/* magic number to check server */
#define NETWORK_MAGIC 0x26111978

/* deal with Nagle algorithm which delays packet
 * sending to prevent lots of small packets */
#define SOCKET_TCP_NODELAY(s) \
do {                          \
 int status = 1;              \
 setsockopt(s,                \
	    IPPROTO_TCP,      \
	    TCP_NODELAY,      \
	    &status,          \
	    sizeof(int));     \
} while ( 0 )



/*******
 * UDP *
 *******/
#elif ( defined NETWORK_UDP )

#include <arpa/inet.h>
/* a connection is a sockaddr_in to server */
struct connection {
  struct sockaddr_in ski;
};

/* the server's port */
#define SERVER_PORT 8256

/* UDP max message size */
#define MAX_MESSAGE_LEN 500



/******
 * GM *
 ******/
#elif ( defined NETWORK_GM )

#include "gm.h"

/* a connection is a port_id and a node_id */
struct connection {
  gm_u16_t node_id;
  gm_u16_t port_id;
};

#define SERVER_PORT_ID 2

#define REQUEST_PRIORITY GM_HIGH_PRIORITY
#define REQUEST_RECV_EVENT GM_HIGH_RECV_EVENT
#define DATA_PRIORITY GM_LOW_PRIORITY
#define DATA_RECV_EVENT GM_RECV_EVENT

#define GM_SIZE_MAX 32
  
#define ANSWER_DATA_SIZE (GM_SIZE_MAX-1)


extern unsigned int preprovided_buffer_size;

extern inline int REQUEST_DATA_SIZE_OF_LENGTH(int length) {
 if ( length <= PREPROVIDED_BUFFER_LENGTH )
    return preprovided_buffer_size;
  else
    return gm_min_size_for_length(length);
}


/*******
 * BIP *
 *******/
#elif ( defined NETWORK_BIP )

/* a connection is a BIP node number */
struct connection {
  int node;
};

#define WORD_SIZE 4
#define ALIGN_TO_WORD_SIZE(x) x = ((x+WORD_SIZE-1) & ~(WORD_SIZE-1))

/* check that PREPROVIDED_BUFFER_LENGTH is aligned */
#if ( PREPROVIDED_BUFFER_LENGTH%WORD_SIZE != 0 )
#error PREPROVIDED_BUFFER_LENGTH must be aligned on word size !
#endif

/* Client's request and server's answer */
#define PROTOCOL_TAG 0
/* Data sent by the server with the answer */
#define ANSWER_DATA_TAG 1


/* Data sent by the client to the server */
#define REQUEST_DATA_TAG_BEGIN 1
#define REQUEST_BIG_DATA_TAG_BEGIN 2

#if (defined BIP_TAG_ECONOMY_ENABLED) && (defined FIXED_SERVER_NODE) && (FIXED_SERVER_NODE == 0)
#define NODE_TO_TAG(node) (node-1)
#else
#define NODE_TO_TAG(node) (node)
#endif

#ifdef BIP_RECVFILE_ENABLED
#define REQUEST_DATA_TAG_FOR_NODE(x) REQUEST_DATA_TAG_BEGIN
#define REQUEST_BIG_DATA_TAG_FOR_NODE(x) (REQUEST_BIG_DATA_TAG_BEGIN+NODE_TO_TAG(x))
#else
#define REQUEST_DATA_TAG_FOR_NODE(x) (REQUEST_DATA_TAG_BEGIN+2*NODE_TO_TAG(x))
#define REQUEST_BIG_DATA_TAG_FOR_NODE(x) (REQUEST_BIG_DATA_TAG_BEGIN+2*NODE_TO_TAG(x))
#endif



/********
 * FAKE *
 ********/
#elif ( defined NETWORK_FAKE )

/* a connection is nothing */
struct connection {
};



#endif /* TCP or UDP or GM or BIP or FAKE */



#endif /* __COMMON_NETWORK_H__ */
