/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __PROTOCOL_H__
#define __PROTOCOL_H__

#include <fcntl.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdint.h>


#include "orfa_common_portable.h"


typedef uint64_t orfa_fd_id_t;
typedef uint64_t orfa_dir_id_t;
typedef uint64_t orfa_connection_id_t;
typedef uint64_t orfa_tag_t;
typedef uint64_t orfa_pid_t;
typedef uint16_t orfa_error_t;
typedef uint64_t orfa_off_t;
typedef int64_t orfa_soff_t;
typedef uint64_t orfa_time_t;
typedef uint32_t orfa_uid_t;
typedef uint32_t orfa_gid_t;
typedef uint64_t orfa_dev_t;
typedef uint32_t orfa_oflags_t;
typedef uint32_t orfa_mode_t;
typedef uint8_t orfa_access_mode_t;
typedef uint8_t orfa_flock_op_t;
typedef uint8_t orfa_bool_t;


/* all kinds of request */
typedef enum {
  BAD_REQ=0,
  FIRST_REQ, /* first bound of good request type */
  OPEN_REQ, CLOSE_REQ,
  PREAD_REQ, PWRITE_REQ,
  FTRUNC_REQ, TRUNC_REQ,
  FSTAT_REQ, STAT_REQ, LSTAT_REQ,
  FSTATFS_REQ, STATFS_REQ,
  MKNOD_REQ, MKDIR_REQ, RMDIR_REQ, UNLINK_REQ,
  RENAME_REQ,
  ACCESS_REQ, FCHMOD_REQ, CHMOD_REQ, 
  FCHOWN_REQ, CHOWN_REQ, LCHOWN_REQ,
  FLOCK_REQ,
  LINK_REQ, SYMLINK_REQ, READLINK_REQ,
  UTIME_REQ,
  FSYNC_REQ, FDATASYNC_REQ,
  FCNTL_SETFL_REQ,
  OPENDIR_REQ, CLOSEDIR_REQ, READDIR_REQ, REWINDDIR_REQ,
#ifdef CONNECTION_ID_ENABLED
  CONNECTION_REQ,
#endif
  LAST_REQ /* last bound of good request type */
} orfa_request_type_t;


/* OPEN_REQ
   this structure
   file_name */
struct open_request_data {
  orfa_oflags_t oflags;
  orfa_mode_t mode;
};

/* CLOSE_REQ
   this structure */
struct close_request_data {
  orfa_fd_id_t fd_id;
};

/* PREAD_REQ
   this structure */
struct pread_request_data {
  orfa_fd_id_t fd_id;
  orfa_off_t nbyte;
  orfa_off_t offset;
};

/* PWRITE_REQ
   this structure
   data to write */
struct pwrite_request_data {
  orfa_fd_id_t fd_id;
  orfa_off_t offset; 
};

/* FTRUNC_REQ
   this structure */
struct ftrunc_request_data {
  orfa_fd_id_t fd_id;
  orfa_off_t length;
};

/* TRUNC_REQ
   this structure
   file_name */
struct trunc_request_data {
  orfa_off_t length;
};

/* FSTAT_REQ
   this structure */
struct fstat_request_data {
  orfa_fd_id_t fd_id;
};

/* FSTATFS_REQ
   this structure */
struct fstatfs_request_data {
  orfa_fd_id_t fd_id;
};

/* MKNOD_REQ
   this structure
   file_name */
struct mknod_request_data {
  orfa_mode_t mode;
  orfa_dev_t dev;
};

/* MKDIR_REQ
   this structure
   file_name */
struct mkdir_request_data {
  orfa_mode_t mode;
};

/* RENAME_REQ
   this structure
   file_name1
   file_name2 */
struct rename_request_data {
  orfa_off_t part_change;
};

/* ACCESS_REQ
   this structure
   file name */
struct access_request_data {
  orfa_access_mode_t mode;
};

/* FCHMOD_REQ
   this structure */
struct fchmod_request_data {
  orfa_fd_id_t fd_id;
  orfa_mode_t mode;
};

/* CHMOD_REQ
   this structure
   file name */
struct chmod_request_data {
  orfa_mode_t mode;
};

/* FCHOWN_REQ
   this structure */
struct fchown_request_data {
  orfa_fd_id_t fd_id;
  orfa_uid_t owner;
  orfa_gid_t group;
};

/* CHOWN_REQ
   this structure
   file name */
struct chown_request_data {
  orfa_uid_t owner;
  orfa_gid_t group;
};

/* LCHOWN_REQ
   this structure
   file name */
struct lchown_request_data {
  orfa_uid_t owner;
  orfa_gid_t group;
};

/* FLOCK_REQ
   this structure */
struct flock_request_data {
  orfa_fd_id_t fd_id;
  orfa_flock_op_t operation;
};

/* LINK_REQ
   this structure
   file_name1
   file_name2 */
struct link_request_data {
  orfa_off_t part_change;
};

/* SYMLINK_REQ
   this structure
   file_name1
   file_name2 */
struct symlink_request_data {
  orfa_off_t part_change;
};

/* READLINK_REQ
   this structure
   file_name */
struct readlink_request_data {
  orfa_off_t bufsiz;
};

/* FSYNC_REQ
   this structure */
struct fsync_request_data {
  orfa_fd_id_t fd_id;
};

/* FDATASYNC_REQ
   this structure */
struct fdatasync_request_data {
  orfa_fd_id_t fd_id;
};

/* UTIME_REQ
   this structure 
   file_name */
struct utime_request_data {
  orfa_bool_t time_defined;
  orfa_time_t actime;
  orfa_time_t modtime;
};

/* FCNTL_SETFL_REQ
   this structure */
struct fcntl_setfl_request_data {
  orfa_fd_id_t fd_id;
  orfa_oflags_t flags;
};

/* CLOSEDIR_REQ
   this structure */
struct closedir_request_data {
  orfa_dir_id_t dir_id;
};

/* READDIR_REQ
   this structure */
struct readdir_request_data {
  orfa_dir_id_t dir_id;
  orfa_off_t count;
  orfa_off_t offset;
};

/* REWINDDIR_REQ
   this structure */
struct rewinddir_request_data {
  orfa_dir_id_t dir_id;
};


#ifdef CONNECTION_ID_ENABLED
/* CONNECTION_REQ
   this structure */
struct connection_request_data {
  enum {
    CONNECTION_OP_NEW,
    CONNECTION_OP_BEFORE_EXEC,
    CONNECTION_OP_EXEC_DONE,
    CONNECTION_OP_FORK_DONE,
    CONNECTION_OP_EXEC_FAILED
  } op;
};
#endif /* CONNECTION_ID_ENABLED */


/* general REQ */
struct request {
  orfa_request_type_t type;
  orfa_soff_t buffer_len;
#ifdef PROTOCOL_DEBUG
  orfa_tag_t tag;
  orfa_pid_t pid;
#endif /* PROTOCOL_DEBUG */
  orfa_uid_t euid;
  orfa_gid_t egid;
#ifdef CONNECTION_ID_ENABLED
  orfa_connection_id_t connection_id;
#endif /* CONNECTION_ID_ENABLED */
  union {
    struct open_request_data open;
    struct close_request_data close;
    struct pread_request_data pread;
    struct pwrite_request_data pwrite;
    struct ftrunc_request_data ftrunc;
    struct trunc_request_data trunc;
    struct fstat_request_data fstat;
    struct fstatfs_request_data fstatfs;
    struct mknod_request_data mknod;
    struct mkdir_request_data mkdir;
    struct rename_request_data rename;
    struct access_request_data access;
    struct fchmod_request_data fchmod;
    struct chmod_request_data chmod;
    struct fchown_request_data fchown;
    struct chown_request_data chown;
    struct lchown_request_data lchown;
    struct flock_request_data flock;
    struct link_request_data link;
    struct symlink_request_data symlink;
    struct readlink_request_data readlink;
    struct fsync_request_data fsync;
    struct fdatasync_request_data fdatasync;
    struct utime_request_data utime;
    struct fcntl_setfl_request_data fcntl_setfl;
    struct closedir_request_data closedir;
    struct readdir_request_data readdir;
    struct rewinddir_request_data rewinddir;
#ifdef CONNECTION_ID_ENABLED
    struct connection_request_data connection;
#endif /* CONNECTION_ID_ENABLED */
  } data;
};



/* all kinds of answer */
typedef enum {
  BAD_ANS=0,
  FIRST_ANS,
  OPEN_ANS, CLOSE_ANS,
  PREAD_ANS, PWRITE_ANS,
  FTRUNC_ANS, TRUNC_ANS,
  FSTAT_ANS, STAT_ANS, LSTAT_ANS,
  FSTATFS_ANS, STATFS_ANS,
  MKNOD_ANS, MKDIR_ANS, RMDIR_ANS, UNLINK_ANS,
  RENAME_ANS,
  ACCESS_ANS, FCHMOD_ANS, CHMOD_ANS,
  FCHOWN_ANS, CHOWN_ANS, LCHOWN_ANS,
  FLOCK_ANS,
  LINK_ANS, SYMLINK_ANS, READLINK_ANS,
  UTIME_ANS,
  FSYNC_ANS, FDATASYNC_ANS,
  FCNTL_SETFL_ANS,
  OPENDIR_ANS, CLOSEDIR_ANS, READDIR_ANS, REWINDDIR_ANS,
#ifdef CONNECTION_ID_ENABLED
  CONNECTION_ANS,
#endif /* CONNECTION_ID_ENABLED */
  LAST_ANS
} orfa_answer_type_t;


/* general ANS */
struct answer {
  orfa_answer_type_t type;
#ifdef PROTOCOL_DEBUG
  orfa_tag_t tag;
  orfa_pid_t pid;
#endif /* PROTOCOL_DEBUG */
  orfa_error_t error;
  orfa_soff_t buffer_len;
  union {
    uint64_t ptr_value;
    int64_t int_value;
    int64_t ssize_t_value;
  } return_value;
};


#endif /* __PROTOCOL_H__ */
