/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include <libaio.h>

#include "orfas_STD_io.h"
#include "orfas_STD_fs.h"
#include "orfas_queue.h"
#include "orfas_portable.h"
#include "orfas_io.h"
#include "orfas_network.h"
#include "orfas_debug.h"



/*************
 * Callbacks *
 *************/

static void AIO_callback_pread(io_context_t ctx, struct iocb *iocb,
			       long res, long res2)
{
  struct server_request *req = dequeue_processing_request_for_key((void*)iocb);
  LOG(4, "Callback for iocb %p\n", iocb);
  free(iocb);

  /* check the return value */
  if ( res < 0 ) {
    req->ans.buffer_len = 0;
    req->ans.return_value.ssize_t_value = -1;
    req->ans.error = convert_errno_to_portable(-res);
  } else {
    req->ans.buffer_len = res;
    req->ans.return_value.ssize_t_value = res;
    resize_io_monovector(req->out_vect, res);
  }

  LOG(5, "read %ld bytes at offset %lld\n",
      (long)res,
      (long long)req->req.data.pread.offset);

  queue_done_request(req);
}


static void AIO_callback_pwrite(io_context_t ctx, struct iocb *iocb,
				long res, long res2)
{
  struct server_request *req = dequeue_processing_request_for_key((void*)iocb);
  LOG(4, "Callback for iocb %p\n", iocb);
  free(iocb);

  /* now free in_vect */
  put_io_vector(req->in_vect);

  /* check the return value */
  if ( res < 0 ) {
    req->ans.return_value.ssize_t_value = -1;
    req->ans.error = convert_errno_to_portable(-res);
  } else
    req->ans.return_value.ssize_t_value = res;

  LOG(5, "write %ld bytes at offset %lld\n",
      (long)res,
      (long long)req->req.data.pwrite.offset);

  queue_done_request(req);
}


static void AIO_callback_fsync(io_context_t ctx, struct iocb *iocb,
			       long res, long res2)
{
  struct server_request *req = dequeue_processing_request_for_key((void*)iocb);
  LOG(4, "Callback for iocb %p\n", iocb);
  free(iocb);
  
  /* check the return value */
  if ( res < 0 ) {
    req->ans.return_value.ssize_t_value = -1;
    req->ans.error = convert_errno_to_portable(-res);
  } else
    req->ans.return_value.ssize_t_value = 0;

  LOG(5, "fsync gets %ld\n",
      (long)res);

  queue_done_request(req);
}


static void AIO_callback_fdatasync(io_context_t ctx, struct iocb *iocb,
				   long res, long res2)
{
  struct server_request *req = dequeue_processing_request_for_key((void*)iocb);
  LOG(4, "Callbask for iocb %p\n", iocb);
  free(iocb);
   
  /* check the return value */
  if ( res < 0 ) {
    req->ans.return_value.ssize_t_value = -1;
    req->ans.error = convert_errno_to_portable(-res);
  } else
    req->ans.return_value.ssize_t_value = 0;
   
  LOG(5, "fsync gets %ld\n",
      (long)res);
   
  queue_done_request(req);
}



/******************
 * Submitting AIO *
 ******************/

static INLINE void AIO_submit_pread(struct server_request *req,
				    int fd, void *buf,
				    size_t count, off64_t offset)
{
  struct iocb *iocb = (struct iocb*) calloc(1, sizeof(struct iocb));

  queue_processing_request(req);
  req->key = iocb;
  LOG(3, "Submit iocb %p for pwrite count %ld at %lld\n",
      iocb, (long)count, (long long)offset);
  io_prep_pread(iocb, fd, buf, count, offset);
  io_set_callback(iocb, AIO_callback_pread);

  AIO_submit(iocb);
}


static INLINE void AIO_submit_pwrite(struct server_request *req,
				     int fd, const void *buf,
				     size_t count, off64_t offset)
{
  struct iocb *iocb = (struct iocb*) calloc(1, sizeof(struct iocb));

  queue_processing_request(req);
  req->key = iocb;
  LOG(3,"Submit iocb %p for pwrite count %ld at %lld\n",
      iocb, (long)count, (long long)offset);
  io_prep_pwrite(iocb, fd, (void*)buf, count, offset);
  io_set_callback(iocb, AIO_callback_pwrite);

  AIO_submit(iocb);
}


static INLINE void AIO_submit_fsync(struct server_request *req,
				    int fd)
{
  struct iocb *iocb = (struct iocb*) calloc(1, sizeof(struct iocb));
   
  queue_processing_request(req);
  req->key = iocb;
  LOG(3, "Submit iocb %p for fsync fd %ld\n",
      iocb, (long)fd);
  io_prep_fsync(iocb, fd);
  io_set_callback(iocb, AIO_callback_fsync);
   
  AIO_submit(iocb);
}


static INLINE void AIO_submit_fdatasync(struct server_request *req,
				        int fd)
{
  struct iocb *iocb = (struct iocb*) calloc(1, sizeof(struct iocb));
   
  queue_processing_request(req);
  req->key = iocb;
  LOG(3, "Submit iocb %p for fdatasync fd %ld\n",
      iocb, (long)fd);
  io_prep_fdsync(iocb, fd);
  io_set_callback(iocb, AIO_callback_fdatasync);
   
  AIO_submit(iocb);
}



/***********************
 * Requests Processing *
 ***********************/

void AIO_process_request(struct server_request *req)
{
  req->out_vect = NULL;
  req->ans.error = 0;
#ifdef PROTOCOL_DEBUG
  req->ans.tag = req->req.tag;
  req->ans.pid = req->req.pid;
#endif /* PROTOCOL_DEBUG */
  req->ans.buffer_len = -1;
	
  setregid(-1, req->req.egid);
  setreuid(-1, req->req.euid);

  switch ( req->req.type )
    {

    /* pread */
    case PREAD_REQ:
      {
	int local_fd;

	req->ans.type = PREAD_ANS;
	
	/* get the real file */
	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.pread.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {
	    LOG(2, "Reading fd %p for local fd %d\n",
		(void*)(uintptr_t)req->req.data.pread.fd_id,
		local_fd);

	    req->out_vect = get_io_monovector_alloc(req->req.data.pread.nbyte);

	    AIO_submit_pread(req,
			     local_fd,
			     VECT_PTR(req->out_vect, 0),
			     req->req.data.pread.nbyte, 
			     req->req.data.pread.offset);
	  }	
	/* end of pread */
	break;
      }


    /* pwrite */
    case PWRITE_REQ:
      {
	int local_fd;

	req->ans.type = PWRITE_ANS;

	/* get the real file */
	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.pwrite.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {
	    LOG(2, "Writing fd %p for local fd %d\n",
		(void*)(uintptr_t)req->req.data.pwrite.fd_id,
		local_fd);

	    AIO_submit_pwrite(req,
			      local_fd,
			      VECT_PTR(req->in_vect, 0),
			      req->req.buffer_len,
			      req->req.data.pwrite.offset);
	  }
	/* end of pwrite */
	break;
      }
    

    /* fsync */
    case FSYNC_REQ:
      {
	int local_fd;

	req->ans.type = FSYNC_ANS;

	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.fsync.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {	

	    LOG(2, "fsync fd %p\n",
		(void*)(uintptr_t)req->req.data.fsync.fd_id);

	    AIO_submit_fsync(req, local_fd);
	  }
	/* end of fsync */
	break;
      }


    /* fdatasync */
    case FDATASYNC_REQ:
      {
	int local_fd;

	req->ans.type = FDATASYNC_ANS;

	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.fdatasync.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {	
	    LOG(2, "fdatasync fd %p\n",
		(void*)(uintptr_t)req->req.data.fdatasync.fd_id);

	    AIO_submit_fdatasync(req, local_fd);
	  }
	/* end of fdatasync */
	break;
      }

     
     default:
      {
	STD_process_easy_request(req);
#ifndef CONNECTION_THREAD_ARCH
	queue_done_request(req);
#endif
        break;
      }
    }

  if ( req->req.type != PWRITE_REQ )
    /* in_vect for WRITE will be freeed after request completion */
    put_io_vector(req->in_vect);

  setreuid(-1, uid);
  setregid(-1, gid);
}
