/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <assert.h>

#include "bip.h"

#include "orfas_network.h"
#include "orfas_queue.h"
#include "orfas_io.h"
#include "orfas_debug.h"



/**********************
 * Static definitions *
 **********************/

static long current_req_recv_id;
static struct server_request *current_req;

/* list of request waiting for their data */
struct waiting_request {
  struct server_request *req;
  long recv_id;
  long not_big_recv_id;
  struct io_vector *not_big_vector;
};

static struct waiting_request *waiting_req;


/********************************
 * Non-exported BIP definitions *
 ********************************/

extern void bip_exit();
extern char bip_nodename[][255];


/**************
 * Send Token *
 **************/

long bip_send_token = 0;
void (*bip_send_callback)(void *data);
void *bip_send_callback_data;


/******************
 * Initialization *
 ******************/

static INLINE void provide_request_receive_buffer()
{
  current_req = get_free_request();
  current_req_recv_id
    = bip_tirecv(PROTOCOL_TAG,
		 (int*)&current_req->req,
		 sizeof(struct request)/WORD_SIZE);

  LOG(4, "Provided buffer %p for request\n", &current_req->req);
}


static INLINE void provide_not_big_data_buffer(unsigned int node)
{
  int tag = REQUEST_DATA_TAG_FOR_NODE(node);
  struct io_vector *vect = get_io_monovector_alloc(PREPROVIDED_BUFFER_LENGTH);

  waiting_req[node].not_big_vector = vect;

  waiting_req[node].not_big_recv_id
    = bip_tirecv(tag,
		 (int*)VECT_PTR(vect, 0),
		 PREPROVIDED_BUFFER_LENGTH/WORD_SIZE);

  /* by default, no data to receive */
  waiting_req[node].recv_id = 0;
}


int BIP_network_init()
{
  int i;

  /* initialize the request receive queue */
  bip_taginit(PROTOCOL_TAG, SERVER_REQUEST_QUEUE_LEN, sizeof(struct request));

  bip_async_init();
  LOG(1, "BIP initialized\n");

#ifdef FIXED_SERVER_NODE
  if ( bip_mynode != FIXED_SERVER_NODE )
    {
      ERROR("Server may be inaccessible with node %d instead of %d ?\n",
	    bip_mynode, FIXED_SERVER_NODE);      
    }
#endif /* FIXED_SERVER_NODE */

  provide_request_receive_buffer();

  waiting_req = calloc(bip_numnodes, sizeof(struct waiting_request));

  for(i=0; i<bip_numnodes; i++)
    if ( i != bip_mynode )
      provide_not_big_data_buffer(i);

  return 0;
}


void BIP_network_finish()
{
  free(waiting_req);

  bip_exit();
}



/***********
 * Sending *
 ***********/

int BIP_network_send(struct server_request *req)
{
  /* copy some data because req can be freed
   * between ans and data sending */
  int node = req->connection.node;
  struct io_vector *out_vect = req->out_vect;
  int data_len = req->ans.buffer_len;
  int ans_type = req->ans.type;

  LOG(2, "Sent answer %p type %d to node %d\n",
      &req->ans, (int)req->ans.type,
      node);

  /* send answer */
  BIP_WAIT_FOR_SEND_TOKEN();
  BIP_TAKE_SEND_TOKEN(bip_tisend(node,
				 PROTOCOL_TAG,
				 (int*)&req->ans,
				 sizeof(struct answer)/WORD_SIZE));
  BIP_FIX_SEND_CALLBACK((void (*)(void*))put_free_request, req);

  /* check ANS type */
  if ( ans_type <= FIRST_ANS
       ||  ans_type >= LAST_ANS )
    {
      ERROR("Cannot come here !!\n");
      assert(0);
    }

  /* at this point, we use copy of req data in
   * ordre to prevent freeing of request. */

  /* send data if necessary */
  if ( data_len >=0 )
    {
#ifdef BIP_SENDFILE_ENABLED
      if ( ans_type == PREAD_ANS )
	{
	  LOG(3, "Sendfile already done\n");
	}
      else
#endif /* BIP_SENDFILE_ENABLED */
	{
	  ALIGN_TO_WORD_SIZE(data_len);
	  
	  out_vect = merge_io_vector(out_vect);

	  BIP_WAIT_FOR_SEND_TOKEN();
	  BIP_TAKE_SEND_TOKEN(bip_tisend(node,
					 ANSWER_DATA_TAG,
					 VECT_PTR(out_vect, 0),
					 data_len/WORD_SIZE));
	  BIP_FIX_SEND_CALLBACK((void (*)(void*))put_io_vector, out_vect);

	  LOG(3, "Sent %d data at %p to node %d\n",
	      (int)data_len, VECT_PTR(out_vect, 0), node);
	}
    }

  return 0;
}



/*********************
 * Handling requests *
 *********************/

static INLINE int handle_req_recv(struct server_request *req, int sender)
{
  int data_len;

  req->connection.node = sender;

  req->in_vect = NULL;
  
  /* Check the type */
  LOG(2, "Recu type %d from node %d\n",
      (int)req->req.type, sender);
  if ( req->req.type <= FIRST_REQ
       ||  req->req.type >= LAST_REQ )
    {
      ERROR("Bad request type %d received, impossible with BIP ?!\n",
	    (int)req->req.type);
      assert(0);
    }

  /* if no data to receive, the request is ready */
  data_len = req->req.buffer_len;
  if ( data_len < 0 )
    {
      queue_ready_request(req);
      return 1;
    }

  /* align data len */
  ALIGN_TO_WORD_SIZE(data_len);

  if ( data_len > PREPROVIDED_BUFFER_LENGTH )
    /* Recv data buffer in a new big buffer */
    {
      struct io_vector *vect = get_io_monovector_alloc(data_len);
      req->in_vect = vect;
      
      /* provide it */
      waiting_req[sender].recv_id
	= bip_tirecv(REQUEST_BIG_DATA_TAG_FOR_NODE(sender),
		     (int*)VECT_PTR(vect, 0),
		     data_len/WORD_SIZE);

      /* tell the client I am ready to receive data */
      BIP_WAIT_FOR_SEND_TOKEN();
      BIP_TAKE_SEND_TOKEN(bip_tisend(sender,
				     PROTOCOL_TAG,
				     NULL,
				     0));
      BIP_FIX_SEND_CALLBACK(NULL, NULL);

      LOG(3, "Waiting for %d BIG data at %p\n",
	  data_len, vect);
    }
  else
    /* Recv data buffer in the preprovided buffer */
    {
      struct io_vector *vect = waiting_req[sender].not_big_vector;
      req->in_vect = vect;

      /* retrieve buffer len */
      VECT_SIZE(vect, 0) = data_len;

      /* use preprovided data recv id */
      waiting_req[sender].recv_id = waiting_req[sender].not_big_recv_id;

      LOG(3, "Waiting for %d data at %p\n",
	  data_len, VECT_PTR(vect, 0));
    }

  /* store the request until the buffer arrives */
  waiting_req[sender].req = req;

  return 0;
}



/*****************
 * Scanning tags *
 *****************/

/* this functions probes all tirecv,
   and returns 1 if a request is ready */
static INLINE int multiprobe()
{
  unsigned int sender;
  int request_ready = 0;

  /* check control channel for new connections */
  bip_check_control_channel();

  /* try the request queue */
  while ( bip_rtestx(current_req_recv_id, &sender) >= 0 )
    /* we received a request */
    {
      struct server_request *req = current_req;

      /* provide a new request */
      provide_request_receive_buffer();

      /* handle this request */
      if ( handle_req_recv(req, sender) )
	request_ready = 1;
    }

  /* try each waiting data queue */
  for(sender=0; sender<bip_numnodes; sender++)
    {
      long recv_id = waiting_req[sender].recv_id;
      int rlen;

      /* try this sender */
      if ( recv_id > 0  &&  (rlen=bip_rtest(recv_id)) >= 0 )
	{
	  struct server_request *req = waiting_req[sender].req;

	  LOG(2, "Recv %d data for req %p\n",
	      rlen*WORD_SIZE, &req->req);

	  if ( recv_id == waiting_req[sender].not_big_recv_id )
	    /* preprovide another buffer */
	    provide_not_big_data_buffer(sender);
	  else
	    /* retake the unused preprovided buffer */
	    waiting_req[sender].recv_id = 0;

	  /* this request is now ready */
	  queue_ready_request(req);

	  request_ready = 1;
	}
    }
  return request_ready;
}


#ifdef AIO_MIXED_WITH_OTHER_ASYNC_API

void BIP_network_test()
{
  multiprobe();
}

#else /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */

void BIP_network_wait()
{
  while ( !multiprobe() );
}

#endif /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */
