/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "orfas_queue.h"
#include "orfas_portable.h"
#include "orfas_io.h"
#include "orfas_network.h"
#include "orfas_debug.h"


/* This is a fake IO module :
 * - the file tree has a depth of DEPTH_MAX
 * - each directory contains DIRECTORY_FILE_ENTRIES files
 *   and DIRECTORY_DIR_ENTRIES dir (except at last depth)
 * - open and opendir give the file number in the directory
 * - write access is impossible
 * - nth file's length is n*FILE_LEN_RATIO
 * - dx/dy/fz is the z-th file of the y-th subdir of the x-th dir
 */


int FAKE_process_init()
{
  return 0;
}

void FAKE_process_finish()
{
}



static INLINE int get_depth(char *file_name)
{
  if ( *(file_name+1) == '\0' )
    return 0;
  return 1 + (strlen(file_name)-(FILE_PREFIX_LEN+2)) / (DIR_PREFIX_LEN+2);
}


static INLINE int get_file_nr(char *file_name)
{
  return atoi(strchr(file_name, '\0')-1);
}

static INLINE off64_t get_file_length(int nr)
{
  return ( (nr+1) * FILE_LEN_RATIO );
}


static int is_dir(char *file_name)
{
  char *end = strchr(file_name, '\0');

  if ( *(file_name+1) == '\0' )
    return 1;

  return ( !strncmp(end-2, DIR_PREFIX, DIR_PREFIX_LEN) );
}


static INLINE struct io_vector *get_file_stat(int nr)
{
  struct io_vector *vect = get_io_monovector_alloc(sizeof(struct portable_stat));

  struct portable_stat *st = VECT_PTR(vect, 0);
  memset(st, 0, sizeof(struct portable_stat));
  st->st_mode = S_IFREG | S_IRUSR | S_IRGRP | S_IROTH;
  st->st_nlink = 1;
  st->st_size = get_file_length(nr);
  st->st_blksize = 512;
  st->st_blocks = st->st_size / st->st_blksize;
  st->st_ino = rand();

  return vect;
}


static INLINE struct io_vector *get_dir_stat()
{
  struct io_vector *vect = get_io_monovector_alloc(sizeof(struct portable_stat));

  struct portable_stat *st = VECT_PTR(vect, 0);
  memset(st, 0, sizeof(struct portable_stat));
  st->st_mode = S_IFDIR | S_IRUSR | S_IXUSR
    | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
  st->st_nlink = 2 + DIRECTORY_DIR_ENTRIES;
  st->st_size = 1024;
  st->st_blksize = 512;
  st->st_blocks = st->st_size / st->st_blksize;
  st->st_ino = rand();

  return vect;
}


void FAKE_process_request(struct server_request *req)
{
  req->out_vect = NULL;
  req->ans.error = 0;
#ifdef PROTOCOL_DEBUG
  req->ans.tag = req->req.tag;
  req->ans.pid = req->req.pid;
#endif /* PROTOCOL_DEBUG */
  req->ans.buffer_len = -1;
	
  switch ( req->req.type )
    {

    /* open */
    case OPEN_REQ:
      {
	req->ans.type = OPEN_ANS;
	LOG(2, "Opening file %s with flags %d\n",
	    (char*)VECT_PTR(req->in_vect, 0),
	    (int)req->req.data.open.oflags);

	if ( ( req->req.data.open.oflags & (P_O_WRONLY|P_O_RDWR|P_O_DIRECTORY) )
	     || is_dir(VECT_PTR(req->in_vect, 0)) )
	  {
	    req->ans.return_value.ptr_value = (uint64_t)NULL;
	    req->ans.buffer_len = 0;
	    req->ans.error = convert_errno_to_portable(EACCES);
	  }
	else
	  {
	    int nr = get_file_nr(VECT_PTR(req->in_vect, 0));
	    req->ans.return_value.ptr_value = (uint64_t)nr;
	    req->ans.buffer_len = sizeof(struct portable_stat);
	    req->out_vect = get_file_stat(nr);
	    LOG(3, "Open of file %s succeded\n", (char*)VECT_PTR(req->in_vect, 0));
	  }
	break;
      }


    /* close */
    case CLOSE_REQ:
      {
	req->ans.type = CLOSE_ANS;
	req->ans.return_value.int_value = 0;
	break;
      }
      

    /* pread */
    case PREAD_REQ:
      {
	int nr = (int)req->req.data.pread.fd_id;
	off64_t offset = req->req.data.pread.offset;
	ssize_t nbyte = req->req.data.pread.nbyte;
	off64_t len = get_file_length(nr);

	if ( nbyte+offset > len )
	  nbyte = len - offset;

	LOG(2, "Reading %ld/%ld data at %lld in %lld\n",
	    (long)nbyte, (long)req->req.data.pread.nbyte,
	    (long long)offset, (long long)len);

	if ( nbyte >= 0 )
	  {
	    req->ans.buffer_len = nbyte;
	    req->ans.return_value.ssize_t_value = nbyte;
	    req->out_vect = get_io_monovector_alloc(nbyte);
	  }
	else
	  {
	    req->ans.buffer_len = 0;
	    req->ans.return_value.ssize_t_value = 0;
	  }

	req->ans.type = PREAD_ANS;
	break;
      }


    /* pwrite */
    case PWRITE_REQ:
      {
	req->ans.type = PWRITE_ANS;
	req->ans.return_value.ssize_t_value = -1;
	req->ans.error = convert_errno_to_portable(EBADF);
	break;
      }
    

    /* ftrunc */
    case FTRUNC_REQ:
      {
	req->ans.type = FTRUNC_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EBADF);
	break;
      }


    /* trunc */
    case TRUNC_REQ:
      {
	req->ans.type = TRUNC_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }

    
    /* fstat */
    case FSTAT_REQ:
      {
	req->ans.type = FSTAT_ANS;
	req->ans.return_value.int_value = 0;
	req->ans.buffer_len = sizeof(struct portable_stat);
	req->out_vect = get_file_stat((int)req->req.data.fstat.fd_id);
	break;
      }


    /* stat */
    case STAT_REQ:
      {
	int nr = get_file_nr(VECT_PTR(req->in_vect, 0));
	req->ans.type = STAT_ANS;
	req->ans.return_value.int_value = 0;
	req->ans.buffer_len = sizeof(struct portable_stat);
	if ( is_dir(VECT_PTR(req->in_vect, 0)) )
	  {
	    LOG(2, "Stat dir %s\n", (char*)VECT_PTR(req->in_vect, 0));
	    req->out_vect = get_dir_stat(nr);
	  }
	else
	  {
	    LOG(2, "Stat file %s\n", (char*)VECT_PTR(req->in_vect, 0));
	    req->out_vect = get_file_stat(nr);
	  }
	break;
      }


    /* lstat */
    case LSTAT_REQ:
      {
	int nr = get_file_nr(VECT_PTR(req->in_vect, 0));
	req->ans.type = LSTAT_ANS;
	req->ans.return_value.int_value = 0;
	req->ans.buffer_len = sizeof(struct portable_stat);
	if ( is_dir(VECT_PTR(req->in_vect, 0)) )
	  {
	    LOG(2, "lstat dir %s\n", (char*)VECT_PTR(req->in_vect, 0));
	    req->out_vect = get_dir_stat(nr);
	  }
	else
	  {
	    LOG(2, "lstat file %s\n", (char*)VECT_PTR(req->in_vect, 0));
	    req->out_vect = get_file_stat(nr);
	  }
	break;
      }


    /* fstatfs */
    case FSTATFS_REQ:
      {
	req->ans.type = FSTATFS_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EBADF);
	break;
      }


    /* statfs */
    case STATFS_REQ:
      {
	req->ans.type = STATFS_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* mknod */
    case MKNOD_REQ:
      {
	req->ans.type = MKNOD_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* mkdir */
    case MKDIR_REQ:
      {
	req->ans.type = MKDIR_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* rmdir */
    case RMDIR_REQ:
      {
	req->ans.type = RMDIR_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* unlink */
    case UNLINK_REQ:
      {
	req->ans.type = UNLINK_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* access */
    case ACCESS_REQ:
      {
	req->ans.type = ACCESS_ANS;
	if ( req->req.data.access.mode & W_OK )
	  {
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EACCES);
	    LOG(2, "Access W denied to %s\n",
		(char*)VECT_PTR(req->in_vect, 0));
	  }
	else if ( (req->req.data.access.mode & X_OK)
		  && !is_dir(VECT_PTR(req->in_vect, 0)) )
	  {
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EACCES);
 	    LOG(2, "Access X denied to file %s\n",
		(char*)VECT_PTR(req->in_vect, 0));
	  }
	else
	  {
	    req->ans.return_value.int_value = 0;	  
 	    LOG(3, "Access to %s\n",
		(char*)VECT_PTR(req->in_vect, 0));
	  }
	break;
      }


    /* fchmod */
    case FCHMOD_REQ:
      {
	req->ans.type = FCHMOD_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EBADF);
	break;
      }


    /* chmod */
    case CHMOD_REQ:
      {
	req->ans.type = CHMOD_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* fchown */
    case FCHOWN_REQ:
      {
	req->ans.type = FCHOWN_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EBADF);
	break;
      }


    /* chown */
    case CHOWN_REQ:
      {
	req->ans.type = CHOWN_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* lchown */
    case LCHOWN_REQ:
      {
	req->ans.type = LCHOWN_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* flock */
    case FLOCK_REQ:
      {
	req->ans.type = FLOCK_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(ENOSYS);
	break;
      }


    /* rename */
    case RENAME_REQ:
      {
	req->ans.type = RENAME_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);	
	break;
      }


    /* link */
    case LINK_REQ:
      {
	req->ans.type = LINK_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* symlink */
    case SYMLINK_REQ:
      {
	req->ans.type = SYMLINK_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* readlink */
    case READLINK_REQ:
      {
	req->ans.type = READLINK_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	break;
      }


    /* fsync */
    case FSYNC_REQ:
      {
	req->ans.type = FSYNC_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EBADF);
	break;
      }


    /* fdatasync */
    case FDATASYNC_REQ:
      {
	req->ans.type = FDATASYNC_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EBADF);
	break;
      }


    /* utime */
    case UTIME_REQ:
      {
	req->ans.type = UTIME_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(EACCES);
	req->ans.error = EACCES;
	break;
      }


    /* fcntl_setfl */
    case FCNTL_SETFL_REQ:
      {
	req->ans.type = FCNTL_SETFL_ANS;
	req->ans.return_value.int_value = -1;
	req->ans.error = convert_errno_to_portable(ENOSYS);
	break;
      }


    /* opendir */
    case OPENDIR_REQ:
      {
	req->ans.type = OPENDIR_ANS;
	if ( is_dir(VECT_PTR(req->in_vect, 0)) )
	  {
	    int d = get_depth(VECT_PTR(req->in_vect, 0)) + 1;
	    req->ans.return_value.ptr_value = (uint64_t)d;
	    LOG(2, "Opendir of %s succeded with %d\n",
		(char*)VECT_PTR(req->in_vect, 0), d);
	  }
	else
	  {
	    req->ans.return_value.ptr_value = (uint64_t)NULL;
	    req->ans.error = convert_errno_to_portable(ENOTDIR);
	    LOG(2, "Opendir of %s failed\n",
		(char*)VECT_PTR(req->in_vect, 0));
	  }
	break;
      }


    /* closedir */
    case CLOSEDIR_REQ:
      {
	req->ans.type = CLOSEDIR_ANS;
	req->ans.return_value.int_value = 0;
	break;
      }


    /* readdir */
    case READDIR_REQ:
      {
	int depth = ((int)req->req.data.readdir.dir_id) - 1;
	off64_t seek = req->req.data.readdir.offset;
	size_t count = req->req.data.readdir.count;
	int indix = 0;
	struct portable_dirent *dirents;
	
	req->ans.type = READDIR_ANS;
	req->out_vect = get_io_monovector_alloc(count*sizeof(struct portable_dirent));

	dirents = VECT_PTR(req->out_vect, 0);
	memset(dirents, 0, VECT_SIZE(req->out_vect, 0));

	if ( count > 0  &&  seek < 1 )
	  {
	    dirents[indix].d_ino = 0;
	    dirents[indix].d_off = seek+1;
	    dirents[indix].d_reclen = 16;
	    dirents[indix].d_type = 0;
	    sprintf(dirents[indix].d_name, "..");
	    indix++;
	    seek++;
	    count--;
	  }

	if ( count > 0  &&  seek < 2 )
	  {
	    dirents[indix].d_ino = 0;
	    dirents[indix].d_off = seek+1;
	    dirents[indix].d_reclen = 16;
	    dirents[indix].d_type = 0;
	    sprintf(dirents[indix].d_name, ".");
	    indix++;
	    seek++;
	    count--;
	  }

	while ( count > 0
		&& seek < 2 + DIRECTORY_FILE_ENTRIES )
	  {
	    dirents[indix].d_ino = 0;
	    dirents[indix].d_off = seek+1;
	    dirents[indix].d_reclen = 16;
	    dirents[indix].d_type = 0;
	    sprintf(dirents[indix].d_name,
		    FILE_PREFIX "%lld",
		    (long long)(seek-2));
	    indix++;
	    seek++;
	    count--;
	  }

	if ( depth < DEPTH_MAX )
	  {
	    while ( count > 0
		    && seek < 2 + DIRECTORY_FILE_ENTRIES + DIRECTORY_DIR_ENTRIES )
	      {
		dirents[indix].d_ino = 0;
		dirents[indix].d_off = seek+1;
		dirents[indix].d_reclen = 16;
		dirents[indix].d_type = 0;
		sprintf(dirents[indix].d_name,
			DIR_PREFIX "%lld",
			(long long)(seek-DIRECTORY_FILE_ENTRIES-2));
		indix++;
		seek++;
		count--;
	      }
	  }

	/* check the return value */
	req->ans.return_value.int_value = indix;
	if ( indix < 0 ) {
	  put_io_vector(req->out_vect);
	  req->out_vect = NULL;
	  req->ans.buffer_len = 0;
	} else {
	  resize_io_monovector(req->out_vect,
			       indix * sizeof(struct portable_dirent));
	  req->ans.buffer_len = indix * sizeof(struct portable_dirent);
	}

	break;
      }


    /* rewinddir */
    case REWINDDIR_REQ:
      {
	req->ans.type = REWINDDIR_ANS;
	req->ans.return_value.int_value = 0;
	break;
      }
      


    /* bad request */
    case BAD_REQ:
      {
	ERROR("Bad request %d => Bad answer\n",
	      (int)req->req.type);
	req->ans.type = BAD_ANS;
	break;
      }

    /* cannot come here because bad request type
     must have been converted to BAD_REQ during receiving*/
    default:
      {
	ERROR("Cannot come here !!\n");
	assert(0);
      }
    }

  put_io_vector(req->in_vect);

#ifndef CONNECTION_THREAD_ARCH
  queue_done_request(req);
#endif
}
