/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <assert.h>

#include "gm.h"

#include "orfas_queue.h"
#include "orfas_debug.h"


static struct gm_port *server_port = NULL;
static unsigned int request_size, answer_size, preprovided_buffer_size;
static int request_shift;



/********
 * Lock *
 ********/

#ifdef QUEUE_THREAD_ARCH

/* This lock is used around gm_call which are not multithread-safe.
 * This could probably be a spinlock. */

#include <pthread.h>
static pthread_mutex_t gm_mutex = PTHREAD_MUTEX_INITIALIZER;
#define GM_LOCK() pthread_mutex_lock(&gm_mutex)
#define GM_UNLOCK() pthread_mutex_unlock(&gm_mutex)

#else

#define GM_LOCK() /* nothing */
#define GM_UNLOCK() /* nothing */

#endif



/*****************************
 * Send Tokens processing *
 *****************************/

static int send_tokens = -1;


#ifdef QUEUE_THREAD_ARCH

/* Queue_Thread_Arch needs a thread who waits for GM events
 * while another one sends completed requests.
 * Send Tokens are also consumed by the latter while they
 * are released by the former.
 *
 * We use a mutex protection because we can wait for a long
 * time before getting a new send token. */

#include <pthread.h>
static pthread_mutex_t send_tokens_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t send_tokens_cond = PTHREAD_COND_INITIALIZER;

#define PUT_SEND_TOKEN()                  \
do {                                      \
pthread_mutex_lock(&send_tokens_mutex);   \
send_tokens++;                            \
LOG(6, "Release a send token, %d free\n", \
    send_tokens);                         \
pthread_cond_signal(&send_tokens_cond);   \
pthread_mutex_unlock(&send_tokens_mutex); \
} while ( 0 )

#define GET_SEND_TOKEN()                  \
do {                                      \
pthread_mutex_lock(&send_tokens_mutex);   \
while ( !send_tokens )                    \
  pthread_cond_wait(&send_tokens_cond,    \
		    &send_tokens_mutex);  \
send_tokens--;                            \
LOG(6, "Get a send token, %d free\n",     \
  send_tokens);                           \
pthread_mutex_unlock(&send_tokens_mutex); \
} while ( 0 )

#else /* QUEUE_THREAD_ARCH */

/* Non-threaded architectures just require us to wait for
 * GM events when no more Send Tokens are free.
 * Send Tokens will not be needed in special GM routine
 * such as gm_unknown so we can call gm_receive without problem. */

static void GM_network_wait_for_send_tokens();

#define PUT_SEND_TOKEN()                  \
do {                                      \
send_tokens++;                            \
LOG(6, "Release a send token, %d free\n", \
    send_tokens);                         \
} while ( 0 )

#define GET_SEND_TOKEN()                \
do {                                    \
if ( !send_tokens ) {                   \
  LOG(6, "Waiting for a send token\n"); \
  GM_network_wait_for_send_tokens();    \
}                                       \
send_tokens--;                          \
LOG(6, "Get a send token, %d free\n",   \
  send_tokens);                         \
} while ( 0 )

#endif /* QUEUE_THREAD_ARCH */



/*****************************
 * Receive Tokens processing *
 *****************************/

static int recv_tokens = -1;

static int delayed_request_providing = 0;


static INLINE int provide_request_receive_buffer();


#define GET_RECV_TOKEN()                 \
do {                                     \
recv_tokens--;                           \
LOG(6, "Get a receive token, %d free\n", \
    recv_tokens);                        \
} while ( 0 )

#define PUT_RECV_TOKEN()                      \
do {                                          \
recv_tokens++;                                \
LOG(6, "Free a receive token, %d free\n",     \
    recv_tokens);                             \
while ( delayed_request_providing-- > 0 ) {   \
  LOG (7, "Try delayed request providing\n"); \
  if ( provide_request_receive_buffer() < 0 ) \
    break;                                    \
} } while ( 0 )



/*******************
 * Buffer managing *
 *******************/

void *GM_buffer_malloc(unsigned int len) {
  return gm_dma_malloc(server_port, len);
}

void GM_buffer_free(void *ptr) {
  gm_dma_free(server_port, ptr);
}


/******************
 * Initialization *
 ******************/

static INLINE int provide_request_receive_buffer()
{
  struct server_request *req;
  if ( recv_tokens >= 2 )
    {
      GET_RECV_TOKEN();
      req = get_free_request();
      gm_provide_receive_buffer(server_port, &req->req,
				request_size, REQUEST_PRIORITY);
      return 0;
    }
  else
    {
      delayed_request_providing++;
      LOG(6, "Delayed request providing, %d delayed\n",
	  delayed_request_providing);
      return -1;
    }
}  


static INLINE void provide_data_receive_buffer(unsigned int length,
					       unsigned int size)
{
  void *buf = gm_dma_malloc(server_port, length);
  GET_RECV_TOKEN();
  gm_provide_receive_buffer(server_port, buf,
			    size, DATA_PRIORITY);
}


int GM_network_init()
{
  int i;
  int n;
  gm_status_t status;
  struct server_request req;
  int total_recv_tokens;

  status = gm_init();
  if ( status != GM_SUCCESS )
    {
      ERROR("Cannot launch GM (%s)\n", gm_strerror(status));
      return -1;
    }

  LOG(1, "GM started\n");
  status = gm_open(&server_port,
		   SERVER_BOARD_NUM, server_port_id,
		   "server_port", GM_API_VERSION);
  if ( status != GM_SUCCESS )
    {
      gm_finalize(); /* NEEDED ? */
      ERROR("Port %d on board %d not open (%s)\n",
	    server_port_id, SERVER_BOARD_NUM, gm_strerror(status));
      return -1;
    }
  LOG(2, "Port open\n");

#ifdef FIXED_NODE_ID
  gm_get_node_id(server_port_id, &i);
  if ( i != FIXED_NODE_ID )
    {
      ERROR("Server may be inaccessible with node id %d instead of %d ?\n",
	    i, FIXED_NODE_ID);
    }
#endif /* FIXED_NODE_ID */

  request_size = gm_min_size_for_length( sizeof(struct request) );  
  answer_size = gm_min_size_for_length( sizeof(struct answer) );  
  request_shift = ((void*)&(req.req)) - ((void*)&req);

  send_tokens = gm_num_send_tokens(server_port);
  recv_tokens = gm_num_receive_tokens(server_port);
  total_recv_tokens = recv_tokens;
  LOG(3, "Got %d send tokens and %d recv tokens\n",
      send_tokens, recv_tokens);

  /* We use 50% of availables token to provide request buffers.
   * Other 30% are used to preprovide dat buffers.
   */

  preprovided_buffer_size = gm_min_size_for_length(PREPROVIDED_BUFFER_LENGTH);

  n = (int)(total_recv_tokens * PREPROVIDED_REQUEST_RECV_TOKEN_RATIO);
  for(i=0; i<n; i++)
    provide_request_receive_buffer();
  LOG(4, "Allocated %d/%d recv tokens for requests (ratio %f)\n",
      n, total_recv_tokens, PREPROVIDED_REQUEST_RECV_TOKEN_RATIO);
  
  n = (int)(total_recv_tokens * PREPROVIDED_DATA_RECV_TOKEN_RATIO);
  for(i=0; i<n; i++)
    provide_data_receive_buffer(PREPROVIDED_BUFFER_LENGTH,
				preprovided_buffer_size);
  LOG(4, "Allocated %d/%d recv tokens for requests (ratio %2.2f)\n",
      n, total_recv_tokens, PREPROVIDED_DATA_RECV_TOKEN_RATIO);

  LOG(4, "Kept %d/%d free recv tokens\n",
      recv_tokens, total_recv_tokens);

  return 0;
}


void GM_network_finish()
{
  gm_close(server_port);
  gm_finalize();
}



/********************
 * Sending callback *
 ********************/

static void resume_sending_callback(struct gm_port *port,
				    void *context,
				    gm_status_t status)
{
  PUT_SEND_TOKEN();
}

static void answer_send_callback(struct gm_port *port,
				 void *context,
				 gm_status_t status)
{
  struct server_request *req = (struct server_request*) context;

  if ( status != GM_SUCCESS )
    {
      ERROR("Answer's callback send status %s. Send canceled.\n", gm_strerror(status));
      gm_resume_sending(port, REQUEST_PRIORITY,
			req->connection.node_id, req->connection.port_id,
			resume_sending_callback, NULL);
    }
  else
    {
      PUT_SEND_TOKEN();
    }

  if ( !--req->send_to_do ) {
    put_io_vector(req->out_vect);
    put_free_request(req);
  }
}


static void data_send_callback(struct gm_port *port,
			       void *context,
			       gm_status_t status)
{
  struct server_request *req = (struct server_request*) context;

  if ( status != GM_SUCCESS )
    {
      ERROR("Answer's data send callback status %s. Send canceled.\n", gm_strerror(status));
      gm_resume_sending(port, DATA_PRIORITY,
			req->connection.node_id, req->connection.port_id,
			resume_sending_callback, NULL);
    }
  else
    {
      PUT_SEND_TOKEN();
    }

  if ( !--req->send_to_do ) {
    put_io_vector(req->out_vect);
    put_free_request(req);
  }
}



/***********
 * Sending *
 ***********/

int GM_network_send(struct server_request *req)
{ 
  /* copy some data because req can be freed
   * between ans and data sending */
  unsigned int node_id = req->connection.node_id;
  unsigned int port_id = req->connection.port_id;
  int data_len = req->ans.buffer_len;
  struct io_vector *out_vect = req->out_vect;

  LOG(2, "Send type %d to node %d port %d\n",
      (int)req->ans.type,
      (int)node_id,
      (int)port_id);

  /* check ANS type */
  if ( req->ans.type <= FIRST_ANS
       ||  req->ans.type >= LAST_ANS )
    {
      ERROR("Cannot come here !!\n");
      assert(0);
    }

  /* send answer */
  req->send_to_do++;
  GET_SEND_TOKEN();
  GM_LOCK();
  gm_send_with_callback(server_port,
			&req->ans,
			answer_size, sizeof(struct answer),
			REQUEST_PRIORITY,
			node_id, port_id,
			answer_send_callback,
			req);
  GM_UNLOCK();

  /* at this point, waiting thread can be schedule,
   * get the send event and free the server_request. */

  /* send data if necessary */
  if ( data_len >= 0 )
    {
      if ( data_len )
	out_vect = merge_io_vector(out_vect);
      else {
	/* GM cannot send empty messages
	   so we send 1 byte instead of 0 */
	put_io_vector(out_vect);
	out_vect = get_io_monovector_alloc(1);
	data_len = 1;
      }
      /* update req for future freeing */
      req->out_vect = out_vect;

      req->send_to_do++;
      GET_SEND_TOKEN();
      GM_LOCK();
      gm_send_with_callback(server_port,
			    VECT_PTR(out_vect, 0), 
			    ANSWER_DATA_SIZE, data_len,
			    DATA_PRIORITY,
			    node_id, port_id,
			    data_send_callback,
			    req);
      GM_UNLOCK();

      LOG(3, "Sent %d data\n",
	  data_len);
    }

  return 0;
}



/*************
 * Receiving *
 *************/

static int network_recv(gm_recv_event_t *event,
			int priority)
{
  if ( priority == REQUEST_PRIORITY )
    /* receive a new request */
    {
      int data_len;
      struct server_request *req =
	((void*)gm_ntohp(event->recv.buffer))
	- request_shift;

      req->connection.node_id = gm_ntoh_u16(event->recv.sender_node_id);
      req->connection.port_id = gm_ntoh_u8(event->recv.sender_port_id);

      req->in_vect = NULL;

      /* Check req->req.type */
      LOG(2, "Recu type %d from %d %d\n",
	  (int)req->req.type,
	  req->connection.node_id, req->connection.port_id);
      if ( req->req.type <= FIRST_REQ
	   ||  req->req.type >= LAST_REQ )
	{
	  ERROR("Bad request type %d received, impossible with GM ?!\n",
		(int)req->req.type);
	  assert(0);
	}
      
      data_len = req->req.buffer_len;
      if ( data_len > 0 )
	{
	  struct io_vector *vect;
	  unsigned int size;

	  size = REQUEST_DATA_SIZE_OF_LENGTH(data_len);

	  vect = dequeue_not_ready_buffer_for_GM_connection(&req->connection);

	  if ( vect )
	    {
	      req->in_vect = vect;
	      VECT_SIZE(vect, 0) = req->req.buffer_len;
	      queue_ready_request(req);
	      LOG(2, "Received request and found buffer in queue\n");
	    }
	  else
	    {
	      if ( size > preprovided_buffer_size )
		provide_data_receive_buffer(gm_max_length_for_size(size), size);
	      queue_not_ready_request(req);
	      LOG(2, "Received request but no buffer found in queue\n");
	    }
	}
      else
	{
	  queue_ready_request(req);
	}
    }
  else
    {
      /* receive data */
      struct server_request *req;
      struct connection connection;
      unsigned int size = gm_ntoh_u8(event->recv.size);
      void *buf = gm_ntohp(event->recv.buffer);
      
      /* provide a new buffer */
      if ( size == preprovided_buffer_size )
	provide_data_receive_buffer(PREPROVIDED_BUFFER_LENGTH, preprovided_buffer_size);
	
      /* get the used buffer */
      connection.node_id = gm_ntoh_u16(event->recv.sender_node_id);
      connection.port_id = gm_ntoh_u8(event->recv.sender_port_id);
      req = dequeue_not_ready_request_for_GM_connection(&connection);

      if ( req )
	{
	  req->in_vect = get_io_monovector(buf, req->req.buffer_len, 0);
	  LOG(2, "Received buffer and found request in queue\n");
	  queue_ready_request(req);
	}
      else
	{
	  size_t len = gm_ntoh_u32(event->recv.length);
	  struct io_vector *vect = get_io_monovector(buf, len, 0);
	 
	  LOG(2, "Received buffer but no request found in queue\n");
	  queue_not_ready_buffer(vect, &connection);
	}
    }

  return 0;
}



#ifdef AIO_MIXED_WITH_OTHER_ASYNC_API

/* this loop is exited only if no more event have
 * arrived. We also goto out only at GM_NO_RECV_EVENT. */
void GM_network_test()
{
  GM_LOCK();
  while ( 1 )
    {  
      gm_recv_event_t *event;

      event = gm_receive(server_port);

      LOG(2, "ev type %d\n", gm_ntoh_u8(event->recv.type));

      switch ( gm_ntoh_u8(event->recv.type) )
	{
	case DATA_RECV_EVENT:
	  {
	    LOG(3, "received event %d (data) at %p len %d\n",
		gm_ntoh_u8(event->recv.type),
		gm_ntohp(event->recv.buffer),
		gm_ntoh_u32(event->recv.length));

	    PUT_RECV_TOKEN();

	    network_recv(event, DATA_PRIORITY);
	    break;
	  }
	case REQUEST_RECV_EVENT:
	  {
	    LOG(3, "received event %d (request) at %p len %d\n",
		gm_ntoh_u8(event->recv.type),
		gm_ntohp(event->recv.buffer),
		gm_ntoh_u32(event->recv.length));
	    
	    PUT_RECV_TOKEN();

	    provide_request_receive_buffer();

	    network_recv(event, REQUEST_PRIORITY);
	    break;
	  }

	case GM_NO_RECV_EVENT:
	  {
	    goto out;
	  }
	
	case _GM_SLEEP_EVENT:
	  {
	    ERROR("_GM_SLEEP_EVENT ?!\n");
	    assert(0);
	  }	    

	default:
	  {
	    gm_unknown(server_port, event);
	  }
	}
    }
 out:
  GM_UNLOCK();
}

#else /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */

/* This loop is exited only if a receive event arrives */
void GM_network_wait()
{
  GM_LOCK();
  while ( 1 )
    {
      gm_recv_event_t *event;

      event = gm_blocking_receive(server_port);

      LOG(2, "ev type %d\n", gm_ntoh_u8(event->recv.type));

      switch ( gm_ntoh_u8(event->recv.type) )
	{
	case DATA_RECV_EVENT:
	  {
	    LOG(3, "received event %d (data) at %p len %d\n",
		gm_ntoh_u8(event->recv.type),
		gm_ntohp(event->recv.buffer),
		gm_ntoh_u32(event->recv.length));
	     
	    PUT_RECV_TOKEN();

	    network_recv(event, DATA_PRIORITY);
	    goto out;
	  }
	case REQUEST_RECV_EVENT:
	  {
	    LOG(3, "received event %d (request) at %p len %d\n",
		gm_ntoh_u8(event->recv.type),
		gm_ntohp(event->recv.buffer),
		gm_ntoh_u32(event->recv.length));
	    
	    PUT_RECV_TOKEN();

	    provide_request_receive_buffer();

	    network_recv(event, REQUEST_PRIORITY);
	    goto out;
	  }

	case GM_NO_RECV_EVENT:
	  {
	    ERROR("Did not receive anything, that's strange...\n");
	    assert(0);
	  }

	case _GM_SLEEP_EVENT:
	  {
	    GM_UNLOCK();
	    gm_unknown(server_port, event);
	    GM_LOCK();
	    break;
	  }	    

	default:
	  {
	    gm_unknown(server_port, event);
	  }
	}
    }
 out:
  GM_UNLOCK();
}

#endif /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */



/****************************************
 * Network Receive Wait For Send Tokens *
 ****************************************/

#ifndef QUEUE_THREAD_ARCH

/* this loop is exited only if send_tokens
 * is non-null. This means that receive events
 * are handled but will not break the while loop. */
static void GM_network_wait_for_send_tokens()
{
  GM_LOCK();
  while ( !send_tokens )
    {
      gm_recv_event_t *event;

      event = gm_blocking_receive(server_port);

      LOG(2, "ev type %d\n", gm_ntoh_u8(event->recv.type));

      switch ( gm_ntoh_u8(event->recv.type) )
	{
	case DATA_RECV_EVENT:
	  {
	    LOG(3, "received event %d (data) at %p len %d\n",
		gm_ntoh_u8(event->recv.type),
		gm_ntohp(event->recv.buffer),
		gm_ntoh_u32(event->recv.length));
	     
	    PUT_RECV_TOKEN();

	    network_recv(event, DATA_PRIORITY);
	    break;
	  }
	case REQUEST_RECV_EVENT:
	  {
	    LOG(3, "received event %d (request) at %p len %d\n",
		gm_ntoh_u8(event->recv.type),
		gm_ntohp(event->recv.buffer),
		gm_ntoh_u32(event->recv.length));
	    
	    PUT_RECV_TOKEN();

	    provide_request_receive_buffer();

	    network_recv(event, REQUEST_PRIORITY);
	    break;
	  }

	case GM_NO_RECV_EVENT:
	  {
	    ERROR("Did not receive anything, that's strange...\n");
	    assert(0);
	  }

	case _GM_SLEEP_EVENT:
	  {
	    GM_UNLOCK();
	    gm_unknown(server_port, event);
	    GM_LOCK();
	    break;
	  }	    

	default:
	  {
	    gm_unknown(server_port, event);
	  }
	}
    }
  GM_UNLOCK();
}

#endif /* ~QUEUE_THREAD_ARCH */
