/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <utime.h>
#include <errno.h>
#include <assert.h>

#include "orfas_network.h"
#include "orfas_io.h"
#include "orfas_portable.h"
#include "orfas_debug.h"

#ifdef CONNECTION_THREAD_ARCH
#include <pthread.h>
#endif



/******************
 * Various Macros *
 ******************/

#define OFFSET_IN_BLOCK(x) ((x)&(MEMORY_BLOCK_SIZE-1))
#define BLOCK_FROM_OFFSET(x) (int)((x)>>MEMORY_BLOCK_BITS)

/* each directory contains two trivial dirents: . and .. */
#define TRIVIAL_DIRENTS_NR 2



/*****************
 * Various types *
 *****************/

typedef enum {
  PATH_NOT_FOUND,
  PATH_NOT_DIR_FOUND,
  FILE_NOT_IN_PATH,
  FILE_FOUND
} lookup_res_t;



/*******************
 * Data structures *
 *******************/

struct RAM_memory_block {
  void *ptr;
};

struct RAM_file_data {
  struct RAM_memory_block *block;
  int block_nr;
  int block_max;
  off64_t length;
};

struct RAM_dirent {
  struct portable_dirent dirent;
};

struct RAM_dir_data {
  struct RAM_dirent **entry;
  int entry_nr;
  int entry_max;
};

union RAM_node_data {
  struct RAM_file_data file;
  struct RAM_dir_data dir;
};



/*******************
 * Node structures *
 *******************/

typedef unsigned long RAM_ino_t;

struct RAM_node {
  /* data */
  orfa_mode_t mode;
  /* owner */
  orfa_uid_t uid;
  orfa_gid_t gid;
  /* times */
  orfa_time_t access_time;
  orfa_time_t modify_time;
  orfa_time_t change_time;

  /* ino */
  RAM_ino_t ino;
  /* specific data */
  union RAM_node_data data;
  /* dentry data */
  struct RAM_node *parent;
  unsigned int indix_for_parent;
  /* hash data */
#if 0
  struct RAM_node *ino_next, *ino_prev;
  unsigned int ino_hash_key;
#endif
  struct RAM_node *name_next, *name_prev;
  unsigned int name_hash_key;
#if CONNECTION_THREAD_ARCH
  /* This lock is used to protect from concurrent access.
   * As it is not used yet, we don't know whether a
   * spinlock is better or not. */
  pthread_mutex_t mutex;
#endif
};


struct RAM_fs {
#if 0
  struct RAM_node **ino_table;
  size_t ino_table_size;
#endif
  struct RAM_node **name_table;
  size_t name_table_size;
  struct RAM_node *root;
  RAM_ino_t ctr;
#if CONNECTION_THREAD_ARCH
  /* This lock is used to protect from concurrent access.
   * As it is not used yet, we don't know whether a
   * spinlock is better or not. */
  pthread_mutex_t mutex;
#endif
};


#if CONNECTION_THREAD_ARCH

#define LOCK_NODE(x) pthread_mutex_lock(&x->mutex)
#define UNLOCK_NODE(x) pthread_mutex_unlock(&x->mutex)
#define LOCK_FS(x) pthread_mutex_lock(&x->mutex)
#define UNLOCK_FS(x) pthread_mutex_unlock(&x->mutex)

#else

#define LOCK_NODE(x) /* nothing */
#define UNLOCK_NODE(x) /* nothing */
#define LOCK_FS(x) /* nothing */
#define UNLOCK_FS(x) /* nothing */

#endif



/**********************
 * Opening structures *
 **********************/

struct RAM_fd {
  struct RAM_node *node;
  uint64_t flags;
#ifdef CONNECTION_ID_ENABLED
  struct connection_id *connection_id;
#endif /* CONNECTION_ID_ENABLED */
  struct RAM_fd *next, *prev;
};

struct RAM_dir {
  struct RAM_node *node;
#ifdef CONNECTION_ID_ENABLED
  struct connection_id *connection_id;
#endif /* CONNECTION_ID_ENABLED */
  struct RAM_dir *next, *prev;
};


/*******************
 * Our file system *
 *******************/

static struct RAM_fs *fs = NULL;

static struct RAM_fd *first_fd = NULL;

static struct RAM_dir *first_dir = NULL;


/*************************
 * Hashing key functions *
 *************************/

static INLINE unsigned int RAM_ino_hash_func(RAM_ino_t ino)
{
  return ino % HASH_VALUE;
}

static INLINE unsigned int __name_hash_func(const char *name)
{
  unsigned int hash = *name;

  if (hash)
    for(name++; *name != '\0'; name++)
      hash = (hash<<5) - hash + *name;

  return hash;
}

static INLINE unsigned int name_hash_func(const char *name, struct RAM_node *parent)
{
  return RAM_ino_hash_func(__name_hash_func(name)+parent->ino);
}



/*********************
 * Hashing functions *
 *********************/

#if 0
static void ino_hash_node(struct RAM_node *node)
{
  unsigned int ino_hash_key = RAM_ino_hash_func(node->ino);

  if ( fs->ino_table[ino_hash_key] )
    fs->ino_table[ino_hash_key]->ino_prev = node;
  node->ino_next = fs->ino_table[ino_hash_key];

  fs->ino_table[ino_hash_key] = node;
  node->ino_prev = NULL;

  node->ino_hash_key = ino_hash_key;
}
#endif 

static void name_hash_node(struct RAM_node *node, char *name)
{
  unsigned int name_hash_key = name_hash_func(name, node->parent);

  if ( fs->name_table[name_hash_key] )
    fs->name_table[name_hash_key]->name_prev = node;
  node->name_next = fs->name_table[name_hash_key];
  
  fs->name_table[name_hash_key] = node;
  node->name_prev = NULL;

  node->name_hash_key = name_hash_key;
}
 
#if 0
static void ino_unhash_node(struct RAM_node *node)
{
  unsigned int ino_hash_key = node->ino_hash_key;
  
  /* unhash from ino_table */
  if ( node->ino_prev )
    node->ino_prev->ino_next = node->ino_next;
  else
    fs->ino_table[ino_hash_key] = node->ino_next;

  if ( node->ino_next )
    node->ino_next->ino_prev = node->ino_prev;
}
#endif

static void name_unhash_node(struct RAM_node *node)
{
  unsigned int name_hash_key = node->name_hash_key;

  /* unhash from name_table */
  if ( node->name_prev )
    node->name_prev->name_next = node->name_next;
  else
    fs->name_table[name_hash_key] = node->name_next;

  if ( node->name_next )
    node->name_next->name_prev = node->name_prev;
}



/**********
 * Parent *
 **********/

static void add_to_parent(struct RAM_node *node, char *name)
{
  struct RAM_dir_data *data = &node->parent->data.dir;
  int nr;

  if ( data->entry_nr == data->entry_max ) {
    /* increase list size if necessary */
    data->entry = (struct RAM_dirent**) realloc(data->entry,
						data->entry_max*2
						*sizeof(struct RAM_dirent*));
    /* initialize it */
    memset(&data->entry[data->entry_max],
	   0, data->entry_max*sizeof(struct RAM_dirent*));

    /* fix its size */
    data->entry_max *= 2;

    /* let's take the first new place */
    nr = data->entry_nr;
  } else {
    /* search the first empty case, we know there's one */
    for(nr=TRIVIAL_DIRENTS_NR; data->entry[nr]!=NULL; nr++);
  }

  /* add the new node */
  data->entry[nr] = (struct RAM_dirent*) network_buffer_malloc(sizeof(struct RAM_dirent));
  node->indix_for_parent = nr;

  strcpy(data->entry[nr]->dirent.d_name, name);
  data->entry[nr]->dirent.d_type = 0;

  /* compute d_off */
  data->entry[nr]->dirent.d_off = data->entry[nr-1]->dirent.d_off;
  data->entry[nr-1]->dirent.d_off = nr;

  /* increase list nr */
  data->entry_nr++;
}


static void remove_from_parent(struct RAM_node *node)
{
  struct RAM_dir_data *data = &node->parent->data.dir;
  int indix = node->indix_for_parent;
  int i;

  /* update d_off of the last entry before this one */
  for(i=indix-1; data->entry[i] == NULL; i--);
  data->entry[i]->dirent.d_off = data->entry[indix]->dirent.d_off;
  
  /* remove the old entry */
  network_buffer_free(data->entry[indix]);
  data->entry[indix] = NULL;
  
  /* decrease list nr */
  data->entry_nr--;
}



/**********
 * Blocks *
 **********/

static INLINE void add_blocks(struct RAM_file_data *data, int nr)
{
  int i;
  
  /* increase block table size if necessary */
  if ( nr > data->block_max )
    {
      int new_max = data->block_max * 2;
      while ( nr > new_max )
        new_max <<= 1;

      data->block = (struct RAM_memory_block*)
	realloc(data->block, new_max * sizeof(struct RAM_memory_block));
      data->block_max = new_max;
    }

  /* malloc new blocks */
  for(i=data->block_nr; i<nr; i++)
      data->block[i].ptr
	= network_buffer_malloc(MEMORY_BLOCK_SIZE);

  /* increase block count */
  data->block_nr = nr;
}


static INLINE void remove_last_blocks(struct RAM_file_data *data, int nr)
{
  int i;

  /* free blocks */
  for(i=nr; i<data->block_nr; i++)
    network_buffer_free(data->block[i].ptr);
  
  /* decrease block count */
  data->block_nr = nr;
}



/******************
 * Node inserting *
 ******************/

static INLINE RAM_ino_t get_new_ino()
{
  static RAM_ino_t next_ino = 0;
  return next_ino++;
}


static INLINE void get_new_file_data(struct RAM_file_data *data)
{
  data->block = (struct RAM_memory_block*)
    malloc(BLOCKS_DEFAULT_MAX * sizeof(struct RAM_memory_block));
  data->block_max = BLOCKS_DEFAULT_MAX;
  data->block_nr = 0;
  data->length = 0;
}


static INLINE void get_new_dir_data(struct RAM_dir_data *data,
				    struct RAM_node *node,
				    struct RAM_node *parent)
{
  data->entry = (struct RAM_dirent**)
    calloc(DIRENTS_DEFAULT_MAX, sizeof(struct RAM_node*));
  data->entry_max = DIRENTS_DEFAULT_MAX;

  data->entry[0] = (struct RAM_dirent*) network_buffer_malloc(sizeof(struct RAM_dirent));
  strcpy(data->entry[0]->dirent.d_name, ".");
  data->entry[0]->dirent.d_type = 0;
  data->entry[0]->dirent.d_off = 1;  
  
  data->entry[1] = (struct RAM_dirent*) network_buffer_malloc(sizeof(struct RAM_dirent));
  strcpy(data->entry[1]->dirent.d_name, "..");
  data->entry[1]->dirent.d_type = 0;
  data->entry[1]->dirent.d_off = -1;

  data->entry_nr = TRIVIAL_DIRENTS_NR;
}


static struct RAM_node *get_new_node(char *name,
				     struct RAM_node *parent,
				     orfa_mode_t mode)
{
  struct RAM_node *node = (struct RAM_node*) malloc( sizeof(struct RAM_node) );

  node->ino = get_new_ino();

  node->parent = parent;
  node->mode = mode;

  if ( S_ISREG(mode) )
    get_new_file_data(&node->data.file);
  else if ( S_ISDIR(mode) )
    get_new_dir_data(&node->data.dir, node, parent);
  else
    assert(0);

#if 0
  ino_hash_node(node);
#endif
  name_hash_node(node, name);

  add_to_parent(node, name);

  return node;
}



/*****************
  * Node removing *
 *****************/

static INLINE void free_file_data(struct RAM_file_data *data)
{
  remove_last_blocks(data, 0);
  free(data->block);
}


static INLINE void free_dir_data(struct RAM_dir_data *data)
{
  network_buffer_free(data->entry[0]);
  network_buffer_free(data->entry[1]);
  free(data->entry);
}


static void free_node_without_parent(struct RAM_node *node)
{
  /* free specific data */
  if ( S_ISREG(node->mode) )
    free_file_data(&node->data.file);
  else if ( S_ISDIR(node->mode) )
    free_dir_data(&node->data.dir);
  else
    assert(0);

  /* free */
  free(node);
}


static void free_node(struct RAM_node *node)
{
  /* remove from parent dirent list */
  remove_from_parent(node);

  /* unhash */
#if 0 
  ino_unhash_node(node);
#endif
  name_unhash_node(node);

  /* real remove */
  free_node_without_parent(node);
}


/**********
 * Lookup *
 **********/
#if 0
static struct RAM_node *lookup_ino(RAM_ino_t ino)
{
  unsigned int hash_key;
  struct RAM_node *node;

  hash_key = RAM_ino_hash_func(ino);
  node = fs->ino_table[hash_key];

  while ( node != NULL )
    {
      if ( node->ino == ino )
	return node;
      node = node->ino_next;
    }
  return NULL;
}
#endif

static struct RAM_node *lookup_name(const char *name, struct RAM_node *parent)
{
  unsigned int hash_key;
  struct RAM_node *node;
  struct RAM_dirent **entry = parent->data.dir.entry;

  hash_key = name_hash_func(name, parent);
  node = fs->name_table[hash_key];

  while ( node != NULL )
    {
      if ( node->parent == parent ) {
	unsigned int indix_for_parent = node->indix_for_parent;
	char *node_name = entry[indix_for_parent]->dirent.d_name;
	if ( !strcmp(node_name, name) )
	  return node;
      }
      node = node->name_next;
    }
  return NULL;
}


static lookup_res_t lookup_path(struct RAM_node **looked_node,
				const char *full_name,
				struct RAM_node **parent_node,
				char **file_name)
{
  char *current_slash, *next_slash;
  int len;
  int finished = 0;

  struct RAM_node *current_node = fs->root;
  struct RAM_node *prev_node = NULL;
  char subname[RAMFS_MAX_FILE_NAME];

  LOG(5, "Lookup for file %s\n", full_name);

  /* the initial '/' is the first one : at the begining */
  current_slash = (char*)full_name;
  subname[0] = '\0';

  if ( *(current_slash+1) == '\0' )
    finished = 1;
      
  while ( !finished )
    {
      if ( !S_ISDIR(current_node->mode) )
	goto not_dir_found;

      next_slash = strchr(current_slash+1, '/');
      if ( next_slash == NULL )
	{
	  len = strlen(current_slash+1);
	  finished = 1;
	}
      else 
	{
	  len = next_slash - current_slash-1;
	}
      
      strncpy(subname, current_slash+1, len);
      subname[len] = '\0';
      current_slash = next_slash;

      LOG(6, "Extracted subname %s\n", subname);

      prev_node = current_node;
      current_node = lookup_name(subname, prev_node);

      if ( current_node == NULL )
	goto not_found;
    }

  /* the file exists */
  LOG(6, "File exists\n");
  if ( parent_node )
    *parent_node = prev_node;
  if ( file_name )
    *file_name = NULL;
  *looked_node = current_node;
  return FILE_FOUND;

 not_found:
  if ( finished )
    /* the path exists but not the file.
     * the only case where we allocate a buffer for file_name
     */
    {
      LOG(6, "The path exists but not the file\n");
      if ( parent_node )
	*parent_node = prev_node;
      if ( file_name )
	*file_name = strdup(subname);
      *looked_node = NULL;
      return FILE_NOT_IN_PATH;
    }
  else
    /* the path doesn't exist */
    {
      LOG(6, "The path doesn't exist\n");
      if ( parent_node )
	*parent_node = NULL;
      if ( file_name )
	*file_name = NULL;
      *looked_node = NULL;
      return PATH_NOT_FOUND;
    }      

 not_dir_found:
  /* a part of the path is not a directory */
  LOG(6, "A subname of the path is not a directory\n");
  if ( parent_node )
    *parent_node = NULL;
  if ( file_name )
    *file_name = NULL;
  *looked_node = NULL;
  return PATH_NOT_DIR_FOUND;
}



/****************
 * FS functions *
 ****************/

static INLINE orfa_time_t get_current_time() {
  orfa_time_t time;
#if 0
  struct timeval timeval;
  gettimeofday(&timeval, NULL);
  time = timeval.tv_sec;
#else
  time = 0;
#endif
  return time;
}

static INLINE struct RAM_node *create_node(const char *name,
					   struct RAM_node *parent,
					   orfa_mode_t mode,
					   orfa_uid_t uid, orfa_gid_t gid)
{
  struct RAM_node *node = get_new_node((char*)name, parent, mode);
  orfa_time_t time = get_current_time();
  node->access_time = time;
  node->change_time = time;
  node->modify_time = time;
  node->uid = uid;
  node->gid = gid;
  return node;
}


static INLINE int remove_node(struct RAM_node *node)
{
  free_node(node);
  return 0;
}


static INLINE orfa_soff_t pread_node(struct RAM_node *node,
				     struct io_vector **buf,
				     orfa_off_t nbyte,
				     orfa_off_t offset)
{
  struct RAM_file_data *data = &node->data.file;
  orfa_off_t begin;
  orfa_off_t end;
  int begin_block;
  int end_block;
  int begin_offset;

  if ( offset >= data->length )
    return 0;

  begin = offset;
  end = offset+nbyte-1;
  if ( end > data->length-1 ) {
    end = data->length-1;
    nbyte = end-begin+1;
  }

  if ( nbyte < 0 )
    return 0;

  begin_block = BLOCK_FROM_OFFSET(begin);
  end_block = BLOCK_FROM_OFFSET(end);
  begin_offset = OFFSET_IN_BLOCK(begin);

  if ( begin_block == end_block )
    {
      *buf = get_io_monovector(data->block[begin_block].ptr + begin_offset,
			       nbyte,
			       1);
      return nbyte;
    }
  else
    {
      struct RAM_memory_block *current_block = &data->block[begin_block];
      struct iovec *current_buffer;
      orfa_off_t remaining_bytes;
      
      /* allocate an output vector */
      *buf = get_io_vector(end_block-begin_block+1, 1);

      /* insert first block */
      current_buffer = VECT_BUF(*buf, 0);
      BUF_PTR(current_buffer) = current_block->ptr + begin_offset;
      BUF_SIZE(current_buffer) = MEMORY_BLOCK_SIZE-begin_offset;

      /* prepare following copy */
      current_block++;
      current_buffer++;
      remaining_bytes = nbyte - (MEMORY_BLOCK_SIZE-begin_offset);
      do {
	orfa_off_t to_copy = remaining_bytes > MEMORY_BLOCK_SIZE
	  ? MEMORY_BLOCK_SIZE 
	  : remaining_bytes;

	/* insert new buffer */
	BUF_PTR(current_buffer) = current_block->ptr;
	BUF_SIZE(current_buffer) = to_copy;

	/* update parameters */
	current_block++;
	current_buffer++;
	remaining_bytes -= to_copy;
      } while ( remaining_bytes > 0 );

      return nbyte;
    }
} 


static INLINE orfa_soff_t pwrite_node(struct RAM_node *node,
				      const char *buf,
				      orfa_off_t nbyte,
				      orfa_off_t offset)
{
  struct RAM_file_data *data = &node->data.file;
  orfa_off_t begin;
  orfa_off_t end;
  int begin_block;
  int end_block;
  int begin_offset;

  begin = offset;
  end = offset+nbyte-1;

  if ( end < begin )
    return 0;

  begin_block = BLOCK_FROM_OFFSET(begin);
  end_block = BLOCK_FROM_OFFSET(end);
  begin_offset = OFFSET_IN_BLOCK(begin);

  if ( end > data->length-1 )
    data->length = end+1;

  if ( end_block >= data->block_nr )
    add_blocks(data, end_block+1);

  if ( begin_block == end_block )
    {
      memcpy(data->block[begin_block].ptr + begin_offset,
	     buf,
	     nbyte);
      return nbyte;
    }
  else
    {
      struct RAM_memory_block *current_block = &data->block[begin_block];
      char *current_bufptr;
      orfa_off_t remaining_bytes;

      /* copy the end of the first block */
      memcpy(current_block->ptr + begin_offset,
	     buf,
	     MEMORY_BLOCK_SIZE-begin_offset);

      /* prepare following copy */
      current_block++;
      remaining_bytes = nbyte - (MEMORY_BLOCK_SIZE-begin_offset);
      current_bufptr = (char*)buf + (MEMORY_BLOCK_SIZE-begin_offset);

      do {
	orfa_off_t to_copy = remaining_bytes > MEMORY_BLOCK_SIZE
	  ? MEMORY_BLOCK_SIZE 
	  : remaining_bytes;

	/* copy data */
	memcpy(current_block->ptr,
	       buf,
	       to_copy);
	/* update parameters */
	current_bufptr += to_copy;
	remaining_bytes -= to_copy;
	current_block++;
      } while ( remaining_bytes > 0 );
     
      return nbyte;
    }
} 


static int truncate_node(struct RAM_node *node, orfa_off_t len)
{
  struct RAM_file_data *data = &node->data.file;
  int new_nr = BLOCK_FROM_OFFSET(len+MEMORY_BLOCK_SIZE-1);

  if ( len < data->length )
    remove_last_blocks(data, new_nr);
  else
    add_blocks(data, new_nr);

  data->length = len;

  return 0;
}


static int stat_node(struct RAM_node *node, struct portable_stat *buf)
{
  buf->st_dev = 0;
  buf->st_ino = node->ino;
  buf->st_mode = node->mode;
  buf->st_uid = 0;
  buf->st_gid = 0;
  buf->st_rdev = 0;

  if ( S_ISREG(node->mode) )
    {
      buf->st_nlink = 1;
      buf->st_size = node->data.file.length;
      buf->st_blksize = MEMORY_BLOCK_SIZE;
    }
  else if ( S_ISDIR(node->mode) )
    {
      buf->st_nlink = 1;
      buf->st_size = node->data.dir.entry_nr;
      buf->st_blksize = 1;
    }
  else
    assert(0);

  buf->st_blocks = buf->st_size / buf->st_blksize;
  convert_stat_time_to_portable(buf, atime, node->access_time);
  convert_stat_time_to_portable(buf, mtime, node->modify_time);
  convert_stat_time_to_portable(buf, ctime, node->change_time);

  return 0;
}


static int chmod_node(struct RAM_node *node, orfa_mode_t mode)
{
  node->mode = mode;

  /* to be finished */

  return 0;
}


static int chown_node(struct RAM_node *node, orfa_uid_t uid, orfa_gid_t gid)
{
  node->uid = uid;
  node->gid = gid;

  return 0;
}


static INLINE int change_time_node(struct RAM_node *node, orfa_time_t actime, orfa_time_t modtime)
{
  node->access_time = actime;
  node->modify_time = modtime;

  return 0;
}


/******************
 * User functions *
 ******************/

/* FD functions */

int RAM_fs_open(const char *name, orfa_oflags_t flags, orfa_mode_t mode, void **fdp)
{
  struct RAM_node *parent, *node;
  char *subname;
  lookup_res_t res;
  struct RAM_fd *fd;

  /* lookup for the node */
  res = lookup_path(&node, name, &parent, &subname);

  /* does the path exist ? */
  if ( res == PATH_NOT_FOUND )
    return -ENOENT;

  if ( res == PATH_NOT_DIR_FOUND )
    return -ENOTDIR;

  /* check creation flags and create */
  if ( res == FILE_NOT_IN_PATH )
    /* the path exists but not the path.
     * subname has been allocated during lookup.
     */
    {
      if ( flags & P_O_CREAT )
	{
	  // CHECK PARENT DIR PERMS
	  node = create_node(subname, parent, mode | S_IFREG, 0, 0);
	  free(subname);
	}
      else
	{
	  free(subname);
	  return -ENOENT;
	}
    }
  else
    /* the file exists */
    {
      if ( (flags & P_O_EXCL) && (flags & P_O_CREAT) )
	return -EEXIST;
      else
	{
	  // CHECK FILE PERMS
	}
    }

  /* file type */
  if ( S_ISDIR(node->mode) )
    {
      if ( flags & O_DIRECTORY )
	return -ENOSYS;
      else
	return -ENOTDIR;
    }
  else if ( !S_ISREG(node->mode) )
    {
      assert(0);
    }

  /* truncate if necessary */
  if ( flags & P_O_TRUNC )
    truncate_node(node, 0);

  /* get an fd */
  fd = (struct RAM_fd*) malloc(sizeof(struct RAM_fd));

  fd->next = first_fd;
  if ( first_fd )
    first_fd->prev = fd;

  first_fd = fd;
  fd->prev = NULL;

  fd->node = node;
  fd->flags = flags;

  *fdp = fd;
  return 0;
}


int RAM_fs_close(void *fdp)
{
  struct RAM_fd *fd = (struct RAM_fd*)fdp;

  if ( fd->prev )
    fd->prev->next = fd->next;
  else
    first_fd = fd->next;

  if ( fd->next )
    fd->next->prev = fd->prev;

  free(fd);
  return 0;
}


orfa_soff_t RAM_fs_pread(void *fd, struct io_vector **buf, orfa_off_t nbyte, orfa_off_t off)
{
  /* check read */
  return pread_node(((struct RAM_fd*)fd)->node, buf, nbyte, off);
}


orfa_soff_t RAM_fs_pwrite(void *fd, const char *buf, orfa_off_t nbyte, orfa_off_t off)
{
  /* check write */
  if ( ((struct RAM_fd*)fd)->flags & P_O_APPEND
       ||  off > ((struct RAM_fd*)fd)->node->data.file.length )
    off = ((struct RAM_fd*)fd)->node->data.file.length;
  return pwrite_node(((struct RAM_fd*)fd)->node, buf, nbyte, off);
}


int RAM_fs_ftruncate(void *fd, orfa_off_t len)
{
  /* check write */
  return truncate_node(((struct RAM_fd*)fd)->node, len);
}


int RAM_fs_fstat(void *fd, struct portable_stat *buf)
{
  return stat_node(((struct RAM_fd*)fd)->node, buf);
}


int RAM_fs_fchmod(void *fd, orfa_mode_t mode)
{
  /* check owner */
  return chmod_node(((struct RAM_fd*)fd)->node, mode);
}


int RAM_fs_fchown(void *fd, orfa_uid_t uid, orfa_gid_t gid)
{
  /* check owner */
  return chown_node(((struct RAM_fd*)fd)->node, uid, gid);
}


int RAM_fs_fsync(void *fd)
{
  return RAM_fs_fsync((struct RAM_fd*)fd);
}


int RAM_fs_fdatasync(void *fd)
{
  return RAM_fs_fsync((struct RAM_fd*)fd);
}


int RAM_fs_flock(void *fd, orfa_flock_op_t operation)
{
  return -ENOSYS;
}


int RAM_fs_fcntl_setfl(void *fd, orfa_oflags_t flags)
{
  return -ENOSYS;
}


/* NAME Functions */

int RAM_fs_truncate(const char *name, orfa_off_t len)
{
  struct RAM_node *node;
  lookup_res_t res;

  res = lookup_path(&node, name, NULL, NULL);
  if ( res != FILE_FOUND )
    {
      if ( res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
	return -ENOENT;
    }

  return truncate_node(node, len);
}


int RAM_fs_stat(const char *name, struct portable_stat *buf)
{
  struct RAM_node *node;
  lookup_res_t res;

  res = lookup_path(&node, name, NULL, NULL);
  if ( res != FILE_FOUND )
    {
      if ( res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
	return -ENOENT;
    }

  return stat_node(node, buf);
}


int RAM_fs_lstat(const char *name, struct portable_stat *buf)
{
  return RAM_fs_stat(name, buf);
}


int RAM_fs_statfs(struct portable_statfs *buf)
{
  return -ENOSYS;
}


int RAM_fs_mknod(const char *name, orfa_mode_t mode, orfa_dev_t dev)
{
  struct RAM_node *parent, *node;
  char *subname;
  lookup_res_t res;

  /* lookup for the node */
  res = lookup_path(&node, name, &parent, &subname);
  if ( res == FILE_FOUND )
    return -EEXIST;

  if ( res == PATH_NOT_FOUND )
    return -ENOENT;

  if ( res == PATH_NOT_DIR_FOUND )
    return -ENOTDIR;

  /* check mode :
   * we cannot create directory with mknod
   * and other kinds are not supported
   */
  if ( !S_ISREG(mode) ) {
    free(subname);
    return -EPERM;
  }

  /* creat_node */
  create_node(subname, parent, mode, 0, 0);
  free(subname);

  return 0;
}


int RAM_fs_mkdir(const char *name, orfa_mode_t mode)
{
  struct RAM_node *parent, *node;
  char *subname;
  lookup_res_t res;

  /* lookup for the node */
  res = lookup_path(&node, name, &parent, &subname);
  if ( res == FILE_FOUND )
    return -EEXIST;

  if ( res == PATH_NOT_FOUND )
    return -ENOENT;

  if ( res == PATH_NOT_DIR_FOUND )
    return -ENOTDIR;

  /* creat_node */
  create_node(subname, parent, mode | S_IFDIR, 0, 0);
  free(subname);

  return 0;
}


int RAM_fs_rmdir(const char *name)
{
  struct RAM_node *node;
  lookup_res_t res;

  res = lookup_path(&node, name, NULL, NULL);

  if ( res != FILE_FOUND )
    {
      if ( res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
	return -ENOENT;
    }

  if ( !S_ISDIR(node->mode) )
    return -ENOTDIR;

  /* do not remove "/" */
  if ( node == fs->root )
    return -EBUSY;

  /* do not remove not empty directory */
  if ( node->data.dir.entry_nr > TRIVIAL_DIRENTS_NR )
    return -ENOTEMPTY;

  return remove_node(node);
}


int RAM_fs_unlink(const char *name)
{
  struct RAM_node *node;
  lookup_res_t res;

  res = lookup_path(&node, name, NULL, NULL);

  if ( res != FILE_FOUND )
    {
      if ( res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
        return -ENOENT;
    }

  if ( !S_ISREG(node->mode) )
    return -EISDIR;

  return remove_node(node);
}


int RAM_fs_access(const char *name, orfa_access_mode_t mode)
{
  struct portable_stat buf;
  struct RAM_node *node;
  lookup_res_t res;

  res = lookup_path(&node, name, NULL, NULL);

  if ( res != FILE_FOUND )
    {
      if ( res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
	return -ENOENT;
    }

  stat_node(node, &buf);

  /* to be finished */

  return 0;
}


int RAM_fs_chmod(const char *name, orfa_mode_t mode)
{
  struct RAM_node *node;
  lookup_res_t res;

  res = lookup_path(&node, name, NULL, NULL);

  if ( res != FILE_FOUND )
    {
      if ( res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
	return -ENOENT;
    }

  return chmod_node(node, mode);
}


int RAM_fs_chown(const char *name, orfa_uid_t uid, orfa_gid_t gid)
{
  struct RAM_node *node;
  lookup_res_t res;

  res = lookup_path(&node, name, NULL, NULL);

  if ( res != FILE_FOUND )
    {
      if ( res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
	return -ENOENT;
    }

  return chown_node(node, uid, gid);
}


int RAM_fs_lchown(const char *name, orfa_uid_t uid, orfa_gid_t gid)
{
  return RAM_fs_chown(name, uid, gid);
}


int RAM_fs_rename(const char *old_name, const char *new_name)
{
  struct RAM_node *old_node, *new_node, *new_parent;
  lookup_res_t old_res, new_res;
  char *new_subname;

  /* check old_name */
  old_res = lookup_path(&old_node, old_name, NULL, NULL);
  if ( old_res != FILE_FOUND )
    {
      if ( old_res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
	return -ENOENT;
    }

  /* do not remove "/" */
  if ( old_node == fs->root )
    return -EBUSY;

  /* check new_name */
  new_res = lookup_path(&new_node, new_name, &new_parent, &new_subname);
  if ( new_res == PATH_NOT_FOUND )
    return -ENOENT;

  if ( new_res == PATH_NOT_DIR_FOUND )
    return -ENOTDIR;

  if ( new_res == FILE_FOUND )
    {
      /* overwrite a directory only with a directory */
      if ( !S_ISDIR(old_node->mode) && S_ISDIR(new_node->mode) )
	return -EISDIR;
  
      if ( S_ISDIR(old_node->mode) )
	{
	  /* overwrite a non-directory only with a non-directory */
	  if ( !S_ISDIR(new_node->mode) )
	    return -ENOTDIR;

	  /* overwrite only an empty directory */
	  if ( new_node->data.dir.entry_nr > TRIVIAL_DIRENTS_NR )
	    return -ENOTEMPTY;
	}
    }

  /* do not move a directory into it self */
  {
    struct RAM_node *tmp = new_parent;
    while ( tmp != fs->root )
      {
	if ( tmp == old_node )
	  return -EINVAL;

	tmp = tmp->parent;
      }
  }

  /* remove the existing new_node */
  if ( new_res == FILE_FOUND )
    remove_node(new_node);

  /* remove the old_node from its path */
  name_unhash_node(old_node);
#if 0
  ino_unhash_node(old_node);
#endif
  remove_from_parent(old_node);

  /* put it in its new path */
  old_node->parent = new_parent;
  name_hash_node(old_node, new_subname);
#if 0
  ino_hash_node(old_node);
#endif
  add_to_parent(old_node, new_subname);

  if ( new_res == FILE_NOT_IN_PATH )
    free(new_subname);

  return 0;
}


int RAM_fs_link(const char *old_name, const char *new_name)
{
  return -EPERM;
}


int RAM_fs_symlink(const char *old_name, const char *new_name)
{
  return -EPERM;
}


int RAM_fs_readlink(const char *name, char *buf, orfa_off_t bufsiz)
{
  return -EINVAL;
}


int RAM_fs_utime(const char *name, orfa_time_t atime, orfa_time_t mtime, orfa_bool_t defined)
{
  struct RAM_node *node;
  lookup_res_t res;

  res = lookup_path(&node, name, NULL, NULL);

  if ( res != FILE_FOUND )
    {
      if ( res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
        return -ENOENT;
    }

  if ( defined )
    return change_time_node(node, atime, mtime);
  else {
    orfa_time_t time = get_current_time();
    return change_time_node(node, time, time);
  }
}


/* DIR Functions */

int RAM_fs_opendir(const char *name, void **dirp)
{
  struct RAM_node *node;
  struct RAM_dir *dir;
  lookup_res_t res;

  res = lookup_path(&node, name, NULL, NULL);

  if ( res != FILE_FOUND )
    {
      if ( res == PATH_NOT_DIR_FOUND )
	return -ENOTDIR;
      else
	return -ENOENT;
    }

  if ( !S_ISDIR(node->mode) )
    return -ENOTDIR;
  
  dir = (struct RAM_dir*) malloc( sizeof(struct RAM_dir) );

  dir->next = first_dir;
  if ( first_dir )
    first_dir->prev = dir;

  first_dir = dir;
  dir->prev = NULL;

  dir->node = node;

  *dirp = (void*)dir;
  return 0;
}


int RAM_fs_preaddir(void *dirp,
		    orfa_off_t off,
		    struct io_vector **d,
		    orfa_off_t count)
{
  struct RAM_dir *dir = (struct RAM_dir*)dirp;
  struct RAM_dir_data *data = &dir->node->data.dir;
  int i;

  if ( off >= data->entry_max )
    return 0;

  if ( count <= 0 )
    return 0;

  *d = get_io_vector(count, 1);

  for(i=0; i<count; i++) {
    if ( off < 0  ||  data->entry[off] == NULL ) {
      count = i;
      break;
    }
    VECT_PTR(*d, i) = &data->entry[off]->dirent;
    VECT_SIZE(*d, i) = sizeof(struct portable_dirent);
    off = data->entry[off]->dirent.d_off;
  }

  VECT_NR(*d) = count;
  return count;
}


int RAM_fs_rewinddir(void *dirp)
{
  return 0;
}


int RAM_fs_closedir(void *dirp)
{
  struct RAM_dir *dir = (struct RAM_dir*)dirp;

  if ( dir->prev )
    dir->prev->next = dir->next;
  else
    first_dir = dir->next;

  if ( dir->next )
    dir->next->prev = dir->prev;

  free(dir);
  return 0;
}



/******************
 * Initialization *
 ******************/

static INLINE void create_root_node()
{
  struct RAM_node *node = (struct RAM_node*) malloc( sizeof(struct RAM_node) );
  
  node->ino = get_new_ino();

  node->parent = node;
  node->mode = S_IFDIR | ACCESSPERMS;

  node->uid = 0;
  node->gid = 0;

  get_new_dir_data(&node->data.dir, node, node);

  name_hash_node(node, "");
#if 0
  ino_hash_node(node);
#endif

  fs->root = node;
}


int RAM_fs_init()
{
  fs = (struct RAM_fs*) calloc(1, sizeof(struct RAM_fs));

  fs->name_table_size = HASH_TABLE_SIZE;
  fs->name_table = (struct RAM_node**)
    calloc(fs->name_table_size, sizeof(struct RAM_node*));
#if 0
  fs->ino_table_size = HASH_TABLE_SIZE;
  fs->ino_table = (struct RAM_node**)
    calloc(fs->ino_table_size, sizeof(struct RAM_node*));
#endif

  create_root_node();

  return 0;
}


void RAM_fs_finish()
{
  int i;
  for(i=0; i<fs->name_table_size; i++)
    {
      struct RAM_node *node, *next;
      for(node=fs->name_table[i]; node!=NULL; node=next)
	{
	  next = node->name_next;
	  free_node_without_parent(node);
	}
    }
  free(fs->name_table);
#if 0
  free(fs->ino_table);
#endif 
  free(fs);
}


#ifdef CONNECTION_ID_ENABLED
void RAM_set_fd_connection_id(void *fd, struct connection_id *id)
{
  ((struct RAM_fd*)fd)->connection_id = id;
}


void RAM_set_dir_connection_id(void *dir, struct connection_id *id)
{
  ((struct RAM_dir*)dir)->connection_id = id;
}


void RAM_close_file_for_connection_id(struct connection_id *connection_id)
{
  struct RAM_dir *tmp_dir = first_dir;
  struct RAM_fd *tmp_fd = first_fd;

  while ( tmp_fd != NULL )
    {
      if ( tmp_fd->connection_id == connection_id )
	RAM_fs_close(tmp_fd);
      tmp_fd = tmp_fd->next;
    }

  while ( tmp_dir != NULL )
    {
      if ( tmp_dir->connection_id == connection_id )
	RAM_fs_closedir(tmp_dir);
      tmp_dir = tmp_dir->next;
    }
}
#endif /* CONNECTION_ID_ENABLED */
