/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <assert.h>

#include "orfas_queue.h"
#include "orfas_portable.h"
#include "orfas_RAM_fs.h"
#include "orfas_network.h"
#include "orfas_debug.h"


int RAM_process_init()
{
  return RAM_fs_init();
}


void RAM_process_finish()
{
  RAM_fs_finish();
}


void RAM_process_request(struct server_request *req)
{
  req->out_vect = NULL;
  req->ans.error = 0;
#ifdef PROTOCOL_DEBUG
  req->ans.tag = req->req.tag;
  req->ans.pid = req->req.pid;
#endif /* PROTOCOL_DEBUG */
  req->ans.buffer_len = -1;
	
  switch ( req->req.type )
    {

    /* open */
    case OPEN_REQ:
      {
	int res;
	void *fd;

	req->ans.type = OPEN_ANS;

	res = RAM_fs_open(VECT_PTR(req->in_vect, 0),
			  req->req.data.open.oflags,
			  req->req.data.open.mode,
			  &fd);
#ifdef CONNECTION_ID_ENABLED
	RAM_set_fd_connection_id(fd, (void*)(uintptr_t)req->req.connection_id);
#endif /* CONNECTION_ID_ENABLED */

	if ( res < 0 ) {
	  /* error opening file */
	  req->ans.buffer_len = 0;
	  req->ans.return_value.ptr_value = (uint64_t)NULL;
	  req->ans.error = convert_errno_to_portable(-res);
	  LOG(2, "Failed open of file %s flags %d mode %d for error %d\n",
	      (char*)VECT_PTR(req->in_vect, 0),
	      (int)req->req.data.open.oflags, 
	      (int)req->req.data.open.mode,
	      -res);
	} else {
	  /* file well open */
	  req->out_vect = get_io_monovector_alloc(sizeof(struct portable_stat));
	  req->ans.return_value.ptr_value = (uint64_t)(uintptr_t)fd;
	  
	  /* stat it */
	  res = RAM_fs_fstat(fd, VECT_PTR(req->out_vect, 0));
	  DEBUG_ASSERT( res >= 0 );
	  
	  req->ans.buffer_len = sizeof(struct portable_stat);
	  
	  LOG(2, "Open fd %p for file named %s\n",
	      fd,
	      (char*)VECT_PTR(req->in_vect, 0));
	}
	/* end of open */
	break;
      }



    /* close */
    case CLOSE_REQ:
      {
	int res;

	req->ans.type = CLOSE_ANS;

	LOG(2, "Closing fd %p\n",
	    (void*)(uintptr_t)req->req.data.close.fd_id);

	res = RAM_fs_close((void*)(uintptr_t)req->req.data.close.fd_id);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else
	  req->ans.return_value.int_value = 0;

	/* end of close */
	break;
      }
      

    /* pread */
    case PREAD_REQ:
      {
	ssize_t res;
	
	req->ans.type = PREAD_ANS;

	LOG(2, "Reading fd %p at %lld\n",
	    (void*)(uintptr_t)req->req.data.pread.fd_id,
	    (long long)req->req.data.pread.offset);

	/* pread allocates a buffer only if
	   it reads something */
	res = RAM_fs_pread((void*)(uintptr_t)req->req.data.pread.fd_id,
			   &req->out_vect,
			   req->req.data.pread.nbyte, 
			   req->req.data.pread.offset);
	LOG(3, "pread gets %ld\n", (long)res);

	if ( res < 0 ) {
	  req->ans.buffer_len = 0;
	  req->ans.return_value.ssize_t_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else {
	  req->ans.buffer_len = res;
	  req->ans.return_value.ssize_t_value = res;
	}
	/* end of pread */
	break;
      }


    /* pwrite */
    case PWRITE_REQ:
      {
	ssize_t res;

	req->ans.type = PWRITE_ANS;

	LOG(2, "Writing fd %p\n",
	    (void*)(uintptr_t)req->req.data.pwrite.fd_id);

	res = RAM_fs_pwrite((void*)(uintptr_t)req->req.data.pwrite.fd_id,
			    VECT_PTR(req->in_vect, 0),
			    VECT_SIZE(req->in_vect, 0),
			    req->req.data.pwrite.offset);
	LOG(3, "pwrite gets %ld\n", (long)res);

	if ( res < 0 ) {
	  req->ans.return_value.ssize_t_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else
	  req->ans.return_value.ssize_t_value = res;

	/* end of pwrite */
	break;
      }
    

    /* ftrunc */
    case FTRUNC_REQ:
      {
	int res;

	req->ans.type = FTRUNC_ANS;

	LOG(2, "ftruncate file %p at %lld\n",
	    (void*)(uintptr_t)req->req.data.ftrunc.fd_id,
	    (long long)req->req.data.ftrunc.length);

	res = RAM_fs_ftruncate((void*)(uintptr_t)req->req.data.ftrunc.fd_id,
			       req->req.data.ftrunc.length);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else
	  req->ans.return_value.int_value = 0;
	  
	/* end of ftrunc */
	break;
      }


    /* trunc */
    case TRUNC_REQ:
      {
	int res;

	req->ans.type = TRUNC_ANS;

	LOG(2, "truncate %s at %lld\n",
	    (char*)req->in_vect,
	    (long long)req->req.data.trunc.length);

	res = RAM_fs_truncate(VECT_PTR(req->in_vect, 0),
			      req->req.data.trunc.length);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else
	  req->ans.return_value.int_value = 0;

	/* end of trunc */
	break;
      }

    
    /* fstat */
    case FSTAT_REQ:
      {
	int res;

	req->ans.type = FSTAT_ANS;
	req->out_vect = get_io_monovector_alloc(sizeof(struct portable_stat));

	res = RAM_fs_fstat((void*)(uintptr_t)req->req.data.fstat.fd_id,
			   VECT_PTR(req->out_vect, 0));

	if ( res < 0 ) {
	  put_io_vector(req->out_vect);
	  req->out_vect = NULL;
	  req->ans.buffer_len = 0;
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else {
	  req->ans.buffer_len = sizeof(struct portable_stat);
	  req->ans.return_value.int_value = 0;
	}

	/* end of fstat */
	break;
      }


    /* stat */
    case STAT_REQ:
      {
	int res;

	req->ans.type = STAT_ANS;
	req->out_vect = get_io_monovector_alloc(sizeof(struct portable_stat));

	res = RAM_fs_stat(VECT_PTR(req->in_vect, 0),
			  VECT_PTR(req->out_vect, 0));

	if ( res < 0 ) {
	  put_io_vector(req->out_vect);
	  req->out_vect = NULL;
	  req->ans.buffer_len = 0;
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else {
	  req->ans.buffer_len = sizeof(struct portable_stat);
	  req->ans.return_value.int_value = 0;
	}

	/* end of stat */
	break;
      }


    /* lstat */
    case LSTAT_REQ:
      {
	int res;

	req->ans.type = LSTAT_ANS;
	req->out_vect = get_io_monovector_alloc(sizeof(struct portable_stat));

	res = RAM_fs_lstat(VECT_PTR(req->in_vect, 0),
			   VECT_PTR(req->out_vect, 0));

	if ( res < 0 ) {
	  put_io_vector(req->out_vect);
	  req->out_vect = NULL;
	  req->ans.buffer_len = 0;
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else {
	  req->ans.buffer_len = sizeof(struct portable_stat);
	  req->ans.return_value.int_value = 0;
	}

	/* end of lstat */
	break;
      }


    /* fstatfs */
    case FSTATFS_REQ:
      {
	int res;

	req->ans.type = FSTATFS_ANS;
	req->out_vect = get_io_monovector_alloc(sizeof(struct portable_statfs));

	res = RAM_fs_statfs(VECT_PTR(req->out_vect, 0));

	if ( res < 0 ) {
	  put_io_vector(req->out_vect);
	  req->out_vect = NULL;
	  req->ans.buffer_len = 0;
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else {
	  req->ans.buffer_len = sizeof(struct portable_statfs);
	  req->ans.return_value.int_value = 0;
	}

	/* end of fstatfs */
	break;
      }


    /* statfs */
    case STATFS_REQ:
      {
	int res;

	req->ans.type = STATFS_ANS;
	req->out_vect = get_io_monovector_alloc(sizeof(struct portable_statfs));

	res = RAM_fs_statfs(VECT_PTR(req->out_vect, 0));

	if ( res < 0 ) {
	  put_io_vector(req->out_vect);
	  req->out_vect = NULL;
	  req->ans.buffer_len = 0;
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else {
	  req->ans.buffer_len = sizeof(struct portable_statfs);
	  req->ans.return_value.int_value = 0;
	}

	/* end of statfs */
	break;
      }


    /* mknod */
    case MKNOD_REQ:
      {
	int res;

	req->ans.type = MKNOD_ANS;

	res = RAM_fs_mknod(VECT_PTR(req->in_vect, 0),
			   req->req.data.mknod.mode,
			   req->req.data.mknod.dev);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of mknod */
	break;
      }


    /* mkdir */
    case MKDIR_REQ:
      {
	int res;

	req->ans.type = MKDIR_ANS;

	res = RAM_fs_mkdir(VECT_PTR(req->in_vect, 0),
			   req->req.data.mkdir.mode);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of mkdir */
	break;
      }


    /* rmdir */
    case RMDIR_REQ:
      {
	int res;

	req->ans.type = RMDIR_ANS;

	res = RAM_fs_rmdir(VECT_PTR(req->in_vect, 0));

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of mknod */
	break;
      }


    /* unlink */
    case UNLINK_REQ:
      {
	int res;

	req->ans.type = UNLINK_ANS;

	res = RAM_fs_unlink(VECT_PTR(req->in_vect, 0));

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of mknod */
	break;
      }


    /* access */
    case ACCESS_REQ:
      {
	int res;

	req->ans.type = ACCESS_ANS;

	res = RAM_fs_access(VECT_PTR(req->in_vect, 0),
			    req->req.data.access.mode);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of access */
	break;
      }


    /* fchmod */
    case FCHMOD_REQ:
      {
	int res;

	req->ans.type = FCHMOD_ANS;

	res = RAM_fs_fchmod((void*)(uintptr_t)req->req.data.fchmod.fd_id,
			    req->req.data.fchmod.mode);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of fchmod */		     
	break;
      }


    /* chmod */
    case CHMOD_REQ:
      {
	int res;

	req->ans.type = CHMOD_ANS;

	res = RAM_fs_chmod(VECT_PTR(req->in_vect, 0),
			   req->req.data.chmod.mode);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of chmod */
	break;
      }


    /* fchown */
    case FCHOWN_REQ:
      {
	int res;

	req->ans.type = FCHOWN_ANS;

	res = RAM_fs_fchown((void*)(uintptr_t)req->req.data.fchown.fd_id,
			    req->req.data.fchown.owner,
			    req->req.data.fchown.group);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of fchown */		     
	break;
      }


    /* chown */
    case CHOWN_REQ:
      {
	int res;

	req->ans.type = CHOWN_ANS;

	res = RAM_fs_chown(VECT_PTR(req->in_vect, 0),
			   req->req.data.chown.owner,
			   req->req.data.chown.group);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of chown */
	break;
      }


    /* lchown */
    case LCHOWN_REQ:
      {
	int res;

	req->ans.type = LCHOWN_ANS;

	res = RAM_fs_lchown(VECT_PTR(req->in_vect, 0),
			    req->req.data.lchown.owner,
			    req->req.data.lchown.group);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of lchown */
	break;
      }


    /* flock */
    case FLOCK_REQ:
      {
	int res;

	req->ans.type = FLOCK_ANS;

	res = RAM_fs_flock((void*)(uintptr_t)req->req.data.flock.fd_id,
			   req->req.data.flock.operation);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of flock */
	break;
      }


    /* rename */
    case RENAME_REQ:
      {
	int res;
	char *oldname = VECT_PTR(req->in_vect, 0);
	char *newname = oldname + req->req.data.rename.part_change;

	req->ans.type = RENAME_ANS;
	
	res = RAM_fs_rename(oldname, newname);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of rename */
	break;
      }


    /* link */
    case LINK_REQ:
      {
	int res;
	char *oldname = VECT_PTR(req->in_vect, 0);
	char *newname = oldname + req->req.data.link.part_change;

	req->ans.type = LINK_ANS;

	res = RAM_fs_link(oldname, newname);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of link */
	break;
      }


    /* symlink */
    case SYMLINK_REQ:
      {
	int res;
	char *oldname = VECT_PTR(req->in_vect, 0);
	char *newname = oldname + req->req.data.symlink.part_change;

	req->ans.type = SYMLINK_ANS;

	res = RAM_fs_symlink(oldname, newname);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of symlink */
	break;
      }


    /* readlink */
    case READLINK_REQ:
      {
	int res;

	req->ans.type = READLINK_ANS;
	req->out_vect = get_io_monovector_alloc(req->req.data.readlink.bufsiz
				     * sizeof(char));

	res = RAM_fs_readlink(VECT_PTR(req->in_vect, 0),
			      VECT_PTR(req->out_vect, 0),
			      req->req.data.readlink.bufsiz);

	if ( res < 0 ) {
	  put_io_vector(req->out_vect);
	  req->out_vect = NULL;
	  req->ans.buffer_len = 0;
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);
	} else {
	  req->ans.buffer_len = res;
	  req->ans.return_value.int_value = res;
	}

	/* end of readlink */
	break;
      }


    /* fsync */
    case FSYNC_REQ:
      {
	int res;

	req->ans.type = FSYNC_ANS;

	res = RAM_fs_fsync((void*)(uintptr_t)req->req.data.fsync.fd_id);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of fsync */
	break;
      }


    /* fdatasync */
    case FDATASYNC_REQ:
      {
	int res;

	req->ans.type = FDATASYNC_ANS;

	res = RAM_fs_fdatasync((void*)(uintptr_t)req->req.data.fdatasync.fd_id);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of fdatasync */
	break;
      }


    /* utime */
    case UTIME_REQ:
      {
	int res;

	req->ans.type = UTIME_ANS;

	res = RAM_fs_utime(VECT_PTR(req->in_vect, 0),
			   req->req.data.utime.actime,
			   req->req.data.utime.modtime,
			   req->req.data.utime.time_defined);
	LOG(2, "utime %s\n",
	    (char*)VECT_PTR(req->in_vect, 0));
	

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of utime */
	break;
      }


    /* fcntl_setfl */
    case FCNTL_SETFL_REQ:
      {
	int res;

	req->ans.type = FCNTL_SETFL_ANS;

	res = RAM_fs_fcntl_setfl((void*)(uintptr_t)req->req.data.fcntl_setfl.fd_id,
				 req->req.data.fcntl_setfl.flags);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of fcntl_setfl */
	break;
      }


    /* opendir */
    case OPENDIR_REQ:
      {
	int res;
	void *dir;
	
	req->ans.type = OPENDIR_ANS;
	
	res = RAM_fs_opendir(VECT_PTR(req->in_vect, 0), &dir);
#ifdef CONNECTION_ID_ENABLED
	RAM_set_dir_connection_id(dir,
				  (void*)(uintptr_t)req->req.connection_id);
#endif /* CONNECTION_ID_ENABLED */

	if ( res < 0 ) {
	  req->ans.return_value.ptr_value = (uint64_t)(uintptr_t)NULL;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.ptr_value = (uint64_t)(uintptr_t)dir;

	/* end of opendir */
	break;
      }


    /* closedir */
    case CLOSEDIR_REQ:
      {
	int res;

	req->ans.type = CLOSEDIR_ANS;
	
	res = RAM_fs_closedir((void*)(uintptr_t)req->req.data.closedir.dir_id);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of closedir */
	break;
      }

    /* readdir */
    case READDIR_REQ:
      {
	int res;

	req->ans.type = READDIR_ANS;

	res = RAM_fs_preaddir((void*)(uintptr_t)req->req.data.readdir.dir_id,
			      req->req.data.readdir.offset,
			      &req->out_vect,
			      req->req.data.readdir.count);
	
	LOG(3, "preaddir gets %d\n", res);

	if ( res == 0 )
	  {
	    req->ans.buffer_len = 0;
	    req->ans.return_value.int_value = 0;
	  }
	else
	  {
	    req->ans.buffer_len = res * sizeof(struct portable_dirent);
	    req->ans.return_value.int_value = res;
	  }

	/* end of readdir */
	break;
      }


    /* rewinddir */
    case REWINDDIR_REQ:
      {
	int res;
	
	req->ans.type = REWINDDIR_ANS;
	
	res = RAM_fs_rewinddir((void*)(uintptr_t)req->req.data.readdir.dir_id);

	if ( res < 0 ) {
	  req->ans.return_value.int_value = -1;
	  req->ans.error = convert_errno_to_portable(-res);  
	} else
	  req->ans.return_value.int_value = 0;

	/* end of rewinddir */
	break;
      }
      


    /* bad request */
    case BAD_REQ:
      {
	ERROR("Bad request %d => Bad answer\n",
	      (int)req->req.type);
	req->ans.type = BAD_ANS;
	break;
      }

    /* cannot come here because bad request type
     must have been converted to BAD_REQ during receiving*/
    default:
      {
	ERROR("Cannot come here !!\n");
	assert(0);
      }
    }

  put_io_vector(req->in_vect);

#ifndef CONNECTION_THREAD_ARCH
  queue_done_request(req);
#endif
}
