/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <assert.h>
#include <stdlib.h>

#include "orfas_debug.h"
#include "orfas_network.h"


/*********
 * Magic *
 *********/

#define FD_MAGIC 0x08081998
#define DIR_MAGIC 0x21071980


/********
 * FILE *
 ********/

struct client_fd {
  DEBUG_INSTR(unsigned int magic);
  int local_fd;
#ifdef CONNECTION_ID_ENABLED
  struct connection_id *connection_id;
#endif /* CONNECTION_ID_ENABLED */
  struct client_fd *prev, *next;
};

static struct client_fd *first_client_fd = NULL;


/*******
 * DIR *
 *******/

struct client_dir {
  DEBUG_INSTR(unsigned int magic);
  void *local_dir;
#ifdef CONNECTION_ID_ENABLED
  struct connection_id *connection_id;
#endif /* CONNECTION_ID_ENABLED */
  struct client_dir *prev, *next;
};

static struct client_dir *first_client_dir = NULL;



/********
 * Init *
 ********/

int STD_process_init() {
  /* umask has already been used by client */
  umask(0000);

  return 0;
}


void STD_process_finish() {
}



/****************
 * FD functions *
 ****************/

void *STD_get_new_fd(int local_fd)
{
  struct client_fd *new = (struct client_fd*) malloc( sizeof(struct client_fd) );

  new->local_fd = local_fd;

  if ( first_client_fd == NULL )
    {
      first_client_fd = new;
      new->prev = NULL;
    }
  else
    {
      struct client_fd *tmp = first_client_fd;
      while ( tmp->next != NULL )
	tmp = tmp->next;

      tmp->next = new;
      new->prev = tmp;
    }
  new->next = NULL;

  DEBUG_INSTR( new->magic = FD_MAGIC );

  return new;
}


int STD_put_old_fd(void *fd)
{
  struct client_fd *tmp = (struct client_fd*) fd;

  if ( tmp->prev )
    tmp->prev->next = tmp->next;
  else
    first_client_fd = tmp->next;
  
  if ( tmp->next )
    tmp->next->prev = tmp->prev;

  DEBUG_INSTR( tmp->magic = 0 );

  free(fd);

  return 0;
}


/* This function returns the real file descriptor
   associated with a virtual fd for the client.
   If undefined, it returns -1 */
int STD_get_local_fd_of_fd(void *fdp)
{
  struct client_fd *fd;
  DEBUG_ASSERT( ((struct client_fd*)fdp)->magic == FD_MAGIC );
  fd = (struct client_fd*)fdp;

  return fd->local_fd;
}



/*****************
 * DIR functions *
 *****************/

void *STD_get_new_dir(void *local_dir)
{
  struct client_dir *new = (struct client_dir*) malloc( sizeof(struct client_dir) );

  new->local_dir = local_dir;

  if ( first_client_dir == NULL )
    {
      first_client_dir = new;
      new->prev = NULL;
    }
  else
    {
      struct client_dir *tmp = first_client_dir;
      while ( tmp->next != NULL )
	tmp = tmp->next;

      tmp->next = new;
      new->prev = tmp;
    }
  new->next = NULL;

  DEBUG_INSTR( new->magic = DIR_MAGIC );

  return new;
}


int STD_put_old_dir(void *dir)
{
  struct client_dir *tmp = (struct client_dir*) dir;

  if ( tmp->prev )
    tmp->prev->next = tmp->next;
  else
    first_client_dir = tmp->next;
  
  if ( tmp->next )
    tmp->next->prev = tmp->prev;

  DEBUG_INSTR( tmp->magic = 0 );

  free(dir);

  return 0;
}


/* This function returns the real file descriptor
   associated with a virtual dir for the client.
   If undefined, it returns -1 */
void *STD_get_local_dir_of_dir(void *dirp)
{
  struct client_dir *dir;
  DEBUG_ASSERT( ((struct client_dir*)dirp)->magic == DIR_MAGIC );
  dir = (struct client_dir*)dirp;

  return dir->local_dir;
}



#ifdef CONNECTION_ID_ENABLED
/****************************
 * Set fd/dir connection id *
 ****************************/

void STD_set_fd_connection_id(void *fdp, struct connection_id *id)
{
  struct client_fd *fd;
  DEBUG_ASSERT( ((struct client_fd*)fdp)->magic == FD_MAGIC );
  fd = (struct client_fd*)fdp;

  fd->connection_id = id;
}

void STD_set_dir_connection_id(void *dirp, struct connection_id *id)
{
  struct client_dir *dir;
  DEBUG_ASSERT( ((struct client_dir*)dirp)->magic == DIR_MAGIC );
  dir = (struct client_dir*)dirp;

  dir->connection_id = id;
}


/****************************************
 * Close file after connection id close *
 *****************************************/

void STD_close_file_for_connection_id(struct connection_id *id)
{
  struct client_fd *fd = first_client_fd;
  struct client_dir *dir = first_client_dir;

  while ( fd != NULL ) {
    struct client_fd *tmp = fd->next;
    
    if ( fd->connection_id == id )
      put_old_fd(fd);
    
    fd = tmp;
  }

  while ( dir != NULL ) {
    struct client_dir *tmp = dir->next;
    
    if ( dir->connection_id == id )
      put_old_dir(dir);
    
    dir = tmp;
  }
}

#endif /* CONNECTION_ID_ENABLED */
