/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <sys/file.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <utime.h>
#include <fcntl.h>

#include "orfas_STD_fs.h"
#include "orfas_queue.h"
#include "orfas_portable.h"
#include "orfas_io.h"
#include "orfas_network.h"
#include "orfas_debug.h"


#ifdef BIP_SENDFILE_ENABLED
#include "bip.h"
#include "orfas_network.h"
#endif /* BIP_SENDFILE_ENABLED */


void STD_process_easy_request(struct server_request *req)
{
  switch ( req->req.type )
    {

    /* open */
    case OPEN_REQ:
      {
	int local_fd;

	req->ans.type = OPEN_ANS;

	/* real open */
	local_fd = open(VECT_PTR(req->in_vect, 0),
			convert_oflags_from_portable(req->req.data.open.oflags),
			req->req.data.open.mode);
	if ( local_fd < 0 )
	  {
	    /* error opening real file */
	    req->ans.return_value.ptr_value = (uint64_t)NULL;
	    req->ans.buffer_len = 0;
	    req->ans.error = convert_errno_to_portable(errno);
	    LOG(2, "failed open of file %s flags %d mode %d for error %m\n",
		(char*)VECT_PTR(req->in_vect, 0),
		(int)req->req.data.open.oflags, 
		(int)req->req.data.open.mode);
	  }
	else
	  {
	    int res;
	    struct portable_stat *pst;
	    void *fd_for_client = STD_get_new_fd(local_fd);

	    req->out_vect = get_io_monovector_alloc(sizeof(struct portable_stat));
	    pst = VECT_PTR(req->out_vect, 0);

#ifdef CONNECTION_ID_ENABLED
	    STD_set_fd_connection_id(fd_for_client,
				     (void*)(uintptr_t)req->req.connection_id);
#endif /* ~CONNECTION_ID_ENABLED */

	    /* if well open, there's a place to store it */
	    req->ans.return_value.ptr_value = (uint64_t)(uintptr_t)fd_for_client;

	    /* stat it */
#if PORTABLE_ENABLED
	    {
	      struct stat64 st;
	      res = fstat64(local_fd, &st);
	      convert_stat_64_to_portable(pst, &st);
	    }
#else
	    res = fstat64(local_fd, pst);
#endif

	    /* check the return value */
	    DEBUG_ASSERT( res >= 0 );

	    req->ans.buffer_len = sizeof(struct portable_stat);

	    LOG(2, "Open fd %p for local fd %d for file named %s\n",
		fd_for_client,
		local_fd,
		(char*)VECT_PTR(req->in_vect, 0));
	  }
	/* end of open */
	break;
      }



    /* close */
    case CLOSE_REQ:
      {
	int local_fd;
	
	req->ans.type = CLOSE_ANS;
	
	/* get the real file */
	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.close.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {
	    int res;

	    LOG(2, "Closing fd %p for local fd %d\n",
		(void*)(uintptr_t)req->req.data.close.fd_id,
		local_fd);
	    res = close(local_fd);
	    STD_put_old_fd((void*)(uintptr_t)req->req.data.close.fd_id);

	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }	
	/* end of close */
	break;
      }
    

    /* ftrunc */
    case FTRUNC_REQ:
      {
	int local_fd;

	req->ans.type = FTRUNC_ANS;

	/* get the real file */
	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.ftrunc.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {	
	    int res;

	    LOG(2, "ftruncate fd %p at %lld\n",
		(void*)(uintptr_t)req->req.data.ftrunc.fd_id,
		(long long)req->req.data.ftrunc.length);

	    /* ftruncate it */
	    res = ftruncate64(local_fd,
			      req->req.data.ftrunc.length);
	    
	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }	
	/* end of ftrunc */
	break;
      }


    /* trunc */
    case TRUNC_REQ:
      {
	int res;

	req->ans.type = TRUNC_ANS;

	LOG(2, "truncate %s at %lld\n",
	    (char*)VECT_PTR(req->in_vect, 0),
	    (long long)req->req.data.trunc.length);

	res = truncate64(VECT_PTR(req->in_vect, 0),
			 req->req.data.trunc.length);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of trunc */
	break;
      }

    
    /* fstat */
    case FSTAT_REQ:
      {
	int local_fd;

	req->ans.type = FSTAT_ANS;

	/* get the real file */
	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.fstat.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.buffer_len = 0;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {	
	    int res;
	    struct portable_stat *pst;

	    req->out_vect = get_io_monovector_alloc(sizeof(struct portable_stat));
	    pst = VECT_PTR(req->out_vect, 0);

	    LOG(2, "fstat fd %p\n",
		(void*)(uintptr_t)req->req.data.fstat.fd_id);

	    /* stat it */
#if PORTABLE_ENABLED
	    {
	      struct stat64 st;
	      res = fstat64(local_fd, &st);
	      convert_stat_64_to_portable(pst, &st);
	    }
#else
	    res = fstat64(local_fd, pst);
#endif

	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      {
		put_io_vector(req->out_vect);
		req->out_vect = NULL;
		req->ans.buffer_len = 0;
		req->ans.error = convert_errno_to_portable(errno);
	      }
	    else
	      req->ans.buffer_len = sizeof(struct portable_stat);
	  }	
	/* end of fstat */
	break;
      }


    /* stat */
    case STAT_REQ:
      {
	int res;
	struct portable_stat *pst;

	req->ans.type = STAT_ANS;
	req->out_vect = get_io_monovector_alloc(sizeof(struct portable_stat));
	pst = VECT_PTR(req->out_vect, 0);

	LOG(2, "stat file %s\n",
	    (char*)VECT_PTR(req->in_vect, 0));

	/* stat it */
#if PORTABLE_ENABLED
	{
	  struct stat64 st;
	  res = stat64(VECT_PTR(req->in_vect, 0), &st);
	  convert_stat_64_to_portable(pst, &st);
	}
#else
	res = stat64(VECT_PTR(req->in_vect, 0), pst);
#endif

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  {
	    put_io_vector(req->out_vect);
	    req->out_vect = NULL;
	    req->ans.buffer_len = 0;
	    req->ans.error = convert_errno_to_portable(errno);
	  }
	else
	  req->ans.buffer_len = sizeof(struct portable_stat);
	/* end of stat */
	break;
      }


    /* lstat */
    case LSTAT_REQ:
      {
	int res;
	struct portable_stat *pst;

	req->ans.type = LSTAT_ANS;
	req->out_vect = get_io_monovector_alloc(sizeof(struct portable_stat));
	pst = VECT_PTR(req->out_vect, 0);
	
	LOG(2, "lstat file %s\n",
	    (char*)VECT_PTR(req->in_vect, 0));

	/* lstat it */
#if PORTABLE_ENABLED
	{
	  struct stat64 st;
	  res = lstat64(VECT_PTR(req->in_vect, 0), &st);
	  convert_stat_64_to_portable(pst, &st);
	}
#else
	res = lstat64(VECT_PTR(req->in_vect, 0), pst);
#endif

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  {
	    put_io_vector(req->out_vect);
	    req->out_vect = NULL;
	    req->ans.buffer_len = 0;
	    req->ans.error = convert_errno_to_portable(errno);
	  }
	else
	  req->ans.buffer_len = sizeof(struct portable_stat);
	/* end of lstat */
	break;
      }


    /* fstatfs */
    case FSTATFS_REQ:
      {
	int local_fd;

	req->ans.type = FSTATFS_ANS;

	/* get the real file */
	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.fstatfs.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.buffer_len = 0;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {	
	    int res;
	    struct portable_statfs *pst;

	    req->out_vect = get_io_monovector_alloc(sizeof(struct portable_statfs));
	    pst = VECT_PTR(req->out_vect, 0);	    

	    LOG(2, "fstatfs fd %p\n",
		(void*)(uintptr_t)req->req.data.fstatfs.fd_id);

	    /* statfs it */
#if PORTABLE_ENABLED
	    {
	      struct statfs64 st;
	      res = fstatfs64(local_fd, &st);
	      convert_statfs_64_to_portable(pst, &st);
	    }
#else
	    res = fstatfs64(local_fd, pst);
#endif

	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      {
		put_io_vector(req->out_vect);
		req->out_vect = NULL;
		req->ans.buffer_len = 0;
		req->ans.error = convert_errno_to_portable(errno);
	      }
	    else
	      req->ans.buffer_len = sizeof(struct portable_statfs);
	  }	
	/* end of fstatfs */
	break;
      }


    /* statfs */
    case STATFS_REQ:
      {
	int res;
	struct portable_statfs *pst;

	req->ans.type = STATFS_ANS;
	req->out_vect = get_io_monovector_alloc(sizeof(struct portable_statfs));
	pst = VECT_PTR(req->out_vect, 0);

	LOG(2, "statfs file %s\n",
	    (char*)VECT_PTR(req->in_vect, 0));

	/* statfs it */
#if PORTABLE_ENABLED
	{
	  struct statfs64 st;
	  res = statfs64(VECT_PTR(req->in_vect, 0), &st);
	  convert_statfs_64_to_portable(pst, &st);
	}
#else
	res = statfs64(VECT_PTR(req->in_vect, 0), pst);
#endif

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  {
	    put_io_vector(req->out_vect);
	    req->out_vect = NULL;
	    req->ans.buffer_len = 0;
	    req->ans.error = convert_errno_to_portable(errno);
	  }
	else
	  req->ans.buffer_len = sizeof(struct portable_statfs);
	/* end of statfs */
	break;
      }


    /* mknod */
    case MKNOD_REQ:
      {
	int res;

	req->ans.type = MKNOD_ANS;

	res = mknod(VECT_PTR(req->in_vect, 0),
		    req->req.data.mknod.mode,
		    req->req.data.mknod.dev);
	LOG(2, "mknod %s mode %d dev %d\n",
	    (char*)VECT_PTR(req->in_vect, 0),
	    (int)req->req.data.mknod.mode,
	    (int)req->req.data.mknod.dev);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of mknod */
	break;
      }


    /* mkdir */
    case MKDIR_REQ:
      {
	int res;

	req->ans.type = MKDIR_ANS;

	res = mkdir(VECT_PTR(req->in_vect, 0),
		    req->req.data.mkdir.mode);
	LOG(2, "mkdir %s mode %d\n",
	    (char*)VECT_PTR(req->in_vect, 0),
	    (int)req->req.data.mkdir.mode);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of mkdir */
	break;
      }


    /* rmdir */
    case RMDIR_REQ:
      {
	int res;

	req->ans.type = RMDIR_ANS;

	res = rmdir(VECT_PTR(req->in_vect, 0));
	LOG(2, "rmdir %s\n",
	    (char*)VECT_PTR(req->in_vect, 0));

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of mknod */
	break;
      }


    /* unlink */
    case UNLINK_REQ:
      {
	int res;

	req->ans.type = UNLINK_ANS;

	res = unlink(VECT_PTR(req->in_vect, 0));
	LOG(2, "unlink %s\n",
	    (char*)VECT_PTR(req->in_vect, 0));

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of mknod */
	break;
      }


    /* access */
    case ACCESS_REQ:
      {
	int res;

	req->ans.type = ACCESS_ANS;

	res = access(VECT_PTR(req->in_vect, 0),
		     req->req.data.access.mode);
	LOG(2, "try to access file %s mode %d\n",
	    (char*)VECT_PTR(req->in_vect, 0),
	    (int)req->req.data.access.mode);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of access */
	break;
      }


    /* fchmod */
    case FCHMOD_REQ:
      {
	int local_fd;

	req->ans.type = FCHMOD_ANS;

	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.fchmod.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {	
	    int res;

	    res = fchmod(local_fd,
			 req->req.data.fchmod.mode);

	    LOG(2, "fchmod fd %p to mode %d result %d\n",
		(void*)(uintptr_t)req->req.data.fchmod.fd_id,
		(int)req->req.data.fchmod.mode,
		res);
	    
	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }
	/* end of fchmod */		     
	break;
      }


    /* chmod */
    case CHMOD_REQ:
      {
	int res;

	req->ans.type = CHMOD_ANS;

	res = chmod(VECT_PTR(req->in_vect, 0),
		    req->req.data.chmod.mode);
	LOG(2, "chmod file %s to mode %d result %d\n",
	    (char*)VECT_PTR(req->in_vect, 0),
	    (int)req->req.data.chmod.mode,
	    res);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of chmod */
	break;
      }


    /* fchown */
    case FCHOWN_REQ:
      {
	int local_fd;

	req->ans.type = FCHOWN_ANS;

	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.fchown.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {	
	    int res;
	    
	    res = fchown(local_fd,
			 req->req.data.fchown.owner,
			 req->req.data.fchown.group);

	    LOG(2, "chown fd %p owner %d group %d\n",
		(void*)(uintptr_t)req->req.data.fchown.fd_id,
		(int)req->req.data.fchown.owner,
		(int)req->req.data.fchown.group);

	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }
	/* end of fchown */		     
	break;
      }


    /* chown */
    case CHOWN_REQ:
      {
	int res;

	req->ans.type = CHOWN_ANS;

	res = chown(VECT_PTR(req->in_vect, 0),
		    req->req.data.chown.owner,
		    req->req.data.chown.group);
	LOG(2, "chown file %s owner %d group %d\n",
	    (char*)VECT_PTR(req->in_vect, 0),
	    (int)req->req.data.chown.owner,
	    (int)req->req.data.chown.group);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of chown */
	break;
      }


    /* lchown */
    case LCHOWN_REQ:
      {
	int res;

	req->ans.type = LCHOWN_ANS;

	res = lchown(VECT_PTR(req->in_vect, 0),
		     req->req.data.lchown.owner,
		     req->req.data.lchown.group);

	LOG(2, "lchown file %s owner %d group %d\n",
	    (char*)VECT_PTR(req->in_vect, 0),
	    (int)req->req.data.lchown.owner,
	    (int)req->req.data.lchown.group);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of lchown */
	break;
      }


    /* flock */
    case FLOCK_REQ:
      {
	int local_fd;

	req->ans.type = FLOCK_ANS;

	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.flock.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {
	    int res;

	    res = flock(local_fd,
			req->req.data.flock.operation);

	    LOG(2, "flock fd %p operation %d\n",
		(void*)(uintptr_t)req->req.data.flock.fd_id,
		(int)req->req.data.flock.operation);

	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }
	/* end of flock */
	break;
      }


    /* rename */
    case RENAME_REQ:
      {
	int res;
	char *oldname = (char*)VECT_PTR(req->in_vect, 0);
	char *newname = oldname + req->req.data.rename.part_change;

	req->ans.type = RENAME_ANS;
	
	res =  rename(oldname, newname);
	LOG(2, "rename %s in %s\n", oldname, newname);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of rename */
	break;
      }


    /* link */
    case LINK_REQ:
      {
	int res;
	char *oldname = (char*)VECT_PTR(req->in_vect, 0);
	char *newname = oldname + req->req.data.link.part_change;

	req->ans.type = LINK_ANS;

	res = link(oldname, newname);
	LOG(2, "link %s to %s\n", oldname, newname);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of link */
	break;
      }


    /* symlink */
    case SYMLINK_REQ:
      {
	int res;
	char *oldname = (char*)VECT_PTR(req->in_vect, 0);
	char *newname = oldname + req->req.data.symlink.part_change;

	req->ans.type = SYMLINK_ANS;

	res = symlink(oldname, newname);
	LOG(2, "symlink %s to %s\n", oldname, newname);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of symlink */
	break;
      }


    /* readlink */
    case READLINK_REQ:
      {
	int res;

	req->ans.type = READLINK_ANS;
	req->out_vect = get_io_monovector_alloc(req->req.data.readlink.bufsiz
				     * sizeof(char));

	res = readlink(VECT_PTR(req->in_vect, 0),
		       VECT_PTR(req->out_vect, 0),
		       req->req.data.readlink.bufsiz);

	LOG(2, "readlink %s gets %d\n",
	    (char*)VECT_PTR(req->in_vect, 0),
	    res);

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  {
	    put_io_vector(req->out_vect);
	    req->out_vect = NULL;
	    req->ans.buffer_len = 0;
	    req->ans.error = convert_errno_to_portable(errno);
	  }
	else
	  {
	    req->ans.buffer_len = res;
	    resize_io_monovector(req->out_vect, res);
	  }
	/* end of readlink */
	break;
      }
      

    /* utime */
    case UTIME_REQ:
      {
	int res;

	req->ans.type = UTIME_ANS;

	if ( req->req.data.utime.time_defined )
	  {
	    struct utimbuf buf;
	    buf.actime = req->req.data.utime.actime;
	    buf.modtime = req->req.data.utime.modtime;
	    res = utime(VECT_PTR(req->in_vect, 0), &buf);
	  }
	else
	  res = utime(VECT_PTR(req->in_vect, 0), NULL);

	LOG(2, "utime %s\n", (char*)VECT_PTR(req->in_vect, 0));

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	/* end of utime */
	break;
      }


    /* fcntl_setfl */
    case FCNTL_SETFL_REQ:
      {
	int local_fd;

	req->ans.type = FCNTL_SETFL_ANS;

	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.fcntl_setfl.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {
	    int res;

	    res = fcntl(local_fd, F_SETFL,
			req->req.data.fcntl_setfl.flags);

	    LOG(2, "fcntl setfl fd %p with flags %d\n", 
		(void*)(uintptr_t)req->req.data.flock.fd_id,
		(int)req->req.data.fcntl_setfl.flags);

	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }
	/* end of fcntl_setfl */
	break;
      }


    /* opendir */
    case OPENDIR_REQ:
      {
	void *local_dir;
	
	req->ans.type = OPENDIR_ANS;
	
	/* real opendir */
	local_dir = opendir(VECT_PTR(req->in_vect, 0));

	if ( local_dir == NULL )
	  {
	    /* error opening real file */
	    req->ans.return_value.ptr_value = (uint64_t)NULL;
	    req->ans.error = convert_errno_to_portable(errno);
	    LOG(2, "Failed open of dir %s for error %d\n",
		(char*)VECT_PTR(req->in_vect, 0),
		errno);
	  }
	else
	  {
	    void *dir_for_client = STD_get_new_dir(local_dir);
#ifdef CONNECTION_ID_ENABLED
	    STD_set_dir_connection_id(dir_for_client,
				      (void*)(uintptr_t)req->req.connection_id);
#endif /* ~CONNECTION_ID_ENABLED */

	    req->ans.return_value.ptr_value
	      = (uint64_t)(uintptr_t)dir_for_client;
	    DEBUG_ASSERT( dir_for_client != NULL );
	    LOG(2, "Opendir %p for dir named %s\n",
		dir_for_client,
		(char*)VECT_PTR(req->in_vect, 0));
	  }
	/* end of opendir */
	break;
      }


    /* closedir */
    case CLOSEDIR_REQ:
      {
	void *local_dir;

	req->ans.type = CLOSEDIR_ANS;
	
	local_dir
	  = STD_get_local_dir_of_dir((void*)(uintptr_t)req->req.data.closedir.dir_id);
	if ( local_dir == NULL )
	  {
	    /* there's no such dir in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }	    
	else
	  {
	    int res;

	    res = closedir(local_dir);
	    STD_put_old_dir((void*)(uintptr_t)req->req.data.closedir.dir_id);

	    /* check the return value */
	    req->ans.return_value.ssize_t_value = res;
	    if ( res < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }
	/* end of closedir */
	break;
      }

    /* readdir */
    case READDIR_REQ:
      {
	DIR *local_dir;

	req->ans.type = READDIR_ANS;
	
	local_dir
	  = STD_get_local_dir_of_dir((void*)(uintptr_t)req->req.data.readdir.dir_id);
	if ( local_dir == NULL )
	  {
	    /* there's no such dir in the server */
	    req->ans.buffer_len = 0;
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }	    
	else
	  {
	    size_t res;
	    struct dirent64 *tmp_dirent;
	    struct portable_dirent *dirents;

	    req->out_vect = get_io_monovector_alloc(req->req.data.readdir.count
						    * sizeof(struct portable_dirent));
	    dirents = VECT_PTR(req->out_vect, 0);

	    /* seek in the dir */
	    seekdir(local_dir,
		    req->req.data.readdir.offset);
	    
	    LOG(2, "Readdir %ld dirents in dir %p after %lld\n",
		(long)req->req.data.readdir.count,
		(void*)(uintptr_t)req->req.data.readdir.dir_id,
		(long long)req->req.data.readdir.offset);

	    /* readdirs */
	    for(res=0; res<req->req.data.readdir.count; res++)
	      {
		/* try to readdir */
		tmp_dirent = readdir64(local_dir);
		if ( tmp_dirent == NULL )
		  break;

		/* store data */
#if PORTABLE_ENABLED
		convert_dirent_64_to_portable(&dirents[res],
					      tmp_dirent);
#else
		memcpy(&dirents[res], tmp_dirent,
		       sizeof(struct dirent));
#endif

		LOG(4, "found %s at %lld\n",
		    tmp_dirent->d_name,
		    (long long)tmp_dirent->d_off);
	      }

	    LOG(3, "Readdir %ld/%ld dirents\n",
		(long)res,
		(long)req->req.data.readdir.count);
	    

	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      {
		put_io_vector(req->out_vect);
		req->out_vect = NULL;
		req->ans.buffer_len = 0;
		req->ans.error = convert_errno_to_portable(errno);
	      }
	    else
	      {
		req->ans.buffer_len = res * sizeof(struct portable_dirent);
		resize_io_monovector(req->out_vect,
				     res * sizeof(struct portable_dirent));
	      }

	    /* end of readdir */
	  }
	/* end of readdir */
	break;
      }


    /* rewinddir */
    case REWINDDIR_REQ:
      {
	int res;
	void * local_dir;
	
	req->ans.type = REWINDDIR_ANS;
	
	local_dir
	  = STD_get_local_dir_of_dir((void*)(uintptr_t)req->req.data.rewinddir.dir_id);

	if ( local_dir == NULL )
	  {
	    /* there's no such dir in the server */
	    res = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }	    
	else
	  {
	    res = 0;
	    rewinddir(local_dir);
	  }

	/* check the return value */
	req->ans.return_value.int_value = res;
	if ( res < 0 )
	  req->ans.error = convert_errno_to_portable(errno);
	LOG(2, "rewinddir %d\n", res);
	/* end of rewinddir */
	break;
      }



    /* bad request */
    case BAD_REQ:
      {
	ERROR("Bad request %d => Bad answer\n",
	      (int)req->req.type);
	req->ans.type = BAD_ANS;
	break;
      }

    /* cannot come here because bad request type
     * must have been converted to BAD_REQ during receiving
     */
    default:
      {
	ERROR("Cannot come here !!\n");
	assert(0);
      }
    }

}



#ifdef STD_IO
void STD_process_request(struct server_request *req)
{
  req->out_vect = NULL;
  req->ans.error = 0;
#ifdef PROTOCOL_DEBUG
  req->ans.tag = req->req.tag;
  req->ans.pid = req->req.pid;
  LOG(1, "Processing request tag %d pid %d\n",
      (int)req->req.tag, (int)req->req.pid);
#endif /* PROTOCOL_DEBUG */
  req->ans.buffer_len = -1;
  
  setregid(-1, req->req.egid);
  setreuid(-1, req->req.euid);

  switch ( req->req.type )
    {

    /* pread */
    case PREAD_REQ:
      {
	int local_fd;

	req->ans.type = PREAD_ANS;

	/* get the real file */
	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.pread.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.buffer_len = 0;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {
	    ssize_t count;

#if (defined BIP_SENDFILE_ENABLED)
	    LOG(2, "Sendfile fd %p for local fd %d\n",
		(void*)(uintptr_t)req->req.data.pread.fd_id,
		local_fd);

	    BIP_WAIT_FOR_SEND_TOKEN();
	    BIP_TAKE_SEND_TOKEN(bip_tisend_file(req->connection.node,
						ANSWER_DATA_TAG,
						local_fd,
						req->req.data.pread.nbyte, 
						req->req.data.pread.offset,
						&count));
	    BIP_FIX_SEND_CALLBACK(NULL, NULL);
	    LOG(3, "Bip_tisendfile read %ld bytes at offset %lld\n",
		(long)count,
		(long long)req->req.data.pread.offset);
#elif (defined TCP_SENDFILE_ENABLED)
	    struct stat64 st;
	    
	    LOG(2, "Sendfile fd %p for local fd %d for %ld bytes at %lld\n",
		(void*)(uintptr_t)req->req.data.pread.fd_id,
		local_fd,
		(long)req->req.data.pread.nbyte,
		(long long)req->req.data.pread.offset);
	     
	    fstat64(local_fd, &st);
	    if ( !(fcntl(local_fd, F_GETFL) & O_WRONLY) ) {
	      /* may be less ugly if flags value were more elegant */
	      count = req->req.data.pread.nbyte;
	      if ( count + req->req.data.pread.offset > st.st_size )
	        count = st.st_size - req->req.data.pread.offset;
	      if ( count < 0 )
	        count = 0;
	    } else {
              LOG(2, "Cannot sendfile because open flags are %d\n", 
		  fcntl(local_fd, F_GETFL));
	      errno = EBADF;
	      count = -1;
	    }

	    LOG(4, "Predict %lld/%lld bytes send\n",
		(long long)count, (long long)req->req.data.pread.nbyte);
#else /* no *_SENDFILE_ENABLED */
	    req->out_vect = get_io_monovector_alloc(req->req.data.pread.nbyte);

	    LOG(2, "Reading fd %p for local fd %d\n",
		(void*)(uintptr_t)req->req.data.pread.fd_id,
		local_fd);

	    count = pread64(local_fd,
			    VECT_PTR(req->out_vect, 0),
			    req->req.data.pread.nbyte, 
			    req->req.data.pread.offset);
	    if ( count >= 0 )
	      resize_io_monovector(req->out_vect, count);
	    else {
	      put_io_vector(req->out_vect);
	      req->out_vect = NULL;
	    }

	    LOG(3, "read %ld bytes at offset %lld\n",
		(long)count,
		(long long)req->req.data.pread.offset);
#endif /* no *_SENDFILE_ENABLED */

	    /* check the return value */
	    req->ans.return_value.ssize_t_value = count;
	    if ( count < 0 )
	      {
		req->ans.buffer_len = 0;
		req->ans.error = convert_errno_to_portable(errno);
	      }
	    else
	      req->ans.buffer_len = count;
	  }
	/* end of pread */
	break;
      }


    /* pwrite */
    case PWRITE_REQ:
      {
	int local_fd;

	req->ans.type = PWRITE_ANS;

	/* get the real file */
	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.pwrite.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {
	    ssize_t count;

	    LOG(2, "Writing fd %p for local fd %d\n",
		(void*)(uintptr_t)req->req.data.pwrite.fd_id,
		local_fd);

	    count = pwrite64(local_fd,
			     VECT_PTR(req->in_vect, 0),
			     req->req.buffer_len, 
			     req->req.data.pwrite.offset);

	    LOG(3, "written %ld bytes at offset %lld\n",
		(long)count,
		(long long)req->req.data.pwrite.offset);

	    /* check the return value */
	    req->ans.return_value.ssize_t_value = count;
	    if ( count < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }
	/* end of pwrite */
	break;
      }
    

    /* fsync */
    case FSYNC_REQ:
      {
	int local_fd;

	req->ans.type = FSYNC_ANS;

	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.fsync.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {	
	    int res;
	    
	    res = fsync(local_fd);
	    LOG(2, "fsync fd %p\n",
		(void*)(uintptr_t)req->req.data.fsync.fd_id);

	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }
	/* end of fsync */
	break;
      }


    /* fdatasync */
    case FDATASYNC_REQ:
      {
	int local_fd;

	req->ans.type = FDATASYNC_ANS;

	local_fd
	  = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.fdatasync.fd_id);
	if ( local_fd < 0 )
	  {
	    /* there's no such fd in the server */
	    req->ans.return_value.int_value = -1;
	    req->ans.error = convert_errno_to_portable(EBADF);
	  }
	else
	  {	
	    int res;
	    
	    res = fdatasync(local_fd);
	    LOG(2, "fdatasync fd %p\n",
		(void*)(uintptr_t)req->req.data.fdatasync.fd_id);

	    /* check the return value */
	    req->ans.return_value.int_value = res;
	    if ( res < 0 )
	      req->ans.error = convert_errno_to_portable(errno);
	  }
	/* end of fdatasync */
	break;
      }


  default:
      {
	 STD_process_easy_request(req);
         break;
      }
    }

  LOG(3, "Got error %d\n", (int)req->ans.error);

  setreuid(-1, uid);
  setreuid(-1, gid);

  put_io_vector(req->in_vect);

#ifndef CONNECTION_THREAD_ARCH
  queue_done_request(req);
#endif
}
#endif /* STD_IO */
