/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/uio.h>
#include <errno.h>
#include <assert.h>

#ifdef TCP_SENDFILE_ENABLED
#include <sys/sendfile.h>
#include "orfas_fs.h"
#endif

#include "orfas_network.h"
#include "orfas_queue.h"
#include "orfas_io.h"
#include "orfas_debug.h"


static int server_socket = -1;



/***********************
 * Polling definitions *
 ***********************/

#ifdef CONNECTION_THREAD_ARCH
/**************************
 * Connection Thread Arch *
 **************************/

/* Connection Thread Architecture needs no global variable
 * because each thread keeps all information about its connection
 */

#include <pthread.h>

#else /* ~CONNECTION_THREAD_ARCH */
/* Other Architectures needs global variables
 * to keep all connections fd and connection_id
 */

# ifdef STD_TCP_POLL
/*********************
 * Standard TCP Poll *
 *********************/

/* Standard Polling requires an array of pollfd structure
 * without holes in it. We got a parallel array of connection_id.
 * Poll gives connection indix back which allow direct 
 * finding of fd and connection_id.
 */

#include <sys/poll.h>

static int connection_max = SERVER_EVENT_GET_MAX+1;
static int connection_nr = 0;

static struct pollfd *pollfd;
#ifdef CONNECTION_ID_ENABLED
static struct connection_id **connection_id;
#endif

#define POLLED2FD(i) (pollfd[i].fd)
#ifdef CONNECTION_ID_ENABLED
#define POLLED2ID(i) (connection_id[i])
#endif


static INLINE void alloc_connection_structs() {
  pollfd = (struct pollfd*)
    malloc(connection_max * sizeof(struct pollfd));
  DEBUG_ASSERT( pollfd != NULL );
#ifdef CONNECTION_ID_ENABLED
  connection_id = (struct connection_id**)
    malloc(connection_max * sizeof(struct connection_id*));
  DEBUG_ASSERT( connection_id != NULL );
#endif
}

static INLINE void grow_connection_structs() {
  connection_max *= 2;
  pollfd = (struct pollfd*)
    realloc(pollfd,
	    connection_max * sizeof(struct pollfd));
  DEBUG_ASSERT( pollfd != NULL );
#ifdef CONNECTION_ID_ENABLED
  connection_id = (struct connection_id**)
    realloc(connection_id,
	    connection_max * sizeof(struct connection_id*));
  DEBUG_ASSERT( connection_id != NULL );
#endif
}

static INLINE void add_connection_poll(int socket) {
  if ( connection_nr == connection_max )
    grow_connection_structs();

  pollfd[connection_nr].events = POLLIN;
  POLLED2FD(connection_nr) = socket;
#ifdef CONNECTION_ID_ENABLED
  POLLED2ID(connection_nr) = NULL;
#endif
   
  connection_nr++;
  
  LOG(3, "Added connection #%d for socket %d\n",
      connection_nr-1, socket);
}

static INLINE void remove_connection_poll(int indix) {
  LOG(3, "Removing connection #%d (socket %d)\n",
      indix, POLLED2FD(indix));
   
  close(POLLED2FD(indix));
#ifdef CONNECTION_ID_ENABLED
  put_connection_id(POLLED2ID(indix));
#endif

  if ( indix != connection_nr-1 ) {
    LOG(4, "Switching with connection #%d (socket %d)\n",
	connection_nr-1, POLLED2FD(connection_nr-1));
    POLLED2FD(indix) = POLLED2FD(connection_nr-1);
#ifdef CONNECTION_ID_ENABLED
    POLLED2ID(indix) = POLLED2ID(connection_nr-1);
#endif
  }

  connection_nr--;
  POLLED2FD(connection_nr) = -1;
#ifdef CONNECTION_ID_ENABLED
  POLLED2ID(connection_nr) = NULL;
#endif    
}

static INLINE void free_connection_structs() {
#ifdef CONNECTION_ID_ENABLED
  free(connection_id);
#endif
  free(pollfd);
}


# elif (defined AIO_TCP_POLL)
/****************
 * AIO TCP Poll *
 ****************/

/* AIO Polling requires a structure per connection with associated
 * iocb and connection_id in it.
 * AIO gives the iocb ptr back. Our structure is aligned with this
 * iocb what allows direct finding of fd (in the iocb)
 * and connection_id.
 */

#include <sys/poll.h>
#include <libaio.h>

struct aio_pollfd {
  struct iocb iocb;
#ifdef CONNECTION_ID_ENABLED
  struct connection_id *connection_id;
#endif
} pollfd_t;

#define IOCB2FD(iocb) ((iocb)->aio_fildes)
#ifdef CONNECTION_ID_ENABLED
#define IOCB2ID(iocb) (((struct aio_pollfd*)iocb)->connection_id)
#endif


static void AIO_recv_callback(io_context_t, struct iocb*, long, long);
static void AIO_connect_callback(io_context_t, struct iocb*, long, long);

static INLINE void alloc_connection_structs() {
}

static INLINE void submit_connection_poll(struct iocb *iocb,
					  io_callback_t cb) {
  int fd = IOCB2FD(iocb);
  io_prep_poll(iocb, fd, POLLIN);
  io_set_callback(iocb, cb);
#ifdef CONNECTION_ID_ENABLED
  IOCB2ID(iocb) = NULL;
#endif
  AIO_submit(iocb);
  LOG(4, "Submit iocb %p for poll on socket %d\n", iocb, fd);  
}

static INLINE void add_connection_poll(int socket) {
  struct iocb *iocb = (struct iocb*) malloc(sizeof(struct iocb));
  LOG(3, "Adding connection on socket %d with iocb %p\n",
      socket, iocb);
#ifdef CONNECTION_ID_ENABLED
  IOCB2FD(iocb) = socket;
#endif
  submit_connection_poll(iocb,
			socket == server_socket
			? AIO_connect_callback
			: AIO_recv_callback);
}

static INLINE void remove_connection_poll(struct iocb *iocb) {
  LOG(3, "Free iocb %p previously used for socket %d\n",
      iocb, IOCB2FD(iocb));
  close(IOCB2FD(iocb));
#ifdef CONNECTION_ID_ENABLED
  put_connection_id(IOCB2ID(iocb));
#endif
  free(iocb);
}

static INLINE void free_connection_structs() {
  /* Cannot easily get mallocated iocb, we do not free them.
   * But they should have been freed by previous client close */
}


# else /* EPOLL_TCP_POLL */
/******************
 * Epoll TCP Poll *
 ******************/

/* Epoll Polling requires a structure per connection with associated
 * fd and connection_id in it.
 * Epoll only gives the fd back. We also need a way to find
 * the corresponding structure.
 */

#include <sys/poll.h>
#include <epoll.h>

static int epoll_fd;

struct epoll_pollfd {
  int fd;
#ifdef CONNECTION_ID_ENABLED
  struct connection_id *id;
#endif
};

#define EPOLLED2FD(ev) (((struct epoll_pollfd*)((ev)->data.ptr))->fd)
#ifdef CONNECTION_ID_ENABLED
#define EPOLLED2ID(ev) (((struct epoll_pollfd*)((ev)->data.ptr))->id)
#endif


static INLINE void alloc_connection_structs() {
  epoll_fd = epoll_create(SERVER_EVENT_QUEUE_SIZE);
  if (epoll_fd < 0) {
    ERROR("epoll_create failed on error %m\n");
  }
}

static INLINE void add_connection_poll(int socket) {
  struct epoll_event ev;
  struct epoll_pollfd *s = (struct epoll_pollfd*) malloc(sizeof(struct epoll_pollfd));

  s->fd = socket;
#ifdef CONNECTION_ID_ENABLED
  s->id = NULL;
#endif

  ev.events = EPOLLIN|EPOLLET;
  ev.data.ptr = s;
  if ( epoll_ctl(epoll_fd, EPOLL_CTL_ADD, socket, &ev) < 0 )
    ERROR("epoll_ctl ADD failed on error %m\n");
  LOG(3, "Added socket %d to epoll list\n", socket);
}

static INLINE void remove_connection_poll(struct epoll_event *ev) {
  if (epoll_ctl(epoll_fd, EPOLL_CTL_DEL, EPOLLED2FD(ev), ev) < 0)
    ERROR("epoll_ctl DEL failed on error %m\n");
  else {
    close(EPOLLED2FD(ev));
#ifdef CONNECTION_ID_ENABLED
    put_connection_id(EPOLLED2ID(ev));
#endif
    free(ev->data.ptr);
    LOG(3, "Removed socket %d from epoll list\n", EPOLLED2FD(ev));
  }
}

static INLINE void free_connection_structs() {
  /* Cannot easily get mallocated struct epoll_pollfd, we do not free them. *
   * But they should have been freed by previous client close */
  close(epoll_fd);
}


# endif /* *_TCP_POLL */

#endif /* *_ARCH */



/******************
 * Initialization *
 ******************/

int TCP_network_init()
{
  struct sockaddr_in server_address = { 0 };

  /* set up Internet address structure for the server */
  //  server_address = (struct sockaddr_in*) calloc(1, sizeof(struct sockaddr_in) );
  server_address.sin_family = AF_INET;
  server_address.sin_port = htons(server_port);
  server_address.sin_addr.s_addr = htonl(INADDR_ANY);
  LOG(1, "Port %d\n", ntohs(server_address.sin_port));
 
  /* open an internet tcp socket */
  if( (server_socket = socket(PF_INET, SOCK_STREAM, 0)) < 0 )
    {
      perror("openserver socket");
      return -1;
    }

  /* bind this socket to the server's Internet address */
  if( bind(server_socket,
	   (struct sockaddr *)&server_address,
	   sizeof(struct sockaddr_in)) < 0 )
    {
      perror("openserver bind");
      return -1;
    }

  /* set up listening backlog for connect requests from clients */
  if( listen(server_socket, LISTENING_DEPTH) < 0 )
    {      
      perror("openserver listen");
      return -1;
    }

  LOG(2, "Now listening on socket %d\n", server_socket);

#ifndef CONNECTION_THREAD_ARCH
  /* set non-blocking to accept multiple connection easily */
  fcntl(server_socket, F_SETFL, fcntl(server_socket, F_GETFL) | O_NONBLOCK);

  /* allocate all connection structures */
  alloc_connection_structs();
  /* initialize a new connection poll structure */
  add_connection_poll(server_socket);
#endif /* ~CONNECTION_THREAD_ARCH */

  return 0;
}



/**********
 * Finish *
 **********/

void TCP_network_finish()
{
#ifndef CONNECTION_THREAD_ARCH
  free_connection_structs();
#endif /* ~CONNECTION_THREAD_ARCH */

  close(server_socket);
}



/*************
 * Receiving *
 *************/

static struct server_request *network_recv(int socket)
{
  struct server_request *req;

  req = get_free_request();
  req->connection.socket = socket;
  req->in_vect = NULL;
  
  LOG(2, "Recving on socket %d\n", socket);

  /* Recv the request */
  if ( recv(socket, &req->req, sizeof(struct request), 0) < sizeof(struct request) )
    {
      LOG(3, "No data received in socket %d\n",
	  socket);
      put_free_request(req);
      return (struct server_request*)NULL;
    }
  
  /* Check the type */
  LOG(3, "Recu type %d\n",
      (int)req->req.type);
  if ( req->req.type <= FIRST_REQ
       ||  req->req.type >= LAST_REQ )
    {
      ERROR("Bad request type %d received, impossible with TCP ?!\n",
	    (int)req->req.type);
      assert(0);
    }

  /* Recv data buffer */
  if ( req->req.buffer_len > 0 )
    {
      int rlen = 0;
      int data_len = req->req.buffer_len;
      char *buf;
      
      LOG(4, "Receiving %d data\n", data_len);
      
      req->in_vect = get_io_monovector_alloc(data_len);
      buf = VECT_PTR(req->in_vect, 0);
      
      while ( data_len > 0)
	{
	  rlen = recv(socket, buf, data_len, 0);
	  LOG(5, "Recv %d/%d data\n", rlen, data_len);
	  data_len -= rlen;
	  buf += rlen;
	}
    }

  return req;
}



/***********
 * Sending *
 ***********/

int TCP_network_send(struct server_request *req)
{
  struct iovec *vect;
  int vect_len = 1;
  int sent;

#ifdef PROTOCOL_DEBUG
  {
#include <sys/poll.h>
    struct pollfd pfd = { fd: req->connection.socket, events: POLLIN, revents: 0};
    poll(&pfd, 1, 0);
    if ( pfd.revents )
      ERROR("events %d on socket %d\n", pfd.revents, req->connection.socket);
  }
#endif /* PROTOCOL_DEBUG */

  if ( req->out_vect  &&  req->ans.buffer_len > 0 )
    vect_len += VECT_NR(req->out_vect);

  vect = (struct iovec*) malloc(vect_len*sizeof(struct iovec));

  vect[0].iov_base = &req->ans;
  vect[0].iov_len = sizeof(struct answer);

  /* check the ANS type */
  LOG(2, "Sent type %d\n",
      (int)req->ans.type);
  if ( req->ans.type <= FIRST_ANS
       ||  req->ans.type >= LAST_ANS )
    {
      ERROR("Cannot come here !!\n");
      assert(0);
    }

  /* Add a data buffer if necessary */
  if ( req->ans.buffer_len > 0 
#ifdef TCP_SENDFILE_ENABLED
       && req->ans.type != PREAD_ANS
#endif
       ) {
    int i;
    for(i=0; i<req->out_vect->nr; i++) {
      vect[i+1].iov_base = VECT_PTR(req->out_vect, i);
      vect[i+1].iov_len = VECT_SIZE(req->out_vect, i);
      LOG(4, "Added buffer #%d of size %ld\n",
	  i, (long)VECT_SIZE(req->out_vect, i));
    }
    
    LOG(3, "Sending %ld data...\n",
	(long)req->ans.buffer_len);
  }
  
  /* Send the answer */
  sent = writev(req->connection.socket, vect, vect_len);
   
#ifdef TCP_SENDFILE_ENABLED
  if ( req->ans.type == PREAD_ANS  &&  req->ans.buffer_len > 0) {
    ssize_t readen;
    off_t offset = req->req.data.pread.offset;
    int fd = STD_get_local_fd_of_fd((void*)(uintptr_t)req->req.data.pread.fd_id);
    DEBUG_ASSERT( fd >= 0 );
    readen = sendfile(req->connection.socket, fd,
		      &offset,
		      req->ans.buffer_len);
    assert(readen == req->ans.buffer_len);
  }
#endif
   
  /* free all */
  put_io_vector(req->out_vect);
  put_free_request(req);
  free(vect);

  return 0;
}



#ifdef CONNECTION_THREAD_ARCH

/*********************
 *********************
 * Connection Thread *
 *********************
 *********************/

static void connection_thread_func(int socket)
{
#ifdef CONNECTION_ID_ENABLED
  struct connection_id *connection_id;
#endif /* ~CONNECTION_ID_ENABLED */

  while ( 1 ) {
    struct server_request *req = network_recv(socket);
    
    if (req == NULL) {
#ifdef CONNECTION_ID_ENABLED
      put_connection_id(connection_id);
#endif /* ~CONNECTION_ID_ENABLED */
      pthread_exit(NULL);
    } else {
#ifdef CONNECTION_ID_ENABLED
      if (req->req.type == CONNECTION_REQ) {
	process_connection_id(req, &connection_id);
	TCP_network_send(req);	
      } else {
	process_request(req);
	TCP_network_send(req);
      }
#else /* CONNECTION_ID_ENABLED */
      process_request(req);
      TCP_network_send(req);
#endif /* ~CONNECTION_ID_ENABLED */
    }
  }
}


void CONNECTION_THREAD_main_connection_thread_func() {
  while ( 1 ) {
    int new_socket;
    pthread_t new_thread;
    new_socket = accept(server_socket, NULL, 0);
    if ( new_socket == -1 ) {
      ERROR("accept failed do to error %m\n");
      break;
    }
    SOCKET_TCP_NODELAY(new_socket);
    pthread_create(&new_thread, NULL,
		   (void*(*)(void*))connection_thread_func,
		   (int*)new_socket);
    LOG(3, "Added new thread to handle new connection on socket %d\n",
	new_socket);
  }
}


#else /* ~CONNECTION_THREAD */

/***********************************
 ***********************************
 * Sequential or Queue Thread Arch *
 ***********************************
 ***********************************/


/***************************
 * Handling new connection *
 **************************/

/* When the server_socket gets a connect request
 * we accept it and add it to the array
 */
static INLINE void connect_new_clients()
{
  int socket_for_client;
  int magic = NETWORK_MAGIC;

  while ( 1 ) {
    socket_for_client = accept(server_socket, NULL, 0);
    if ( socket_for_client < 0 ) {
      if ( errno == EAGAIN   ||  errno == EWOULDBLOCK )
	break;
      else {
	ERROR("accept failed due to error %m\n");
	break;
      }
    }
    SOCKET_TCP_NODELAY(socket_for_client);
    add_connection_poll(socket_for_client);
    write(socket_for_client, &magic, sizeof(int));
    LOG(2, "Added a new client on socket %d\n",
	socket_for_client);
  }
}



/*******************
 * Handling events *
 *******************/

# ifdef STD_TCP_POLL

/******************************
 * Using Standard TCP Polling *
 ******************************/

static INLINE void STD_handle_events(int res) {
  int i, fd;
  int events;
  int old_nr = connection_nr;
   
  events = pollfd[0].revents;
  if (events) {
    if (events != POLLIN) {
      ERROR("Strange event %d polled from main connection\n",
	    events);
      assert(0);
    }
    connect_new_clients();
    res--;
  }

  for(i=old_nr-1; i>0 && res>0; i--) {
    events = pollfd[i].revents;
    fd = POLLED2FD(i);
    if (events) {
      struct server_request *req;
      if ( events == POLLIN  &&  (req=network_recv(fd)) != NULL ) {
#ifdef CONNECTION_ID_ENABLED
	if ( req->req.type == CONNECTION_REQ ) {
	  process_connection_id(req, &POLLED2ID(i));
	  queue_done_request(req);
	} else {
	  queue_ready_request(req);
	}
#else /* CONNECTION_ID_ENABLED */
	queue_ready_request(req);
#endif /* ~CONNECTION_ID_ENABLED */
      } else {
	remove_connection_poll(i);
      }
      res--;
    } 
  }
}


#  ifdef AIO_MIXED_WITH_OTHER_ASYNC_API

void TCP_network_test() {
  int res = poll(pollfd, connection_nr, 0);
  LOG(1, "Poll gets %d revents\n", res);

  if (res > 0)
    STD_handle_events(res);
}


#  else

void TCP_network_wait() {
  int res;

  res = poll(pollfd, connection_nr, -1);
  LOG(1, "Poll gets %d revents\n", res);

  if (res < 0) {
    /* poll interrupt by a signal.
     * return to main loop */
    return;
  }

  STD_handle_events(res);
}


#  endif

# elif (defined AIO_TCP_POLL)

/******************************
 * Using AIO TCP Polling *
 ******************************/

static void AIO_connect_callback(io_context_t ctx,
				 struct iocb *iocb,
				 long res, long res2)
{
  if ( res != POLLIN ) {
    ERROR("Strange event %ld polled from main connection\n",
	  res);
    assert(0);
  }
  connect_new_clients();
  submit_connection_poll(iocb, AIO_connect_callback);
}


static void AIO_recv_callback(io_context_t ctx,
			      struct iocb *iocb,
			      long res, long res2)
{
  int fd = IOCB2FD(iocb);
  struct server_request *req;

  LOG(2, "Polled on socket %d gets %ld\n",
      fd, res);
  
  if ( res == POLLIN  &&  (req=network_recv(fd)) != NULL ) {
#ifdef CONNECTION_ID_ENABLED
    if ( req->req.type == CONNECTION_REQ ) {
      process_connection_id(req, &IOCB2ID(iocb));
      queue_done_request(req);
    } else {
      submit_connection_poll(iocb, AIO_recv_callback);
      queue_ready_request(req);
    }
#else /* CONNECTION_ID_ENABLED */
    submit_connection_poll(iocb, AIO_recv_callback);
    queue_ready_request(req);
#endif /* ~CONNECTION_ID_ENABLED */
  } else {
    remove_connection_poll(iocb);      
  }
}


/*****************
 * Wait/Test functions are combined with
 * other AIO calls (disk IO) in server.c
 */


# else /* EPOLL_TCP_POLL */

/******************************
 * Using Epoll TCP Polling *
 ******************************/

static INLINE void EPOLL_handle_events(struct epoll_event ev[], int res) {
  int i;
  int events, fd;

  for(i=0; i<res; i++) {
    events = ev[i].events;
    fd = EPOLLED2FD(&ev[i]);

    if (fd == server_socket) {
      if (events != EPOLLIN) {
	ERROR("Strange event %d polled from main connection\n",
	      events);
	assert(0);
      }
      connect_new_clients();
    } else {
      struct server_request *req;
      if ( events == EPOLLIN  &&  (req=network_recv(fd)) != NULL ) {
#ifdef CONNECTION_ID_ENABLED
	if ( req->req.type == CONNECTION_REQ ) {
	  process_connection_id(req, &EPOLLED2ID(&ev[i]));
	  queue_done_request(req);
	} else {
	  queue_ready_request(req);
	}
#else /* CONNECTION_ID_ENABLED */
	queue_ready_request(req);
#endif /* ~CONNECTION_ID_ENABLED */
      } else {
	remove_connection_poll(&ev[i]);
      }
    }
  }
}


#  ifdef AIO_MIXED_WITH_OTHER_ASYNC_API

void TCP_network_test() {
  struct epoll_event events[SERVER_EVENT_GET_MAX];
  int res;

 try_again:
  res = epoll_wait(epoll_fd, events, SERVER_EVENT_GET_MAX, 0);
  LOG(1, "epoll_wait gets %d revents\n", res);

  if ( res < 0 ) {
    ERROR("epoll_wait failed with error %m\n");
    assert(0);
  }

  if ( !res )
    return;

  EPOLL_handle_events(events, res);

  if (res == SERVER_EVENT_GET_MAX) {
    LOG(3, "Got max events, try again...\n");
    goto try_again;
  }
}


#  else

void TCP_network_wait() {
  struct epoll_event events[SERVER_EVENT_GET_MAX];
  int res;
  
  res = epoll_wait(epoll_fd, events, SERVER_EVENT_GET_MAX, -1);
  LOG(1, "epoll_wait gets %d revents\n", res);

  if ( res < 0 ) {
    ERROR("epoll_wait failed with error %m\n");
    assert(0);
  }

  EPOLL_handle_events(events, res);

  while (res == SERVER_EVENT_GET_MAX) {
    LOG(3, "Got max events, try again...\n");

    res = epoll_wait(epoll_fd, events, SERVER_EVENT_GET_MAX, 0);
    LOG(1, "epoll_wait gets %d revents\n", res);

    if ( res < 0 ) {
      ERROR("epoll_wait failed with error %m\n");
      assert(0);
    }

    if ( !res )
      return;

    EPOLL_handle_events(events, res);
  }
}


#  endif

# endif /* *_TCP_POLL */

#endif /* ~CONNECTION_THREAD_ARCH */
