/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/uio.h>
#include <assert.h>

#include "orfas_queue.h"
#include "orfas_io.h"
#include "orfas_debug.h"


static int server_socket = -1;


/**********************
 * Polling definition *
 **********************/

static struct pollfd pollfd;


/******************
 * Initialization *
 ******************/

int UDP_network_init()
{
  struct sockaddr_in server_address = { 0 };
  char server_name[MAXHOSTNAMELEN+1];
  struct hostent *server_hostent;
  
  /* get the internet name of the local host node on which we are running */
  if ( gethostname(server_name, MAXHOSTNAMELEN) < 0 )
    {
      perror("openserver gethostname");
      return -1;
    }

  /* get structure for local interface server is to use */
  /* do not free server_hostent, it is allocated by malloc ! */
  if( (server_hostent = gethostbyname(server_name)) == NULL )
    {
      herror(server_name);
      return -1;
    }

  /* set up Internet address structure for the server */
  server_address.sin_family = server_hostent->h_addrtype;  /* should be AF_INET */
  server_address.sin_port = htons(server_port);
  LOG(1, "Port %d\n", ntohs(server_address.sin_port));

  if( server_name == NULL )
    server_address.sin_addr.s_addr = htonl(INADDR_ANY); /* use any interface */
  else
    memcpy(&(server_address.sin_addr),
	   server_hostent->h_addr,
	   server_hostent->h_length);
 
  /* open an internet tcp socket */
  if( (server_socket = socket(PF_INET, SOCK_DGRAM, 0)) < 0 )
    {
      perror("openserver socket");
      return -1;
    }
  
  /* bind this socket to the server's Internet address */
  if( bind(server_socket,
	   (struct sockaddr *)&server_address,
	   sizeof(struct sockaddr_in)) < 0 )
    {
      perror("openserver bind");
      return -1;
    }
 
  pollfd.fd = server_socket;
  pollfd.events = POLLIN;

  return 0;
}


void UDP_network_finish()
{
  close(server_socket);
}



/***********
 * Sending *
 ***********/

int UDP_network_send(struct server_request *req)
{
  struct iovec vect[2];
  struct msghdr message;

  vect[0].iov_base = &req->ans;
  vect[0].iov_len = sizeof(struct answer);

  message.msg_name = &req->connection.ski;
  message.msg_namelen = sizeof(struct connection);
  message.msg_iov = vect;
  message.msg_iovlen = 1;
  message.msg_control = NULL;
  message.msg_controllen = 0;
  message.msg_flags = 0;

  /* check the ANS type */
  LOG(1, "Sent type %d\n",
      (int)req->ans.type);
  if ( req->ans.type == BAD_ANS )
    {
      ERROR("Bad request => Bad answer %d\n",
	    (int)req->ans.type);
    }
  else if ( req->ans.type <= FIRST_ANS
	    ||  req->ans.type >= LAST_ANS )
    {
      ERROR("Cannot come here !!\n");
      assert(0);
    }

  /* Add a data buffer if necessary */
  if ( req->ans.buffer_len > 0 )
    {
      char *remaining_buf;
      size_t remaining_len;
      size_t to_send;
      int vector_indix = 0;

      remaining_buf = VECT_PTR(req->out_vect, 0);
      remaining_len = VECT_SIZE(req->out_vect, 0);

      LOG(3, "%ld data to send...\n",
	  (long)req->ans.buffer_len);

      /* send the answer and the begining of data */
      to_send = remaining_len > MAX_MESSAGE_LEN - sizeof(struct answer)
	? MAX_MESSAGE_LEN - sizeof(struct answer)
	: remaining_len;

      message.msg_iovlen = 2;
      vect[1].iov_base = remaining_buf;
      vect[1].iov_len = to_send;
      LOG(3, "Send answer and begining of data\n");
      sendmsg(server_socket, &message, 0);

      while ( remaining_len > to_send 
	      ||  vector_indix < VECT_NR(req->out_vect)-1 ) {
	/* update buffer parameters */
	remaining_len -= to_send;
	if ( !remaining_len ) {
	  vector_indix++;
	  remaining_buf = VECT_PTR(req->out_vect, vector_indix);
	  remaining_len = VECT_SIZE(req->out_vect, vector_indix);
	} else
	  remaining_buf += to_send;
 
	/* send next part of data */
	to_send = remaining_len > MAX_MESSAGE_LEN
	  ? MAX_MESSAGE_LEN
	  : remaining_len;
	sendto(server_socket,
	       remaining_buf, to_send,
	       0,
	       message.msg_name, message.msg_namelen);
	LOG(3, "Send %ld next data\n",
	    (long)to_send);
      }
    }
  else
    {
      LOG(3, "Just send the answer\n");      /* just send a message */
      sendmsg(server_socket, &message, 0);
    }
  
  /* free all */
  put_io_vector(req->out_vect);
  put_free_request(req);

  return 0;
}



/*************
 * Receiving *
 *************/

static INLINE int network_recv()
{
  struct iovec vect[2];
  struct msghdr message;
  
  struct server_request *req;

  req = get_free_request();
  req->in_vect = NULL;

  vect[0].iov_base = &(req->req);
  vect[0].iov_len = sizeof(struct request);

  message.msg_name = &req->connection.ski;
  message.msg_namelen = sizeof(struct connection);
  message.msg_iov = vect;
  message.msg_iovlen = 1;
  message.msg_control = NULL;
  message.msg_controllen = 0;
  message.msg_flags = 0;

  /* Recv the request */
  if ( recvmsg(server_socket, &message, MSG_PEEK) < 0 )
    return -1;

  /* Check the type */
  LOG(1, "Recu type %d\n", 
      (int)req->req.type);
  if ( req->req.type <= FIRST_REQ
       ||  req->req.type >= LAST_REQ )
    {
      ERROR("Bad request type received %d\n",
	    (int)req->req.type);
      req->req.type = BAD_REQ;
      /* receive it to be empty the queue */
      recvmsg(server_socket, &message, 0);
    }
  else
    {
      /* Add a data buffer if necessary */
      if ( req->req.buffer_len > 0 )
	{
	  char *remaining_buf;
	  size_t remaining_len;
	  size_t to_recv;
	  
	  req->in_vect = get_io_monovector_alloc(req->req.buffer_len);

	  remaining_buf = VECT_PTR(req->in_vect, 0);
	  remaining_len = req->req.buffer_len;

	  LOG(3, "%ld data to recv...\n",
	      (long)remaining_len);

	  /* recv the answer and the begining of data */
	  to_recv = remaining_len > MAX_MESSAGE_LEN - sizeof(struct request)
	    ? MAX_MESSAGE_LEN - sizeof(struct request)
	    : remaining_len;

	  message.msg_iovlen = 2;
	  vect[1].iov_base = remaining_buf;
	  vect[1].iov_len = to_recv;
	  LOG(3, "Recv request and begining of data\n");
	  recvmsg(server_socket, &message, 0);
	 
	  while ( remaining_len > to_recv ) {
	    /* update buffer parameters */
	    remaining_buf += to_recv;
	    remaining_len -= to_recv;

	    /* recv next part of data */
	    to_recv = remaining_len > MAX_MESSAGE_LEN
	      ? MAX_MESSAGE_LEN
	      : remaining_len;
	    to_recv = read(server_socket, remaining_buf, to_recv);
	    LOG(3, "Recv %ld next data\n",
		(long)to_recv);
	  };
	}
      else
	{
	  LOG(2, "Just recv the request\n");
	  /* just receive a message */    
	  recvmsg(server_socket, &message, 0);
	}
    }

  queue_ready_request(req);

  return 0;
}


#ifdef AIO_MIXED_WITH_OTHER_ASYNC_API

void UDP_network_test()
{
  int res = poll(&pollfd, 1, 0);
  if ( res < 0 )
    return;
  network_recv();
}

#else /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */

void UDP_network_wait()
{
  int res = poll(&pollfd, 1, -1);
  if ( res < 0 )
    return;
  network_recv();
}

#endif /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */
