/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __ORFAS_SERVER_H__
#define __ORFAS_SERVER_H__


/* check if IO type is well defined */
#if ( STD_IO + AIO_IO + FAKE_IO + RAM_IO != 1 )
#error "You must define one IO type !"
#endif /* STD_IO or AIO_IO or FAKE_IO or RAM_IO */

/* check if arch type is well defined */
#if ( SEQUENTIAL_ARCH + QUEUE_THREAD_ARCH + CONNECTION_THREAD_ARCH != 1 )
#error "You must define one ARCH type !"
#endif /* SEQUENTIAL_ARCH or QUEUE_THREAD_ARCH or CONNECTION_THREAD_ARCH */

/* with TCP, check if TCP poll type is well defined */
#if (defined NETWORK_TCP) && ( STD_TCP_POLL + EPOLL_TCP_POLL + AIO_TCP_POLL != 1 )
#error "You must define one TCP Poll Arch type !"
#endif /* NETWORK_TCP and ( STD_TCP_POLL or EPOLL_TCP_POLL or AIO_TCP_POLL ) */

/* globals */
#define SERVER 1
#define _GNU_SOURCE 1

/* AIO configuration */
#if (defined AIO_IO) || (defined AIO_TCP_POLL)
# define NEED_AIO
#endif

#if (defined AIO_IO) && (!defined AIO_TCP_POLL) && (!defined QUEUE_THREAD_ARCH)
# define AIO_MIXED_WITH_OTHER_ASYNC_API
#endif


/* maximal number of real FD open by the server */
#define FD_MAX 8192

/* number of server_request which are allocated at once */
#define QUEUE_ALLOCATING_NUMBER 10


/* initial uid/gid */
#include <sys/types.h>
extern uid_t uid;
extern gid_t gid;


/* debug */
#if CLIENT_DEBUG_LEVEL >= 0
#define INLINE /* nothing */
#else
#define INLINE inline
#endif


#ifdef FAKE_IO
/***********************
 * FAKE_IO definitions *
 ***********************/

#define DIRECTORY_FILE_ENTRIES 3
#define DIRECTORY_DIR_ENTRIES 3
#define DEPTH_MAX 2

#define FILE_LEN_RATIO 428143657

#define DIR_PREFIX "d"
#define DIR_PREFIX_LEN 1
#define FILE_PREFIX "f"
#define FILE_PREFIX_LEN 1


#elif (defined RAM_IO)
/**********************
 * RAM_IO definitions *
 **********************/

#define RAMFS_FD_MAX 256

#define MEMORY_BLOCK_BITS 16
#define MEMORY_BLOCK_SIZE (1UL << (MEMORY_BLOCK_BITS))
#define BLOCKS_DEFAULT_MAX 32 /* 2 Mo per file */

#define DIRENTS_DEFAULT_MAX 128

#define HASH_TABLE_SIZE 3141
#define RAMFS_MAX_FILE_NAME 256
#define HASH_VALUE HASH_TABLE_SIZE


#endif /* RAM_IO or FAKE_IO */



#ifdef NETWORK_TCP
/*******************
 * TCP definitions *
 *******************/

#define LISTENING_DEPTH 32


#elif ( defined NETWORK_GM )
/******************
 * GM definitions *
 ******************/

#define SERVER_BOARD_NUM 0

#define PREPROVIDED_REQUEST_RECV_TOKEN_RATIO 0.4
#define PREPROVIDED_DATA_RECV_TOKEN_RATIO 0.4


#elif ( defined NETWORK_BIP )
/*******************
 * BIP definitions *
 *******************/

#define SERVER_REQUEST_QUEUE_LEN 16


#elif ( defined NETWORK_FAKE )
/********************
 * FAKE definitions *
 ********************/

#error Need not to compile server with FAKE_NETWORK

#endif

/************************
 * Command line options *
 ************************/

#if ( defined NETWORK_TCP ) || ( defined NETWORK_UDP )
extern unsigned long server_port;
#elif ( defined NETWORK_GM )
extern unsigned int server_port_id;
#endif


#ifdef NEED_AIO
/*******************
 * AIO definitions *
 *******************/

#include <libaio.h>
extern void AIO_submit(struct iocb *iocb);
extern void AIO_cancel(struct iocb *iocb);

#endif /* NEED_AIO */



/************************************
 * general server request structure *
 ************************************/

#include "orfa_protocol.h"
#include "orfa_common_network.h"

struct server_request {
  struct request req;
  struct io_vector *in_vect;

  struct answer ans;
  struct io_vector *out_vect;

  struct connection connection;

#ifdef AIO_IO
  void *key;
#endif
#ifdef NETWORK_GM
  int send_to_do;
#endif

  struct server_request *prev, *next;
};


#endif /* __ORFAS_SERVER_H__ */
