/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>

#include "orfas_network.h"
#include "orfas_debug.h"
#include "orfas_io.h"


#ifdef CONNECTION_ID_ENABLED


struct connection_id {
  int count;
};



static INLINE struct connection_id *add_connection_id() {
  struct connection_id *new;
  new = (struct connection_id*) malloc(sizeof(struct connection_id));
  new->count = 0;
  LOG(6, "Added new connection_id %p\n", new);
  return new;
}


static INLINE void get_connection_id(struct connection_id *id) {
  id->count++;
  LOG(5, "Got connection_id %p with count %d\n", id, id->count);
}


static INLINE void real_remove_connection_id(struct connection_id *old) {
  LOG(5, "Freeing connection_id %p\n", old);
  close_file_for_connection_id(old);
  free(old);
}


void put_connection_id(struct connection_id *id) {
  id->count--;
  LOG(4, "Releaseing connection_id %p to count %d\n", id, id->count);
  if (!id->count)
    real_remove_connection_id(id);
}


void process_connection_id(struct server_request *req, struct connection_id **idptr) {
  struct connection_id *id;

  LOG(4, "Process connection_id\n");

  DEBUG_ASSERT( req->req.type == CONNECTION_REQ );

  switch ( req->req.data.connection.op ) {
  case CONNECTION_OP_NEW: {
    id = add_connection_id();
    *idptr = id;
    LOG(4, "Setting connection_id to %p\n", id);
    get_connection_id(id);
    req->ans.return_value.ptr_value = (orfa_connection_id_t)(uintptr_t)id;
    break;
  }
  case CONNECTION_OP_FORK_DONE: {
    id = (struct connection_id*)(uintptr_t)req->req.connection_id;
    *idptr = id;
    LOG(4, "Setting connection_id to %p\n", id);    
    get_connection_id(id);
    req->ans.return_value.ptr_value = (orfa_connection_id_t)(uintptr_t)id;
    break;
  }
  case CONNECTION_OP_BEFORE_EXEC: {
    id = (struct connection_id*)(uintptr_t)req->req.connection_id;
    get_connection_id(id);
    req->ans.return_value.ptr_value = (orfa_connection_id_t)(uintptr_t)id;
    break;
  }
  case CONNECTION_OP_EXEC_DONE: {
    id = (struct connection_id*)(uintptr_t)req->req.connection_id;
    *idptr = id;
    LOG(4, "Setting connection_id to %p\n", id);
    req->ans.return_value.ptr_value = (orfa_connection_id_t)(uintptr_t)id;
    break;
  }
  case CONNECTION_OP_EXEC_FAILED: {
    id = (struct connection_id*)(uintptr_t)req->req.connection_id;
    put_connection_id(id);
    req->ans.return_value.ptr_value = (orfa_connection_id_t)(uintptr_t)id;
    break;
  }
  }

  req->ans.type = CONNECTION_ANS;
#ifdef PROTOCOL_DEBUG
  req->ans.tag = req->req.tag;
  req->ans.pid = req->req.pid;
#endif /* PROTOCOL_DEBUG */
}


#endif /* CONNECTION_ID_ENABLED */

