/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>

#include "orfas_network.h"
#include "orfas_debug.h"
#include "orfas_io.h"



/*****************
 * FREE REQUESTS *
 *****************/

static struct server_request *free_request_queue = NULL;

#if ( defined CONNECTION_THREAD_ARCH  ||  defined QUEUE_THREAD_ARCH )

/* This lock protect against concurrent access.
 * This could be a spinlock */

#include <pthread.h>
static pthread_mutex_t free_mutex = PTHREAD_MUTEX_INITIALIZER;

#endif


static void alloc_new_free_request()
{
  int i;
  struct server_request *new = (struct server_request*)
    network_buffer_malloc(QUEUE_ALLOCATING_NUMBER * sizeof(struct server_request));

  for(i=0; i<QUEUE_ALLOCATING_NUMBER-1; i++)
    new[i].next = &new[i+1];
  new[QUEUE_ALLOCATING_NUMBER-1].next = NULL;

  free_request_queue = &new[0];

  LOG(4, "Allocated %d new request\n", QUEUE_ALLOCATING_NUMBER);
}


void put_free_request(struct server_request *req)
{
  LOG(4, "Put free request %p\n", req);

#if ( defined CONNECTION_THREAD_ARCH  ||  defined QUEUE_THREAD_ARCH )
  pthread_mutex_lock(&free_mutex);
#endif

  req->next = free_request_queue;
  free_request_queue = req;

#if ( defined CONNECTION_THREAD_ARCH  ||  defined QUEUE_THREAD_ARCH )
  pthread_mutex_unlock(&free_mutex);
#endif
}


struct server_request* get_free_request()
{
  struct server_request *req;

#if ( defined CONNECTION_THREAD_ARCH  ||  defined QUEUE_THREAD_ARCH )
  pthread_mutex_lock(&free_mutex);
#endif

  if ( free_request_queue == NULL )
    alloc_new_free_request();

  req = free_request_queue;
  free_request_queue = free_request_queue->next;

#if ( defined CONNECTION_THREAD_ARCH  ||  defined QUEUE_THREAD_ARCH )
  pthread_mutex_unlock(&free_mutex);
#endif

  memset(req, 0, sizeof(struct server_request));
  LOG(4, "Get free request %p\n", req);
  return req;
}



#ifndef CONNECTION_THREAD_ARCH
/****************************************
 * GENERIC QUEUE MANIPULATION FUNCTIONS *
 ****************************************/

/* adding a request at the end of a queue */
static void queue_generic_request(struct server_request *req, struct server_request **queue)
{
  if ( !*queue )
    {
      *queue = req;
      req->prev = NULL;
    }
  else
    {
      struct server_request *tmp = NULL;

      tmp = *queue;
      while ( tmp->next != NULL )
	tmp = tmp->next;

      tmp->next = req;
      req->prev = tmp;
    }
  req->next = NULL;
}


/* removing and returning the first request in a queue */
static struct server_request *dequeue_generic_request(struct server_request **queue)
{
  if ( !*queue )
    return NULL;
  else
    {
      struct server_request *tmp = *queue;
      *queue = tmp->next;
      if ( *queue )
	(*queue)->prev = NULL;
      return tmp;
    }
}



/******************
 * READY REQUESTS *
 ******************/

static struct server_request *ready_request_queue = NULL;

#ifdef QUEUE_THREAD_ARCH

/* This lock protects against concurrent access.
 * This have to be a mutex due to the necessity
 * to wait for a new ready request.
 * But this strategie could change. */

#include <pthread.h>
static pthread_mutex_t ready_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t ready_cond = PTHREAD_COND_INITIALIZER;

#endif


void queue_ready_request(struct server_request *req)
{
  LOG(4, "Queue ready request %p\n", req);
#ifdef QUEUE_THREAD_ARCH
  pthread_mutex_lock(&ready_mutex);
#endif
  queue_generic_request(req, &ready_request_queue);
#ifdef QUEUE_THREAD_ARCH
  pthread_cond_signal(&ready_cond);
  pthread_mutex_unlock(&ready_mutex);
#endif
}

struct server_request *dequeue_ready_request()
{
  struct server_request *req;
#ifdef QUEUE_THREAD_ARCH
  pthread_mutex_lock(&ready_mutex);
  while ( ( req=dequeue_generic_request(&ready_request_queue) )
	  == NULL )
    pthread_cond_wait(&ready_cond, &ready_mutex);

  pthread_mutex_unlock(&ready_mutex);
#else
  req = dequeue_generic_request(&ready_request_queue);
#endif
  LOG(4, "Dequeue ready request %p\n", req);
  return req;
}



/*****************
 * DONE REQUESTS *
 *****************/

static struct server_request *done_request_queue = NULL;

#ifdef QUEUE_THREAD_ARCH

/* This lock protects against concurrent access.
 * This have to be a mutex due to the necessity
 * to wait for a new done request.
 * But this strategie could change. */

#include <pthread.h>
static pthread_mutex_t done_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t done_cond = PTHREAD_COND_INITIALIZER;

#endif


void queue_done_request(struct server_request *req)
{
  LOG(4, "Queue done request %p\n", req);
#ifdef QUEUE_THREAD_ARCH
  pthread_mutex_lock(&done_mutex);
#endif
  queue_generic_request(req, &done_request_queue);
#ifdef QUEUE_THREAD_ARCH
  pthread_cond_signal(&done_cond);
  pthread_mutex_unlock(&done_mutex);
#endif
}

struct server_request *dequeue_done_request()
{
  struct server_request *req;
#ifdef QUEUE_THREAD_ARCH
  pthread_mutex_lock(&done_mutex);
  while ( ( req=dequeue_generic_request(&done_request_queue) )
	  == NULL )
    pthread_cond_wait(&done_cond, &done_mutex);

  pthread_mutex_unlock(&done_mutex);
#else
  req = dequeue_generic_request(&done_request_queue);
#endif
  LOG(4, "Dequeue done request %p\n", req);
  return req;
}


#endif /* ~CONNECTION_THREAD_ARCH */



#ifdef AIO_IO
/***********************
 * PROCESSING REQUESTS *
 ***********************/

static struct server_request *processing_request_queue = NULL;

#ifdef QUEUE_THREAD_ARCH

/* This lock protects against concurrent access.
 * This could be a spinlock even if dequeueing
 * can be long. */

#include <pthread.h>
static pthread_mutex_t processing_mutex = PTHREAD_MUTEX_INITIALIZER;

#endif


void queue_processing_request(struct server_request *req)
{
  LOG(4, "Queue processing request %p\n", req);
#ifdef QUEUE_THREAD_ARCH
  pthread_mutex_lock(&processing_mutex);
#endif
  queue_generic_request(req, &processing_request_queue);
#ifdef QUEUE_THREAD_ARCH
  pthread_mutex_unlock(&processing_mutex);
#endif
}


/* this function looks in the not_ready_request_queue
   for a request and removes it.
   on error, it returns NULL */
struct server_request *dequeue_processing_request_for_key(void *key)
{
  struct server_request *tmp = NULL;

#ifdef QUEUE_THREAD_ARCH
  pthread_mutex_lock(&processing_mutex);
#endif
  tmp = processing_request_queue;
  while ( tmp != NULL )
    {
      if ( key == tmp->key )
	{
	  if ( tmp->prev )
	    tmp->prev->next = tmp->next;
	  else
	    processing_request_queue = tmp->next;

	  if ( tmp->next )
	    tmp->next->prev = tmp->prev;

	  LOG(4, "Dequeue processing request %p\n", tmp);
	  break;
	}
      tmp = tmp->next;
    }
  /* the request MUST be there */
  DEBUG_ASSERT( tmp != NULL );

#ifdef QUEUE_THREAD_ARCH
  pthread_mutex_unlock(&processing_mutex);
#endif
  return tmp;
}


#endif /* AIO_IO */



#ifdef NETWORK_GM
/**********************
 * NOT READY REQUESTS *
 **********************/

/* No lock is needed here because everything is done
 * in the same thread (the receiver one). */


static struct server_request *not_ready_request_queue = NULL;

struct not_ready_buffer {
  struct io_vector *vect;
  struct not_ready_buffer *next, *prev;
  struct connection connection;
};

static struct not_ready_buffer *not_ready_buffer_queue = NULL;


void queue_not_ready_request(struct server_request *req)
{
  LOG(4, "Queue not ready request %p\n", req);
  queue_generic_request(req, &not_ready_request_queue);
}


/* this function looks in the not_ready_request_queue
   for a request and removes it.
   on error, it returns NULL */
struct server_request *dequeue_not_ready_request_for_GM_connection(struct connection *connection)
{
  struct server_request *tmp = NULL;

  tmp = not_ready_request_queue;
  while ( tmp != NULL )
    {
      if ( !memcmp(&tmp->connection, connection, sizeof(struct connection)) )
	{
	  if ( tmp->prev )
	    tmp->prev->next = tmp->next;
	  else
	    not_ready_request_queue = tmp->next;

	  if ( tmp->next )
	    tmp->next->prev = tmp->prev;

	  LOG(4, "Dequeue not ready request %p\n", tmp);
 	  return tmp;
	}
      tmp = tmp->next;
    }
  return NULL;
}


void queue_not_ready_buffer(struct io_vector *vect, struct connection *connection)
{
  struct not_ready_buffer *tmp = not_ready_buffer_queue;
  struct not_ready_buffer *buf = malloc(sizeof(struct not_ready_buffer));
  buf->vect = vect;
  memcpy(&buf->connection, connection, sizeof(struct connection));

  buf->next = NULL;

  LOG(4, "Queue not ready buffer %p for vector %p\n",
      buf, vect);

  if ( tmp )
    {
      while ( tmp->next != NULL )
	tmp = tmp->next;

      tmp->next = buf;
      buf->prev = tmp;
    }
  else
    {
      not_ready_buffer_queue = buf;
      buf->prev = NULL;
    }
}


struct io_vector *dequeue_not_ready_buffer_for_GM_connection(struct connection *connection)
{
  struct not_ready_buffer *tmp = NULL;

  tmp = not_ready_buffer_queue;
  while ( tmp != NULL )
    {
      if ( !memcmp(&tmp->connection, connection, sizeof(struct connection)) )
	{
	  struct io_vector *vect = tmp->vect;

	  LOG(4, "Dequeue not ready buffer %p for vector %p\n",
	      tmp, vect);

	  if ( tmp->prev )
	    tmp->prev->next = tmp->next;
	  else
	    not_ready_buffer_queue = tmp->next;

	  if ( tmp->next )
	    tmp->next->prev = tmp->prev;

	  free(tmp);

 	  return vect;
	}
      tmp = tmp->next;
    }
  return NULL;
}

#endif /* NETWORK_GM */


