/*
 * ORFA - A user-level remote file system access protocol
 * Copyright (C) 2002 Brice Goglin <Brice.Goglin@ens-lyon.fr>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "orfas_network.h"
#include "orfas_queue.h"
#include "orfas_io.h"
#include "orfas_debug.h"


/* initial uid/gid */
uid_t uid;
gid_t gid;


/* just a counter to inform user */
static int request_nr = 0;


/* initialize some LOG needs */
INIT_LOGS();



/************************
 * Command Line Parsing *
 ************************/

static char *callname;

#if ( defined NETWORK_TCP ) || ( defined NETWORK_UDP )
unsigned long server_port = SERVER_PORT;
#elif ( defined NETWORK_GM )
unsigned int server_port_id = SERVER_PORT_ID;
#endif

static
void exit_with_usage()
{
  fprintf(stderr,
          "Usage: %s [--port|-p N] [--path|-P path]\n"
          "  -p N       set server port\n"
	  "  -P path    set server path\n",
          callname);
  exit(-1);
}


static inline
void parse_cmdline(int argc, char *const argv[])
{
  char c;
  static struct option long_opts[] = {
    { "port", 1, NULL, 'p' },
    { "path", 1, NULL, 'P' },
    { NULL, 0, NULL, 0 },
  };

  callname = strrchr(argv[0], '/')+1;

  /* option parsing loop */
  while ( (c=getopt_long(argc, argv, "p:P:", long_opts, NULL))
          != EOF ) {
    switch ( c ) {

    case 'p': {
      long port;
      char *strtolres;

      port = strtol(optarg, &strtolres, 10);

      if ( !*optarg || *strtolres ) {
        fprintf(stderr, "Unrecognized port %s\n", optarg);
        exit_with_usage();
      }

#if ( defined NETWORK_TCP ) || ( defined NETWORK_UDP )
      fprintf(stderr, "Using port %ld\n", port);
      server_port = port;
#elif ( defined NETWORK_GM )
      fprintf(stderr, "Using port %ld\n", port);
      server_port_id = port;
#else
      fprintf(stderr, "Specifying port useless without TCP, UDP or GM\n");
#endif
      break;
    }

    case 'P': {
#if ( defined STD_IO ) || ( defined AIO_IO )
      fprintf(stderr, "Using path %s\n", optarg);
      chroot(optarg);
#else
      fprintf(stderr, "Specifying path useless without STD or AIO I/O\n");
#endif
      break;
    }

    case '?':
    default:
      exit_with_usage();
    }
  }
}


/*******
 * AIO *
 *******/

#ifdef NEED_AIO

/* These functions handle all AIO processing,
 * concerning both TCP networking and disk IO
 */

#include <libaio.h>

io_context_t aio_context = NULL;

#if 0
static int aio_queue_nr_max = SERVER_EVENT_QUEUE_SIZE;
#endif


static int AIO_init()
{
  int res = 0;
  if ( !aio_context ) {
    res = io_setup(SERVER_EVENT_QUEUE_SIZE, &aio_context);
    if ( res < 0 ) {
      ERROR("io_setup failed on %d with error %m\n", res);
    }
  }
  return res;
}


static void AIO_finish()
{
  if ( aio_context )
    {
      io_destroy(aio_context);
      aio_context = NULL;
    }
}


void AIO_submit(struct iocb *iocb)
{
  if ( io_submit(aio_context, 1, &iocb) != 1 ) {
#if 0
    if (errno == EGAIN) {
      aio_queue_nr_max *= 2;
      LOG(5, "Increasing aio queue size to %d\n", aio_queue_nr_max);
      io_queue_grow(aio_context, aio_queue_nr_max);
      if ( io_submit(aio_context, 1, &iocb) != 1 )
        ERROR("io_submit failed again on error %m\n");
    } else
#endif
      ERROR("io_submit failed on error %m\n");
  }
}


void AIO_cancel(struct iocb *iocb)
{
  struct io_event res;
  io_cancel(aio_context, iocb, &res);
}


static INLINE void AIO_handle_events(struct io_event events[], int nr)
{
  int i;
  for(i=0; i<nr; i++)
    {
      io_callback_t callback = (io_callback_t)events[i].data;
      struct iocb *iocb = (struct iocb*)events[i].obj;
      int res1 = events[i].res;
      int res2 = events[i].res2;
      LOG(3, "Handling event #%d\n", i);
      callback(aio_context, iocb, res1, res2);
    }
}


#ifdef AIO_MIXED_WITH_OTHER_ASYNC_API
/* When mixing AIO with non compatible asynchron
 * interface (like GM or BIP), we have to not block
 * AIO event waiting
 */
static void AIO_test()
{
  static struct timespec timeout = {0, 0};
  struct io_event events[SERVER_EVENT_GET_MAX];
  int res;
  
 try_again:
  LOG(1, "Testing AIO events\n");
  res = io_getevents(aio_context, 1, SERVER_EVENT_GET_MAX, events, &timeout);

  LOG(2, "io_getevents returns %d\n", res);

  if ( res < 0 ) {
    ERROR("io_getevents failed on error %m\n");
    assert(0);
  }

  if ( !res )
    return;
  
  AIO_handle_events(events, res);
  
  if ( res == SERVER_EVENT_GET_MAX ) {
    LOG(3, "Got max events, try again...\n");
    goto try_again;
  }
}

#else /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */
/* When all asynchron calls are handled by AIO, we
 * can block on AIO event waiting
 */
static void AIO_wait()
{
  struct io_event events[SERVER_EVENT_GET_MAX];
  int res;

  LOG(1, "Waiting for AIO events\n");
  res = io_getevents(aio_context, 1, SERVER_EVENT_GET_MAX, events, NULL);

  LOG(2, "io_getevents returns %d\n", res);

  if ( res < 0 ) {
    ERROR("io_getevents failed on error %m\n");
    assert(0);
  }

  if ( !res )
    return;

  AIO_handle_events(events, res);

  while ( res == SERVER_EVENT_GET_MAX ) {
    static struct timespec timeout = {0, 0};

    LOG(3, "Got max events, try again...\n");

    LOG(1, "Testing AIO events\n");
    res = io_getevents(aio_context, 1, SERVER_EVENT_GET_MAX, events, &timeout);
    
    LOG(2, "io_getevents returns %d\n", res);

    if ( res < 0 ) {
      ERROR("io_getevents failed on error %m\n");
      assert(0);
    }
    
    if ( !res )
      return;

    AIO_handle_events(events, res);
  }
}
#endif /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */

#endif /* NEED_AIO */



/***********
 * Signals *
 ***********/

static void sigint_handler(int nothing)
{
  printf("Request #%d\n", request_nr);

  process_finish();
  network_finish();
#ifdef NEED_AIO
  AIO_finish();
#endif
  exit(0);
}


static void sigterm_handler(int nothing)
{
  printf("Request #%d\n", request_nr);
}


static void fix_sig_handler()
{
  struct sigaction sa;

  sa.sa_handler = sigterm_handler;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGTERM, &sa, NULL);

  sa.sa_handler = sigint_handler;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGINT, &sa, NULL);
}



/*****************
 * Queue threads *
 *****************/

#ifdef QUEUE_THREAD_ARCH

#ifndef AIO_TCP_POLL
/* AIO TCP polling is totally handled
 * by callbacks implied by the initial socket poll submit
 * Using AIO to poll on TCP sockets do not require this thread.
 */
static pthread_t recv_thread;

static void recv_thread_func(void)
{
  while ( 1 )
    {
      network_wait();
    }
}
#endif /* ~AIO_TCP_POLL */


static pthread_t process_thread;

static void process_thread_func(void)
{
  while ( 1 )
    {
      struct server_request *req;
      req = dequeue_ready_request();
      process_request(req);
    }
}


static pthread_t send_thread;

static void send_thread_func(void)
{
  while ( 1 )
    {
      struct server_request *req;
      req = dequeue_done_request();
      network_send(req);
      request_nr++;
    }
}

static INLINE void QUEUE_THREAD_process()
{
  /* create jobs */
#ifndef AIO_TCP_POLL
  pthread_create(&recv_thread, NULL, (void *(*)(void*))recv_thread_func, NULL);
#endif
  pthread_create(&process_thread, NULL, (void *(*)(void*))process_thread_func, NULL);
  pthread_create(&send_thread, NULL, (void *(*)(void*))send_thread_func, NULL);

  while ( 1 )
    {
# ifdef NEED_AIO
      /* When using AIO, we need a thread to handle AIO events */
      AIO_wait();
# else /* ~NEED_AIO */
      /* Elsewhere the main thread does nothing */
      sched_yield();
# endif /* ~NEED_AIO */
    }
}


#elif (defined SEQUENTIAL_ARCH)

/**************
 * Sequential *
 **************/

static INLINE void dequeueing()
{
  struct server_request *req;

  /* process all ready requests */
  while ( (req = dequeue_ready_request()) != NULL )
    process_request(req);
  
  /* send all done requests */
  while ( (req = dequeue_done_request()) != NULL )
    {
      network_send(req);
      request_nr++;
    }
}


static INLINE void SEQUENTIAL_process()
{
  while ( 1 )
    {
#ifdef NEED_AIO
# ifdef AIO_MIXED_WITH_OTHER_ASYNC_API
      /* is there done requests or TCP/UDP AIO receive events */
      AIO_test();

      /* is there GM/BIP or not-AIO TCP/UDP receive events */
      network_test();
# else /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */
      /* is there done requests or TCP/UDP AIO receive events */
      AIO_wait();
# endif /* ~AIO_MIXED_WITH_OTHER_ASYNC_API */
#else /* ~NEED_AIO */
      network_wait();
#endif /* ~NEED_AIO */
      /* process request and send done requests */
      dequeueing();
    }
}


#else /* CONNECTION_THREAD_ARCH */

/*********************
 * Connection thread *
 *********************/

static INLINE void CONNECTION_THREAD_process() {
  CONNECTION_THREAD_main_connection_thread_func();
}

#endif /* CONNECTION_THREAD_ARCH */



/********
 * Main *
 ********/

int main(int argc, char *const argv[])
{
  struct rlimit rlim;

  parse_cmdline(argc, argv);

  uid = getuid();
  gid = getgid();

  rlim.rlim_cur = FD_MAX;
  rlim.rlim_max = FD_MAX;
  if ( setrlimit(RLIMIT_NOFILE, &rlim) < 0 ) {
    getrlimit(RLIMIT_NOFILE, &rlim);
    rlim.rlim_cur = rlim.rlim_max;
    setrlimit(RLIMIT_NOFILE, &rlim);
    ERROR("Cannot setrlimit ENOFILE to %d, reverted to %ld\n",
	  FD_MAX, (long)rlim.rlim_max);
  }

#ifdef NEED_AIO
  if ( AIO_init() < 0 )
    goto eerror1;
#endif

  if ( network_init() < 0 )
    goto eerror2;

  if ( process_init() < 0 )
    goto eerror3;
 
  fix_sig_handler();

#ifdef QUEUE_THREAD_ARCH
  QUEUE_THREAD_process();
#elif CONNECTION_THREAD_ARCH
  CONNECTION_THREAD_process();
#else /* SEQUENTIAL_ARCH */
  SEQUENTIAL_process();
#endif

  process_finish();
 eerror3:

  network_finish();
 eerror2:

#ifdef NEED_AIO
  AIO_finish();
 eerror1:
#endif

  return 0;
}
